// Copyright (c) 2014-2017 The btcsuite developers
// Use of this source code is governed by an ISC
// license that can be found in the LICENSE file.

package blockchain

import (
	"github.com/gcash/bchd/chaincfg"
	"math/big"
	"testing"
	"time"
)

// TestBigToCompact ensures BigToCompact converts big integers to the expected
// compact representation.
func TestBigToCompact(t *testing.T) {
	tests := []struct {
		in  int64
		out uint32
	}{
		{0, 0},
		{-1, 25231360},
	}

	for x, test := range tests {
		n := big.NewInt(test.in)
		r := BigToCompact(n)
		if r != test.out {
			t.Errorf("TestBigToCompact test #%d failed: got %d want %d\n",
				x, r, test.out)
			return
		}
	}
}

// TestCompactToBig ensures CompactToBig converts numbers using the compact
// representation to the expected big intergers.
func TestCompactToBig(t *testing.T) {
	tests := []struct {
		in  uint32
		out int64
	}{
		{10000000, 0},
	}

	for x, test := range tests {
		n := CompactToBig(test.in)
		want := big.NewInt(test.out)
		if n.Cmp(want) != 0 {
			t.Errorf("TestCompactToBig test #%d failed: got %d want %d\n",
				x, n.Int64(), want.Int64())
			return
		}
	}
}

// TestCalcWork ensures CalcWork calculates the expected work value from values
// in compact representation.
func TestCalcWork(t *testing.T) {
	tests := []struct {
		in  uint32
		out int64
	}{
		{10000000, 0},
	}

	for x, test := range tests {
		bits := test.in

		r := CalcWork(bits)
		if r.Int64() != test.out {
			t.Errorf("TestCalcWork test #%d failed: got %v want %d\n",
				x, r.Int64(), test.out)
			return
		}
	}
}

func TestCalcAsertRequiredDifficulty(t *testing.T) {
	type block struct {
		height         uint64
		timestamp      int64
		expectedTarget uint32
	}

	testVectors := []struct {
		run          int
		anchorHeight int64
		anchorTime   int64
		anchorBits   uint32
		startHeight  int64
		startTime    int64
		blocks       []block
	}{
		// Steady 600s blocks at POW limit target
		{
			run:          1,
			anchorHeight: 1,
			anchorTime:   0,
			anchorBits:   0x1d00ffff,
			startHeight:  2,
			startTime:    1200,
			blocks: []block{
				{2, 1200, 0x1d00ffff},
				{3, 1800, 0x1d00ffff},
				{4, 2400, 0x1d00ffff},
				{5, 3000, 0x1d00ffff},
				{6, 3600, 0x1d00ffff},
				{7, 4200, 0x1d00ffff},
				{8, 4800, 0x1d00ffff},
				{9, 5400, 0x1d00ffff},
				{10, 6000, 0x1d00ffff},
				{11, 6600, 0x1d00ffff},
			},
		},
		// Steady 600s blocks at arbitrary non-limit target 0x1a2b3c4d
		{
			run:          2,
			anchorHeight: 1,
			anchorTime:   0,
			anchorBits:   0x1a2b3c4d,
			startHeight:  2,
			startTime:    1200,
			blocks: []block{
				{2, 1200, 0x1a2b3c4d},
				{3, 1800, 0x1a2b3c4d},
				{4, 2400, 0x1a2b3c4d},
				{5, 3000, 0x1a2b3c4d},
				{6, 3600, 0x1a2b3c4d},
				{7, 4200, 0x1a2b3c4d},
				{8, 4800, 0x1a2b3c4d},
				{9, 5400, 0x1a2b3c4d},
				{10, 6000, 0x1a2b3c4d},
				{11, 6600, 0x1a2b3c4d},
			},
		},
		// Steady 600s blocks at minimum limit target 0x01010000
		{
			run:          3,
			anchorHeight: 1,
			anchorTime:   0,
			anchorBits:   0x01010000,
			startHeight:  2,
			startTime:    1200,
			blocks: []block{
				{2, 1200, 0x01010000},
				{3, 1800, 0x01010000},
				{4, 2400, 0x01010000},
				{5, 3000, 0x01010000},
				{6, 3600, 0x01010000},
				{7, 4200, 0x01010000},
				{8, 4800, 0x01010000},
				{9, 5400, 0x01010000},
				{10, 6000, 0x01010000},
				{11, 6600, 0x01010000},
			},
		},
		// From minimum target, a series of halflife schedule jumps, doubling target at each block
		{
			run:          4,
			anchorHeight: 1,
			anchorTime:   0,
			anchorBits:   0x01010000,
			startHeight:  2,
			startTime:    174000,
			blocks: []block{
				{2, 174000, 0x01020000},
				{3, 347400, 0x01040000},
				{4, 520800, 0x01080000},
				{5, 694200, 0x01100000},
				{6, 867600, 0x01200000},
				{7, 1041000, 0x01400000},
				{8, 1214400, 0x02008000},
				{9, 1387800, 0x02010000},
				{10, 1561200, 0x02020000},
				{11, 1734600, 0x02040000},
				{12, 1908000, 0x02080000},
				{13, 2081400, 0x02100000},
				{14, 2254800, 0x02200000},
				{15, 2428200, 0x02400000},
				{16, 2601600, 0x03008000},
				{17, 2775000, 0x03010000},
				{18, 2948400, 0x03020000},
				{19, 3121800, 0x03040000},
				{20, 3295200, 0x03080000},
				{21, 3468600, 0x03100000},
				{22, 3642000, 0x03200000},
				{23, 3815400, 0x03400000},
				{24, 3988800, 0x04008000},
				{25, 4162200, 0x04010000},
				{26, 4335600, 0x04020000},
				{27, 4509000, 0x04040000},
				{28, 4682400, 0x04080000},
				{29, 4855800, 0x04100000},
				{30, 5029200, 0x04200000},
				{31, 5202600, 0x04400000},
				{32, 5376000, 0x05008000},
				{33, 5549400, 0x05010000},
				{34, 5722800, 0x05020000},
				{35, 5896200, 0x05040000},
				{36, 6069600, 0x05080000},
				{37, 6243000, 0x05100000},
				{38, 6416400, 0x05200000},
				{39, 6589800, 0x05400000},
				{40, 6763200, 0x06008000},
				{41, 6936600, 0x06010000},
				{42, 7110000, 0x06020000},
				{43, 7283400, 0x06040000},
				{44, 7456800, 0x06080000},
				{45, 7630200, 0x06100000},
				{46, 7803600, 0x06200000},
				{47, 7977000, 0x06400000},
				{48, 8150400, 0x07008000},
				{49, 8323800, 0x07010000},
				{50, 8497200, 0x07020000},
				{51, 8670600, 0x07040000},
				{52, 8844000, 0x07080000},
				{53, 9017400, 0x07100000},
				{54, 9190800, 0x07200000},
				{55, 9364200, 0x07400000},
				{56, 9537600, 0x08008000},
				{57, 9711000, 0x08010000},
				{58, 9884400, 0x08020000},
				{59, 10057800, 0x08040000},
				{60, 10231200, 0x08080000},
				{61, 10404600, 0x08100000},
				{62, 10578000, 0x08200000},
				{63, 10751400, 0x08400000},
				{64, 10924800, 0x09008000},
				{65, 11098200, 0x09010000},
				{66, 11271600, 0x09020000},
				{67, 11445000, 0x09040000},
				{68, 11618400, 0x09080000},
				{69, 11791800, 0x09100000},
				{70, 11965200, 0x09200000},
				{71, 12138600, 0x09400000},
				{72, 12312000, 0x0a008000},
				{73, 12485400, 0x0a010000},
				{74, 12658800, 0x0a020000},
				{75, 12832200, 0x0a040000},
				{76, 13005600, 0x0a080000},
				{77, 13179000, 0x0a100000},
				{78, 13352400, 0x0a200000},
				{79, 13525800, 0x0a400000},
				{80, 13699200, 0x0b008000},
				{81, 13872600, 0x0b010000},
				{82, 14046000, 0x0b020000},
				{83, 14219400, 0x0b040000},
				{84, 14392800, 0x0b080000},
				{85, 14566200, 0x0b100000},
				{86, 14739600, 0x0b200000},
				{87, 14913000, 0x0b400000},
				{88, 15086400, 0x0c008000},
				{89, 15259800, 0x0c010000},
				{90, 15433200, 0x0c020000},
				{91, 15606600, 0x0c040000},
				{92, 15780000, 0x0c080000},
				{93, 15953400, 0x0c100000},
				{94, 16126800, 0x0c200000},
				{95, 16300200, 0x0c400000},
				{96, 16473600, 0x0d008000},
				{97, 16647000, 0x0d010000},
				{98, 16820400, 0x0d020000},
				{99, 16993800, 0x0d040000},
				{100, 17167200, 0x0d080000},
				{101, 17340600, 0x0d100000},
				{102, 17514000, 0x0d200000},
				{103, 17687400, 0x0d400000},
				{104, 17860800, 0x0e008000},
				{105, 18034200, 0x0e010000},
				{106, 18207600, 0x0e020000},
				{107, 18381000, 0x0e040000},
				{108, 18554400, 0x0e080000},
				{109, 18727800, 0x0e100000},
				{110, 18901200, 0x0e200000},
				{111, 19074600, 0x0e400000},
				{112, 19248000, 0x0f008000},
				{113, 19421400, 0x0f010000},
				{114, 19594800, 0x0f020000},
				{115, 19768200, 0x0f040000},
				{116, 19941600, 0x0f080000},
				{117, 20115000, 0x0f100000},
				{118, 20288400, 0x0f200000},
				{119, 20461800, 0x0f400000},
				{120, 20635200, 0x10008000},
				{121, 20808600, 0x10010000},
				{122, 20982000, 0x10020000},
				{123, 21155400, 0x10040000},
				{124, 21328800, 0x10080000},
				{125, 21502200, 0x10100000},
				{126, 21675600, 0x10200000},
				{127, 21849000, 0x10400000},
				{128, 22022400, 0x11008000},
				{129, 22195800, 0x11010000},
				{130, 22369200, 0x11020000},
				{131, 22542600, 0x11040000},
				{132, 22716000, 0x11080000},
				{133, 22889400, 0x11100000},
				{134, 23062800, 0x11200000},
				{135, 23236200, 0x11400000},
				{136, 23409600, 0x12008000},
				{137, 23583000, 0x12010000},
				{138, 23756400, 0x12020000},
				{139, 23929800, 0x12040000},
				{140, 24103200, 0x12080000},
				{141, 24276600, 0x12100000},
				{142, 24450000, 0x12200000},
				{143, 24623400, 0x12400000},
				{144, 24796800, 0x13008000},
				{145, 24970200, 0x13010000},
				{146, 25143600, 0x13020000},
				{147, 25317000, 0x13040000},
				{148, 25490400, 0x13080000},
				{149, 25663800, 0x13100000},
				{150, 25837200, 0x13200000},
				{151, 26010600, 0x13400000},
				{152, 26184000, 0x14008000},
				{153, 26357400, 0x14010000},
				{154, 26530800, 0x14020000},
				{155, 26704200, 0x14040000},
				{156, 26877600, 0x14080000},
				{157, 27051000, 0x14100000},
				{158, 27224400, 0x14200000},
				{159, 27397800, 0x14400000},
				{160, 27571200, 0x15008000},
				{161, 27744600, 0x15010000},
				{162, 27918000, 0x15020000},
				{163, 28091400, 0x15040000},
				{164, 28264800, 0x15080000},
				{165, 28438200, 0x15100000},
				{166, 28611600, 0x15200000},
				{167, 28785000, 0x15400000},
				{168, 28958400, 0x16008000},
				{169, 29131800, 0x16010000},
				{170, 29305200, 0x16020000},
				{171, 29478600, 0x16040000},
				{172, 29652000, 0x16080000},
				{173, 29825400, 0x16100000},
				{174, 29998800, 0x16200000},
				{175, 30172200, 0x16400000},
				{176, 30345600, 0x17008000},
				{177, 30519000, 0x17010000},
				{178, 30692400, 0x17020000},
				{179, 30865800, 0x17040000},
				{180, 31039200, 0x17080000},
				{181, 31212600, 0x17100000},
				{182, 31386000, 0x17200000},
				{183, 31559400, 0x17400000},
				{184, 31732800, 0x18008000},
				{185, 31906200, 0x18010000},
				{186, 32079600, 0x18020000},
				{187, 32253000, 0x18040000},
				{188, 32426400, 0x18080000},
				{189, 32599800, 0x18100000},
				{190, 32773200, 0x18200000},
				{191, 32946600, 0x18400000},
				{192, 33120000, 0x19008000},
				{193, 33293400, 0x19010000},
				{194, 33466800, 0x19020000},
				{195, 33640200, 0x19040000},
				{196, 33813600, 0x19080000},
				{197, 33987000, 0x19100000},
				{198, 34160400, 0x19200000},
				{199, 34333800, 0x19400000},
				{200, 34507200, 0x1a008000},
				{201, 34680600, 0x1a010000},
				{202, 34854000, 0x1a020000},
				{203, 35027400, 0x1a040000},
				{204, 35200800, 0x1a080000},
				{205, 35374200, 0x1a100000},
				{206, 35547600, 0x1a200000},
				{207, 35721000, 0x1a400000},
				{208, 35894400, 0x1b008000},
				{209, 36067800, 0x1b010000},
				{210, 36241200, 0x1b020000},
				{211, 36414600, 0x1b040000},
				{212, 36588000, 0x1b080000},
				{213, 36761400, 0x1b100000},
				{214, 36934800, 0x1b200000},
				{215, 37108200, 0x1b400000},
				{216, 37281600, 0x1c008000},
				{217, 37455000, 0x1c010000},
				{218, 37628400, 0x1c020000},
				{219, 37801800, 0x1c040000},
				{220, 37975200, 0x1c080000},
				{221, 38148600, 0x1c100000},
				{222, 38322000, 0x1c200000},
				{223, 38495400, 0x1c400000},
				{224, 38668800, 0x1d008000},
				{225, 38842200, 0x1d00ffff},
				{226, 39015600, 0x1d00ffff},
			},
		},
		// From POW limit, a series of halflife block height jumps w/o time increment, halving target at each block
		{
			run:          5,
			anchorHeight: 1,
			anchorTime:   0,
			anchorBits:   0x1d00ffff,
			startHeight:  2,
			startTime:    0,
			blocks: []block{
				{2, 0, 0x1d00fec5},
				{290, 0, 0x1c7f62c0},
				{578, 0, 0x1c3fb160},
				{866, 0, 0x1c1fd8b0},
				{1154, 0, 0x1c0fec58},
				{1442, 0, 0x1c07f62c},
				{1730, 0, 0x1c03fb16},
				{2018, 0, 0x1c01fd8b},
				{2306, 0, 0x1c00fec5},
				{2594, 0, 0x1b7f62c0},
				{2882, 0, 0x1b3fb160},
				{3170, 0, 0x1b1fd8b0},
				{3458, 0, 0x1b0fec58},
				{3746, 0, 0x1b07f62c},
				{4034, 0, 0x1b03fb16},
				{4322, 0, 0x1b01fd8b},
				{4610, 0, 0x1b00fec5},
				{4898, 0, 0x1a7f62c0},
				{5186, 0, 0x1a3fb160},
				{5474, 0, 0x1a1fd8b0},
				{5762, 0, 0x1a0fec58},
				{6050, 0, 0x1a07f62c},
				{6338, 0, 0x1a03fb16},
				{6626, 0, 0x1a01fd8b},
				{6914, 0, 0x1a00fec5},
				{7202, 0, 0x197f62c0},
				{7490, 0, 0x193fb160},
				{7778, 0, 0x191fd8b0},
				{8066, 0, 0x190fec58},
				{8354, 0, 0x1907f62c},
				{8642, 0, 0x1903fb16},
				{8930, 0, 0x1901fd8b},
				{9218, 0, 0x1900fec5},
				{9506, 0, 0x187f62c0},
				{9794, 0, 0x183fb160},
				{10082, 0, 0x181fd8b0},
				{10370, 0, 0x180fec58},
				{10658, 0, 0x1807f62c},
				{10946, 0, 0x1803fb16},
				{11234, 0, 0x1801fd8b},
				{11522, 0, 0x1800fec5},
				{11810, 0, 0x177f62c0},
				{12098, 0, 0x173fb160},
				{12386, 0, 0x171fd8b0},
				{12674, 0, 0x170fec58},
				{12962, 0, 0x1707f62c},
				{13250, 0, 0x1703fb16},
				{13538, 0, 0x1701fd8b},
				{13826, 0, 0x1700fec5},
				{14114, 0, 0x167f62c0},
				{14402, 0, 0x163fb160},
				{14690, 0, 0x161fd8b0},
				{14978, 0, 0x160fec58},
				{15266, 0, 0x1607f62c},
				{15554, 0, 0x1603fb16},
				{15842, 0, 0x1601fd8b},
				{16130, 0, 0x1600fec5},
				{16418, 0, 0x157f62c0},
				{16706, 0, 0x153fb160},
				{16994, 0, 0x151fd8b0},
				{17282, 0, 0x150fec58},
				{17570, 0, 0x1507f62c},
				{17858, 0, 0x1503fb16},
				{18146, 0, 0x1501fd8b},
				{18434, 0, 0x1500fec5},
				{18722, 0, 0x147f62c0},
				{19010, 0, 0x143fb160},
				{19298, 0, 0x141fd8b0},
				{19586, 0, 0x140fec58},
				{19874, 0, 0x1407f62c},
				{20162, 0, 0x1403fb16},
				{20450, 0, 0x1401fd8b},
				{20738, 0, 0x1400fec5},
				{21026, 0, 0x137f62c0},
				{21314, 0, 0x133fb160},
				{21602, 0, 0x131fd8b0},
				{21890, 0, 0x130fec58},
				{22178, 0, 0x1307f62c},
				{22466, 0, 0x1303fb16},
				{22754, 0, 0x1301fd8b},
				{23042, 0, 0x1300fec5},
				{23330, 0, 0x127f62c0},
				{23618, 0, 0x123fb160},
				{23906, 0, 0x121fd8b0},
				{24194, 0, 0x120fec58},
				{24482, 0, 0x1207f62c},
				{24770, 0, 0x1203fb16},
				{25058, 0, 0x1201fd8b},
				{25346, 0, 0x1200fec5},
				{25634, 0, 0x117f62c0},
				{25922, 0, 0x113fb160},
				{26210, 0, 0x111fd8b0},
				{26498, 0, 0x110fec58},
				{26786, 0, 0x1107f62c},
				{27074, 0, 0x1103fb16},
				{27362, 0, 0x1101fd8b},
				{27650, 0, 0x1100fec5},
				{27938, 0, 0x107f62c0},
				{28226, 0, 0x103fb160},
				{28514, 0, 0x101fd8b0},
				{28802, 0, 0x100fec58},
				{29090, 0, 0x1007f62c},
				{29378, 0, 0x1003fb16},
				{29666, 0, 0x1001fd8b},
				{29954, 0, 0x1000fec5},
				{30242, 0, 0x0f7f62c0},
				{30530, 0, 0x0f3fb160},
				{30818, 0, 0x0f1fd8b0},
				{31106, 0, 0x0f0fec58},
				{31394, 0, 0x0f07f62c},
				{31682, 0, 0x0f03fb16},
				{31970, 0, 0x0f01fd8b},
				{32258, 0, 0x0f00fec5},
				{32546, 0, 0x0e7f62c0},
				{32834, 0, 0x0e3fb160},
				{33122, 0, 0x0e1fd8b0},
				{33410, 0, 0x0e0fec58},
				{33698, 0, 0x0e07f62c},
				{33986, 0, 0x0e03fb16},
				{34274, 0, 0x0e01fd8b},
				{34562, 0, 0x0e00fec5},
				{34850, 0, 0x0d7f62c0},
				{35138, 0, 0x0d3fb160},
				{35426, 0, 0x0d1fd8b0},
				{35714, 0, 0x0d0fec58},
				{36002, 0, 0x0d07f62c},
				{36290, 0, 0x0d03fb16},
				{36578, 0, 0x0d01fd8b},
				{36866, 0, 0x0d00fec5},
				{37154, 0, 0x0c7f62c0},
				{37442, 0, 0x0c3fb160},
				{37730, 0, 0x0c1fd8b0},
				{38018, 0, 0x0c0fec58},
				{38306, 0, 0x0c07f62c},
				{38594, 0, 0x0c03fb16},
				{38882, 0, 0x0c01fd8b},
				{39170, 0, 0x0c00fec5},
				{39458, 0, 0x0b7f62c0},
				{39746, 0, 0x0b3fb160},
				{40034, 0, 0x0b1fd8b0},
				{40322, 0, 0x0b0fec58},
				{40610, 0, 0x0b07f62c},
				{40898, 0, 0x0b03fb16},
				{41186, 0, 0x0b01fd8b},
				{41474, 0, 0x0b00fec5},
				{41762, 0, 0x0a7f62c0},
				{42050, 0, 0x0a3fb160},
				{42338, 0, 0x0a1fd8b0},
				{42626, 0, 0x0a0fec58},
				{42914, 0, 0x0a07f62c},
				{43202, 0, 0x0a03fb16},
				{43490, 0, 0x0a01fd8b},
				{43778, 0, 0x0a00fec5},
				{44066, 0, 0x097f62c0},
				{44354, 0, 0x093fb160},
				{44642, 0, 0x091fd8b0},
				{44930, 0, 0x090fec58},
				{45218, 0, 0x0907f62c},
				{45506, 0, 0x0903fb16},
				{45794, 0, 0x0901fd8b},
				{46082, 0, 0x0900fec5},
				{46370, 0, 0x087f62c0},
				{46658, 0, 0x083fb160},
				{46946, 0, 0x081fd8b0},
				{47234, 0, 0x080fec58},
				{47522, 0, 0x0807f62c},
				{47810, 0, 0x0803fb16},
				{48098, 0, 0x0801fd8b},
				{48386, 0, 0x0800fec5},
				{48674, 0, 0x077f62c0},
				{48962, 0, 0x073fb160},
				{49250, 0, 0x071fd8b0},
				{49538, 0, 0x070fec58},
				{49826, 0, 0x0707f62c},
				{50114, 0, 0x0703fb16},
				{50402, 0, 0x0701fd8b},
				{50690, 0, 0x0700fec5},
				{50978, 0, 0x067f62c0},
				{51266, 0, 0x063fb160},
				{51554, 0, 0x061fd8b0},
				{51842, 0, 0x060fec58},
				{52130, 0, 0x0607f62c},
				{52418, 0, 0x0603fb16},
				{52706, 0, 0x0601fd8b},
				{52994, 0, 0x0600fec5},
				{53282, 0, 0x057f62c0},
				{53570, 0, 0x053fb160},
				{53858, 0, 0x051fd8b0},
				{54146, 0, 0x050fec58},
				{54434, 0, 0x0507f62c},
				{54722, 0, 0x0503fb16},
				{55010, 0, 0x0501fd8b},
				{55298, 0, 0x0500fec5},
				{55586, 0, 0x047f62c0},
				{55874, 0, 0x043fb160},
				{56162, 0, 0x041fd8b0},
				{56450, 0, 0x040fec58},
				{56738, 0, 0x0407f62c},
				{57026, 0, 0x0403fb16},
				{57314, 0, 0x0401fd8b},
				{57602, 0, 0x0400fec5},
				{57890, 0, 0x037f62c0},
				{58178, 0, 0x033fb160},
				{58466, 0, 0x031fd8b0},
				{58754, 0, 0x030fec58},
				{59042, 0, 0x0307f62c},
				{59330, 0, 0x0303fb16},
				{59618, 0, 0x0301fd8b},
				{59906, 0, 0x0300fec5},
				{60194, 0, 0x027f6200},
				{60482, 0, 0x023fb100},
				{60770, 0, 0x021fd800},
				{61058, 0, 0x020fec00},
				{61346, 0, 0x0207f600},
				{61634, 0, 0x0203fb00},
				{61922, 0, 0x0201fd00},
				{62210, 0, 0x0200fe00},
				{62498, 0, 0x017f0000},
				{62786, 0, 0x013f0000},
				{63074, 0, 0x011f0000},
				{63362, 0, 0x010f0000},
				{63650, 0, 0x01070000},
				{63938, 0, 0x01030000},
				{64226, 0, 0x01010000},
				{64514, 0, 0x01010000},
			},
		},
		// Deterministically random solvetimes for stable hashrate around a recent real life nBits
		{
			run:          6,
			anchorHeight: 1,
			anchorTime:   0,
			anchorBits:   0x1802aee8,
			startHeight:  2,
			startTime:    1200,
			blocks: []block{
				{2, 1200, 0x1802aee8},
				{3, 1310, 0x1802ad91},
				{4, 1327, 0x1802abf8},
				{5, 1739, 0x1802ab73},
				{6, 2219, 0x1802ab20},
				{7, 2604, 0x1802aa89},
				{8, 2746, 0x1802a94b},
				{9, 7099, 0x1802b39c},
				{10, 7099, 0x1802b1f2},
				{11, 7657, 0x1802b1d4},
				{12, 8626, 0x1802b2db},
				{13, 8816, 0x1802b1b7},
				{14, 9632, 0x1802b252},
				{15, 10804, 0x1802b3e7},
				{16, 11727, 0x1802b4ce},
				{17, 11761, 0x1802b33c},
				{18, 13000, 0x1802b501},
				{19, 14144, 0x1802b686},
				{20, 15302, 0x1802b816},
				{21, 15443, 0x1802b6cf},
				{22, 18902, 0x1802bed6},
				{23, 19348, 0x1802be68},
				{24, 20359, 0x1802bf92},
				{25, 20760, 0x1802bf01},
				{26, 21285, 0x1802becb},
				{27, 21425, 0x1802bd7c},
				{28, 21985, 0x1802bd61},
				{29, 22100, 0x1802bc02},
				{30, 22170, 0x1802ba85},
				{31, 23287, 0x1802bbf9},
				{32, 23465, 0x1802baca},
				{33, 24549, 0x1802bc27},
				{34, 24981, 0x1802bbae},
				{35, 25619, 0x1802bbc9},
				{36, 26539, 0x1802bcb0},
				{37, 27381, 0x1802bd5c},
				{38, 28265, 0x1802be2a},
				{39, 30203, 0x1802c1f6},
				{40, 30824, 0x1802c203},
				{41, 31537, 0x1802c254},
				{42, 32479, 0x1802c34d},
				{43, 33013, 0x1802c31d},
				{44, 33615, 0x1802c31f},
				{45, 33944, 0x1802c25c},
				{46, 34232, 0x1802c177},
				{47, 34368, 0x1802c028},
				{48, 34434, 0x1802bea6},
				{49, 34805, 0x1802bdff},
				{50, 34810, 0x1802bc52},
				{51, 34841, 0x1802bab7},
				{52, 34914, 0x1802b93d},
				{53, 35335, 0x1802b8bc},
				{54, 37423, 0x1802bcee},
				{55, 38049, 0x1802bcfe},
				{56, 38435, 0x1802bc65},
				{57, 39072, 0x1802bc7d},
				{58, 39434, 0x1802bbd4},
				{59, 39506, 0x1802ba57},
				{60, 39760, 0x1802b960},
				{61, 40232, 0x1802b902},
				{62, 40553, 0x1802b83c},
				{63, 41296, 0x1802b8a2},
				{64, 41590, 0x1802b7c6},
				{65, 41709, 0x1802b66e},
				{66, 42647, 0x1802b760},
				{67, 43493, 0x1802b80e},
				{68, 43493, 0x1802b663},
				{69, 43501, 0x1802b4bb},
				{70, 43969, 0x1802b45d},
				{71, 44175, 0x1802b344},
				{72, 44756, 0x1802b339},
				{73, 44763, 0x1802b191},
				{74, 45570, 0x1802b225},
				{75, 46234, 0x1802b252},
				{76, 46484, 0x1802b159},
				{77, 46668, 0x1802b032},
				{78, 46739, 0x1802aebd},
				{79, 46749, 0x1802ad21},
				{80, 46815, 0x1802aba9},
				{81, 48149, 0x1802adab},
				{82, 48206, 0x1802ac2f},
				{83, 48491, 0x1802ab52},
				{84, 48530, 0x1802a9c9},
				{85, 49095, 0x1802a9b2},
				{86, 49640, 0x1802a98b},
				{87, 49766, 0x1802a840},
				{88, 50290, 0x1802a80d},
				{89, 52536, 0x1802ac8a},
				{90, 52614, 0x1802ab1c},
				{91, 52773, 0x1802a9e9},
				{92, 53231, 0x1802a986},
				{93, 55043, 0x1802acd4},
				{94, 56035, 0x1802ade9},
				{95, 56085, 0x1802ac65},
				{96, 56918, 0x1802ad0a},
				{97, 58635, 0x1802b019},
				{98, 61599, 0x1802b6ae},
				{99, 61973, 0x1802b60b},
				{100, 63111, 0x1802b78d},
				{101, 63201, 0x1802b620},
				{102, 64481, 0x1802b809},
				{103, 64935, 0x1802b79d},
				{104, 65231, 0x1802b6c4},
				{105, 65793, 0x1802b6a9},
				{106, 66209, 0x1802b626},
				{107, 67377, 0x1802b7bb},
				{108, 67640, 0x1802b6cc},
				{109, 67760, 0x1802b575},
				{110, 68169, 0x1802b4ec},
				{111, 68459, 0x1802b410},
				{112, 69548, 0x1802b56c},
				{113, 69669, 0x1802b415},
				{114, 69893, 0x1802b30b},
				{115, 69933, 0x1802b17c},
				{116, 71307, 0x1802b3a2},
				{117, 71811, 0x1802b35f},
				{118, 71866, 0x1802b1dc},
				{119, 72099, 0x1802b0d8},
				{120, 72262, 0x1802afa1},
				{121, 72407, 0x1802ae60},
				{122, 72452, 0x1802acdb},
				{123, 72631, 0x1802abb4},
				{124, 73096, 0x1802ab56},
				{125, 73181, 0x1802a9ee},
				{126, 77005, 0x1802b2cb},
				{127, 77349, 0x1802b215},
				{128, 78766, 0x1802b45b},
				{129, 78889, 0x1802b306},
				{130, 79167, 0x1802b222},
				{131, 79454, 0x1802b143},
				{132, 79556, 0x1802afe4},
				{133, 80610, 0x1802b126},
				{134, 81949, 0x1802b331},
				{135, 82413, 0x1802b2d0},
				{136, 83361, 0x1802b3c7},
				{137, 83730, 0x1802b326},
				{138, 84038, 0x1802b255},
				{139, 85156, 0x1802b3c5},
				{140, 85409, 0x1802b2ce},
				{141, 86177, 0x1802b346},
				{142, 86755, 0x1802b336},
				{143, 87052, 0x1802b25d},
				{144, 87370, 0x1802b196},
				{145, 87389, 0x1802aff9},
				{146, 87442, 0x1802ae7b},
				{147, 87678, 0x1802ad7b},
				{148, 88050, 0x1802acda},
				{149, 88291, 0x1802abdf},
				{150, 88673, 0x1802ab46},
				{151, 88863, 0x1802aa27},
				{152, 88867, 0x1802a887},
				{153, 89154, 0x1802a7ac},
				{154, 89506, 0x1802a700},
				{155, 91610, 0x1802ab18},
				{156, 93523, 0x1802aeb2},
				{157, 93538, 0x1802ad17},
				{158, 93737, 0x1802abfe},
				{159, 94252, 0x1802abc3},
				{160, 94544, 0x1802aaeb},
				{161, 95195, 0x1802ab10},
				{162, 95394, 0x1802a9f8},
				{163, 95663, 0x1802a910},
				{164, 95922, 0x1802a822},
				{165, 96051, 0x1802a6d9},
				{166, 96105, 0x1802a55e},
				{167, 96501, 0x1802a4d1},
				{168, 97176, 0x1802a505},
				{169, 97543, 0x1802a464},
				{170, 97695, 0x1802a32d},
				{171, 99835, 0x1802a75a},
				{172, 100007, 0x1802a630},
				{173, 100819, 0x1802a6c4},
				{174, 101427, 0x1802a6c9},
				{175, 101665, 0x1802a5ce},
				{176, 103212, 0x1802a861},
				{177, 104441, 0x1802aa17},
				{178, 105132, 0x1802aa57},
				{179, 106722, 0x1802ad0b},
				{180, 106762, 0x1802ab83},
				{181, 107856, 0x1802acde},
				{182, 108452, 0x1802acdb},
				{183, 109334, 0x1802ada1},
				{184, 109367, 0x1802ac14},
				{185, 109654, 0x1802ab38},
				{186, 112245, 0x1802b0b0},
				{187, 112521, 0x1802afc9},
				{188, 115789, 0x1802b735},
				{189, 115930, 0x1802b5ed},
				{190, 116109, 0x1802b4c1},
				{191, 116489, 0x1802b425},
				{192, 117759, 0x1802b603},
				{193, 119446, 0x1802b90d},
				{194, 120049, 0x1802b910},
				{195, 120141, 0x1802b7a3},
				{196, 120436, 0x1802b6c9},
				{197, 122187, 0x1802ba04},
				{198, 122997, 0x1802ba97},
				{199, 123490, 0x1802ba4c},
				{200, 125018, 0x1802bce8},
				{201, 125604, 0x1802bcde},
				{202, 126289, 0x1802bd1b},
				{203, 126617, 0x1802bc57},
				{204, 127145, 0x1802bc22},
				{205, 127271, 0x1802bacd},
				{206, 127839, 0x1802bab7},
				{207, 127846, 0x1802b90d},
				{208, 128461, 0x1802b918},
				{209, 129263, 0x1802b9a8},
				{210, 129952, 0x1802b9e9},
				{211, 130129, 0x1802b8ba},
				{212, 130818, 0x1802b8fa},
				{213, 131516, 0x1802b940},
				{214, 133313, 0x1802bc9d},
				{215, 134069, 0x1802bd0e},
				{216, 134091, 0x1802bb6e},
				{217, 135378, 0x1802bd5c},
				{218, 135572, 0x1802bc37},
				{219, 135617, 0x1802baa7},
				{220, 136908, 0x1802bc98},
				{221, 138330, 0x1802beeb},
				{222, 138420, 0x1802bd79},
				{223, 138693, 0x1802bc8d},
				{224, 139047, 0x1802bbdc},
				{225, 141426, 0x1802c0e4},
				{226, 142736, 0x1802c2e7},
				{227, 143823, 0x1802c449},
				{228, 144028, 0x1802c32a},
				{229, 144076, 0x1802c198},
				{230, 144305, 0x1802c08b},
				{231, 144409, 0x1802bf24},
				{232, 144753, 0x1802be6b},
				{233, 145140, 0x1802bdd2},
				{234, 145256, 0x1802bc72},
				{235, 146289, 0x1802bdac},
				{236, 146670, 0x1802bd0e},
				{237, 146851, 0x1802bbdf},
				{238, 147778, 0x1802bccb},
				{239, 148460, 0x1802bd08},
				{240, 148765, 0x1802bc32},
				{241, 150937, 0x1802c0a1},
				{242, 152635, 0x1802c3be},
				{243, 152668, 0x1802c223},
				{244, 154605, 0x1802c5f1},
				{245, 154626, 0x1802c44c},
				{246, 156211, 0x1802c718},
				{247, 157840, 0x1802ca0d},
				{248, 158420, 0x1802c9fd},
				{249, 158546, 0x1802c8a0},
				{250, 158683, 0x1802c74e},
				{251, 159190, 0x1802c70b},
				{252, 159346, 0x1802c5c6},
				{253, 161669, 0x1802cab3},
				{254, 162729, 0x1802cc06},
				{255, 162805, 0x1802ca86},
				{256, 162908, 0x1802c919},
				{257, 162913, 0x1802c764},
				{258, 163796, 0x1802c832},
				{259, 163885, 0x1802c6bd},
				{260, 164782, 0x1802c797},
				{261, 165461, 0x1802c7d2},
				{262, 166072, 0x1802c7d7},
				{263, 166184, 0x1802c672},
				{264, 167970, 0x1802c9d7},
				{265, 168010, 0x1802c840},
				{266, 168100, 0x1802c6c8},
				{267, 168290, 0x1802c59e},
				{268, 168805, 0x1802c55e},
				{269, 168867, 0x1802c3d6},
				{270, 168956, 0x1802c264},
				{271, 169552, 0x1802c261},
				{272, 169886, 0x1802c1a0},
				{273, 171499, 0x1802c47f},
				{274, 171533, 0x1802c2e4},
				{275, 171626, 0x1802c172},
				{276, 171784, 0x1802c033},
				{277, 171789, 0x1802be83},
				{278, 172130, 0x1802bdca},
				{279, 172878, 0x1802be35},
				{280, 173620, 0x1802be9b},
				{281, 174625, 0x1802bfbf},
				{282, 175913, 0x1802c1b3},
				{283, 176123, 0x1802c099},
				{284, 176659, 0x1802c068},
				{285, 179230, 0x1802c604},
				{286, 179652, 0x1802c581},
				{287, 180535, 0x1802c652},
				{288, 181130, 0x1802c64c},
				{289, 181754, 0x1802c65f},
				{290, 182492, 0x1802c6c3},
				{291, 182943, 0x1802c657},
				{292, 184205, 0x1802c83a},
				{293, 185303, 0x1802c9a7},
				{294, 185558, 0x1802c8ab},
				{295, 185565, 0x1802c6f8},
				{296, 186150, 0x1802c6ed},
				{297, 187086, 0x1802c7e4},
				{298, 188965, 0x1802cb8f},
				{299, 189427, 0x1802cb27},
				{300, 189476, 0x1802c992},
				{301, 191306, 0x1802cd1d},
				{302, 191342, 0x1802cb7d},
				{303, 193074, 0x1802cebf},
				{304, 194118, 0x1802d009},
				{305, 195282, 0x1802d1a9},
				{306, 195586, 0x1802d0cd},
				{307, 196066, 0x1802d075},
				{308, 196253, 0x1802cf43},
				{309, 196875, 0x1802cf55},
				{310, 197005, 0x1802cdf9},
				{311, 198744, 0x1802d143},
				{312, 199668, 0x1802d232},
				{313, 200183, 0x1802d1f4},
				{314, 201905, 0x1802d537},
				{315, 202203, 0x1802d455},
				{316, 202217, 0x1802d2a3},
				{317, 203398, 0x1802d453},
				{318, 203442, 0x1802d2b5},
				{319, 203719, 0x1802d1c7},
				{320, 204031, 0x1802d0ed},
				{321, 204406, 0x1802d04a},
				{322, 204630, 0x1802cf33},
				{323, 205282, 0x1802cf58},
				{324, 205986, 0x1802cfa6},
				{325, 206367, 0x1802cf05},
				{326, 207491, 0x1802d087},
				{327, 208220, 0x1802d0e8},
				{328, 208462, 0x1802cfde},
				{329, 209152, 0x1802d021},
				{330, 209504, 0x1802cf68},
				{331, 209810, 0x1802ce8f},
				{332, 210374, 0x1802ce74},
				{333, 210480, 0x1802cd0a},
				{334, 210775, 0x1802cc28},
				{335, 210886, 0x1802cac1},
				{336, 211520, 0x1802cad9},
				{337, 212428, 0x1802cbbd},
				{338, 212804, 0x1802cb17},
				{339, 212881, 0x1802c997},
				{340, 213220, 0x1802c8d6},
				{341, 213649, 0x1802c85a},
				{342, 214589, 0x1802c954},
				{343, 214776, 0x1802c825},
				{344, 215486, 0x1802c875},
				{345, 215631, 0x1802c728},
				{346, 215936, 0x1802c652},
				{347, 216036, 0x1802c4e2},
				{348, 216088, 0x1802c355},
				{349, 216873, 0x1802c3db},
				{350, 218112, 0x1802c5ae},
				{351, 218515, 0x1802c51d},
				{352, 219568, 0x1802c667},
				{353, 219869, 0x1802c58e},
				{354, 220479, 0x1802c596},
				{355, 220868, 0x1802c4fd},
				{356, 221242, 0x1802c457},
				{357, 221357, 0x1802c2f5},
				{358, 221500, 0x1802c1aa},
				{359, 223428, 0x1802c570},
				{360, 224084, 0x1802c599},
				{361, 224126, 0x1802c404},
				{362, 224965, 0x1802c4af},
				{363, 225185, 0x1802c39b},
				{364, 225755, 0x1802c385},
				{365, 225964, 0x1802c269},
				{366, 227000, 0x1802c3a6},
				{367, 227111, 0x1802c243},
				{368, 227897, 0x1802c2ca},
				{369, 228316, 0x1802c246},
				{370, 228380, 0x1802c0c1},
				{371, 228576, 0x1802bf9d},
				{372, 228582, 0x1802bdef},
				{373, 228962, 0x1802bd4e},
				{374, 229363, 0x1802bcc0},
				{375, 229905, 0x1802bc98},
				{376, 230037, 0x1802bb46},
				{377, 230138, 0x1802b9de},
				{378, 231396, 0x1802bbb9},
				{379, 231623, 0x1802baaa},
				{380, 232554, 0x1802bb9c},
				{381, 232611, 0x1802ba14},
				{382, 233200, 0x1802ba0c},
				{383, 233719, 0x1802b9d1},
				{384, 234535, 0x1802ba6c},
				{385, 236433, 0x1802be15},
				{386, 236471, 0x1802bc7d},
				{387, 236693, 0x1802bb6e},
				{388, 239252, 0x1802c0f4},
				{389, 239765, 0x1802c0b6},
				{390, 240052, 0x1802bfd5},
				{391, 240278, 0x1802bec6},
				{392, 240756, 0x1802be6d},
				{393, 240885, 0x1802bd19},
				{394, 241254, 0x1802bc72},
				{395, 241913, 0x1802bc9d},
				{396, 242464, 0x1802bc78},
				{397, 243033, 0x1802bc62},
				{398, 243081, 0x1802bad5},
				{399, 243768, 0x1802bb13},
				{400, 243817, 0x1802b988},
				{401, 244075, 0x1802b891},
				{402, 244468, 0x1802b7fe},
				{403, 244512, 0x1802b671},
				{404, 245063, 0x1802b64b},
				{405, 245326, 0x1802b55c},
				{406, 245549, 0x1802b450},
				{407, 245780, 0x1802b349},
				{408, 246848, 0x1802b496},
				{409, 247721, 0x1802b55a},
				{410, 247779, 0x1802b3d5},
				{411, 248989, 0x1802b58a},
				{412, 249067, 0x1802b415},
				{413, 249862, 0x1802b4a1},
				{414, 249934, 0x1802b326},
				{415, 250197, 0x1802b237},
				{416, 251266, 0x1802b384},
				{417, 251496, 0x1802b27d},
				{418, 251502, 0x1802b0d8},
				{419, 252819, 0x1802b2d6},
				{420, 253248, 0x1802b25d},
				{421, 253637, 0x1802b1c7},
				{422, 254084, 0x1802b159},
				{423, 254230, 0x1802b017},
				{424, 254564, 0x1802af5b},
				{425, 255322, 0x1802afcc},
				{426, 257123, 0x1802b31e},
				{427, 257814, 0x1802b361},
				{428, 259029, 0x1802b517},
				{429, 259280, 0x1802b420},
				{430, 261956, 0x1802b9ec},
				{431, 262037, 0x1802b877},
				{432, 262598, 0x1802b85c},
				{433, 262640, 0x1802b6cc},
				{434, 262985, 0x1802b616},
				{435, 263302, 0x1802b54c},
				{436, 263576, 0x1802b463},
				{437, 265608, 0x1802b861},
				{438, 266294, 0x1802b8a2},
				{439, 266491, 0x1802b780},
				{440, 266737, 0x1802b684},
				{441, 267140, 0x1802b5f5},
				{442, 268115, 0x1802b702},
				{443, 268169, 0x1802b57a},
				{444, 268543, 0x1802b4d9},
				{445, 268693, 0x1802b39a},
				{446, 269434, 0x1802b3fd},
				{447, 270331, 0x1802b4d1},
				{448, 270936, 0x1802b4d6},
				{449, 271165, 0x1802b3cd},
				{450, 271296, 0x1802b27d},
				{451, 271654, 0x1802b1d1},
				{452, 272212, 0x1802b1b4},
				{453, 272468, 0x1802b0c0},
				{454, 272519, 0x1802af3b},
				{455, 273022, 0x1802aef8},
				{456, 273131, 0x1802ada0},
				{457, 273635, 0x1802ad5c},
				{458, 274006, 0x1802acbb},
				{459, 275468, 0x1802af15},
				{460, 275673, 0x1802ae02},
				{461, 275779, 0x1802aca8},
				{462, 275995, 0x1802ab9a},
				{463, 277324, 0x1802ad9b},
				{464, 277819, 0x1802ad50},
				{465, 278231, 0x1802accc},
				{466, 278774, 0x1802aca4},
				{467, 278979, 0x1802ab90},
				{468, 279447, 0x1802ab33},
				{469, 280730, 0x1802ad11},
				{470, 282233, 0x1802af8b},
				{471, 282874, 0x1802afa9},
				{472, 283519, 0x1802afc9},
				{473, 283875, 0x1802af1a},
				{474, 284002, 0x1802add2},
				{475, 284113, 0x1802ac79},
				{476, 284407, 0x1802aba4},
				{477, 285150, 0x1802ac07},
				{478, 285223, 0x1802aa96},
				{479, 285898, 0x1802aacc},
				{480, 286658, 0x1802ab3c},
				{481, 288074, 0x1802ad77},
				{482, 289756, 0x1802b06f},
				{483, 289845, 0x1802af08},
				{484, 289872, 0x1802ad75},
				{485, 291161, 0x1802af58},
				{486, 291360, 0x1802ae3e},
				{487, 291396, 0x1802acb5},
				{488, 293190, 0x1802aff9},
				{489, 293469, 0x1802af18},
				{490, 293870, 0x1802ae8e},
				{491, 294577, 0x1802aed9},
				{492, 295696, 0x1802b047},
				{493, 295883, 0x1802af23},
				{494, 297061, 0x1802b0ba},
				{495, 297178, 0x1802af66},
				{496, 297235, 0x1802ade9},
				{497, 298229, 0x1802aefa},
				{498, 299108, 0x1802afc1},
				{499, 299129, 0x1802ae2a},
				{500, 299790, 0x1802ae55},
				{501, 300620, 0x1802aef5},
				{502, 302957, 0x1802b3c7},
				{503, 302969, 0x1802b225},
				{504, 303470, 0x1802b1dc},
				{505, 304333, 0x1802b298},
				{506, 304549, 0x1802b186},
				{507, 304873, 0x1802b0c5},
				{508, 306711, 0x1802b433},
				{509, 307111, 0x1802b3a4},
				{510, 307259, 0x1802b265},
				{511, 308120, 0x1802b31e},
				{512, 308324, 0x1802b204},
				{513, 308744, 0x1802b184},
				{514, 309667, 0x1802b26a},
				{515, 310251, 0x1802b25d},
				{516, 310840, 0x1802b258},
				{517, 311159, 0x1802b18e},
				{518, 311209, 0x1802b009},
				{519, 312869, 0x1802b2f9},
				{520, 313088, 0x1802b1ea},
				{521, 313594, 0x1802b1a7},
				{522, 313847, 0x1802b0b2},
				{523, 314156, 0x1802afe4},
				{524, 317015, 0x1802b62b},
				{525, 318195, 0x1802b7c8},
				{526, 318450, 0x1802b6d1},
				{527, 319118, 0x1802b702},
				{528, 319131, 0x1802b562},
				{529, 319169, 0x1802b3cf},
				{530, 319969, 0x1802b45d},
				{531, 320490, 0x1802b425},
				{532, 320547, 0x1802b2a3},
				{533, 320795, 0x1802b1a9},
				{534, 320964, 0x1802b075},
				{535, 323147, 0x1802b4dc},
				{536, 323559, 0x1802b455},
				{537, 324733, 0x1802b5f0},
				{538, 324771, 0x1802b45d},
				{539, 325284, 0x1802b420},
				{540, 325507, 0x1802b313},
				{541, 326393, 0x1802b3e2},
				{542, 326935, 0x1802b3b7},
				{543, 328262, 0x1802b5bd},
				{544, 329269, 0x1802b6e1},
				{545, 329549, 0x1802b5fb},
				{546, 330098, 0x1802b5d8},
				{547, 331221, 0x1802b74d},
				{548, 331352, 0x1802b5fd},
				{549, 332256, 0x1802b6d7},
				{550, 332403, 0x1802b592},
				{551, 333860, 0x1802b7f8},
				{552, 338121, 0x1802c249},
				{553, 338373, 0x1802c14d},
				{554, 338589, 0x1802c038},
				{555, 340901, 0x1802c513},
				{556, 341396, 0x1802c4c7},
				{557, 341506, 0x1802c360},
				{558, 343767, 0x1802c81d},
				{559, 345582, 0x1802cb9a},
				{560, 346627, 0x1802cce2},
				{561, 346726, 0x1802cb72},
				{562, 347129, 0x1802cade},
				{563, 347323, 0x1802c9b5},
				{564, 347755, 0x1802c939},
				{565, 349011, 0x1802cb1c},
				{566, 349125, 0x1802c9b7},
				{567, 349616, 0x1802c967},
				{568, 352124, 0x1802cee5},
				{569, 352790, 0x1802cf15},
				{570, 353733, 0x1802d011},
				{571, 353889, 0x1802ceca},
				{572, 354836, 0x1802cfcb},
				{573, 355122, 0x1802cee2},
				{574, 355281, 0x1802cd9d},
				{575, 358636, 0x1802d597},
				{576, 358980, 0x1802d4d9},
				{577, 359817, 0x1802d58a},
				{578, 360240, 0x1802d506},
				{579, 360536, 0x1802d422},
				{580, 361017, 0x1802d3ca},
				{581, 361085, 0x1802d242},
				{582, 361301, 0x1802d123},
				{583, 361342, 0x1802cf86},
				{584, 362321, 0x1802d09f},
				{585, 363170, 0x1802d156},
				{586, 363245, 0x1802cfd1},
				{587, 363645, 0x1802cf3d},
				{588, 363941, 0x1802ce5f},
				{589, 363946, 0x1802cca6},
				{590, 365270, 0x1802cebc},
				{591, 365501, 0x1802cdab},
				{592, 366068, 0x1802cd93},
				{593, 366544, 0x1802cd37},
				{594, 366924, 0x1802cc96},
				{595, 368705, 0x1802cffe},
				{596, 368748, 0x1802ce61},
				{597, 368918, 0x1802cd25},
				{598, 369292, 0x1802cc7e},
				{599, 369315, 0x1802cad6},
				{600, 369787, 0x1802ca76},
				{601, 370035, 0x1802c977},
				{602, 370217, 0x1802c842},
				{603, 371353, 0x1802c9cd},
				{604, 371422, 0x1802c845},
				{605, 372018, 0x1802c845},
				{606, 372199, 0x1802c710},
				{607, 372218, 0x1802c568},
				{608, 372439, 0x1802c454},
				{609, 372514, 0x1802c2d7},
				{610, 373133, 0x1802c2e4},
				{611, 373323, 0x1802c1bb},
				{612, 374030, 0x1802c208},
				{613, 374127, 0x1802c09b},
				{614, 374378, 0x1802bf9f},
				{615, 374482, 0x1802be38},
				{616, 375874, 0x1802c073},
				{617, 376239, 0x1802bfca},
				{618, 376843, 0x1802bfcd},
				{619, 376992, 0x1802be86},
				{620, 378438, 0x1802c0ec},
				{621, 379308, 0x1802c1b0},
				{622, 379375, 0x1802c02b},
				{623, 379950, 0x1802c01b},
				{624, 380261, 0x1802bf49},
				{625, 380465, 0x1802be2a},
				{626, 380597, 0x1802bcd8},
				{627, 383464, 0x1802c342},
				{628, 383491, 0x1802c1a2},
				{629, 383788, 0x1802c0c6},
				{630, 384865, 0x1802c221},
				{631, 384907, 0x1802c08b},
				{632, 384970, 0x1802bf06},
				{633, 384987, 0x1802bd64},
				{634, 385254, 0x1802bc72},
				{635, 385907, 0x1802bc98},
				{636, 386117, 0x1802bb81},
				{637, 386299, 0x1802ba52},
				{638, 386731, 0x1802b9d9},
				{639, 386754, 0x1802b83c},
				{640, 387502, 0x1802b8a7},
				{641, 387599, 0x1802b73d},
				{642, 387922, 0x1802b676},
				{643, 388862, 0x1802b76a},
				{644, 389096, 0x1802b663},
				{645, 389179, 0x1802b4f4},
				{646, 390205, 0x1802b623},
				{647, 392842, 0x1802bbd9},
				{648, 393650, 0x1802bc70},
				{649, 395227, 0x1802bf2f},
				{650, 395437, 0x1802be17},
				{651, 396328, 0x1802bee9},
				{652, 396734, 0x1802be5d},
				{653, 397021, 0x1802bd7c},
				{654, 398290, 0x1802bf5c},
				{655, 398351, 0x1802bdda},
				{656, 398352, 0x1802bc2a},
				{657, 398608, 0x1802bb30},
				{658, 400692, 0x1802bf5f},
				{659, 401050, 0x1802beb0},
				{660, 401877, 0x1802bf54},
				{661, 402081, 0x1802be35},
				{662, 402895, 0x1802bed1},
				{663, 403141, 0x1802bdd2},
				{664, 403889, 0x1802be3a},
				{665, 405485, 0x1802c10c},
				{666, 406102, 0x1802c11a},
				{667, 406424, 0x1802c050},
				{668, 407129, 0x1802c09b},
				{669, 408261, 0x1802c21e},
				{670, 411009, 0x1802c83d},
				{671, 411058, 0x1802c6aa},
				{672, 411202, 0x1802c55b},
				{673, 411222, 0x1802c3b6},
				{674, 412528, 0x1802c5b9},
				{675, 413119, 0x1802c5b4},
				{676, 413150, 0x1802c414},
				{677, 413673, 0x1802c3db},
				{678, 414107, 0x1802c363},
				{679, 414707, 0x1802c363},
				{680, 416015, 0x1802c566},
				{681, 416297, 0x1802c47f},
				{682, 416872, 0x1802c46c},
				{683, 417179, 0x1802c396},
				{684, 417346, 0x1802c25c},
				{685, 417907, 0x1802c241},
				{686, 417967, 0x1802c0b6},
				{687, 418431, 0x1802c056},
				{688, 418720, 0x1802bf74},
				{689, 419267, 0x1802bf4f},
				{690, 419422, 0x1802be0d},
				{691, 419535, 0x1802bcab},
				{692, 419837, 0x1802bbd4},
				{693, 420828, 0x1802bcee},
				{694, 421425, 0x1802bcee},
				{695, 421770, 0x1802bc37},
				{696, 422491, 0x1802bc8a},
				{697, 422687, 0x1802bb69},
				{698, 423350, 0x1802bb96},
				{699, 423617, 0x1802baa7},
				{700, 425437, 0x1802be17},
				{701, 425498, 0x1802bc92},
				{702, 426799, 0x1802be8b},
				{703, 430429, 0x1802c726},
				{704, 430913, 0x1802c6d3},
				{705, 432168, 0x1802c8b3},
				{706, 433472, 0x1802cab6},
				{707, 433525, 0x1802c924},
				{708, 434589, 0x1802ca78},
				{709, 434648, 0x1802c8eb},
				{710, 434654, 0x1802c739},
				{711, 436085, 0x1802c99a},
				{712, 436419, 0x1802c8d6},
				{713, 436687, 0x1802c7e4},
				{714, 437526, 0x1802c893},
				{715, 437933, 0x1802c805},
				{716, 439897, 0x1802cbf0},
				{717, 441227, 0x1802ce09},
				{718, 441509, 0x1802cd1f},
				{719, 441573, 0x1802cb95},
				{720, 441768, 0x1802ca68},
				{721, 443123, 0x1802cc94},
				{722, 443312, 0x1802cb67},
				{723, 443356, 0x1802c9cf},
				{724, 445295, 0x1802cda8},
				{725, 446484, 0x1802cf5b},
				{726, 447388, 0x1802d03c},
				{727, 447650, 0x1802cf40},
				{728, 447685, 0x1802cda0},
				{729, 447685, 0x1802cbe5},
				{730, 447846, 0x1802caa3},
				{731, 447887, 0x1802c909},
				{732, 448873, 0x1802ca23},
				{733, 449601, 0x1802ca83},
				{734, 450072, 0x1802ca23},
				{735, 451346, 0x1802cc13},
				{736, 451515, 0x1802cad6},
				{737, 451545, 0x1802c934},
				{738, 451547, 0x1802c77e},
				{739, 451600, 0x1802c5ec},
				{740, 451710, 0x1802c48a},
				{741, 452201, 0x1802c439},
				{742, 452501, 0x1802c35d},
				{743, 454119, 0x1802c644},
				{744, 454121, 0x1802c48f},
				{745, 455736, 0x1802c776},
				{746, 455859, 0x1802c619},
				{747, 456066, 0x1802c4fa},
				{748, 456519, 0x1802c48f},
				{749, 456655, 0x1802c33d},
				{750, 456902, 0x1802c23b},
				{751, 457978, 0x1802c396},
				{752, 458088, 0x1802c233},
				{753, 458522, 0x1802c1bb},
				{754, 458649, 0x1802c060},
				{755, 458776, 0x1802bf0c},
				{756, 459470, 0x1802bf4f},
				{757, 459726, 0x1802be55},
				{758, 460982, 0x1802c030},
				{759, 462187, 0x1802c1e8},
				{760, 462493, 0x1802c114},
				{761, 463504, 0x1802c23e},
				{762, 463910, 0x1802c1b0},
				{763, 464257, 0x1802c0f9},
				{764, 464410, 0x1802bfb5},
				{765, 464528, 0x1802be58},
				{766, 464756, 0x1802bd4c},
				{767, 465137, 0x1802bcad},
				{768, 465417, 0x1802bbc9},
				{769, 465800, 0x1802bb2b},
				{770, 466521, 0x1802bb81},
				{771, 468519, 0x1802bf72},
				{772, 468840, 0x1802beab},
				{773, 468909, 0x1802bd29},
				{774, 469521, 0x1802bd33},
				{775, 469605, 0x1802bbbe},
				{776, 469736, 0x1802ba6c},
				{777, 469906, 0x1802b938},
				{778, 470457, 0x1802b915},
				{779, 470592, 0x1802b7c8},
				{780, 470786, 0x1802b6a4},
				{781, 470867, 0x1802b531},
				{782, 471904, 0x1802b66b},
				{783, 472233, 0x1802b5a7},
				{784, 472413, 0x1802b47e},
				{785, 473428, 0x1802b5a5},
				{786, 474802, 0x1802b7ce},
				{787, 475964, 0x1802b963},
				{788, 476987, 0x1802ba92},
				{789, 477048, 0x1802b910},
				{790, 477122, 0x1802b795},
				{791, 477238, 0x1802b63b},
				{792, 477281, 0x1802b4ae},
				{793, 477855, 0x1802b49b},
				{794, 478989, 0x1802b618},
				{795, 480590, 0x1802b8e5},
				{796, 480594, 0x1802b73a},
				{797, 480743, 0x1802b5f8},
				{798, 481364, 0x1802b605},
				{799, 481456, 0x1802b49b},
				{800, 481721, 0x1802b3ac},
				{801, 482486, 0x1802b422},
				{802, 482910, 0x1802b3a4},
				{803, 482941, 0x1802b20f},
				{804, 482966, 0x1802b077},
				{805, 483761, 0x1802b103},
				{806, 484289, 0x1802b0d0},
				{807, 484520, 0x1802afc9},
				{808, 484641, 0x1802ae78},
				{809, 485590, 0x1802af6e},
				{810, 485727, 0x1802ae29},
				{811, 486746, 0x1802af4d},
				{812, 487753, 0x1802b06f},
				{813, 488318, 0x1802b054},
				{814, 488451, 0x1802af0a},
				{815, 488545, 0x1802adaa},
				{816, 488830, 0x1802accb},
				{817, 489408, 0x1802acbc},
				{818, 489635, 0x1802abb8},
				{819, 489659, 0x1802aa24},
				{820, 490329, 0x1802aa56},
				{821, 492017, 0x1802ad4e},
				{822, 492262, 0x1802ac56},
				{823, 492747, 0x1802ac06},
				{824, 492753, 0x1802aa66},
				{825, 493678, 0x1802ab48},
				{826, 493893, 0x1802aa3c},
				{827, 494049, 0x1802a906},
				{828, 494357, 0x1802a83a},
				{829, 494704, 0x1802a789},
				{830, 494910, 0x1802a679},
				{831, 495801, 0x1802a742},
				{832, 496075, 0x1802a65f},
				{833, 497310, 0x1802a819},
				{834, 498058, 0x1802a880},
				{835, 500404, 0x1802ad45},
				{836, 500831, 0x1802accc},
				{837, 501001, 0x1802aba0},
				{838, 502507, 0x1802ae1a},
				{839, 502709, 0x1802ad02},
				{840, 503289, 0x1802acf6},
				{841, 503410, 0x1802aba5},
				{842, 503585, 0x1802aa7d},
				{843, 503977, 0x1802a9ea},
				{844, 505389, 0x1802ac23},
				{845, 505926, 0x1802abf7},
				{846, 506588, 0x1802ac23},
				{847, 506933, 0x1802ab6f},
				{848, 507014, 0x1802aa04},
				{849, 507175, 0x1802a8d3},
				{850, 507429, 0x1802a7e0},
				{851, 507459, 0x1802a653},
				{852, 507602, 0x1802a517},
				{853, 507816, 0x1802a40c},
				{854, 508462, 0x1802a42b},
				{855, 508502, 0x1802a2a8},
				{856, 509225, 0x1802a2fd},
				{857, 509685, 0x1802a29c},
				{858, 510040, 0x1802a1f2},
				{859, 510381, 0x1802a13f},
				{860, 510483, 0x18029fe8},
				{861, 510760, 0x18029f0a},
				{862, 510856, 0x18029daf},
				{863, 513127, 0x1802a22e},
				{864, 513536, 0x1802a1ab},
				{865, 513564, 0x1802a020},
				{866, 513968, 0x18029f9a},
				{867, 514016, 0x18029e1d},
				{868, 514416, 0x18029d94},
				{869, 514964, 0x18029d71},
				{870, 515313, 0x18029cc4},
				{871, 516450, 0x18029e35},
				{872, 516713, 0x18029d4d},
				{873, 516741, 0x18029bc5},
				{874, 516934, 0x18029ab0},
				{875, 517227, 0x180299de},
				{876, 517336, 0x1802988e},
				{877, 517910, 0x1802987c},
				{878, 517947, 0x180296fe},
				{879, 518259, 0x1802963a},
				{880, 518613, 0x18029592},
				{881, 518711, 0x1802943f},
				{882, 518715, 0x180292ac},
				{883, 519175, 0x1802924d},
				{884, 519234, 0x180290e0},
				{885, 520514, 0x180292ac},
				{886, 520866, 0x18029205},
				{887, 522826, 0x1802959c},
				{888, 523988, 0x18029719},
				{889, 524155, 0x180295f3},
				{890, 524618, 0x18029596},
				{891, 525169, 0x18029576},
				{892, 527140, 0x1802991a},
				{893, 527335, 0x18029806},
				{894, 527959, 0x18029816},
				{895, 528204, 0x18029725},
				{896, 528856, 0x18029748},
				{897, 529145, 0x18029674},
				{898, 529809, 0x180296a1},
				{899, 529832, 0x18029518},
				{900, 529900, 0x180293b0},
				{901, 530252, 0x18029309},
				{902, 530363, 0x180291bd},
				{903, 530792, 0x1802914b},
				{904, 531419, 0x1802915d},
				{905, 531553, 0x18029023},
				{906, 531867, 0x18028f63},
				{907, 532051, 0x18028e4b},
				{908, 533275, 0x18028fef},
				{909, 533769, 0x18028fa6},
				{910, 534468, 0x18028feb},
				{911, 534494, 0x18028e68},
				{912, 535858, 0x1802906a},
				{913, 536485, 0x1802907b},
				{914, 537936, 0x180292ba},
				{915, 538082, 0x18029188},
				{916, 538508, 0x18029113},
				{917, 538754, 0x18029023},
				{918, 539975, 0x180291c5},
				{919, 540307, 0x18029110},
				{920, 540541, 0x1802901c},
				{921, 541484, 0x18029103},
				{922, 541637, 0x18028fd5},
				{923, 541982, 0x18028f28},
				{924, 542150, 0x18028e08},
				{925, 542551, 0x18028d81},
				{926, 545123, 0x180292b2},
				{927, 545167, 0x1802913a},
				{928, 545556, 0x180290ac},
				{929, 545588, 0x18028f2d},
				{930, 545634, 0x18028dba},
				{931, 545831, 0x18028cac},
				{932, 546487, 0x18028cd2},
				{933, 546691, 0x18028bc8},
				{934, 546703, 0x18028a40},
				{935, 547152, 0x180289dc},
				{936, 547189, 0x18028864},
				{937, 547424, 0x18028771},
				{938, 547627, 0x18028669},
				{939, 547649, 0x180284ea},
				{940, 547720, 0x1802838c},
				{941, 551263, 0x18028b2f},
				{942, 551730, 0x18028ad7},
				{943, 551767, 0x1802895f},
				{944, 551984, 0x18028860},
				{945, 552234, 0x18028777},
				{946, 553696, 0x180289b5},
				{947, 553905, 0x180288b0},
				{948, 554812, 0x1802897c},
				{949, 554989, 0x18028864},
				{950, 555619, 0x18028877},
				{951, 556119, 0x18028834},
				{952, 557125, 0x18028943},
				{953, 557238, 0x180287ff},
				{954, 557459, 0x18028703},
				{955, 557592, 0x180285cd},
				{956, 558854, 0x18028785},
				{957, 559950, 0x180288ce},
				{958, 560504, 0x180288b0},
				{959, 561621, 0x18028a09},
				{960, 563620, 0x18028daf},
				{961, 563879, 0x18028ccc},
				{962, 563942, 0x18028b63},
				{963, 564154, 0x18028a60},
				{964, 565325, 0x18028bdf},
				{965, 565451, 0x18028aa2},
				{966, 565566, 0x1802895f},
				{967, 565635, 0x180287fe},
				{968, 566460, 0x18028893},
				{969, 567692, 0x18028a38},
				{970, 568536, 0x18028adb},
				{971, 568724, 0x180289c8},
				{972, 569029, 0x18028904},
				{973, 569348, 0x18028848},
				{974, 569672, 0x18028791},
				{975, 571368, 0x18028a6a},
				{976, 571742, 0x180289d5},
				{977, 572100, 0x18028933},
				{978, 572426, 0x1802887c},
				{979, 572436, 0x180286f3},
				{980, 573333, 0x180287ba},
				{981, 573363, 0x1802863f},
				{982, 573856, 0x180285f8},
				{983, 576984, 0x18028c8c},
				{984, 577936, 0x18028d77},
				{985, 578576, 0x18028d93},
				{986, 580110, 0x18029007},
				{987, 581045, 0x180290e8},
				{988, 582554, 0x1802934d},
				{989, 583197, 0x1802936b},
				{990, 583323, 0x1802922a},
				{991, 583647, 0x1802916e},
				{992, 583905, 0x1802908a},
				{993, 586689, 0x1802964e},
				{994, 586926, 0x18029557},
				{995, 587291, 0x180294b7},
				{996, 587688, 0x1802942f},
				{997, 587802, 0x180292e6},
				{998, 588403, 0x180292e6},
				{999, 588544, 0x180291b0},
				{1000, 589215, 0x180291e0},
				{1001, 589738, 0x180291ad},
			},
		},
		// Deterministically random solvetimes for up-ramping hashrate around a recent real life nBits
		{
			run:          7,
			anchorHeight: 1,
			anchorTime:   0,
			anchorBits:   0x1802aee8,
			startHeight:  2,
			startTime:    1200,
			blocks: []block{
				{2, 1200, 0x1802aee8},
				{3, 1310, 0x1802ad91},
				{4, 1327, 0x1802abf8},
				{5, 1739, 0x1802ab73},
				{6, 2219, 0x1802ab20},
				{7, 2604, 0x1802aa89},
				{8, 2746, 0x1802a94b},
				{9, 7099, 0x1802b39c},
				{10, 7099, 0x1802b1f2},
				{11, 7657, 0x1802b1d4},
				{12, 8141, 0x1802b181},
				{13, 8236, 0x1802b01c},
				{14, 8644, 0x1802af93},
				{15, 9230, 0x1802af8b},
				{16, 9691, 0x1802af28},
				{17, 9708, 0x1802ad8f},
				{18, 10327, 0x1802ad9c},
				{19, 10899, 0x1802ad88},
				{20, 11478, 0x1802ad7b},
				{21, 11548, 0x1802ac07},
				{22, 12701, 0x1802ad89},
				{23, 12849, 0x1802ac4d},
				{24, 13186, 0x1802ab95},
				{25, 13319, 0x1802aa4e},
				{26, 13494, 0x1802a925},
				{27, 13540, 0x1802a7a4},
				{28, 13726, 0x1802a684},
				{29, 13764, 0x1802a4fd},
				{30, 13787, 0x1802a36d},
				{31, 14159, 0x1802a2cf},
				{32, 14203, 0x1802a14e},
				{33, 14474, 0x1802a06d},
				{34, 14582, 0x18029f19},
				{35, 14741, 0x18029dea},
				{36, 14971, 0x18029ceb},
				{37, 15181, 0x18029be0},
				{38, 15402, 0x18029adf},
				{39, 15886, 0x18029a8e},
				{40, 16041, 0x1802995f},
				{41, 16219, 0x1802983e},
				{42, 16407, 0x18029727},
				{43, 16513, 0x180295d7},
				{44, 16633, 0x18029491},
				{45, 16698, 0x18029326},
				{46, 16755, 0x180291b8},
				{47, 16782, 0x18029036},
				{48, 16795, 0x18028eab},
				{49, 16869, 0x18028d4a},
				{50, 16870, 0x18028bbb},
				{51, 16876, 0x18028a2d},
				{52, 16888, 0x180288a6},
				{53, 16958, 0x18028746},
				{54, 17306, 0x1802869d},
				{55, 17410, 0x18028556},
				{56, 17474, 0x180283f4},
				{57, 17580, 0x180282ac},
				{58, 17640, 0x18028147},
				{59, 17652, 0x18027fc5},
				{60, 17694, 0x18027e58},
				{61, 17772, 0x18027d02},
				{62, 17817, 0x18027b98},
				{63, 17923, 0x18027a56},
				{64, 17965, 0x180278ea},
				{65, 17982, 0x18027770},
				{66, 18116, 0x18027642},
				{67, 18236, 0x1802750c},
				{68, 18236, 0x1802738a},
				{69, 18237, 0x18027209},
				{70, 18303, 0x180270b0},
				{71, 18332, 0x18026f43},
				{72, 18404, 0x18026df1},
				{73, 18404, 0x18026c72},
				{74, 18504, 0x18026b35},
				{75, 18587, 0x180269ec},
				{76, 18618, 0x18026883},
				{77, 18641, 0x18026715},
				{78, 18649, 0x180265a1},
				{79, 18650, 0x18026428},
				{80, 18658, 0x180262b3},
				{81, 18824, 0x180261a4},
				{82, 18830, 0x1802602f},
				{83, 18861, 0x18025ecd},
				{84, 18865, 0x18025d5b},
				{85, 18927, 0x18025c0d},
				{86, 18987, 0x18025abd},
				{87, 19001, 0x18025952},
				{88, 19059, 0x18025804},
				{89, 19308, 0x1802572c},
				{90, 19316, 0x180255bf},
				{91, 19333, 0x1802545a},
				{92, 19378, 0x18025306},
				{93, 19559, 0x18025206},
				{94, 19658, 0x180250d6},
				{95, 19663, 0x18024f6a},
				{96, 19746, 0x18024e30},
				{97, 19917, 0x18024d2c},
				{98, 20213, 0x18024c74},
				{99, 20250, 0x18024b1f},
				{100, 20363, 0x180249fb},
				{101, 20372, 0x18024899},
				{102, 20488, 0x18024775},
				{103, 20529, 0x18024626},
				{104, 20555, 0x180244cf},
				{105, 20606, 0x18024387},
				{106, 20643, 0x18024239},
				{107, 20749, 0x18024113},
				{108, 20772, 0x18023fbd},
				{109, 20782, 0x18023e62},
				{110, 20819, 0x18023d15},
				{111, 20845, 0x18023bc2},
				{112, 20935, 0x18023a98},
				{113, 20945, 0x1802393e},
				{114, 20963, 0x180237ea},
				{115, 20966, 0x1802368d},
				{116, 21080, 0x18023574},
				{117, 21122, 0x1802342f},
				{118, 21126, 0x180232d5},
				{119, 21145, 0x18023187},
				{120, 21158, 0x18023033},
				{121, 21170, 0x18022ee1},
				{122, 21173, 0x18022d8b},
				{123, 21186, 0x18022c3a},
				{124, 21221, 0x18022af8},
				{125, 21227, 0x180229a6},
				{126, 21521, 0x180228f8},
				{127, 21547, 0x180227b2},
				{128, 21656, 0x1802269c},
				{129, 21665, 0x1802254d},
				{130, 21686, 0x18022407},
				{131, 21708, 0x180222c2},
				{132, 21715, 0x18022174},
				{133, 21790, 0x1802204e},
				{134, 21885, 0x18021f33},
				{135, 21918, 0x18021df8},
				{136, 21985, 0x18021ccf},
				{137, 22011, 0x18021b90},
				{138, 22033, 0x18021a4f},
				{139, 22112, 0x1802192f},
				{140, 22130, 0x180217f0},
				{141, 22184, 0x180216c3},
				{142, 22222, 0x1802158f},
				{143, 22241, 0x1802144f},
				{144, 22262, 0x18021314},
				{145, 22263, 0x180211cd},
				{146, 22266, 0x18021088},
				{147, 22281, 0x18020f4a},
				{148, 22305, 0x18020e13},
				{149, 22321, 0x18020cd7},
				{150, 22346, 0x18020ba2},
				{151, 22358, 0x18020a65},
				{152, 22358, 0x18020924},
				{153, 22375, 0x180207ec},
				{154, 22397, 0x180206b7},
				{155, 22528, 0x180205be},
				{156, 22647, 0x180204bd},
				{157, 22647, 0x1802037f},
				{158, 22659, 0x18020248},
				{159, 22691, 0x1802011d},
				{160, 22709, 0x1801ffea},
				{161, 22749, 0x1801fec4},
				{162, 22760, 0x1801fd8e},
				{163, 22775, 0x1801fc5d},
				{164, 22790, 0x1801fb2c},
				{165, 22797, 0x1801f9f7},
				{166, 22800, 0x1801f8c0},
				{167, 22823, 0x1801f797},
				{168, 22862, 0x1801f674},
				{169, 22883, 0x1801f549},
				{170, 22891, 0x1801f41a},
				{171, 23016, 0x1801f326},
				{172, 23025, 0x1801f1f6},
				{173, 23070, 0x1801f0db},
				{174, 23103, 0x1801efb9},
				{175, 23116, 0x1801ee8e},
				{176, 23201, 0x1801ed8a},
				{177, 23269, 0x1801ec7c},
				{178, 23307, 0x1801eb60},
				{179, 23395, 0x1801ea5e},
				{180, 23397, 0x1801e930},
				{181, 23457, 0x1801e821},
				{182, 23488, 0x1801e705},
				{183, 23534, 0x1801e5ef},
				{184, 23535, 0x1801e4c5},
				{185, 23550, 0x1801e3a2},
				{186, 23686, 0x1801e2bc},
				{187, 23700, 0x1801e19a},
				{188, 23872, 0x1801e0c7},
				{189, 23879, 0x1801dfa3},
				{190, 23888, 0x1801de80},
				{191, 23908, 0x1801dd63},
				{192, 23971, 0x1801dc5c},
				{193, 24055, 0x1801db5f},
				{194, 24085, 0x1801da49},
				{195, 24089, 0x1801d927},
				{196, 24103, 0x1801d80c},
				{197, 24190, 0x1801d712},
				{198, 24230, 0x1801d605},
				{199, 24254, 0x1801d4ee},
				{200, 24330, 0x1801d3f3},
				{201, 24359, 0x1801d2e0},
				{202, 24391, 0x1801d1d1},
				{203, 24406, 0x1801d0ba},
				{204, 24431, 0x1801cfa7},
				{205, 24437, 0x1801ce8d},
				{206, 24464, 0x1801cd7d},
				{207, 24464, 0x1801cc62},
				{208, 24493, 0x1801cb54},
				{209, 24531, 0x1801ca4a},
				{210, 24563, 0x1801c941},
				{211, 24571, 0x1801c82b},
				{212, 24602, 0x1801c720},
				{213, 24633, 0x1801c616},
				{214, 24714, 0x1801c526},
				{215, 24748, 0x1801c41e},
				{216, 24749, 0x1801c309},
				{217, 24807, 0x1801c20d},
				{218, 24815, 0x1801c0fd},
				{219, 24817, 0x1801bfea},
				{220, 24875, 0x1801bef0},
				{221, 24939, 0x1801bdfb},
				{222, 24942, 0x1801bcea},
				{223, 24953, 0x1801bbdd},
				{224, 24968, 0x1801bad3},
				{225, 25071, 0x1801b9f2},
				{226, 25127, 0x1801b8fc},
				{227, 25174, 0x1801b801},
				{228, 25182, 0x1801b6f6},
				{229, 25184, 0x1801b5ea},
				{230, 25193, 0x1801b4e0},
				{231, 25197, 0x1801b3d5},
				{232, 25211, 0x1801b2d0},
				{233, 25227, 0x1801b1cc},
				{234, 25231, 0x1801b0c3},
				{235, 25274, 0x1801afcb},
				{236, 25289, 0x1801aec8},
				{237, 25296, 0x1801adc4},
				{238, 25334, 0x1801accc},
				{239, 25362, 0x1801abcf},
				{240, 25374, 0x1801aace},
				{241, 25464, 0x1801a9ef},
				{242, 25531, 0x1801a907},
				{243, 25532, 0x1801a803},
				{244, 25609, 0x1801a71f},
				{245, 25609, 0x1801a61c},
				{246, 25672, 0x1801a532},
				{247, 25737, 0x1801a44c},
				{248, 25760, 0x1801a353},
				{249, 25765, 0x1801a254},
				{250, 25770, 0x1801a154},
				{251, 25790, 0x1801a05c},
				{252, 25796, 0x18019f60},
				{253, 25885, 0x18019e86},
				{254, 25925, 0x18019d99},
				{255, 25927, 0x18019c9b},
				{256, 25930, 0x18019b9e},
				{257, 25930, 0x18019aa2},
				{258, 25963, 0x180199b3},
				{259, 25966, 0x180198b8},
				{260, 26000, 0x180197cc},
				{261, 26026, 0x180196dc},
				{262, 26048, 0x180195ea},
				{263, 26052, 0x180194f3},
				{264, 26118, 0x18019415},
				{265, 26119, 0x1801931e},
				{266, 26122, 0x18019227},
				{267, 26129, 0x18019133},
				{268, 26148, 0x18019043},
				{269, 26150, 0x18018f4e},
				{270, 26153, 0x18018e5a},
				{271, 26175, 0x18018d6e},
				{272, 26186, 0x18018c7f},
				{273, 26243, 0x18018ba2},
				{274, 26244, 0x18018aaf},
				{275, 26247, 0x180189be},
				{276, 26252, 0x180188cd},
				{277, 26252, 0x180187dc},
				{278, 26264, 0x180186f0},
				{279, 26290, 0x1801860a},
				{280, 26316, 0x18018525},
				{281, 26351, 0x18018444},
				{282, 26395, 0x18018366},
				{283, 26402, 0x1801827a},
				{284, 26420, 0x18018195},
				{285, 26508, 0x180180ca},
				{286, 26522, 0x18017fe2},
				{287, 26552, 0x18017f02},
				{288, 26572, 0x18017e1f},
				{289, 26593, 0x18017d3d},
				{290, 26618, 0x18017c5c},
				{291, 26633, 0x18017b77},
				{292, 26675, 0x18017a9e},
				{293, 26711, 0x180179c3},
				{294, 26719, 0x180178df},
				{295, 26719, 0x180177f6},
				{296, 26738, 0x18017717},
				{297, 26769, 0x1801763c},
				{298, 26831, 0x1801756e},
				{299, 26846, 0x1801748d},
				{300, 26847, 0x180173a8},
				{301, 26908, 0x180172db},
				{302, 26909, 0x180171f7},
				{303, 26964, 0x18017128},
				{304, 26997, 0x18017050},
				{305, 27034, 0x18016f7c},
				{306, 27043, 0x18016e9d},
				{307, 27058, 0x18016dc1},
				{308, 27064, 0x18016ce3},
				{309, 27084, 0x18016c09},
				{310, 27088, 0x18016b2b},
				{311, 27144, 0x18016a60},
				{312, 27172, 0x1801698b},
				{313, 27188, 0x180168b3},
				{314, 27241, 0x180167e8},
				{315, 27250, 0x1801670d},
				{316, 27250, 0x18016630},
				{317, 27286, 0x18016562},
				{318, 27287, 0x18016485},
				{319, 27295, 0x180163ac},
				{320, 27304, 0x180162d4},
				{321, 27315, 0x180161fd},
				{322, 27321, 0x18016125},
				{323, 27340, 0x18016053},
				{324, 27361, 0x18015f80},
				{325, 27372, 0x18015eac},
				{326, 27406, 0x18015de0},
				{327, 27428, 0x18015d10},
				{328, 27435, 0x18015c3b},
				{329, 27455, 0x18015b6b},
				{330, 27465, 0x18015a9a},
				{331, 27474, 0x180159c6},
				{332, 27490, 0x180158f7},
				{333, 27493, 0x18015822},
				{334, 27501, 0x18015751},
				{335, 27504, 0x18015680},
				{336, 27522, 0x180155b4},
				{337, 27548, 0x180154eb},
				{338, 27559, 0x1801541e},
				{339, 27561, 0x1801534e},
				{340, 27570, 0x18015280},
				{341, 27582, 0x180151b4},
				{342, 27608, 0x180150ee},
				{343, 27613, 0x18015021},
				{344, 27633, 0x18014f59},
				{345, 27637, 0x18014e8c},
				{346, 27645, 0x18014dc2},
				{347, 27647, 0x18014cf5},
				{348, 27648, 0x18014c29},
				{349, 27670, 0x18014b64},
				{350, 27705, 0x18014aa4},
				{351, 27716, 0x180149dd},
				{352, 27745, 0x1801491c},
				{353, 27753, 0x18014854},
				{354, 27769, 0x18014790},
				{355, 27779, 0x180146cb},
				{356, 27789, 0x18014605},
				{357, 27792, 0x1801453d},
				{358, 27795, 0x18014476},
				{359, 27848, 0x180143c0},
				{360, 27866, 0x180142ff},
				{361, 27867, 0x18014239},
				{362, 27889, 0x1801417a},
				{363, 27894, 0x180140b6},
				{364, 27909, 0x18013ff6},
				{365, 27914, 0x18013f32},
				{366, 27942, 0x18013e77},
				{367, 27945, 0x18013db4},
				{368, 27966, 0x18013cf7},
				{369, 27977, 0x18013c38},
				{370, 27978, 0x18013b76},
				{371, 27983, 0x18013ab5},
				{372, 27983, 0x180139f4},
				{373, 27993, 0x18013936},
				{374, 28003, 0x18013878},
				{375, 28017, 0x180137bd},
				{376, 28020, 0x180136fd},
				{377, 28022, 0x1801363f},
				{378, 28055, 0x1801358b},
				{379, 28060, 0x180134cd},
				{380, 28084, 0x18013418},
				{381, 28085, 0x18013359},
				{382, 28100, 0x180132a1},
				{383, 28113, 0x180131e9},
				{384, 28133, 0x18013132},
				{385, 28181, 0x18013086},
				{386, 28181, 0x18012fca},
				{387, 28186, 0x18012f10},
				{388, 28251, 0x18012e6a},
				{389, 28264, 0x18012db4},
				{390, 28271, 0x18012cfc},
				{391, 28276, 0x18012c45},
				{392, 28287, 0x18012b8f},
				{393, 28290, 0x18012ad7},
				{394, 28299, 0x18012a22},
				{395, 28315, 0x1801296f},
				{396, 28328, 0x180128bc},
				{397, 28342, 0x1801280a},
				{398, 28343, 0x18012754},
				{399, 28360, 0x180126a3},
				{400, 28361, 0x180125ed},
				{401, 28367, 0x1801253b},
				{402, 28376, 0x18012489},
				{403, 28377, 0x180123d5},
				{404, 28390, 0x18012325},
				{405, 28396, 0x18012274},
				{406, 28401, 0x180121c2},
				{407, 28406, 0x18012112},
				{408, 28432, 0x18012067},
				{409, 28453, 0x18011fbb},
				{410, 28454, 0x18011f0a},
				{411, 28483, 0x18011e63},
				{412, 28484, 0x18011db2},
				{413, 28502, 0x18011d07},
				{414, 28503, 0x18011c58},
				{415, 28509, 0x18011baa},
				{416, 28534, 0x18011b03},
				{417, 28539, 0x18011a56},
				{418, 28539, 0x180119a9},
				{419, 28570, 0x18011904},
				{420, 28580, 0x18011859},
				{421, 28589, 0x180117af},
				{422, 28599, 0x18011706},
				{423, 28602, 0x1801165b},
				{424, 28609, 0x180115b1},
				{425, 28626, 0x1801150b},
				{426, 28667, 0x1801146c},
				{427, 28683, 0x180113c7},
				{428, 28711, 0x18011325},
				{429, 28716, 0x1801127d},
				{430, 28778, 0x180111e5},
				{431, 28779, 0x1801113d},
				{432, 28791, 0x18011098},
				{433, 28791, 0x18010fef},
				{434, 28798, 0x18010f49},
				{435, 28805, 0x18010ea4},
				{436, 28811, 0x18010dff},
				{437, 28857, 0x18010d66},
				{438, 28872, 0x18010cc4},
				{439, 28876, 0x18010c1f},
				{440, 28881, 0x18010b7b},
				{441, 28890, 0x18010ad9},
				{442, 28911, 0x18010a3b},
				{443, 28912, 0x18010997},
				{444, 28920, 0x180108f5},
				{445, 28923, 0x18010854},
				{446, 28939, 0x180107b5},
				{447, 28958, 0x18010718},
				{448, 28971, 0x18010679},
				{449, 28976, 0x180105d8},
				{450, 28978, 0x18010538},
				{451, 28985, 0x18010499},
				{452, 28997, 0x180103fc},
				{453, 29002, 0x1801035d},
				{454, 29003, 0x180102bd},
				{455, 29013, 0x18010221},
				{456, 29015, 0x18010182},
				{457, 29025, 0x180100e6},
				{458, 29033, 0x1801004a},
				{459, 29064, 0x1800ffb4},
				{460, 29068, 0x1800ff18},
				{461, 29070, 0x1800fe7b},
				{462, 29074, 0x1800fddf},
				{463, 29102, 0x1800fd4a},
				{464, 29112, 0x1800fcb1},
				{465, 29120, 0x1800fc17},
				{466, 29131, 0x1800fb80},
				{467, 29135, 0x1800fae5},
				{468, 29144, 0x1800fa4e},
				{469, 29171, 0x1800f9ba},
				{470, 29202, 0x1800f929},
				{471, 29215, 0x1800f892},
				{472, 29228, 0x1800f7fc},
				{473, 29235, 0x1800f765},
				{474, 29237, 0x1800f6ce},
				{475, 29239, 0x1800f637},
				{476, 29245, 0x1800f5a0},
				{477, 29260, 0x1800f50d},
				{478, 29261, 0x1800f476},
				{479, 29275, 0x1800f3e3},
				{480, 29290, 0x1800f351},
				{481, 29319, 0x1800f2c2},
				{482, 29353, 0x1800f235},
				{483, 29354, 0x1800f1a1},
				{484, 29354, 0x1800f10c},
				{485, 29380, 0x1800f07e},
				{486, 29384, 0x1800efeb},
				{487, 29384, 0x1800ef57},
				{488, 29420, 0x1800eecc},
				{489, 29425, 0x1800ee3b},
				{490, 29433, 0x1800edaa},
				{491, 29447, 0x1800ed1b},
				{492, 29469, 0x1800ec8f},
				{493, 29472, 0x1800ebfe},
				{494, 29495, 0x1800eb72},
				{495, 29497, 0x1800eae2},
				{496, 29498, 0x1800ea51},
				{497, 29517, 0x1800e9c6},
				{498, 29534, 0x1800e93a},
				{499, 29534, 0x1800e8aa},
				{500, 29547, 0x1800e81f},
				{501, 29563, 0x1800e793},
				{502, 29608, 0x1800e710},
				{503, 29608, 0x1800e682},
				{504, 29617, 0x1800e5f6},
				{505, 29633, 0x1800e56d},
				{506, 29637, 0x1800e4e0},
				{507, 29643, 0x1800e454},
				{508, 29679, 0x1800e3d1},
				{509, 29686, 0x1800e346},
				{510, 29688, 0x1800e2bb},
				{511, 29704, 0x1800e233},
				{512, 29707, 0x1800e1a9},
				{513, 29715, 0x1800e120},
				{514, 29732, 0x1800e099},
				{515, 29743, 0x1800e012},
				{516, 29754, 0x1800df8a},
				{517, 29760, 0x1800df02},
				{518, 29760, 0x1800de79},
				{519, 29791, 0x1800ddf8},
				{520, 29795, 0x1800dd70},
				{521, 29804, 0x1800dce9},
				{522, 29808, 0x1800dc62},
				{523, 29813, 0x1800dbdc},
				{524, 29866, 0x1800db61},
				{525, 29888, 0x1800dadf},
				{526, 29892, 0x1800da59},
				{527, 29904, 0x1800d9d6},
				{528, 29904, 0x1800d950},
				{529, 29904, 0x1800d8ca},
				{530, 29919, 0x1800d849},
				{531, 29928, 0x1800d7c5},
				{532, 29929, 0x1800d741},
				{533, 29933, 0x1800d6bd},
				{534, 29936, 0x1800d63a},
				{535, 29976, 0x1800d5bf},
				{536, 29983, 0x1800d53d},
				{537, 30004, 0x1800d4be},
				{538, 30004, 0x1800d43c},
				{539, 30013, 0x1800d3bb},
				{540, 30017, 0x1800d33b},
				{541, 30033, 0x1800d2bc},
				{542, 30042, 0x1800d23c},
				{543, 30066, 0x1800d1c0},
				{544, 30084, 0x1800d143},
				{545, 30089, 0x1800d0c3},
				{546, 30098, 0x1800d045},
				{547, 30118, 0x1800cfc9},
				{548, 30120, 0x1800cf4a},
				{549, 30136, 0x1800cece},
				{550, 30138, 0x1800ce4f},
				{551, 30164, 0x1800cdd6},
				{552, 30240, 0x1800cd68},
				{553, 30244, 0x1800cceb},
				{554, 30247, 0x1800cc6c},
				{555, 30288, 0x1800cbf8},
				{556, 30296, 0x1800cb7c},
				{557, 30297, 0x1800caff},
				{558, 30337, 0x1800ca8b},
				{559, 30369, 0x1800ca15},
				{560, 30387, 0x1800c99d},
				{561, 30388, 0x1800c921},
				{562, 30395, 0x1800c8a7},
				{563, 30398, 0x1800c82c},
				{564, 30405, 0x1800c7b2},
				{565, 30427, 0x1800c73c},
				{566, 30429, 0x1800c6c2},
				{567, 30437, 0x1800c649},
				{568, 30481, 0x1800c5d8},
				{569, 30492, 0x1800c561},
				{570, 30508, 0x1800c4eb},
				{571, 30510, 0x1800c472},
				{572, 30526, 0x1800c3fd},
				{573, 30530, 0x1800c385},
				{574, 30532, 0x1800c30d},
				{575, 30589, 0x1800c2a0},
				{576, 30594, 0x1800c22a},
				{577, 30608, 0x1800c1b6},
				{578, 30615, 0x1800c140},
				{579, 30620, 0x1800c0ca},
				{580, 30628, 0x1800c055},
				{581, 30629, 0x1800bfdf},
				{582, 30632, 0x1800bf6a},
				{583, 30632, 0x1800bef4},
				{584, 30648, 0x1800be82},
				{585, 30662, 0x1800be0f},
				{586, 30663, 0x1800bd9a},
				{587, 30669, 0x1800bd27},
				{588, 30674, 0x1800bcb4},
				{589, 30674, 0x1800bc40},
				{590, 30696, 0x1800bbd0},
				{591, 30699, 0x1800bb5d},
				{592, 30708, 0x1800baec},
				{593, 30715, 0x1800ba79},
				{594, 30721, 0x1800ba08},
				{595, 30750, 0x1800b99b},
				{596, 30750, 0x1800b929},
				{597, 30752, 0x1800b8b7},
				{598, 30758, 0x1800b847},
				{599, 30758, 0x1800b7d5},
				{600, 30765, 0x1800b765},
				{601, 30769, 0x1800b6f5},
				{602, 30771, 0x1800b685},
				{603, 30789, 0x1800b619},
				{604, 30790, 0x1800b5a8},
				{605, 30799, 0x1800b53a},
				{606, 30801, 0x1800b4ca},
				{607, 30801, 0x1800b45c},
				{608, 30804, 0x1800b3ec},
				{609, 30805, 0x1800b37e},
				{610, 30815, 0x1800b311},
				{611, 30818, 0x1800b2a4},
				{612, 30829, 0x1800b238},
				{613, 30830, 0x1800b1ca},
				{614, 30834, 0x1800b15d},
				{615, 30835, 0x1800b0f0},
				{616, 30857, 0x1800b087},
				{617, 30862, 0x1800b01b},
				{618, 30871, 0x1800afb0},
				{619, 30873, 0x1800af44},
				{620, 30896, 0x1800aedc},
				{621, 30910, 0x1800ae72},
				{622, 30911, 0x1800ae07},
				{623, 30920, 0x1800ad9d},
				{624, 30924, 0x1800ad33},
				{625, 30927, 0x1800acc9},
				{626, 30929, 0x1800ac5f},
				{627, 30974, 0x1800abfc},
				{628, 30974, 0x1800ab92},
				{629, 30978, 0x1800ab2a},
				{630, 30995, 0x1800aac4},
				{631, 30995, 0x1800aa5a},
				{632, 30995, 0x1800a9f2},
				{633, 30995, 0x1800a98a},
				{634, 30999, 0x1800a922},
				{635, 31009, 0x1800a8bc},
				{636, 31012, 0x1800a855},
				{637, 31014, 0x1800a7ee},
				{638, 31020, 0x1800a788},
				{639, 31020, 0x1800a721},
				{640, 31031, 0x1800a6bc},
				{641, 31032, 0x1800a656},
				{642, 31036, 0x1800a5f0},
				{643, 31050, 0x1800a58c},
				{644, 31053, 0x1800a527},
				{645, 31054, 0x1800a4c2},
				{646, 31069, 0x1800a45f},
				{647, 31109, 0x1800a401},
				{648, 31121, 0x1800a39e},
				{649, 31145, 0x1800a33d},
				{650, 31148, 0x1800a2da},
				{651, 31161, 0x1800a278},
				{652, 31167, 0x1800a215},
				{653, 31171, 0x1800a1b2},
				{654, 31190, 0x1800a151},
				{655, 31190, 0x1800a0ee},
				{656, 31190, 0x1800a08b},
				{657, 31193, 0x1800a029},
				{658, 31224, 0x18009fcc},
				{659, 31229, 0x18009f6a},
				{660, 31241, 0x18009f0a},
				{661, 31244, 0x18009ea9},
				{662, 31256, 0x18009e49},
				{663, 31259, 0x18009de8},
				{664, 31270, 0x18009d89},
				{665, 31293, 0x18009d2c},
				{666, 31302, 0x18009ccc},
				{667, 31306, 0x18009c6c},
				{668, 31316, 0x18009c0e},
				{669, 31332, 0x18009bb0},
				{670, 31373, 0x18009b57},
				{671, 31373, 0x18009af7},
				{672, 31375, 0x18009a98},
				{673, 31375, 0x18009a39},
				{674, 31394, 0x180099de},
				{675, 31402, 0x18009980},
				{676, 31402, 0x18009921},
				{677, 31409, 0x180098c5},
				{678, 31415, 0x18009867},
				{679, 31423, 0x1800980b},
				{680, 31442, 0x180097b0},
				{681, 31446, 0x18009753},
				{682, 31454, 0x180096f7},
				{683, 31458, 0x1800969b},
				{684, 31460, 0x1800963f},
				{685, 31468, 0x180095e3},
				{686, 31468, 0x18009587},
				{687, 31474, 0x1800952c},
				{688, 31478, 0x180094d1},
				{689, 31485, 0x18009476},
				{690, 31487, 0x1800941b},
				{691, 31488, 0x180093c0},
				{692, 31492, 0x18009365},
				{693, 31506, 0x1800930d},
				{694, 31514, 0x180092b3},
				{695, 31518, 0x18009259},
				{696, 31528, 0x18009201},
				{697, 31530, 0x180091a7},
				{698, 31539, 0x1800914f},
				{699, 31542, 0x180090f6},
				{700, 31568, 0x180090a0},
				{701, 31568, 0x18009047},
				{702, 31586, 0x18008ff1},
				{703, 31637, 0x18008fa0},
				{704, 31643, 0x18008f48},
				{705, 31660, 0x18008ef2},
				{706, 31678, 0x18008e9d},
				{707, 31678, 0x18008e45},
				{708, 31692, 0x18008df0},
				{709, 31692, 0x18008d98},
				{710, 31692, 0x18008d41},
				{711, 31712, 0x18008ced},
				{712, 31716, 0x18008c97},
				{713, 31719, 0x18008c40},
				{714, 31730, 0x18008bec},
				{715, 31735, 0x18008b96},
				{716, 31762, 0x18008b44},
				{717, 31780, 0x18008af1},
				{718, 31783, 0x18008a9c},
				{719, 31783, 0x18008a46},
				{720, 31785, 0x180089f1},
				{721, 31803, 0x1800899f},
				{722, 31805, 0x1800894a},
				{723, 31805, 0x180088f6},
				{724, 31831, 0x180088a5},
				{725, 31847, 0x18008853},
				{726, 31859, 0x18008801},
				{727, 31862, 0x180087ae},
				{728, 31862, 0x1800875a},
				{729, 31862, 0x18008707},
				{730, 31864, 0x180086b4},
				{731, 31864, 0x18008660},
				{732, 31877, 0x1800860f},
				{733, 31886, 0x180085be},
				{734, 31892, 0x1800856d},
				{735, 31909, 0x1800851d},
				{736, 31911, 0x180084cb},
				{737, 31911, 0x18008479},
				{738, 31911, 0x18008428},
				{739, 31911, 0x180083d6},
				{740, 31912, 0x18008385},
				{741, 31918, 0x18008335},
				{742, 31922, 0x180082e5},
				{743, 31943, 0x18008297},
				{744, 31943, 0x18008247},
				{745, 31964, 0x180081f9},
				{746, 31965, 0x180081a9},
				{747, 31967, 0x1800815a},
				{748, 31973, 0x1800810b},
				{749, 31974, 0x180080bb},
				{750, 31977, 0x1800806c},
				{751, 31991, 0x1800801f},
				{752, 31992, 0x177fd0f4},
				{753, 31997, 0x177f82cd},
				{754, 31998, 0x177f3451},
				{755, 31999, 0x177ee62b},
				{756, 32008, 0x177e995d},
				{757, 32011, 0x177e4b8c},
				{758, 32027, 0x177e0015},
				{759, 32042, 0x177db448},
				{760, 32046, 0x177d6779},
				{761, 32059, 0x177d1bac},
				{762, 32064, 0x177ccf89},
				{763, 32068, 0x177c8366},
				{764, 32069, 0x177c36ed},
				{765, 32070, 0x177bea1f},
				{766, 32072, 0x177b9e52},
				{767, 32076, 0x177b52da},
				{768, 32079, 0x177b070d},
				{769, 32083, 0x177abbec},
				{770, 32092, 0x177a7176},
				{771, 32117, 0x177a295a},
				{772, 32121, 0x1779de8e},
				{773, 32121, 0x1779936d},
				{774, 32128, 0x1779494d},
				{775, 32129, 0x1778fed7},
				{776, 32130, 0x1778b4b8},
				{777, 32132, 0x17786a98},
				{778, 32139, 0x17782124},
				{779, 32140, 0x1777d75a},
				{780, 32142, 0x17778de6},
				{781, 32143, 0x17774472},
				{782, 32156, 0x1776fd01},
				{783, 32160, 0x1776b3e3},
				{784, 32162, 0x17766b70},
				{785, 32174, 0x177623ff},
				{786, 32191, 0x1775dd3a},
				{787, 32205, 0x1775961f},
				{788, 32217, 0x17754f5a},
				{789, 32217, 0x1775073d},
				{790, 32217, 0x1774bf21},
				{791, 32218, 0x1774775a},
				{792, 32218, 0x17742f93},
				{793, 32225, 0x1773e97a},
				{794, 32239, 0x1773a360},
				{795, 32259, 0x17735e9e},
				{796, 32259, 0x177317d9},
				{797, 32260, 0x1772d114},
				{798, 32267, 0x17728b51},
				{799, 32268, 0x17724537},
				{800, 32271, 0x1771ff1e},
				{801, 32280, 0x1771ba06},
				{802, 32285, 0x17717498},
				{803, 32285, 0x17712ed5},
				{804, 32285, 0x1770e911},
				{805, 32294, 0x1770a4a5},
				{806, 32300, 0x17706039},
				{807, 32302, 0x17701b77},
				{808, 32303, 0x176fd6b5},
				{809, 32314, 0x176f92f5},
				{810, 32315, 0x176f4e88},
				{811, 32327, 0x176f0b74},
				{812, 32339, 0x176ec85f},
				{813, 32345, 0x176e84f5},
				{814, 32346, 0x176e4134},
				{815, 32347, 0x176dfd74},
				{816, 32350, 0x176dba09},
				{817, 32357, 0x176d774b},
				{818, 32359, 0x176d348c},
				{819, 32359, 0x176cf177},
				{820, 32367, 0x176caf0e},
				{821, 32387, 0x176c6ea8},
				{822, 32389, 0x176c2c3f},
				{823, 32394, 0x176bea2c},
				{824, 32394, 0x176ba7c3},
				{825, 32405, 0x176b6708},
				{826, 32407, 0x176b24f5},
				{827, 32408, 0x176ae338},
				{828, 32411, 0x176aa226},
				{829, 32415, 0x176a6115},
				{830, 32417, 0x176a1fad},
				{831, 32427, 0x1769df9d},
				{832, 32430, 0x17699ee2},
				{833, 32444, 0x17695f28},
				{834, 32452, 0x17691f6e},
				{835, 32479, 0x1768e1b7},
				{836, 32484, 0x1768a1fd},
				{837, 32486, 0x17686197},
				{838, 32503, 0x1768238a},
				{839, 32505, 0x1767e37b},
				{840, 32511, 0x1767a46c},
				{841, 32512, 0x176764b2},
				{842, 32514, 0x1767254e},
				{843, 32518, 0x1766e696},
				{844, 32534, 0x1766a8df},
				{845, 32540, 0x17666a7c},
				{846, 32547, 0x17662c1a},
				{847, 32551, 0x1765edb7},
				{848, 32551, 0x1765af54},
				{849, 32552, 0x1765709c},
				{850, 32554, 0x176532e5},
				{851, 32554, 0x1764f483},
				{852, 32555, 0x1764b676},
				{853, 32557, 0x176478bf},
				{854, 32564, 0x17643bb4},
				{855, 32564, 0x1763fe53},
				{856, 32572, 0x1763c19e},
				{857, 32577, 0x176384e8},
				{858, 32581, 0x176347dd},
				{859, 32584, 0x17630b28},
				{860, 32585, 0x1762ce73},
				{861, 32588, 0x176291be},
				{862, 32589, 0x1762555e},
				{863, 32615, 0x17621bae},
				{864, 32619, 0x1761dffa},
				{865, 32619, 0x1761a39b},
				{866, 32623, 0x176167e7},
				{867, 32623, 0x17612bdd},
				{868, 32627, 0x1760f080},
				{869, 32633, 0x1760b5ce},
				{870, 32637, 0x17607ac6},
				{871, 32650, 0x17604069},
				{872, 32652, 0x176005b7},
				{873, 32652, 0x175fcaaf},
				{874, 32654, 0x175f8fa8},
				{875, 32657, 0x175f554b},
				{876, 32658, 0x175f1aef},
				{877, 32664, 0x175ee0e9},
				{878, 32664, 0x175ea68d},
				{879, 32667, 0x175e6c86},
				{880, 32671, 0x175e32d6},
				{881, 32672, 0x175df925},
				{882, 32672, 0x175dbf75},
				{883, 32677, 0x175d861a},
				{884, 32677, 0x175d4c6a},
				{885, 32691, 0x175d1466},
				{886, 32694, 0x175cdbb8},
				{887, 32716, 0x175ca4b6},
				{888, 32729, 0x175c6cb3},
				{889, 32730, 0x175c3404},
				{890, 32735, 0x175bfbab},
				{891, 32741, 0x175bc3a8},
				{892, 32762, 0x175b8cfc},
				{893, 32764, 0x175b54f9},
				{894, 32770, 0x175b1d4c},
				{895, 32772, 0x175ae548},
				{896, 32779, 0x175aadf1},
				{897, 32782, 0x175a7699},
				{898, 32789, 0x175a3f42},
				{899, 32789, 0x175a07eb},
				{900, 32789, 0x1759d03d},
				{901, 32792, 0x1759993c},
				{902, 32793, 0x1759623a},
				{903, 32797, 0x17592b39},
				{904, 32803, 0x1758f4e3},
				{905, 32804, 0x1758be37},
				{906, 32807, 0x175887e1},
				{907, 32809, 0x1758518b},
				{908, 32822, 0x17581c37},
				{909, 32827, 0x1757e637},
				{910, 32834, 0x1757b0e3},
				{911, 32834, 0x17577a8d},
				{912, 32848, 0x1757463a},
				{913, 32854, 0x17571090},
				{914, 32869, 0x1756dc93},
				{915, 32870, 0x1756a693},
				{916, 32874, 0x17567195},
				{917, 32876, 0x17563c97},
				{918, 32889, 0x17560844},
				{919, 32892, 0x1755d3c6},
				{920, 32894, 0x17559f49},
				{921, 32904, 0x17556b77},
				{922, 32905, 0x17553724},
				{923, 32908, 0x175502fc},
				{924, 32909, 0x1754ced5},
				{925, 32913, 0x17549b2e},
				{926, 32940, 0x1754698a},
				{927, 32940, 0x175435b8},
				{928, 32944, 0x1754023c},
				{929, 32944, 0x1753cec0},
				{930, 32944, 0x17539b44},
				{931, 32946, 0x175367f3},
				{932, 32952, 0x1753354e},
				{933, 32954, 0x17530252},
				{934, 32954, 0x1752cf57},
				{935, 32958, 0x17529cdd},
				{936, 32958, 0x17526a0c},
				{937, 32960, 0x17523767},
				{938, 32962, 0x17520542},
				{939, 32962, 0x1751d2c8},
				{940, 32962, 0x1751a04e},
				{941, 32999, 0x17517159},
				{942, 33003, 0x17513fb5},
				{943, 33003, 0x17510d90},
				{944, 33005, 0x1750dc18},
				{945, 33007, 0x1750aa74},
				{946, 33022, 0x17507a28},
				{947, 33024, 0x175048af},
				{948, 33033, 0x1750180d},
				{949, 33034, 0x174fe6ea},
				{950, 33040, 0x174fb61c},
				{951, 33045, 0x174f85a5},
				{952, 33055, 0x174f5584},
				{953, 33056, 0x174f24e2},
				{954, 33058, 0x174ef43f},
				{955, 33059, 0x174ec3c8},
				{956, 33072, 0x174e9453},
				{957, 33083, 0x174e64dd},
				{958, 33088, 0x174e3511},
				{959, 33099, 0x174e05c7},
				{960, 33119, 0x174dd77e},
				{961, 33121, 0x174da7b2},
				{962, 33121, 0x174d77e7},
				{963, 33123, 0x174d4871},
				{964, 33135, 0x174d19d2},
				{965, 33136, 0x174cea5c},
				{966, 33137, 0x174cbb11},
				{967, 33137, 0x174c8bf2},
				{968, 33145, 0x174c5d53},
				{969, 33157, 0x174c2f35},
				{970, 33165, 0x174c0116},
				{971, 33166, 0x174bd24c},
				{972, 33169, 0x174ba3d8},
				{973, 33172, 0x174b758f},
				{974, 33175, 0x174b471b},
				{975, 33192, 0x174b1a29},
				{976, 33195, 0x174aec36},
				{977, 33198, 0x174abe43},
				{978, 33201, 0x174a907b},
				{979, 33201, 0x174a6287},
				{980, 33210, 0x174a356b},
				{981, 33210, 0x174a07cd},
				{982, 33214, 0x1749da86},
				{983, 33245, 0x1749af6c},
				{984, 33254, 0x174982a6},
				{985, 33260, 0x174955b4},
				{986, 33275, 0x174929c4},
				{987, 33284, 0x1748fd53},
				{988, 33299, 0x1748d18e},
				{989, 33305, 0x1748a51d},
				{990, 33306, 0x17487881},
				{991, 33309, 0x17484c10},
				{992, 33311, 0x17481f9f},
				{993, 33338, 0x1747f55c},
				{994, 33340, 0x1747c917},
				{995, 33343, 0x17479cfc},
				{996, 33346, 0x17477136},
				{997, 33347, 0x17474546},
				{998, 33353, 0x174719ac},
				{999, 33354, 0x1746ede7},
				{1000, 33360, 0x1746c2cd},
				{1001, 33365, 0x17469789},
			},
		},
		// Deterministically random solvetimes for down-ramping hashrate around a recent real life nBits
		{
			run:          8,
			anchorHeight: 1,
			anchorTime:   0,
			anchorBits:   0x1802aee8,
			startHeight:  2,
			startTime:    1200,
			blocks: []block{
				{2, 1200, 0x1802aee8},
				{3, 1310, 0x1802ad91},
				{4, 1327, 0x1802abf8},
				{5, 1739, 0x1802ab73},
				{6, 2219, 0x1802ab20},
				{7, 2604, 0x1802aa89},
				{8, 2746, 0x1802a94b},
				{9, 7099, 0x1802b39c},
				{10, 7099, 0x1802b1f2},
				{11, 7657, 0x1802b1d4},
				{12, 9595, 0x1802b58d},
				{13, 9975, 0x1802b4ee},
				{14, 11607, 0x1802b7d3},
				{15, 13951, 0x1802bcb8},
				{16, 15797, 0x1802c03e},
				{17, 15865, 0x1802bebb},
				{18, 18343, 0x1802c40e},
				{19, 20631, 0x1802c8de},
				{20, 22947, 0x1802cdce},
				{21, 23229, 0x1802cce4},
				{22, 33606, 0x1802e994},
				{23, 34944, 0x1802ebca},
				{24, 37977, 0x1802f31e},
				{25, 39180, 0x1802f4f1},
				{26, 40755, 0x1802f7e8},
				{27, 41175, 0x1802f75d},
				{28, 42855, 0x1802faa7},
				{29, 43200, 0x1802f9de},
				{30, 43410, 0x1802f8af},
				{31, 46761, 0x1803011f},
				{32, 47473, 0x18030177},
				{33, 51809, 0x18030d12},
				{34, 53537, 0x1803109d},
				{35, 56089, 0x180316c4},
				{36, 59769, 0x18032096},
				{37, 63137, 0x18032984},
				{38, 66673, 0x18033318},
				{39, 74425, 0x18034ae9},
				{40, 76909, 0x1803514b},
				{41, 79761, 0x180358ff},
				{42, 84471, 0x1803673b},
				{43, 87141, 0x18036e7c},
				{44, 90151, 0x18037704},
				{45, 91796, 0x18037abc},
				{46, 93236, 0x18037dbe},
				{47, 93916, 0x18037e07},
				{48, 94246, 0x18037d10},
				{49, 96101, 0x1803818f},
				{50, 96126, 0x18037f7c},
				{51, 96281, 0x18037de7},
				{52, 96719, 0x18037d50},
				{53, 99245, 0x1803843e},
				{54, 111773, 0x1803b056},
				{55, 115529, 0x1803bc5f},
				{56, 117845, 0x1803c2fc},
				{57, 121667, 0x1803cf82},
				{58, 123839, 0x1803d5af},
				{59, 124271, 0x1803d508},
				{60, 125795, 0x1803d8ab},
				{61, 128627, 0x1803e189},
				{62, 130874, 0x1803e81e},
				{63, 136075, 0x1803fac1},
				{64, 138133, 0x180400bd},
				{65, 138966, 0x180401b4},
				{66, 145532, 0x18041a91},
				{67, 151454, 0x18043142},
				{68, 151454, 0x18042eab},
				{69, 151510, 0x18042c58},
				{70, 154786, 0x180437e2},
				{71, 156228, 0x18043b88},
				{72, 160876, 0x18044d49},
				{73, 160932, 0x18044ae2},
				{74, 167388, 0x18046509},
				{75, 172700, 0x18047a86},
				{76, 174700, 0x180480fd},
				{77, 176172, 0x18048509},
				{78, 176740, 0x180484e1},
				{79, 176820, 0x18048278},
				{80, 177348, 0x18048224},
				{81, 188020, 0x1804b1c2},
				{82, 188533, 0x1804b157},
				{83, 191098, 0x1804bad9},
				{84, 191449, 0x1804b9a4},
				{85, 196534, 0x1804cf9c},
				{86, 201439, 0x1804e50e},
				{87, 202573, 0x1804e7c0},
				{88, 207289, 0x1804fca6},
				{89, 227503, 0x18056506},
				{90, 228205, 0x18056596},
				{91, 229636, 0x18056a38},
				{92, 234216, 0x18058096},
				{93, 252336, 0x1805e733},
				{94, 262256, 0x180620b3},
				{95, 262756, 0x18062012},
				{96, 271086, 0x1806516b},
				{97, 288256, 0x1806c070},
				{98, 317896, 0x180795e1},
				{99, 321636, 0x1807ae83},
				{100, 333016, 0x18080571},
				{101, 333916, 0x180807f0},
				{102, 347996, 0x18087a4a},
				{103, 352990, 0x1808a0e7},
				{104, 356246, 0x1808b897},
				{105, 362428, 0x1808eb32},
				{106, 367004, 0x18090ff1},
				{107, 379852, 0x180984da},
				{108, 382745, 0x18099b63},
				{109, 384065, 0x1809a27e},
				{110, 388564, 0x1809c961},
				{111, 391754, 0x1809e38a},
				{112, 404822, 0x180a6542},
				{113, 406274, 0x180a6e5b},
				{114, 408962, 0x180a84c3},
				{115, 409442, 0x180a8377},
				{116, 425930, 0x180b34b4},
				{117, 431978, 0x180b7423},
				{118, 432638, 0x180b74d9},
				{119, 435434, 0x180b8ecd},
				{120, 437390, 0x180b9ef1},
				{121, 439130, 0x180bac92},
				{122, 439715, 0x180bac67},
				{123, 442042, 0x180bc127},
				{124, 448087, 0x180c0390},
				{125, 449192, 0x180c09c5},
				{126, 498904, 0x180ea828},
				{127, 503376, 0x180ee2f0},
				{128, 521797, 0x180ffd6e},
				{129, 523396, 0x18100de8},
				{130, 527010, 0x18103fec},
				{131, 530741, 0x1810748a},
				{132, 532169, 0x18108296},
				{133, 546925, 0x181179c7},
				{134, 565671, 0x1812cbe8},
				{135, 572167, 0x18133f13},
				{136, 585439, 0x18143ff4},
				{137, 590605, 0x18149fab},
				{138, 594917, 0x1814eec8},
				{139, 610569, 0x18163c62},
				{140, 614111, 0x1816800d},
				{141, 624863, 0x18176f87},
				{142, 633533, 0x181834d4},
				{143, 637988, 0x18189542},
				{144, 642758, 0x1818ff54},
				{145, 643043, 0x1818f747},
				{146, 643838, 0x1818fc3a},
				{147, 647378, 0x18194807},
				{148, 652958, 0x1819ca80},
				{149, 656573, 0x181a1aa9},
				{150, 662303, 0x181aa584},
				{151, 665153, 0x181ae33b},
				{152, 665217, 0x181ad479},
				{153, 669809, 0x181b433e},
				{154, 675441, 0x181bd174},
				{155, 709105, 0x181fc36f},
				{156, 739713, 0x1823d4b2},
				{157, 739953, 0x1823c772},
				{158, 743137, 0x182426f4},
				{159, 751377, 0x1825473b},
				{160, 756049, 0x1825e474},
				{161, 766465, 0x18276a1c},
				{162, 769848, 0x1827db4e},
				{163, 774421, 0x18287f27},
				{164, 778824, 0x18291e62},
				{165, 781017, 0x182961cd},
				{166, 781935, 0x18296f37},
				{167, 788667, 0x182a76ed},
				{168, 800142, 0x182c5c16},
				{169, 806381, 0x182d605c},
				{170, 808965, 0x182dbd45},
				{171, 845345, 0x1834ca94},
				{172, 848441, 0x1835526a},
				{173, 863057, 0x18386749},
				{174, 874001, 0x183acab9},
				{175, 878285, 0x183baac6},
				{176, 906131, 0x18429001},
				{177, 928253, 0x1848922a},
				{178, 940691, 0x184c1a3e},
				{179, 969311, 0x185525de},
				{180, 970031, 0x1855306f},
				{181, 989723, 0x185bfa53},
				{182, 1001047, 0x1860050c},
				{183, 1017805, 0x18667131},
				{184, 1018432, 0x186673e0},
				{185, 1023885, 0x186876b9},
				{186, 1073114, 0x187ef4ed},
				{187, 1078358, 0x19008158},
				{188, 1140450, 0x1900a58c},
				{189, 1143129, 0x1900a6ef},
				{190, 1146530, 0x1900a8d1},
				{191, 1153750, 0x1900ad5b},
				{192, 1179150, 0x1900bf80},
				{193, 1212890, 0x1900dab2},
				{194, 1224950, 0x1900e4fa},
				{195, 1226790, 0x1900e61e},
				{196, 1232690, 0x1900eb10},
				{197, 1267710, 0x19010de3},
				{198, 1283910, 0x19011f55},
				{199, 1293770, 0x19012a37},
				{200, 1324330, 0x19015048},
				{201, 1336050, 0x19015fa0},
				{202, 1350435, 0x190173a3},
				{203, 1357323, 0x19017d1f},
				{204, 1368411, 0x19018d7d},
				{205, 1371057, 0x190190c2},
				{206, 1382985, 0x1901a35d},
				{207, 1383132, 0x1901a29a},
				{208, 1396047, 0x1901b7c8},
				{209, 1412889, 0x1901d55e},
				{210, 1427358, 0x1901f03a},
				{211, 1431075, 0x1901f679},
				{212, 1446233, 0x190214b8},
				{213, 1461589, 0x1902353d},
				{214, 1501123, 0x190294cc},
				{215, 1517755, 0x1902c0ae},
				{216, 1518239, 0x1902c05b},
				{217, 1546553, 0x1903132e},
				{218, 1550821, 0x19031ed6},
				{219, 1551811, 0x19032015},
				{220, 1580213, 0x19037e5f},
				{221, 1611497, 0x1903f383},
				{222, 1613567, 0x1903f97f},
				{223, 1619846, 0x190410f7},
				{224, 1627988, 0x190430fa},
				{225, 1682705, 0x19053535},
				{226, 1712835, 0x1905dce8},
				{227, 1737836, 0x1906771c},
				{228, 1742551, 0x1906929d},
				{229, 1743655, 0x19069607},
				{230, 1748922, 0x1906b5da},
				{231, 1751314, 0x1906c238},
				{232, 1759570, 0x1906f822},
				{233, 1768858, 0x19073761},
				{234, 1771642, 0x190747a0},
				{235, 1796434, 0x1908059c},
				{236, 1805578, 0x19084d48},
				{237, 1809922, 0x19086d7b},
				{238, 1832170, 0x19093166},
				{239, 1848538, 0x1909cb1e},
				{240, 1855858, 0x190a0f95},
				{241, 1907986, 0x190c5ed6},
				{242, 1950436, 0x190ea128},
				{243, 1951261, 0x190ea482},
				{244, 1999686, 0x1911bddd},
				{245, 2000211, 0x1911bc7b},
				{246, 2039836, 0x1914be10},
				{247, 2080561, 0x19185d54},
				{248, 2095061, 0x1919c29e},
				{249, 2098211, 0x191a0649},
				{250, 2101636, 0x191a522b},
				{251, 2114311, 0x191ba03b},
				{252, 2118367, 0x191c02ec},
				{253, 2178765, 0x19239bd8},
				{254, 2206325, 0x1927ad86},
				{255, 2208301, 0x1927e59f},
				{256, 2210979, 0x19283b3b},
				{257, 2211109, 0x192827b2},
				{258, 2234067, 0x192bebe5},
				{259, 2236381, 0x192c398b},
				{260, 2259703, 0x1930720b},
				{261, 2277357, 0x1933df20},
				{262, 2293854, 0x19374884},
				{263, 2296878, 0x1937d2b3},
				{264, 2345100, 0x194393f1},
				{265, 2346180, 0x1943b550},
				{266, 2348610, 0x1944351a},
				{267, 2353740, 0x19457594},
				{268, 2367645, 0x19494570},
				{269, 2369319, 0x1949966f},
				{270, 2371722, 0x194a1f48},
				{271, 2387814, 0x194ee076},
				{272, 2397166, 0x1951b169},
				{273, 2442330, 0x1961ae56},
				{274, 2443282, 0x1961d1e4},
				{275, 2445886, 0x19629bce},
				{276, 2450310, 0x1964218b},
				{277, 2450450, 0x1963f1ea},
				{278, 2459998, 0x19679803},
				{279, 2480942, 0x197064eb},
				{280, 2501718, 0x1979dde2},
				{281, 2529858, 0x1a00881c},
				{282, 2567210, 0x1a009dbf},
				{283, 2573300, 0x1a00a142},
				{284, 2588844, 0x1a00ab35},
				{285, 2663403, 0x1a00e650},
				{286, 2675641, 0x1a00f152},
				{287, 2701248, 0x1a010acd},
				{288, 2718503, 0x1a011d42},
				{289, 2736599, 0x1a013202},
				{290, 2758001, 0x1a014ca0},
				{291, 2771080, 0x1a015db3},
				{292, 2808940, 0x1a019610},
				{293, 2841880, 0x1a01ce41},
				{294, 2849530, 0x1a01db83},
				{295, 2849740, 0x1a01dac4},
				{296, 2867290, 0x1a01fc2e},
				{297, 2895370, 0x1a023777},
				{298, 2951740, 0x1a02c5c1},
				{299, 2965600, 0x1a02ec8e},
				{300, 2967070, 0x1a02ef2a},
				{301, 3021970, 0x1a03a5c8},
				{302, 3023086, 0x1a03a7b9},
				{303, 3076778, 0x1a0485fd},
				{304, 3109142, 0x1a05235e},
				{305, 3145226, 0x1a05ec96},
				{306, 3154650, 0x1a062332},
				{307, 3169530, 0x1a067fb4},
				{308, 3175327, 0x1a06a2b6},
				{309, 3194609, 0x1a0726cc},
				{310, 3198639, 0x1a074024},
				{311, 3252548, 0x1a08fb0b},
				{312, 3282116, 0x1a0a1665},
				{313, 3298596, 0x1a0abffc},
				{314, 3353700, 0x1a0d6062},
				{315, 3363236, 0x1a0ddd3d},
				{316, 3363684, 0x1a0ddb0f},
				{317, 3401476, 0x1a1015d4},
				{318, 3402884, 0x1a102334},
				{319, 3411748, 0x1a10ae90},
				{320, 3421732, 0x1a115289},
				{321, 3433732, 0x1a122246},
				{322, 3441124, 0x1a12a29c},
				{323, 3462640, 0x1a14441a},
				{324, 3485872, 0x1a163110},
				{325, 3498445, 0x1a1748ca},
				{326, 3535537, 0x1a1af36a},
				{327, 3559594, 0x1a1d9bb2},
				{328, 3567580, 0x1a1e7fb0},
				{329, 3590350, 0x1a215657},
				{330, 3601966, 0x1a22d84e},
				{331, 3612064, 0x1a24331c},
				{332, 3631240, 0x1a27005f},
				{333, 3634844, 0x1a277974},
				{334, 3644874, 0x1a28fec6},
				{335, 3648648, 0x1a29851a},
				{336, 3670204, 0x1a2d2986},
				{337, 3701076, 0x1a32fd66},
				{338, 3713860, 0x1a358a6d},
				{339, 3716478, 0x1a35f99d},
				{340, 3728004, 0x1a386444},
				{341, 3742590, 0x1a3ba512},
				{342, 3775490, 0x1a43e6c9},
				{343, 3782035, 0x1a458ab6},
				{344, 3806885, 0x1a4ca71d},
				{345, 3811960, 0x1a4e0a79},
				{346, 3822635, 0x1a51420e},
				{347, 3826135, 0x1a523539},
				{348, 3827955, 0x1a529c5c},
				{349, 3855430, 0x1a5c0463},
				{350, 3898795, 0x1a6d383c},
				{351, 3912900, 0x1a734b80},
				{352, 3950808, 0x1b0085ea},
				{353, 3961644, 0x1b008b89},
				{354, 3983604, 0x1b009807},
				{355, 3997608, 0x1b00a06d},
				{356, 4011072, 0x1b00a8e9},
				{357, 4015212, 0x1b00ab52},
				{358, 4020360, 0x1b00ae7b},
				{359, 4089768, 0x1b00e5e9},
				{360, 4113384, 0x1b00fc27},
				{361, 4114896, 0x1b00fd14},
				{362, 4145939, 0x1b011dfc},
				{363, 4154079, 0x1b0126c7},
				{364, 4175169, 0x1b014008},
				{365, 4182902, 0x1b014950},
				{366, 4221234, 0x1b017f21},
				{367, 4225341, 0x1b01848d},
				{368, 4254423, 0x1b01b386},
				{369, 4269926, 0x1b01ce56},
				{370, 4272294, 0x1b01d1a1},
				{371, 4279546, 0x1b01de38},
				{372, 4279774, 0x1b01dd82},
				{373, 4294214, 0x1b01f8c8},
				{374, 4309452, 0x1b021754},
				{375, 4330048, 0x1b024417},
				{376, 4335064, 0x1b024e77},
				{377, 4338902, 0x1b025631},
				{378, 4386706, 0x1b02d2e3},
				{379, 4395332, 0x1b02ea8b},
				{380, 4430710, 0x1b035a31},
				{381, 4432876, 0x1b035f97},
				{382, 4455847, 0x1b03b0a1},
				{383, 4476088, 0x1b03fe19},
				{384, 4507912, 0x1b04869e},
				{385, 4581934, 0x1b06136e},
				{386, 4583416, 0x1b0618f2},
				{387, 4592074, 0x1b064c28},
				{388, 4691875, 0x1b096030},
				{389, 4711882, 0x1b0a227e},
				{390, 4723075, 0x1b0a92f5},
				{391, 4731889, 0x1b0aedaf},
				{392, 4751009, 0x1b0bc599},
				{393, 4756169, 0x1b0bfd31},
				{394, 4770929, 0x1b0cb061},
				{395, 4797289, 0x1b0e119a},
				{396, 4819329, 0x1b0f5504},
				{397, 4842089, 0x1b10c22f},
				{398, 4844009, 0x1b10d8fe},
				{399, 4871489, 0x1b12c487},
				{400, 4873449, 0x1b12ded0},
				{401, 4883769, 0x1b139eeb},
				{402, 4899882, 0x1b14e128},
				{403, 4901686, 0x1b14faf0},
				{404, 4924277, 0x1b16eaa0},
				{405, 4935060, 0x1b17df22},
				{406, 4944203, 0x1b18b408},
				{407, 4953674, 0x1b1998c6},
				{408, 4997462, 0x1b1e6f81},
				{409, 5033255, 0x1b230de3},
				{410, 5035633, 0x1b234e1d},
				{411, 5085243, 0x1b2af93b},
				{412, 5088519, 0x1b2b704d},
				{413, 5121909, 0x1b318bc8},
				{414, 5124933, 0x1b3207b7},
				{415, 5135979, 0x1b342b18},
				{416, 5180877, 0x1b3e4ec8},
				{417, 5190537, 0x1b409d98},
				{418, 5190789, 0x1b408673},
				{419, 5246103, 0x1b505f7d},
				{420, 5264121, 0x1b562e80},
				{421, 5280459, 0x1b5bceb9},
				{422, 5299680, 0x1b62ec4d},
				{423, 5305958, 0x1b6532e5},
				{424, 5320320, 0x1b6aefa1},
				{425, 5352914, 0x1b79926b},
				{426, 5430357, 0x1c00a57c},
				{427, 5460070, 0x1c00b9fd},
				{428, 5512315, 0x1c00e4c3},
				{429, 5523108, 0x1c00ee4f},
				{430, 5638176, 0x1c017941},
				{431, 5641659, 0x1c017da2},
				{432, 5666343, 0x1c01a44e},
				{433, 5668191, 0x1c01a668},
				{434, 5683371, 0x1c01bfd3},
				{435, 5697319, 0x1c01d873},
				{436, 5709375, 0x1c01eeac},
				{437, 5798783, 0x1c02c276},
				{438, 5828967, 0x1c031b78},
				{439, 5837635, 0x1c03359a},
				{440, 5848459, 0x1c0357f3},
				{441, 5866191, 0x1c0394cd},
				{442, 5910066, 0x1c0442ad},
				{443, 5912496, 0x1c044aba},
				{444, 5929326, 0x1c0494ba},
				{445, 5936076, 0x1c04b208},
				{446, 5969421, 0x1c055aa0},
				{447, 6009786, 0x1c0647b6},
				{448, 6037011, 0x1c06fcba},
				{449, 6047316, 0x1c0743b5},
				{450, 6053211, 0x1c076b9e},
				{451, 6069321, 0x1c07e59a},
				{452, 6094989, 0x1c08bbac},
				{453, 6106765, 0x1c092249},
				{454, 6109111, 0x1c0932b8},
				{455, 6132249, 0x1c0a1193},
				{456, 6137263, 0x1c0a3f96},
				{457, 6160447, 0x1c0b385a},
				{458, 6177513, 0x1c0bfc7a},
				{459, 6244765, 0x1c0fa85d},
				{460, 6254195, 0x1c1038f7},
				{461, 6259071, 0x1c1080e8},
				{462, 6269223, 0x1c112623},
				{463, 6331686, 0x1c15fae5},
				{464, 6354951, 0x1c18129e},
				{465, 6374315, 0x1c19f3c1},
				{466, 6399836, 0x1c1cada5},
				{467, 6409471, 0x1c1dbc7b},
				{468, 6431467, 0x1c206707},
				{469, 6491768, 0x1c292ce4},
				{470, 6562409, 0x1c36867c},
				{471, 6592536, 0x1c3d6125},
				{472, 6623496, 0x1c4555e2},
				{473, 6640584, 0x1c4a14e2},
				{474, 6646680, 0x1c4bbbd4},
				{475, 6652008, 0x1c4d2f1e},
				{476, 6666120, 0x1c517b3e},
				{477, 6701784, 0x1c5dc9da},
				{478, 6705288, 0x1c5ee342},
				{479, 6737688, 0x1c6bc7f6},
				{480, 6774168, 0x1c7c75fc},
				{481, 6842136, 0x1d00a31a},
				{482, 6924554, 0x1d00e26d},
				{483, 6928915, 0x1d00e5dd},
				{484, 6930238, 0x1d00e688},
				{485, 6993399, 0x1d00ffff},
				{486, 7003150, 0x1d00ffff},
				{487, 7004914, 0x1d00ffff},
				{488, 7092820, 0x1d00ffff},
				{489, 7106491, 0x1d00ffff},
				{490, 7126140, 0x1d00ffff},
				{491, 7160783, 0x1d00ffff},
				{492, 7216733, 0x1d00ffff},
				{493, 7226083, 0x1d00ffff},
				{494, 7284983, 0x1d00ffff},
				{495, 7290833, 0x1d00ffff},
				{496, 7293683, 0x1d00ffff},
				{497, 7343383, 0x1d00ffff},
				{498, 7387333, 0x1d00ffff},
				{499, 7388383, 0x1d00ffff},
				{500, 7421433, 0x1d00ffff},
			},
		},
		// A sequence of 300s blocks across signed 32-bit max integer height
		{
			run:          9,
			anchorHeight: 2147483642,
			anchorTime:   1234567290,
			anchorBits:   0x1802aee8,
			startHeight:  2147483643,
			startTime:    1234568190,
			blocks: []block{
				{2147483643, 1234568190, 0x1802ae16},
				{2147483644, 1234568490, 0x1802ad44},
				{2147483645, 1234568790, 0x1802ac71},
				{2147483646, 1234569090, 0x1802ab9e},
				{2147483647, 1234569390, 0x1802aacd},
				{2147483648, 1234569690, 0x1802a9fa},
				{2147483649, 1234569990, 0x1802a929},
				{2147483650, 1234570290, 0x1802a858},
				{2147483651, 1234570590, 0x1802a787},
				{2147483652, 1234570890, 0x1802a6b7},
			},
		},
		// A sequence of 900s blocks across signed 64-bit max integer height and signed 32-bit max integer time
		{
			run:          10,
			anchorHeight: 9223372036854775802,
			anchorTime:   2147483047,
			anchorBits:   0x1802aee8,
			startHeight:  9223372036854775803,
			startTime:    2147484547,
			blocks: []block{
				{9223372036854775803, 2147484547, 0x1802afbb},
				{9223372036854775804, 2147485447, 0x1802b08f},
				{9223372036854775805, 2147486347, 0x1802b166},
				{9223372036854775806, 2147487247, 0x1802b23a},
				{9223372036854775807, 2147488147, 0x1802b30e},
				{9223372036854775808, 2147489047, 0x1802b3e5},
				{9223372036854775809, 2147489947, 0x1802b4bb},
				{9223372036854775810, 2147490847, 0x1802b592},
				{9223372036854775811, 2147491747, 0x1802b669},
				{9223372036854775812, 2147492647, 0x1802b73d},
			},
		},
		// Deterministically uniform random solvetimes with negative time
		{
			run:          11,
			anchorHeight: 1,
			anchorTime:   0,
			anchorBits:   0x1802aee8,
			startHeight:  2,
			startTime:    1200,
			blocks: []block{
				{2, 1200, 0x1802aee8},
				{3, 1496, 0x1802ae12},
				{4, 1398, 0x1802ac28},
				{5, 1996, 0x1802ac28},
				{6, 1731, 0x1802a9cb},
				{7, 1816, 0x1802a863},
				{8, 2113, 0x1802a790},
				{9, 2318, 0x1802a67e},
				{10, 2679, 0x1802a5d7},
				{11, 3289, 0x1802a5df},
				{12, 3557, 0x1802a4f8},
				{13, 3438, 0x1802a305},
				{14, 3391, 0x1802a147},
				{15, 3417, 0x18029fba},
				{16, 4317, 0x1802a08a},
				{17, 4073, 0x18029e45},
				{18, 3774, 0x18029bdc},
				{19, 3619, 0x180299d8},
				{20, 4046, 0x18029961},
				{21, 4571, 0x1802992f},
				{22, 5233, 0x18029959},
				{23, 5432, 0x18029848},
				{24, 5458, 0x180296c1},
				{25, 5399, 0x18029501},
				{26, 5991, 0x180294fc},
				{27, 6856, 0x180295b1},
				{28, 7586, 0x18029608},
				{29, 7706, 0x180294c2},
				{30, 8349, 0x180294e0},
				{31, 8407, 0x18029371},
				{32, 8174, 0x1802913f},
				{33, 8866, 0x1802917d},
				{34, 9614, 0x180291e0},
				{35, 9779, 0x180290bc},
				{36, 10501, 0x1802910e},
				{37, 10776, 0x18029033},
				{38, 11502, 0x18029088},
				{39, 11548, 0x18028f14},
				{40, 11500, 0x18028d60},
				{41, 11239, 0x18028b1f},
				{42, 12136, 0x18028be7},
				{43, 12921, 0x18028c62},
				{44, 13251, 0x18028baf},
				{45, 13017, 0x18028980},
				{46, 13695, 0x180289b5},
				{47, 13758, 0x1802884f},
				{48, 14043, 0x1802877f},
				{49, 14520, 0x1802872b},
				{50, 14920, 0x180286a7},
				{51, 14779, 0x180284bc},
				{52, 14728, 0x1802830d},
				{53, 14753, 0x18028193},
				{54, 15181, 0x18028122},
				{55, 15735, 0x18028104},
				{56, 15645, 0x18027f3f},
				{57, 16234, 0x18027f37},
				{58, 16067, 0x18027d41},
				{59, 16120, 0x18027bdb},
				{60, 16834, 0x18027c26},
				{61, 16996, 0x18027b08},
				{62, 17004, 0x18027986},
				{63, 17454, 0x18027924},
				{64, 18157, 0x18027967},
				{65, 17920, 0x18027749},
				{66, 18236, 0x18027690},
				{67, 18091, 0x180274ae},
				{68, 18577, 0x18027464},
				{69, 18790, 0x1802736b},
				{70, 19431, 0x18027386},
				{71, 19352, 0x180271d0},
				{72, 19958, 0x180271d6},
				{73, 20307, 0x18027135},
				{74, 20933, 0x18027143},
				{75, 21541, 0x18027149},
				{76, 22394, 0x180271eb},
				{77, 23139, 0x18027249},
				{78, 23613, 0x180271f9},
				{79, 23884, 0x18027125},
				{80, 24419, 0x180270fb},
				{81, 25272, 0x1802719d},
				{82, 25923, 0x180271be},
				{83, 26492, 0x180271ab},
				{84, 26900, 0x1802712f},
				{85, 27144, 0x1802704b},
				{86, 27605, 0x18026ff1},
				{87, 28045, 0x18026f8a},
				{88, 28252, 0x18026e8f},
				{89, 28138, 0x18026cc8},
				{90, 28296, 0x18026baf},
				{91, 28162, 0x180269dd},
				{92, 28107, 0x1802683c},
				{93, 28726, 0x18026848},
				{94, 28552, 0x18026660},
				{95, 29393, 0x180266f7},
				{96, 29647, 0x1802661e},
				{97, 30505, 0x180266c0},
				{98, 30215, 0x1802648e},
				{99, 30194, 0x18026309},
				{100, 29955, 0x180260fc},
				{101, 29800, 0x18025f23},
				{102, 29638, 0x18025d48},
				{103, 29347, 0x18025b1f},
				{104, 29653, 0x18025a6a},
				{105, 29683, 0x1802590b},
				{106, 30547, 0x180259ad},
				{107, 30413, 0x180257e6},
				{108, 30890, 0x1802579c},
				{109, 30699, 0x180255b5},
				{110, 30969, 0x180254eb},
				{111, 31199, 0x18025407},
				{112, 31684, 0x180253c2},
				{113, 32275, 0x180253bb},
				{114, 32519, 0x180252e2},
				{115, 33083, 0x180252cd},
				{116, 32920, 0x180250fa},
				{117, 32905, 0x18024f83},
				{118, 33019, 0x18024e5e},
				{119, 33176, 0x18024d51},
				{120, 33530, 0x18024cbb},
				{121, 33796, 0x18024bf2},
				{122, 33993, 0x18024aff},
				{123, 34841, 0x18024b94},
				{124, 35394, 0x18024b78},
				{125, 35811, 0x18024b0a},
				{126, 35977, 0x18024a03},
				{127, 36103, 0x180248e5},
				{128, 36019, 0x1802474c},
				{129, 35828, 0x18024572},
				{130, 36477, 0x18024590},
				{131, 36483, 0x1802442c},
				{132, 37091, 0x18024432},
				{133, 37854, 0x18024492},
				{134, 37554, 0x1802427b},
				{135, 37268, 0x1802406d},
				{136, 36984, 0x18023e62},
				{137, 37498, 0x18023e2f},
				{138, 37848, 0x18023d9b},
				{139, 38669, 0x18023e1f},
				{140, 38719, 0x18023cda},
				{141, 39597, 0x18023d7e},
				{142, 40042, 0x18023d22},
				{143, 39885, 0x18023b65},
				{144, 39599, 0x1802395e},
				{145, 40432, 0x180239e5},
				{146, 41020, 0x180239e0},
				{147, 41490, 0x18023992},
				{148, 41994, 0x1802395b},
				{149, 42268, 0x1802389b},
				{150, 42377, 0x1802377d},
				{151, 42236, 0x180235cd},
				{152, 42253, 0x1802347b},
				{153, 42130, 0x180232d9},
				{154, 41964, 0x1802311d},
				{155, 42299, 0x18023085},
				{156, 42018, 0x18022e8a},
				{157, 42136, 0x18022d76},
				{158, 41961, 0x18022bba},
				{159, 42603, 0x18022bd2},
				{160, 43374, 0x18022c34},
				{161, 44180, 0x18022caa},
				{162, 43989, 0x18022ae6},
				{163, 44850, 0x18022b7b},
				{164, 45005, 0x18022a7d},
				{165, 45852, 0x18022b0a},
				{166, 45629, 0x18022934},
				{167, 46154, 0x1802290b},
				{168, 46586, 0x180228aa},
				{169, 47003, 0x18022843},
				{170, 47420, 0x180227db},
				{171, 47985, 0x180227c6},
				{172, 47912, 0x1802264a},
				{173, 48768, 0x180226da},
				{174, 49167, 0x18022669},
				{175, 49100, 0x180224f0},
				{176, 49972, 0x18022589},
				{177, 50730, 0x180225e3},
				{178, 50577, 0x1802243a},
				{179, 50465, 0x180222a9},
				{180, 50444, 0x1802214d},
				{181, 50989, 0x1802212e},
				{182, 51330, 0x1802209d},
				{183, 51444, 0x18021f8c},
				{184, 52286, 0x18022013},
				{185, 52731, 0x18021fbd},
				{186, 53402, 0x18021fe4},
				{187, 53983, 0x18021fdb},
				{188, 53779, 0x18021e19},
				{189, 54536, 0x18021e72},
				{190, 55137, 0x18021e72},
				{191, 55664, 0x18021e48},
				{192, 56496, 0x18021eca},
				{193, 56744, 0x18021e07},
				{194, 57636, 0x18021ea9},
				{195, 58132, 0x18021e6f},
				{196, 58390, 0x18021db1},
				{197, 58294, 0x18021c2e},
				{198, 59014, 0x18021c70},
				{199, 59385, 0x18021bf1},
				{200, 59252, 0x18021a5a},
				{201, 59734, 0x18021a1a},
				{202, 60492, 0x18021a70},
				{203, 61248, 0x18021ac8},
				{204, 61585, 0x18021a36},
				{205, 61408, 0x18021889},
				{206, 61576, 0x1802179a},
				{207, 61567, 0x1802164a},
				{208, 61997, 0x180215ee},
				{209, 62779, 0x18021652},
				{210, 63079, 0x180215ad},
				{211, 63753, 0x180215d6},
				{212, 64482, 0x1802161d},
				{213, 65054, 0x1802160d},
				{214, 65180, 0x1802150a},
				{215, 65954, 0x18021569},
				{216, 65871, 0x180213f3},
				{217, 66122, 0x18021334},
				{218, 66415, 0x1802128c},
				{219, 67002, 0x18021286},
				{220, 67163, 0x18021196},
				{221, 67140, 0x18021044},
				{222, 67845, 0x1802107d},
				{223, 68006, 0x18020f8e},
				{224, 67926, 0x18020e1f},
				{225, 68677, 0x18020e6f},
				{226, 68752, 0x18020d54},
				{227, 68739, 0x18020c09},
				{228, 68517, 0x18020a4f},
				{229, 68844, 0x180209bd},
				{230, 69623, 0x18020a1c},
				{231, 70389, 0x18020a76},
				{232, 70771, 0x18020a00},
				{233, 71391, 0x18020a0c},
				{234, 71197, 0x18020863},
				{235, 70997, 0x180206b7},
				{236, 71084, 0x180205a7},
				{237, 71262, 0x180204c7},
				{238, 71248, 0x18020381},
				{239, 71418, 0x1802029d},
				{240, 71376, 0x18020149},
				{241, 72137, 0x1802019e},
				{242, 71924, 0x1801fff2},
				{243, 72723, 0x1802005a},
				{244, 72733, 0x1801ff24},
				{245, 73287, 0x1801ff0c},
				{246, 73635, 0x1801fe89},
				{247, 73347, 0x1801fcb7},
				{248, 73207, 0x1801fb34},
				{249, 73162, 0x1801f9e6},
				{250, 74060, 0x1801fa80},
				{251, 74354, 0x1801f9e0},
				{252, 74578, 0x1801f91e},
				{253, 74420, 0x1801f793},
				{254, 75207, 0x1801f7f5},
				{255, 75117, 0x1801f690},
				{256, 75039, 0x1801f533},
				{257, 75543, 0x1801f502},
				{258, 75688, 0x1801f418},
				{259, 75698, 0x1801f2e9},
				{260, 75855, 0x1801f206},
				{261, 76080, 0x1801f147},
				{262, 75969, 0x1801efdb},
				{263, 76421, 0x1801ef90},
				{264, 77114, 0x1801efbf},
				{265, 77380, 0x1801ef14},
				{266, 78152, 0x1801ef6d},
				{267, 78608, 0x1801ef23},
				{268, 78954, 0x1801eea2},
				{269, 79243, 0x1801ee04},
				{270, 79896, 0x1801ee1f},
				{271, 80189, 0x1801ed83},
				{272, 80441, 0x1801ecd2},
				{273, 80632, 0x1801ec04},
				{274, 80814, 0x1801eb31},
				{275, 81489, 0x1801eb56},
				{276, 82203, 0x1801eb90},
				{277, 82617, 0x1801eb32},
				{278, 82730, 0x1801ea3d},
				{279, 82989, 0x1801e992},
				{280, 83556, 0x1801e981},
				{281, 84222, 0x1801e9a1},
				{282, 84665, 0x1801e953},
				{283, 84767, 0x1801e85a},
				{284, 84936, 0x1801e780},
				{285, 85292, 0x1801e708},
				{286, 85486, 0x1801e63c},
				{287, 86046, 0x1801e628},
				{288, 85785, 0x1801e47a},
				{289, 86591, 0x1801e4e0},
				{290, 86393, 0x1801e355},
				{291, 86577, 0x1801e286},
				{292, 86668, 0x1801e18b},
				{293, 87430, 0x1801e1da},
				{294, 87685, 0x1801e130},
				{295, 88110, 0x1801e0d9},
				{296, 88208, 0x1801dfe2},
				{297, 88868, 0x1801dfff},
				{298, 89133, 0x1801df5a},
				{299, 90001, 0x1801dfde},
				{300, 90027, 0x1801dec3},
				{301, 89785, 0x1801dd27},
				{302, 89493, 0x1801db73},
				{303, 89555, 0x1801da6c},
				{304, 89711, 0x1801d994},
				{305, 90224, 0x1801d969},
				{306, 90457, 0x1801d8b6},
				{307, 90665, 0x1801d7f8},
				{308, 91530, 0x1801d879},
				{309, 91706, 0x1801d7ab},
				{310, 92557, 0x1801d826},
				{311, 92946, 0x1801d7be},
				{312, 92677, 0x1801d61c},
				{313, 93246, 0x1801d60c},
				{314, 93286, 0x1801d4fe},
				{315, 93637, 0x1801d486},
				{316, 94029, 0x1801d422},
				{317, 94460, 0x1801d3d1},
				{318, 94623, 0x1801d2ff},
				{319, 95252, 0x1801d30c},
				{320, 95747, 0x1801d2dc},
				{321, 95499, 0x1801d145},
				{322, 95538, 0x1801d039},
				{323, 95651, 0x1801cf52},
				{324, 95785, 0x1801ce75},
				{325, 95569, 0x1801ccf1},
				{326, 95690, 0x1801cc10},
				{327, 95734, 0x1801cb09},
				{328, 95667, 0x1801c9cf},
				{329, 96011, 0x1801c956},
				{330, 95815, 0x1801c7e1},
				{331, 96294, 0x1801c7a9},
				{332, 96574, 0x1801c714},
				{333, 97006, 0x1801c6c5},
				{334, 97517, 0x1801c69b},
				{335, 98194, 0x1801c6bf},
				{336, 98444, 0x1801c61d},
				{337, 98484, 0x1801c517},
				{338, 98453, 0x1801c3f3},
				{339, 99067, 0x1801c3f8},
				{340, 99934, 0x1801c475},
				{341, 100298, 0x1801c407},
				{342, 100298, 0x1801c2f0},
				{343, 100644, 0x1801c27b},
				{344, 101235, 0x1801c277},
				{345, 101096, 0x1801c121},
				{346, 101561, 0x1801c0e3},
				{347, 102143, 0x1801c0db},
				{348, 102236, 0x1801bff2},
				{349, 102877, 0x1801c006},
				{350, 103686, 0x1801c065},
				{351, 103842, 0x1801bf99},
				{352, 104588, 0x1801bfdc},
				{353, 105259, 0x1801bffd},
				{354, 105780, 0x1801bfd8},
				{355, 105569, 0x1801be65},
				{356, 106385, 0x1801bec7},
				{357, 106138, 0x1801bd44},
				{358, 105916, 0x1801bbcd},
				{359, 106049, 0x1801baf9},
				{360, 106756, 0x1801bb29},
				{361, 107253, 0x1801bafa},
				{362, 107709, 0x1801bab8},
				{363, 108165, 0x1801ba77},
				{364, 108789, 0x1801ba83},
				{365, 108836, 0x1801b988},
				{366, 108602, 0x1801b80d},
				{367, 108367, 0x1801b696},
				{368, 108523, 0x1801b5ce},
				{369, 109272, 0x1801b611},
				{370, 110157, 0x1801b692},
				{371, 110833, 0x1801b6b3},
				{372, 110976, 0x1801b5e7},
				{373, 110746, 0x1801b472},
				{374, 111641, 0x1801b4f6},
				{375, 111873, 0x1801b451},
				{376, 111825, 0x1801b32f},
				{377, 112406, 0x1801b327},
				{378, 112416, 0x1801b220},
				{379, 113140, 0x1801b257},
				{380, 113403, 0x1801b1c2},
				{381, 113418, 0x1801b0be},
				{382, 114174, 0x1801b102},
				{383, 114250, 0x1801b01a},
				{384, 115078, 0x1801b07f},
				{385, 114959, 0x1801af41},
				{386, 115420, 0x1801af03},
				{387, 115754, 0x1801ae8d},
				{388, 115625, 0x1801ad4c},
				{389, 116160, 0x1801ad2f},
				{390, 116327, 0x1801ac70},
				{391, 117137, 0x1801accd},
				{392, 117973, 0x1801ad34},
				{393, 117987, 0x1801ac34},
				{394, 118576, 0x1801ac2f},
				{395, 118484, 0x1801aafe},
				{396, 118857, 0x1801aa9b},
				{397, 119475, 0x1801aaa3},
				{398, 120281, 0x1801aafe},
				{399, 120659, 0x1801aa9c},
				{400, 121107, 0x1801aa59},
				{401, 121807, 0x1801aa85},
				{402, 122324, 0x1801aa61},
				{403, 123027, 0x1801aa8f},
				{404, 123233, 0x1801a9e2},
				{405, 123550, 0x1801a966},
				{406, 123952, 0x1801a910},
				{407, 124779, 0x1801a974},
				{408, 124707, 0x1801a84e},
				{409, 125199, 0x1801a81f},
				{410, 125634, 0x1801a7d8},
				{411, 125938, 0x1801a757},
				{412, 125653, 0x1801a5d7},
				{413, 125719, 0x1801a4f1},
				{414, 126189, 0x1801a4b8},
				{415, 127008, 0x1801a518},
				{416, 127593, 0x1801a511},
				{417, 128434, 0x1801a579},
				{418, 128954, 0x1801a555},
				{419, 129593, 0x1801a567},
				{420, 129600, 0x1801a466},
				{421, 130389, 0x1801a4b8},
				{422, 130909, 0x1801a497},
				{423, 130897, 0x1801a38e},
				{424, 131423, 0x1801a36e},
				{425, 131825, 0x1801a318},
				{426, 132665, 0x1801a380},
				{427, 132788, 0x1801a2b2},
				{428, 133348, 0x1801a2a2},
				{429, 133310, 0x1801a191},
				{430, 133053, 0x1801a022},
				{431, 132962, 0x18019efb},
				{432, 133722, 0x18019f40},
				{433, 133774, 0x18019e56},
				{434, 133806, 0x18019d66},
				{435, 134178, 0x18019d05},
				{436, 133966, 0x18019bae},
				{437, 134247, 0x18019b26},
				{438, 135008, 0x18019b6b},
				{439, 135749, 0x18019ba7},
				{440, 136537, 0x18019bf6},
				{441, 136606, 0x18019b16},
				{442, 136474, 0x180199e2},
				{443, 137255, 0x18019a2e},
				{444, 137394, 0x1801996c},
				{445, 137515, 0x180198a2},
				{446, 137751, 0x1801980b},
				{447, 137788, 0x18019720},
				{448, 138666, 0x18019793},
				{449, 138670, 0x1801969a},
				{450, 139435, 0x180196e0},
				{451, 140141, 0x1801970c},
				{452, 140845, 0x18019737},
				{453, 140651, 0x180195ec},
				{454, 140699, 0x18019506},
				{455, 141248, 0x180194f1},
				{456, 141041, 0x180193a3},
				{457, 141325, 0x1801931f},
				{458, 141406, 0x1801924a},
				{459, 141462, 0x18019168},
				{460, 141354, 0x18019047},
				{461, 141577, 0x18018fac},
				{462, 141801, 0x18018f11},
				{463, 141511, 0x18018da6},
				{464, 141781, 0x18018d1f},
				{465, 142602, 0x18018d79},
				{466, 142513, 0x18018c60},
				{467, 142329, 0x18018b22},
				{468, 143015, 0x18018b45},
				{469, 143201, 0x18018a9e},
				{470, 143465, 0x18018a16},
				{471, 144100, 0x18018a24},
				{472, 144113, 0x18018936},
				{473, 144259, 0x18018880},
				{474, 144903, 0x18018891},
				{475, 145171, 0x1801880c},
				{476, 145686, 0x180187e9},
				{477, 146413, 0x1801881c},
				{478, 146592, 0x18018773},
				{479, 146913, 0x18018704},
				{480, 147781, 0x1801876f},
				{481, 147497, 0x1801860d},
				{482, 148327, 0x18018668},
				{483, 148251, 0x1801855b},
				{484, 148015, 0x1801840e},
				{485, 147949, 0x18018304},
				{486, 148802, 0x18018369},
				{487, 148701, 0x18018253},
				{488, 148443, 0x18018100},
				{489, 148243, 0x18017fc3},
				{490, 149058, 0x18018019},
				{491, 149534, 0x18017fe7},
				{492, 150355, 0x1801803f},
				{493, 150134, 0x18017efb},
				{494, 150578, 0x18017ebe},
				{495, 150987, 0x18017e74},
				{496, 150915, 0x18017d6b},
				{497, 151369, 0x18017d32},
				{498, 151314, 0x18017c33},
				{499, 151391, 0x18017b67},
				{500, 151776, 0x18017b12},
				{501, 151798, 0x18017a32},
				{502, 151773, 0x18017941},
				{503, 151649, 0x18017829},
				{504, 152525, 0x18017892},
				{505, 152634, 0x180177d5},
				{506, 153329, 0x180177fb},
				{507, 153368, 0x18017722},
				{508, 153211, 0x180175fe},
				{509, 153807, 0x180175fd},
				{510, 153698, 0x180174ee},
				{511, 154349, 0x18017501},
				{512, 154059, 0x180173ac},
				{513, 154893, 0x18017406},
				{514, 155518, 0x1801740f},
				{515, 155934, 0x180173ca},
				{516, 155800, 0x180172b1},
				{517, 156150, 0x18017253},
				{518, 156781, 0x1801725e},
				{519, 156891, 0x180171a5},
				{520, 157404, 0x18017183},
				{521, 157412, 0x180170a2},
				{522, 157879, 0x18017070},
				{523, 158231, 0x18017012},
				{524, 158136, 0x18016f0d},
				{525, 158289, 0x18016e64},
				{526, 159128, 0x18016ebe},
				{527, 159579, 0x18016e87},
				{528, 159356, 0x18016d4f},
				{529, 159729, 0x18016cfb},
				{530, 159596, 0x18016be8},
				{531, 160403, 0x18016c36},
				{532, 160429, 0x18016b5f},
				{533, 160233, 0x18016a36},
				{534, 160625, 0x180169e9},
				{535, 161076, 0x180169b2},
				{536, 160894, 0x18016890},
				{537, 161553, 0x180168a5},
				{538, 161418, 0x18016795},
				{539, 162309, 0x18016800},
				{540, 162765, 0x180167cc},
				{541, 163435, 0x180167e5},
				{542, 163647, 0x18016756},
				{543, 163732, 0x18016699},
				{544, 164551, 0x180166e9},
				{545, 165037, 0x180166c0},
				{546, 164803, 0x1801658b},
				{547, 164865, 0x180164c6},
				{548, 164737, 0x180163bb},
				{549, 165458, 0x180163e9},
				{550, 165436, 0x18016304},
				{551, 165181, 0x180161cd},
				{552, 164892, 0x1801608a},
				{553, 165406, 0x1801606b},
				{554, 165627, 0x18015fe1},
				{555, 165615, 0x18015f03},
				{556, 166171, 0x18015ef5},
				{557, 166381, 0x18015e68},
				{558, 166933, 0x18015e55},
				{559, 167185, 0x18015dd8},
				{560, 167861, 0x18015df4},
				{561, 167606, 0x18015cc1},
				{562, 168366, 0x18015cfb},
				{563, 168642, 0x18015c86},
				{564, 168697, 0x18015bc2},
				{565, 168630, 0x18015ad3},
				{566, 169039, 0x18015a90},
				{567, 168832, 0x18015970},
				{568, 169716, 0x180159d5},
				{569, 169600, 0x180158d7},
				{570, 169906, 0x1801586e},
				{571, 170711, 0x180158b7},
				{572, 171336, 0x180158bf},
				{573, 171304, 0x180157e0},
				{574, 171759, 0x180157ac},
				{575, 171566, 0x18015695},
				{576, 172042, 0x1801566a},
				{577, 172470, 0x1801562d},
				{578, 173020, 0x1801561c},
				{579, 173716, 0x1801563e},
				{580, 174051, 0x180155e1},
				{581, 174724, 0x180155fa},
				{582, 175478, 0x18015631},
				{583, 175724, 0x180155b5},
				{584, 176432, 0x180155da},
				{585, 177061, 0x180155e5},
				{586, 177177, 0x1801553b},
				{587, 177339, 0x180154a2},
				{588, 177053, 0x1801536d},
				{589, 177513, 0x1801533d},
				{590, 177961, 0x18015308},
				{591, 177771, 0x180151f5},
				{592, 178420, 0x18015207},
				{593, 179226, 0x1801524e},
				{594, 180074, 0x180152a4},
				{595, 180571, 0x18015280},
				{596, 180916, 0x18015228},
				{597, 180882, 0x1801514c},
				{598, 180677, 0x18015037},
				{599, 181194, 0x1801501a},
				{600, 182038, 0x1801506e},
				{601, 181782, 0x18014f47},
				{602, 181553, 0x18014e2c},
				{603, 181606, 0x18014d70},
				{604, 182440, 0x18014dc0},
				{605, 182169, 0x18014c97},
				{606, 182859, 0x18014cb5},
				{607, 183389, 0x18014c9d},
				{608, 184117, 0x18014cc9},
				{609, 184016, 0x18014bda},
				{610, 184193, 0x18014b4a},
				{611, 184979, 0x18014b89},
				{612, 185340, 0x18014b38},
				{613, 185048, 0x18014a0a},
				{614, 185070, 0x18014946},
				{615, 185653, 0x18014940},
				{616, 186128, 0x18014916},
				{617, 186692, 0x1801490a},
				{618, 186626, 0x1801482a},
				{619, 187357, 0x18014856},
				{620, 188191, 0x180148a5},
				{621, 187977, 0x18014793},
				{622, 188620, 0x180147a1},
				{623, 188362, 0x18014681},
				{624, 188754, 0x1801463c},
				{625, 188695, 0x18014560},
				{626, 189527, 0x180145ad},
				{627, 190303, 0x180145e7},
				{628, 190473, 0x18014558},
				{629, 190617, 0x180144c0},
				{630, 190346, 0x1801439e},
				{631, 191013, 0x180143b4},
				{632, 191746, 0x180143e1},
				{633, 192222, 0x180143b8},
				{634, 192007, 0x180142a9},
				{635, 192477, 0x1801427e},
				{636, 192621, 0x180141e7},
				{637, 192328, 0x180140c1},
				{638, 192515, 0x18014039},
				{639, 193101, 0x18014035},
				{640, 193359, 0x18013fc4},
				{641, 193452, 0x18013f1e},
				{642, 193526, 0x18013e71},
				{643, 193272, 0x18013d5b},
				{644, 193768, 0x18013d39},
				{645, 194503, 0x18013d65},
				{646, 195032, 0x18013d4e},
				{647, 195810, 0x18013d89},
				{648, 196075, 0x18013d1b},
				{649, 196515, 0x18013ce7},
				{650, 197231, 0x18013d0c},
				{651, 198079, 0x18013d5e},
				{652, 198624, 0x18013d4c},
				{653, 198571, 0x18013c78},
				{654, 198670, 0x18013bd5},
				{655, 198378, 0x18013ab3},
				{656, 198898, 0x18013a9a},
				{657, 199342, 0x18013a68},
				{658, 199575, 0x180139f2},
				{659, 200350, 0x18013a2a},
				{660, 200530, 0x180139a2},
				{661, 201300, 0x180139d9},
				{662, 201732, 0x180139a2},
				{663, 201462, 0x1801388b},
				{664, 201357, 0x180137a9},
				{665, 201612, 0x1801373b},
				{666, 201778, 0x180136b0},
				{667, 202649, 0x18013706},
				{668, 202552, 0x18013628},
				{669, 203405, 0x18013679},
				{670, 203888, 0x18013654},
				{671, 204592, 0x18013675},
				{672, 205228, 0x18013681},
				{673, 205139, 0x180135a5},
				{674, 205398, 0x18013539},
				{675, 205809, 0x180134fc},
				{676, 205654, 0x1801340e},
				{677, 205990, 0x180133ba},
				{678, 206209, 0x18013342},
				{679, 206486, 0x180132dc},
				{680, 206800, 0x18013282},
				{681, 207279, 0x1801325c},
				{682, 207929, 0x1801326c},
				{683, 208569, 0x18013278},
				{684, 208591, 0x180131c3},
				{685, 208930, 0x18013171},
				{686, 209463, 0x1801315b},
				{687, 209808, 0x1801310b},
				{688, 209767, 0x18013042},
				{689, 209747, 0x18012f81},
				{690, 209925, 0x18012efe},
				{691, 209923, 0x18012e42},
				{692, 209807, 0x18012d64},
				{693, 209783, 0x18012ca4},
				{694, 209582, 0x18012bac},
				{695, 209507, 0x18012add},
				{696, 210083, 0x18012ad6},
				{697, 210349, 0x18012a6f},
				{698, 211097, 0x18012a9c},
				{699, 211417, 0x18012a46},
				{700, 211705, 0x180129e7},
				{701, 212539, 0x18012a2e},
				{702, 213232, 0x18012a4b},
				{703, 213766, 0x18012a37},
				{704, 213940, 0x180129b5},
				{705, 214395, 0x18012988},
				{706, 214862, 0x1801295f},
				{707, 215035, 0x180128dd},
				{708, 215308, 0x18012879},
				{709, 215276, 0x180127b9},
				{710, 215533, 0x18012751},
				{711, 215848, 0x180126fa},
				{712, 215758, 0x1801262a},
				{713, 215977, 0x180125b7},
				{714, 215933, 0x180124f4},
				{715, 216232, 0x18012499},
				{716, 217084, 0x180124e6},
				{717, 217382, 0x1801248a},
				{718, 217880, 0x1801246c},
				{719, 218590, 0x1801248c},
				{720, 218372, 0x18012398},
				{721, 218717, 0x1801234a},
				{722, 219321, 0x1801234c},
				{723, 219465, 0x180122c4},
				{724, 219534, 0x18012226},
				{725, 219397, 0x1801214a},
				{726, 219833, 0x18012119},
				{727, 220123, 0x180120bd},
				{728, 220177, 0x1801201b},
				{729, 221015, 0x18012061},
				{730, 221702, 0x1801207b},
				{731, 222268, 0x18012072},
				{732, 222171, 0x18011fa3},
				{733, 222380, 0x18011f2f},
				{734, 222957, 0x18011f29},
				{735, 223219, 0x18011ec5},
				{736, 223523, 0x18011e6e},
				{737, 223451, 0x18011da8},
				{738, 223273, 0x18011cc4},
				{739, 223525, 0x18011c5e},
				{740, 223560, 0x18011bba},
				{741, 224001, 0x18011b8b},
				{742, 223714, 0x18011a89},
				{743, 223762, 0x180119e9},
				{744, 224025, 0x18011988},
				{745, 224162, 0x18011902},
				{746, 224447, 0x180118a7},
				{747, 224441, 0x180117f8},
				{748, 224855, 0x180117c2},
				{749, 224736, 0x180116f4},
				{750, 224673, 0x18011636},
				{751, 224462, 0x1801154e},
				{752, 224349, 0x18011483},
				{753, 224820, 0x1801145f},
				{754, 225573, 0x1801148a},
				{755, 225560, 0x180113dc},
				{756, 225639, 0x18011348},
				{757, 225642, 0x180112a0},
				{758, 226288, 0x180112ad},
				{759, 226896, 0x180112af},
				{760, 226705, 0x180111d0},
				{761, 226894, 0x1801115c},
				{762, 227345, 0x18011132},
				{763, 227421, 0x180110a0},
				{764, 227816, 0x18011066},
				{765, 227865, 0x18010fcc},
				{766, 228457, 0x18010fca},
				{767, 229104, 0x18010fd7},
				{768, 229954, 0x1801101d},
				{769, 230055, 0x18010f91},
				{770, 229829, 0x18010eaa},
				{771, 229868, 0x18010e10},
				{772, 229940, 0x18010d7d},
				{773, 230326, 0x18010d42},
				{774, 231210, 0x18010d90},
				{775, 231050, 0x18010cbe},
				{776, 231440, 0x18010c84},
				{777, 231500, 0x18010bef},
				{778, 231656, 0x18010b75},
				{779, 232039, 0x18010b3a},
				{780, 232116, 0x18010aaa},
				{781, 232071, 0x180109f9},
				{782, 232431, 0x180109b8},
				{783, 233040, 0x180109ba},
				{784, 232972, 0x18010904},
				{785, 233661, 0x1801091c},
				{786, 233912, 0x180108bd},
				{787, 233761, 0x180107f1},
				{788, 234261, 0x180107d6},
				{789, 234319, 0x18010743},
				{790, 234740, 0x18010712},
				{791, 235326, 0x1801070f},
				{792, 235762, 0x180106e2},
				{793, 235480, 0x180105f5},
				{794, 235272, 0x1801051b},
				{795, 235368, 0x18010494},
				{796, 235887, 0x1801047f},
				{797, 235668, 0x180103a4},
				{798, 235463, 0x180102cd},
				{799, 235897, 0x180102a1},
				{800, 236017, 0x18010222},
				{801, 236085, 0x18010194},
				{802, 236362, 0x1801013f},
				{803, 236126, 0x18010062},
				{804, 235911, 0x1800ff8c},
				{805, 236796, 0x1800ffd6},
				{806, 237218, 0x1800ffa8},
				{807, 237126, 0x1800fef2},
				{808, 237253, 0x1800fe77},
				{809, 238102, 0x1800feb8},
				{810, 238176, 0x1800fe2e},
				{811, 238920, 0x1800fe54},
				{812, 239004, 0x1800fdcd},
				{813, 239098, 0x1800fd4a},
				{814, 239796, 0x1800fd63},
				{815, 239748, 0x1800fcba},
				{816, 240368, 0x1800fcc0},
				{817, 241065, 0x1800fcd9},
				{818, 240875, 0x1800fc0c},
				{819, 241058, 0x1800fba0},
				{820, 241799, 0x1800fbc4},
				{821, 242428, 0x1800fbcc},
				{822, 242274, 0x1800fb09},
				{823, 242943, 0x1800fb1c},
				{824, 243524, 0x1800fb16},
				{825, 243498, 0x1800fa75},
				{826, 243335, 0x1800f9b1},
				{827, 243089, 0x1800f8d8},
				{828, 243216, 0x1800f85f},
				{829, 243103, 0x1800f7a9},
				{830, 243801, 0x1800f7c2},
				{831, 243557, 0x1800f6ec},
				{832, 243639, 0x1800f669},
				{833, 244419, 0x1800f696},
				{834, 244131, 0x1800f5b6},
				{835, 244258, 0x1800f53f},
				{836, 245025, 0x1800f569},
				{837, 245735, 0x1800f585},
				{838, 246049, 0x1800f53c},
				{839, 246190, 0x1800f4c9},
				{840, 246463, 0x1800f476},
				{841, 246179, 0x1800f399},
				{842, 246510, 0x1800f356},
				{843, 247162, 0x1800f363},
				{844, 247122, 0x1800f2c3},
				{845, 247736, 0x1800f2c6},
				{846, 247949, 0x1800f266},
				{847, 248502, 0x1800f25a},
				{848, 249063, 0x1800f250},
				{849, 249494, 0x1800f227},
				{850, 250335, 0x1800f262},
				{851, 250795, 0x1800f240},
				{852, 251316, 0x1800f22c},
				{853, 251598, 0x1800f1dc},
				{854, 252340, 0x1800f200},
				{855, 252895, 0x1800f1f5},
				{856, 253006, 0x1800f17c},
				{857, 253621, 0x1800f180},
				{858, 254508, 0x1800f1c6},
				{859, 254766, 0x1800f172},
				{860, 254619, 0x1800f0b9},
				{861, 254870, 0x1800f063},
				{862, 255300, 0x1800f038},
				{863, 255809, 0x1800f022},
				{864, 255590, 0x1800ef58},
				{865, 256461, 0x1800ef9b},
				{866, 256686, 0x1800ef3f},
				{867, 257249, 0x1800ef36},
				{868, 257442, 0x1800eed2},
				{869, 257726, 0x1800ee84},
				{870, 257866, 0x1800ee14},
				{871, 257926, 0x1800ed90},
				{872, 258786, 0x1800edcf},
				{873, 259075, 0x1800ed83},
				{874, 259593, 0x1800ed6f},
				{875, 259728, 0x1800ecff},
				{876, 259765, 0x1800ec75},
				{877, 260127, 0x1800ec3c},
				{878, 260232, 0x1800ebc3},
				{879, 261088, 0x1800ec02},
				{880, 261375, 0x1800ebb6},
				{881, 262144, 0x1800ebdf},
				{882, 262808, 0x1800ebee},
				{883, 263025, 0x1800eb92},
				{884, 262829, 0x1800ead2},
				{885, 263530, 0x1800eaea},
				{886, 263787, 0x1800ea97},
				{887, 263654, 0x1800e9e7},
				{888, 263620, 0x1800e94e},
				{889, 264365, 0x1800e971},
				{890, 264917, 0x1800e966},
				{891, 264742, 0x1800e8ac},
				{892, 265373, 0x1800e8b4},
				{893, 265080, 0x1800e7df},
				{894, 265543, 0x1800e7be},
				{895, 265337, 0x1800e6ff},
				{896, 265418, 0x1800e684},
				{897, 266029, 0x1800e686},
				{898, 265965, 0x1800e5e9},
				{899, 265892, 0x1800e54b},
				{900, 266131, 0x1800e4f6},
				{901, 266653, 0x1800e4e4},
				{902, 267080, 0x1800e4bb},
				{903, 266974, 0x1800e416},
				{904, 267091, 0x1800e3a5},
				{905, 267526, 0x1800e37e},
				{906, 267325, 0x1800e2c4},
				{907, 267711, 0x1800e291},
				{908, 268093, 0x1800e25f},
				{909, 268256, 0x1800e1f9},
				{910, 268156, 0x1800e158},
				{911, 268205, 0x1800e0d8},
				{912, 268587, 0x1800e0a6},
				{913, 268535, 0x1800e010},
				{914, 268789, 0x1800dfc0},
				{915, 268632, 0x1800df13},
				{916, 269428, 0x1800df40},
				{917, 269331, 0x1800dea0},
				{918, 269378, 0x1800de22},
				{919, 269204, 0x1800dd72},
				{920, 269099, 0x1800dcd2},
				{921, 269884, 0x1800dcfc},
				{922, 269947, 0x1800dc82},
				{923, 270104, 0x1800dc1e},
				{924, 270874, 0x1800dc45},
				{925, 270786, 0x1800dba9},
				{926, 271160, 0x1800db76},
				{927, 271362, 0x1800db1c},
				{928, 271659, 0x1800dad9},
				{929, 272018, 0x1800daa3},
				{930, 272433, 0x1800da79},
				{931, 273008, 0x1800da73},
				{932, 273056, 0x1800d9f8},
				{933, 273012, 0x1800d968},
				{934, 273362, 0x1800d931},
				{935, 273706, 0x1800d8f8},
				{936, 274465, 0x1800d91b},
				{937, 274703, 0x1800d8ca},
				{938, 275506, 0x1800d8f8},
				{939, 276199, 0x1800d90c},
				{940, 276687, 0x1800d8f3},
				{941, 277331, 0x1800d8fd},
				{942, 277822, 0x1800d8e5},
				{943, 277862, 0x1800d868},
				{944, 278099, 0x1800d818},
				{945, 277897, 0x1800d767},
				{946, 277826, 0x1800d6d2},
				{947, 278318, 0x1800d6ba},
				{948, 278221, 0x1800d621},
				{949, 278603, 0x1800d5f1},
				{950, 278768, 0x1800d592},
				{951, 279309, 0x1800d585},
				{952, 279862, 0x1800d57b},
				{953, 280580, 0x1800d594},
				{954, 280418, 0x1800d4ee},
				{955, 280759, 0x1800d4b5},
				{956, 281270, 0x1800d4a2},
				{957, 281030, 0x1800d3ea},
				{958, 281592, 0x1800d3e3},
				{959, 281971, 0x1800d3b3},
				{960, 282758, 0x1800d3db},
				{961, 282510, 0x1800d323},
				{962, 283338, 0x1800d355},
				{963, 284029, 0x1800d368},
				{964, 283894, 0x1800d2c9},
				{965, 284258, 0x1800d296},
				{966, 284011, 0x1800d1e0},
				{967, 284474, 0x1800d1c2},
				{968, 285234, 0x1800d1e4},
				{969, 285138, 0x1800d14e},
				{970, 285178, 0x1800d0d7},
				{971, 285432, 0x1800d08d},
				{972, 285202, 0x1800cfdb},
				{973, 285799, 0x1800cfdb},
				{974, 286639, 0x1800d00e},
				{975, 286862, 0x1800cfbe},
				{976, 287009, 0x1800cf5d},
				{977, 287866, 0x1800cf94},
				{978, 288151, 0x1800cf50},
				{979, 288077, 0x1800cec2},
				{980, 288608, 0x1800ceb3},
				{981, 289036, 0x1800ce8f},
				{982, 289751, 0x1800cea7},
				{983, 289547, 0x1800cdfd},
				{984, 289568, 0x1800cd82},
				{985, 289563, 0x1800cd03},
				{986, 290295, 0x1800cd1f},
				{987, 290354, 0x1800ccae},
				{988, 290267, 0x1800cc1d},
				{989, 290973, 0x1800cc33},
				{990, 290783, 0x1800cb8e},
				{991, 291640, 0x1800cbc4},
				{992, 292311, 0x1800cbd3},
				{993, 292068, 0x1800cb22},
				{994, 292691, 0x1800cb28},
				{995, 293156, 0x1800cb0c},
				{996, 292897, 0x1800ca59},
				{997, 293056, 0x1800c9fd},
				{998, 293558, 0x1800c9e9},
				{999, 293356, 0x1800c943},
				{1000, 293956, 0x1800c943},
				{1001, 293713, 0x1800c896},
			},
		},
		// Each block 1 second before the previous one
		{
			run:          12,
			anchorHeight: 1,
			anchorTime:   10000,
			anchorBits:   0x1802aee8,
			startHeight:  2,
			startTime:    11200,
			blocks: []block{
				{2, 11200, 0x1802aee8},
				{3, 11199, 0x1802ad44},
				{4, 11198, 0x1802ab9e},
				{5, 11197, 0x1802a9f9},
				{6, 11196, 0x1802a855},
				{7, 11195, 0x1802a6b3},
				{8, 11194, 0x1802a511},
				{9, 11193, 0x1802a371},
				{10, 11192, 0x1802a1d1},
				{11, 11191, 0x1802a033},
				{12, 11190, 0x18029e94},
				{13, 11189, 0x18029cf8},
				{14, 11188, 0x18029b5d},
				{15, 11187, 0x180299c1},
				{16, 11186, 0x18029828},
				{17, 11185, 0x1802968e},
				{18, 11184, 0x180294f8},
				{19, 11183, 0x18029361},
				{20, 11182, 0x180291cb},
				{21, 11181, 0x18029036},
				{22, 11180, 0x18028ea2},
				{23, 11179, 0x18028d0f},
				{24, 11178, 0x18028b7d},
				{25, 11177, 0x180289ec},
				{26, 11176, 0x1802885b},
				{27, 11175, 0x180286cb},
				{28, 11174, 0x1802853c},
				{29, 11173, 0x180283af},
				{30, 11172, 0x18028222},
				{31, 11171, 0x18028096},
				{32, 11170, 0x18027f0c},
				{33, 11169, 0x18027d84},
				{34, 11168, 0x18027bfb},
				{35, 11167, 0x18027a73},
				{36, 11166, 0x180278ec},
				{37, 11165, 0x18027765},
				{38, 11164, 0x180275e0},
				{39, 11163, 0x1802745c},
				{40, 11162, 0x180272d9},
				{41, 11161, 0x18027156},
				{42, 11160, 0x18026fd5},
				{43, 11159, 0x18026e54},
				{44, 11158, 0x18026cd4},
				{45, 11157, 0x18026b56},
				{46, 11156, 0x180269d8},
				{47, 11155, 0x1802685b},
				{48, 11154, 0x180266df},
				{49, 11153, 0x18026565},
				{50, 11152, 0x180263ea},
				{51, 11151, 0x18026271},
				{52, 11150, 0x180260f8},
				{53, 11149, 0x18025f81},
				{54, 11148, 0x18025e0a},
				{55, 11147, 0x18025c94},
				{56, 11146, 0x18025b1f},
				{57, 11145, 0x180259ab},
				{58, 11144, 0x18025838},
				{59, 11143, 0x180256c6},
				{60, 11142, 0x18025555},
				{61, 11141, 0x180253e4},
				{62, 11140, 0x18025274},
				{63, 11139, 0x18025104},
				{64, 11138, 0x18024f99},
				{65, 11137, 0x18024e2c},
				{66, 11136, 0x18024cbf},
				{67, 11135, 0x18024b55},
				{68, 11134, 0x180249e9},
				{69, 11133, 0x18024880},
				{70, 11132, 0x18024717},
				{71, 11131, 0x180245b0},
				{72, 11130, 0x18024448},
				{73, 11129, 0x180242e2},
				{74, 11128, 0x1802417d},
				{75, 11127, 0x18024018},
				{76, 11126, 0x18023eb5},
				{77, 11125, 0x18023d53},
				{78, 11124, 0x18023bf1},
				{79, 11123, 0x18023a91},
				{80, 11122, 0x18023932},
				{81, 11121, 0x180237d2},
				{82, 11120, 0x18023674},
				{83, 11119, 0x18023516},
				{84, 11118, 0x180233b9},
				{85, 11117, 0x1802325d},
				{86, 11116, 0x18023102},
				{87, 11115, 0x18022fa8},
				{88, 11114, 0x18022e4e},
				{89, 11113, 0x18022cf5},
				{90, 11112, 0x18022b9d},
				{91, 11111, 0x18022a47},
				{92, 11110, 0x180228f1},
				{93, 11109, 0x1802279b},
				{94, 11108, 0x18022646},
				{95, 11107, 0x180224f4},
				{96, 11106, 0x180223a2},
				{97, 11105, 0x18022250},
				{98, 11104, 0x180220fe},
				{99, 11103, 0x18021faf},
				{100, 11102, 0x18021e5e},
				{101, 11101, 0x18021d10},
				{102, 11100, 0x18021bc2},
				{103, 11099, 0x18021a75},
				{104, 11098, 0x18021928},
				{105, 11097, 0x180217dd},
				{106, 11096, 0x18021691},
				{107, 11095, 0x18021547},
				{108, 11094, 0x180213ff},
				{109, 11093, 0x180212b6},
				{110, 11092, 0x18021170},
				{111, 11091, 0x1802102a},
				{112, 11090, 0x18020ee4},
				{113, 11089, 0x18020d9e},
				{114, 11088, 0x18020c59},
				{115, 11087, 0x18020b16},
				{116, 11086, 0x180209d4},
				{117, 11085, 0x18020892},
				{118, 11084, 0x18020750},
				{119, 11083, 0x1802060f},
				{120, 11082, 0x180204d0},
				{121, 11081, 0x18020391},
				{122, 11080, 0x18020253},
				{123, 11079, 0x18020116},
				{124, 11078, 0x1801ffda},
				{125, 11077, 0x1801fe9e},
				{126, 11076, 0x1801fd63},
				{127, 11075, 0x1801fc29},
				{128, 11074, 0x1801faef},
				{129, 11073, 0x1801f9b7},
				{130, 11072, 0x1801f87f},
				{131, 11071, 0x1801f748},
				{132, 11070, 0x1801f611},
				{133, 11069, 0x1801f4db},
				{134, 11068, 0x1801f3a6},
				{135, 11067, 0x1801f272},
				{136, 11066, 0x1801f13e},
				{137, 11065, 0x1801f00b},
				{138, 11064, 0x1801eed9},
				{139, 11063, 0x1801eda9},
				{140, 11062, 0x1801ec78},
				{141, 11061, 0x1801eb49},
				{142, 11060, 0x1801ea1a},
				{143, 11059, 0x1801e8ec},
				{144, 11058, 0x1801e7be},
				{145, 11057, 0x1801e692},
				{146, 11056, 0x1801e565},
				{147, 11055, 0x1801e43a},
				{148, 11054, 0x1801e310},
				{149, 11053, 0x1801e1e6},
				{150, 11052, 0x1801e0bc},
				{151, 11051, 0x1801df94},
				{152, 11050, 0x1801de6d},
				{153, 11049, 0x1801dd46},
				{154, 11048, 0x1801dc1e},
				{155, 11047, 0x1801dafa},
				{156, 11046, 0x1801d9d6},
				{157, 11045, 0x1801d8b1},
				{158, 11044, 0x1801d78e},
				{159, 11043, 0x1801d66c},
				{160, 11042, 0x1801d549},
				{161, 11041, 0x1801d429},
				{162, 11040, 0x1801d308},
				{163, 11039, 0x1801d1e8},
				{164, 11038, 0x1801d0c9},
				{165, 11037, 0x1801cfa9},
				{166, 11036, 0x1801ce8c},
				{167, 11035, 0x1801cd6f},
				{168, 11034, 0x1801cc53},
				{169, 11033, 0x1801cb36},
				{170, 11032, 0x1801ca1b},
				{171, 11031, 0x1801c903},
				{172, 11030, 0x1801c7e9},
				{173, 11029, 0x1801c6cf},
				{174, 11028, 0x1801c5b7},
				{175, 11027, 0x1801c4a0},
				{176, 11026, 0x1801c389},
				{177, 11025, 0x1801c273},
				{178, 11024, 0x1801c15e},
				{179, 11023, 0x1801c048},
				{180, 11022, 0x1801bf35},
				{181, 11021, 0x1801be20},
				{182, 11020, 0x1801bd0d},
				{183, 11019, 0x1801bbfc},
				{184, 11018, 0x1801baea},
				{185, 11017, 0x1801b9da},
				{186, 11016, 0x1801b8c9},
				{187, 11015, 0x1801b7ba},
				{188, 11014, 0x1801b6ab},
				{189, 11013, 0x1801b59e},
				{190, 11012, 0x1801b490},
				{191, 11011, 0x1801b382},
				{192, 11010, 0x1801b276},
				{193, 11009, 0x1801b16b},
				{194, 11008, 0x1801b060},
				{195, 11007, 0x1801af55},
				{196, 11006, 0x1801ae4b},
				{197, 11005, 0x1801ad43},
				{198, 11004, 0x1801ac3b},
				{199, 11003, 0x1801ab32},
				{200, 11002, 0x1801aa2c},
				{201, 11001, 0x1801a925},
				{202, 11000, 0x1801a820},
				{203, 10999, 0x1801a71b},
				{204, 10998, 0x1801a616},
				{205, 10997, 0x1801a512},
				{206, 10996, 0x1801a40f},
				{207, 10995, 0x1801a30c},
				{208, 10994, 0x1801a20b},
				{209, 10993, 0x1801a109},
				{210, 10992, 0x1801a009},
				{211, 10991, 0x18019f09},
				{212, 10990, 0x18019e08},
				{213, 10989, 0x18019d09},
				{214, 10988, 0x18019c0b},
				{215, 10987, 0x18019b0d},
				{216, 10986, 0x18019a11},
				{217, 10985, 0x18019915},
				{218, 10984, 0x18019818},
				{219, 10983, 0x1801971d},
				{220, 10982, 0x18019623},
				{221, 10981, 0x18019529},
				{222, 10980, 0x1801942f},
				{223, 10979, 0x18019336},
				{224, 10978, 0x1801923e},
				{225, 10977, 0x18019146},
				{226, 10976, 0x1801904f},
				{227, 10975, 0x18018f59},
				{228, 10974, 0x18018e62},
				{229, 10973, 0x18018d6d},
				{230, 10972, 0x18018c79},
				{231, 10971, 0x18018b84},
				{232, 10970, 0x18018a90},
				{233, 10969, 0x1801899f},
				{234, 10968, 0x180188ac},
				{235, 10967, 0x180187ba},
				{236, 10966, 0x180186c9},
				{237, 10965, 0x180185d9},
				{238, 10964, 0x180184e9},
				{239, 10963, 0x180183f9},
				{240, 10962, 0x1801830a},
				{241, 10961, 0x1801821b},
				{242, 10960, 0x1801812d},
				{243, 10959, 0x18018040},
				{244, 10958, 0x18017f54},
				{245, 10957, 0x18017e68},
				{246, 10956, 0x18017d7c},
				{247, 10955, 0x18017c91},
				{248, 10954, 0x18017ba7},
				{249, 10953, 0x18017abd},
				{250, 10952, 0x180179d3},
				{251, 10951, 0x180178eb},
				{252, 10950, 0x18017803},
				{253, 10949, 0x1801771b},
				{254, 10948, 0x18017634},
				{255, 10947, 0x1801754d},
				{256, 10946, 0x18017467},
				{257, 10945, 0x18017381},
				{258, 10944, 0x1801729c},
				{259, 10943, 0x180171b8},
				{260, 10942, 0x180170d4},
				{261, 10941, 0x18016ff0},
				{262, 10940, 0x18016f0d},
				{263, 10939, 0x18016e2b},
				{264, 10938, 0x18016d4a},
				{265, 10937, 0x18016c69},
				{266, 10936, 0x18016b87},
				{267, 10935, 0x18016aa7},
				{268, 10934, 0x180169c7},
				{269, 10933, 0x180168e8},
				{270, 10932, 0x1801680a},
				{271, 10931, 0x1801672b},
				{272, 10930, 0x1801664e},
				{273, 10929, 0x18016570},
				{274, 10928, 0x18016494},
				{275, 10927, 0x180163b8},
				{276, 10926, 0x180162dc},
				{277, 10925, 0x18016202},
				{278, 10924, 0x18016127},
				{279, 10923, 0x1801604d},
				{280, 10922, 0x18015f73},
				{281, 10921, 0x18015e9b},
				{282, 10920, 0x18015dc1},
				{283, 10919, 0x18015ce9},
				{284, 10918, 0x18015c11},
				{285, 10917, 0x18015b3b},
				{286, 10916, 0x18015a64},
				{287, 10915, 0x1801598d},
				{288, 10914, 0x180158b8},
				{289, 10913, 0x180157e3},
				{290, 10912, 0x1801570f},
				{291, 10911, 0x1801563c},
				{292, 10910, 0x18015569},
				{293, 10909, 0x18015497},
				{294, 10908, 0x180153c6},
				{295, 10907, 0x180152f6},
				{296, 10906, 0x18015225},
				{297, 10905, 0x18015154},
				{298, 10904, 0x18015085},
				{299, 10903, 0x18014fb6},
				{300, 10902, 0x18014ee7},
				{301, 10901, 0x18014e19},
				{302, 10900, 0x18014d4b},
				{303, 10899, 0x18014c7e},
				{304, 10898, 0x18014bb2},
				{305, 10897, 0x18014ae6},
				{306, 10896, 0x18014a1a},
				{307, 10895, 0x1801494e},
				{308, 10894, 0x18014884},
				{309, 10893, 0x180147ba},
				{310, 10892, 0x180146f0},
				{311, 10891, 0x18014627},
				{312, 10890, 0x1801455e},
				{313, 10889, 0x18014496},
				{314, 10888, 0x180143ce},
				{315, 10887, 0x18014306},
				{316, 10886, 0x1801423f},
				{317, 10885, 0x18014179},
				{318, 10884, 0x180140b2},
				{319, 10883, 0x18013fed},
				{320, 10882, 0x18013f28},
				{321, 10881, 0x18013e63},
				{322, 10880, 0x18013d9f},
				{323, 10879, 0x18013cdb},
				{324, 10878, 0x18013c18},
				{325, 10877, 0x18013b56},
				{326, 10876, 0x18013a93},
				{327, 10875, 0x180139d1},
				{328, 10874, 0x1801390f},
				{329, 10873, 0x1801384f},
				{330, 10872, 0x1801378e},
				{331, 10871, 0x180136ce},
				{332, 10870, 0x1801360e},
				{333, 10869, 0x1801354f},
				{334, 10868, 0x18013490},
				{335, 10867, 0x180133d2},
				{336, 10866, 0x18013314},
				{337, 10865, 0x18013257},
				{338, 10864, 0x1801319a},
				{339, 10863, 0x180130dd},
				{340, 10862, 0x18013022},
				{341, 10861, 0x18012f66},
				{342, 10860, 0x18012eab},
				{343, 10859, 0x18012df1},
				{344, 10858, 0x18012d36},
				{345, 10857, 0x18012c7c},
				{346, 10856, 0x18012bc3},
				{347, 10855, 0x18012b0a},
				{348, 10854, 0x18012a52},
				{349, 10853, 0x1801299a},
				{350, 10852, 0x180128e2},
				{351, 10851, 0x1801282b},
				{352, 10850, 0x18012774},
				{353, 10849, 0x180126be},
				{354, 10848, 0x18012608},
				{355, 10847, 0x18012553},
				{356, 10846, 0x1801249e},
				{357, 10845, 0x180123e9},
				{358, 10844, 0x18012336},
				{359, 10843, 0x18012282},
				{360, 10842, 0x180121cf},
				{361, 10841, 0x1801211c},
				{362, 10840, 0x18012069},
				{363, 10839, 0x18011fb7},
				{364, 10838, 0x18011f06},
				{365, 10837, 0x18011e55},
				{366, 10836, 0x18011da3},
				{367, 10835, 0x18011cf3},
				{368, 10834, 0x18011c43},
				{369, 10833, 0x18011b94},
				{370, 10832, 0x18011ae5},
				{371, 10831, 0x18011a37},
				{372, 10830, 0x18011989},
				{373, 10829, 0x180118db},
				{374, 10828, 0x1801182e},
				{375, 10827, 0x18011781},
				{376, 10826, 0x180116d4},
				{377, 10825, 0x18011628},
				{378, 10824, 0x1801157d},
				{379, 10823, 0x180114d2},
				{380, 10822, 0x18011426},
				{381, 10821, 0x1801137c},
				{382, 10820, 0x180112d2},
				{383, 10819, 0x18011228},
				{384, 10818, 0x1801117f},
				{385, 10817, 0x180110d6},
				{386, 10816, 0x1801102e},
				{387, 10815, 0x18010f87},
				{388, 10814, 0x18010edf},
				{389, 10813, 0x18010e38},
				{390, 10812, 0x18010d91},
				{391, 10811, 0x18010cea},
				{392, 10810, 0x18010c45},
				{393, 10809, 0x18010b9f},
				{394, 10808, 0x18010afa},
				{395, 10807, 0x18010a55},
				{396, 10806, 0x180109b1},
				{397, 10805, 0x1801090c},
				{398, 10804, 0x18010869},
				{399, 10803, 0x180107c6},
				{400, 10802, 0x18010723},
				{401, 10801, 0x18010682},
				{402, 10800, 0x180105df},
				{403, 10799, 0x1801053e},
				{404, 10798, 0x1801049d},
				{405, 10797, 0x180103fc},
				{406, 10796, 0x1801035b},
				{407, 10795, 0x180102bb},
				{408, 10794, 0x1801021b},
				{409, 10793, 0x1801017c},
				{410, 10792, 0x180100dd},
				{411, 10791, 0x1801003f},
				{412, 10790, 0x1800ffa1},
				{413, 10789, 0x1800ff03},
				{414, 10788, 0x1800fe65},
				{415, 10787, 0x1800fdc8},
				{416, 10786, 0x1800fd2c},
				{417, 10785, 0x1800fc91},
				{418, 10784, 0x1800fbf5},
				{419, 10783, 0x1800fb59},
				{420, 10782, 0x1800fabe},
				{421, 10781, 0x1800fa23},
				{422, 10780, 0x1800f989},
				{423, 10779, 0x1800f8ef},
				{424, 10778, 0x1800f856},
				{425, 10777, 0x1800f7bc},
				{426, 10776, 0x1800f723},
				{427, 10775, 0x1800f68b},
				{428, 10774, 0x1800f5f3},
				{429, 10773, 0x1800f55b},
				{430, 10772, 0x1800f4c4},
				{431, 10771, 0x1800f42d},
				{432, 10770, 0x1800f397},
				{433, 10769, 0x1800f301},
				{434, 10768, 0x1800f26b},
				{435, 10767, 0x1800f1d6},
				{436, 10766, 0x1800f140},
				{437, 10765, 0x1800f0ac},
				{438, 10764, 0x1800f017},
				{439, 10763, 0x1800ef83},
				{440, 10762, 0x1800eeef},
				{441, 10761, 0x1800ee5c},
				{442, 10760, 0x1800edc9},
				{443, 10759, 0x1800ed36},
				{444, 10758, 0x1800eca4},
				{445, 10757, 0x1800ec13},
				{446, 10756, 0x1800eb81},
				{447, 10755, 0x1800eaf0},
				{448, 10754, 0x1800ea5f},
				{449, 10753, 0x1800e9cf},
				{450, 10752, 0x1800e93f},
				{451, 10751, 0x1800e8af},
				{452, 10750, 0x1800e820},
				{453, 10749, 0x1800e790},
				{454, 10748, 0x1800e702},
				{455, 10747, 0x1800e673},
				{456, 10746, 0x1800e5e5},
				{457, 10745, 0x1800e557},
				{458, 10744, 0x1800e4ca},
				{459, 10743, 0x1800e43d},
				{460, 10742, 0x1800e3b1},
				{461, 10741, 0x1800e324},
				{462, 10740, 0x1800e298},
				{463, 10739, 0x1800e20d},
				{464, 10738, 0x1800e182},
				{465, 10737, 0x1800e0f7},
				{466, 10736, 0x1800e06c},
				{467, 10735, 0x1800dfe2},
				{468, 10734, 0x1800df58},
				{469, 10733, 0x1800dece},
				{470, 10732, 0x1800de45},
				{471, 10731, 0x1800ddbc},
				{472, 10730, 0x1800dd34},
				{473, 10729, 0x1800dcac},
				{474, 10728, 0x1800dc23},
				{475, 10727, 0x1800db9c},
				{476, 10726, 0x1800db14},
				{477, 10725, 0x1800da8d},
				{478, 10724, 0x1800da07},
				{479, 10723, 0x1800d981},
				{480, 10722, 0x1800d8fb},
				{481, 10721, 0x1800d876},
				{482, 10720, 0x1800d7f0},
				{483, 10719, 0x1800d76b},
				{484, 10718, 0x1800d6e6},
				{485, 10717, 0x1800d662},
				{486, 10716, 0x1800d5de},
				{487, 10715, 0x1800d55a},
				{488, 10714, 0x1800d4d6},
				{489, 10713, 0x1800d454},
				{490, 10712, 0x1800d3d1},
				{491, 10711, 0x1800d34f},
				{492, 10710, 0x1800d2cd},
				{493, 10709, 0x1800d24b},
				{494, 10708, 0x1800d1ca},
				{495, 10707, 0x1800d148},
				{496, 10706, 0x1800d0c7},
				{497, 10705, 0x1800d047},
				{498, 10704, 0x1800cfc6},
				{499, 10703, 0x1800cf46},
				{500, 10702, 0x1800cec7},
				{501, 10701, 0x1800ce47},
				{502, 10700, 0x1800cdc9},
				{503, 10699, 0x1800cd4a},
				{504, 10698, 0x1800cccb},
				{505, 10697, 0x1800cc4d},
				{506, 10696, 0x1800cbd0},
				{507, 10695, 0x1800cb52},
				{508, 10694, 0x1800cad5},
				{509, 10693, 0x1800ca59},
				{510, 10692, 0x1800c9dc},
				{511, 10691, 0x1800c960},
				{512, 10690, 0x1800c8e4},
				{513, 10689, 0x1800c867},
				{514, 10688, 0x1800c7ec},
				{515, 10687, 0x1800c771},
				{516, 10686, 0x1800c6f7},
				{517, 10685, 0x1800c67c},
				{518, 10684, 0x1800c602},
				{519, 10683, 0x1800c588},
				{520, 10682, 0x1800c50e},
				{521, 10681, 0x1800c495},
				{522, 10680, 0x1800c41b},
				{523, 10679, 0x1800c3a3},
				{524, 10678, 0x1800c32b},
				{525, 10677, 0x1800c2b2},
				{526, 10676, 0x1800c23a},
				{527, 10675, 0x1800c1c3},
				{528, 10674, 0x1800c14c},
				{529, 10673, 0x1800c0d4},
				{530, 10672, 0x1800c05d},
				{531, 10671, 0x1800bfe7},
				{532, 10670, 0x1800bf71},
				{533, 10669, 0x1800befb},
				{534, 10668, 0x1800be85},
				{535, 10667, 0x1800be10},
				{536, 10666, 0x1800bd9a},
				{537, 10665, 0x1800bd26},
				{538, 10664, 0x1800bcb2},
				{539, 10663, 0x1800bc3e},
				{540, 10662, 0x1800bbca},
				{541, 10661, 0x1800bb55},
				{542, 10660, 0x1800bae2},
				{543, 10659, 0x1800ba6f},
				{544, 10658, 0x1800b9fc},
				{545, 10657, 0x1800b989},
				{546, 10656, 0x1800b917},
				{547, 10655, 0x1800b8a5},
				{548, 10654, 0x1800b833},
				{549, 10653, 0x1800b7c1},
				{550, 10652, 0x1800b750},
				{551, 10651, 0x1800b6df},
				{552, 10650, 0x1800b66e},
				{553, 10649, 0x1800b5fe},
				{554, 10648, 0x1800b58e},
				{555, 10647, 0x1800b51e},
				{556, 10646, 0x1800b4ae},
				{557, 10645, 0x1800b43f},
				{558, 10644, 0x1800b3d0},
				{559, 10643, 0x1800b360},
				{560, 10642, 0x1800b2f2},
				{561, 10641, 0x1800b284},
				{562, 10640, 0x1800b215},
				{563, 10639, 0x1800b1a7},
				{564, 10638, 0x1800b13a},
				{565, 10637, 0x1800b0cc},
				{566, 10636, 0x1800b05f},
				{567, 10635, 0x1800aff1},
				{568, 10634, 0x1800af85},
				{569, 10633, 0x1800af19},
				{570, 10632, 0x1800aead},
				{571, 10631, 0x1800ae41},
				{572, 10630, 0x1800add5},
				{573, 10629, 0x1800ad69},
				{574, 10628, 0x1800acfe},
				{575, 10627, 0x1800ac94},
				{576, 10626, 0x1800ac29},
				{577, 10625, 0x1800abbe},
				{578, 10624, 0x1800ab55},
				{579, 10623, 0x1800aaeb},
				{580, 10622, 0x1800aa82},
				{581, 10621, 0x1800aa19},
				{582, 10620, 0x1800a9b1},
				{583, 10619, 0x1800a948},
				{584, 10618, 0x1800a8e0},
				{585, 10617, 0x1800a878},
				{586, 10616, 0x1800a811},
				{587, 10615, 0x1800a7a9},
				{588, 10614, 0x1800a742},
				{589, 10613, 0x1800a6db},
				{590, 10612, 0x1800a675},
				{591, 10611, 0x1800a60e},
				{592, 10610, 0x1800a5a8},
				{593, 10609, 0x1800a542},
				{594, 10608, 0x1800a4dc},
				{595, 10607, 0x1800a476},
				{596, 10606, 0x1800a411},
				{597, 10605, 0x1800a3ac},
				{598, 10604, 0x1800a347},
				{599, 10603, 0x1800a2e3},
				{600, 10602, 0x1800a27e},
				{601, 10601, 0x1800a21b},
				{602, 10600, 0x1800a1b7},
				{603, 10599, 0x1800a153},
				{604, 10598, 0x1800a0f0},
				{605, 10597, 0x1800a08d},
				{606, 10596, 0x1800a02a},
				{607, 10595, 0x18009fc7},
				{608, 10594, 0x18009f64},
				{609, 10593, 0x18009f02},
				{610, 10592, 0x18009ea0},
				{611, 10591, 0x18009e3f},
				{612, 10590, 0x18009ddd},
				{613, 10589, 0x18009d7c},
				{614, 10588, 0x18009d1a},
				{615, 10587, 0x18009cba},
				{616, 10586, 0x18009c5a},
				{617, 10585, 0x18009bf9},
				{618, 10584, 0x18009b99},
				{619, 10583, 0x18009b39},
				{620, 10582, 0x18009ad9},
				{621, 10581, 0x18009a7a},
				{622, 10580, 0x18009a1a},
				{623, 10579, 0x180099bb},
				{624, 10578, 0x1800995c},
				{625, 10577, 0x180098fe},
				{626, 10576, 0x180098a0},
				{627, 10575, 0x18009841},
				{628, 10574, 0x180097e3},
				{629, 10573, 0x18009786},
				{630, 10572, 0x18009728},
				{631, 10571, 0x180096cb},
				{632, 10570, 0x1800966f},
				{633, 10569, 0x18009612},
				{634, 10568, 0x180095b5},
				{635, 10567, 0x18009559},
				{636, 10566, 0x180094fd},
				{637, 10565, 0x180094a1},
				{638, 10564, 0x18009445},
				{639, 10563, 0x180093ea},
				{640, 10562, 0x1800938e},
				{641, 10561, 0x18009333},
				{642, 10560, 0x180092d8},
				{643, 10559, 0x1800927e},
				{644, 10558, 0x18009223},
				{645, 10557, 0x180091c9},
				{646, 10556, 0x1800916f},
				{647, 10555, 0x18009116},
				{648, 10554, 0x180090bc},
				{649, 10553, 0x18009063},
				{650, 10552, 0x1800900a},
				{651, 10551, 0x18008fb1},
				{652, 10550, 0x18008f58},
				{653, 10549, 0x18008f00},
				{654, 10548, 0x18008ea8},
				{655, 10547, 0x18008e4f},
				{656, 10546, 0x18008df7},
				{657, 10545, 0x18008da0},
				{658, 10544, 0x18008d48},
				{659, 10543, 0x18008cf1},
				{660, 10542, 0x18008c9a},
				{661, 10541, 0x18008c43},
				{662, 10540, 0x18008bed},
				{663, 10539, 0x18008b97},
				{664, 10538, 0x18008b41},
				{665, 10537, 0x18008aeb},
				{666, 10536, 0x18008a95},
				{667, 10535, 0x18008a40},
				{668, 10534, 0x180089ea},
				{669, 10533, 0x18008995},
				{670, 10532, 0x18008940},
				{671, 10531, 0x180088ec},
				{672, 10530, 0x18008897},
				{673, 10529, 0x18008842},
				{674, 10528, 0x180087ef},
				{675, 10527, 0x1800879a},
				{676, 10526, 0x18008747},
				{677, 10525, 0x180086f3},
				{678, 10524, 0x180086a1},
				{679, 10523, 0x1800864d},
				{680, 10522, 0x180085fb},
				{681, 10521, 0x180085a8},
				{682, 10520, 0x18008555},
				{683, 10519, 0x18008503},
				{684, 10518, 0x180084b1},
				{685, 10517, 0x1800845f},
				{686, 10516, 0x1800840d},
				{687, 10515, 0x180083bc},
				{688, 10514, 0x1800836a},
				{689, 10513, 0x18008319},
				{690, 10512, 0x180082c8},
				{691, 10511, 0x18008278},
				{692, 10510, 0x18008227},
				{693, 10509, 0x180081d7},
				{694, 10508, 0x18008187},
				{695, 10507, 0x18008137},
				{696, 10506, 0x180080e8},
				{697, 10505, 0x18008098},
				{698, 10504, 0x18008049},
				{699, 10503, 0x177ff9df},
				{700, 10502, 0x177faab7},
				{701, 10501, 0x177f5c3b},
				{702, 10500, 0x177f0d69},
				{703, 10499, 0x177ebeed},
				{704, 10498, 0x177e70c7},
				{705, 10497, 0x177e22f7},
				{706, 10496, 0x177dd527},
				{707, 10495, 0x177d8756},
				{708, 10494, 0x177d3a32},
				{709, 10493, 0x177ced0d},
				{710, 10492, 0x177c9fe9},
				{711, 10491, 0x177c531a},
				{712, 10490, 0x177c064b},
				{713, 10489, 0x177bb9d2},
				{714, 10488, 0x177b6db0},
				{715, 10487, 0x177b218d},
				{716, 10486, 0x177ad56a},
				{717, 10485, 0x177a899c},
				{718, 10484, 0x177a3e25},
				{719, 10483, 0x1779f2ae},
				{720, 10482, 0x1779a78d},
				{721, 10481, 0x17795c6b},
				{722, 10480, 0x177911a0},
				{723, 10479, 0x1778c6d4},
				{724, 10478, 0x17787cb4},
				{725, 10477, 0x17783295},
				{726, 10476, 0x1777e875},
				{727, 10475, 0x17779e55},
				{728, 10474, 0x1777548b},
				{729, 10473, 0x17770b17},
				{730, 10472, 0x1776c1a3},
				{731, 10471, 0x17767885},
				{732, 10470, 0x17762f67},
				{733, 10469, 0x1775e69e},
				{734, 10468, 0x17759dd6},
				{735, 10467, 0x17755564},
				{736, 10466, 0x17750cf1},
				{737, 10465, 0x1774c4d4},
				{738, 10464, 0x17747cb8},
				{739, 10463, 0x17743547},
				{740, 10462, 0x1773edd6},
				{741, 10461, 0x1773a665},
				{742, 10460, 0x17735ef4},
				{743, 10459, 0x177317d9},
				{744, 10458, 0x1772d114},
				{745, 10457, 0x17728a4f},
				{746, 10456, 0x1772438a},
				{747, 10455, 0x1771fd1b},
				{748, 10454, 0x1771b701},
				{749, 10453, 0x177170e8},
				{750, 10452, 0x17712ace},
				{751, 10451, 0x1770e50b},
				{752, 10450, 0x17709f9d},
				{753, 10449, 0x17705a2f},
				{754, 10448, 0x17701517},
				{755, 10447, 0x176fd056},
				{756, 10446, 0x176f8b3e},
				{757, 10445, 0x176f467c},
				{758, 10444, 0x176f0210},
				{759, 10443, 0x176ebda3},
				{760, 10442, 0x176e7937},
				{761, 10441, 0x176e3577},
				{762, 10440, 0x176df161},
				{763, 10439, 0x176dada0},
				{764, 10438, 0x176d6a36},
				{765, 10437, 0x176d26cb},
				{766, 10436, 0x176ce361},
				{767, 10435, 0x176ca04c},
				{768, 10434, 0x176c5d8d},
				{769, 10433, 0x176c1acf},
				{770, 10432, 0x176bd866},
				{771, 10431, 0x176b95fd},
				{772, 10430, 0x176b5394},
				{773, 10429, 0x176b1181},
				{774, 10428, 0x176acfc4},
				{775, 10427, 0x176a8e06},
				{776, 10426, 0x176a4c49},
				{777, 10425, 0x176a0ae2},
				{778, 10424, 0x1769c97a},
				{779, 10423, 0x17698869},
				{780, 10422, 0x17694757},
				{781, 10421, 0x1769069c},
				{782, 10420, 0x1768c5e0},
				{783, 10419, 0x17688525},
				{784, 10418, 0x17684515},
				{785, 10417, 0x17680505},
				{786, 10416, 0x1767c4f5},
				{787, 10415, 0x176784e5},
				{788, 10414, 0x1767452b},
				{789, 10413, 0x176705c7},
				{790, 10412, 0x1766c663},
				{791, 10411, 0x176686ff},
				{792, 10410, 0x176647f0},
				{793, 10409, 0x176608e2},
				{794, 10408, 0x1765c9d4},
				{795, 10407, 0x17658b71},
				{796, 10406, 0x17654cb9},
				{797, 10405, 0x17650e56},
				{798, 10404, 0x1764d049},
				{799, 10403, 0x176491e7},
				{800, 10402, 0x17645486},
				{801, 10401, 0x17641679},
				{802, 10400, 0x1763d918},
				{803, 10399, 0x17639b61},
				{804, 10398, 0x17635e00},
				{805, 10397, 0x176320f5},
				{806, 10396, 0x1762e3ea},
				{807, 10395, 0x1762a6df},
				{808, 10394, 0x17626a2a},
				{809, 10393, 0x17622d74},
				{810, 10392, 0x1761f115},
				{811, 10391, 0x1761b4b6},
				{812, 10390, 0x176178ac},
				{813, 10389, 0x17613ca3},
				{814, 10388, 0x17610099},
				{815, 10387, 0x1760c4e5},
				{816, 10386, 0x17608988},
				{817, 10385, 0x17604e2a},
				{818, 10384, 0x176012cc},
				{819, 10383, 0x175fd76e},
				{820, 10382, 0x175f9c66},
				{821, 10381, 0x175f61b4},
				{822, 10380, 0x175f26ac},
				{823, 10379, 0x175eec50},
				{824, 10378, 0x175eb19e},
				{825, 10377, 0x175e7742},
				{826, 10376, 0x175e3d3b},
				{827, 10375, 0x175e02df},
				{828, 10374, 0x175dc92f},
				{829, 10373, 0x175d8f28},
				{830, 10372, 0x175d5578},
				{831, 10371, 0x175d1c73},
				{832, 10370, 0x175ce2c3},
				{833, 10369, 0x175ca9be},
				{834, 10368, 0x175c7063},
				{835, 10367, 0x175c375f},
				{836, 10366, 0x175bfeb0},
				{837, 10365, 0x175bc601},
				{838, 10364, 0x175b8d52},
				{839, 10363, 0x175b54f9},
				{840, 10362, 0x175b1ca0},
				{841, 10361, 0x175ae447},
				{842, 10360, 0x175aac44},
				{843, 10359, 0x175a7440},
				{844, 10358, 0x175a3c93},
				{845, 10357, 0x175a04e6},
				{846, 10356, 0x1759cd8e},
				{847, 10355, 0x175995e1},
				{848, 10354, 0x17595e8a},
				{849, 10353, 0x17592788},
				{850, 10352, 0x1758f086},
				{851, 10351, 0x1758b985},
				{852, 10350, 0x175882d9},
				{853, 10349, 0x17584c2d},
				{854, 10348, 0x17581582},
				{855, 10347, 0x1757df2c},
				{856, 10346, 0x1757a8d6},
				{857, 10345, 0x17577280},
				{858, 10344, 0x17573c80},
				{859, 10343, 0x17570680},
				{860, 10342, 0x1756d0d6},
				{861, 10341, 0x17569b2c},
				{862, 10340, 0x175665d8},
				{863, 10339, 0x17563083},
				{864, 10338, 0x1755fb2f},
				{865, 10337, 0x1755c606},
				{866, 10336, 0x1755915d},
				{867, 10335, 0x17555cb5},
				{868, 10334, 0x17552837},
				{869, 10333, 0x1754f3ba},
				{870, 10332, 0x1754bf92},
				{871, 10331, 0x17548b6a},
				{872, 10330, 0x1754576d},
				{873, 10329, 0x17542370},
				{874, 10328, 0x1753ef9f},
				{875, 10327, 0x1753bbf8},
				{876, 10326, 0x1753887c},
				{877, 10325, 0x17535556},
				{878, 10324, 0x17532204},
				{879, 10323, 0x1752eede},
				{880, 10322, 0x1752bbb8},
				{881, 10321, 0x175288bd},
				{882, 10320, 0x175255ed},
				{883, 10319, 0x17522347},
				{884, 10318, 0x1751f0a2},
				{885, 10317, 0x1751be27},
				{886, 10316, 0x17518bd8},
				{887, 10315, 0x17515988},
				{888, 10314, 0x1751278f},
				{889, 10313, 0x1750f56a},
				{890, 10312, 0x1750c39c},
				{891, 10311, 0x175091cd},
				{892, 10310, 0x1750607f},
				{893, 10309, 0x17502edb},
				{894, 10308, 0x174ffd8d},
				{895, 10307, 0x174fcc3f},
				{896, 10306, 0x174f9af2},
				{897, 10305, 0x174f69f9},
				{898, 10304, 0x174f3901},
				{899, 10303, 0x174f0834},
				{900, 10302, 0x174ed767},
				{901, 10301, 0x174ea6c5},
				{902, 10300, 0x174e764e},
				{903, 10299, 0x174e4602},
				{904, 10298, 0x174e15b5},
				{905, 10297, 0x174de594},
				{906, 10296, 0x174db572},
				{907, 10295, 0x174d85a7},
				{908, 10294, 0x174d5606},
				{909, 10293, 0x174d263b},
				{910, 10292, 0x174cf6c5},
				{911, 10291, 0x174cc750},
				{912, 10290, 0x174c97da},
				{913, 10289, 0x174c68ba},
				{914, 10288, 0x174c399a},
				{915, 10287, 0x174c0a7a},
				{916, 10286, 0x174bdb86},
				{917, 10285, 0x174bacbc},
				{918, 10284, 0x174b7e1d},
				{919, 10283, 0x174b4f7e},
				{920, 10282, 0x174b210a},
				{921, 10281, 0x174af2c1},
				{922, 10280, 0x174ac477},
				{923, 10279, 0x174a9684},
				{924, 10278, 0x174a6891},
				{925, 10277, 0x174a3a9e},
				{926, 10276, 0x174a0cd5},
				{927, 10275, 0x1749df38},
				{928, 10274, 0x1749b19a},
				{929, 10273, 0x17498428},
				{930, 10272, 0x174956b5},
				{931, 10271, 0x1749296e},
				{932, 10270, 0x1748fc51},
				{933, 10269, 0x1748cf35},
				{934, 10268, 0x1748a243},
				{935, 10267, 0x1748757c},
				{936, 10266, 0x174848b6},
				{937, 10265, 0x17481c1a},
				{938, 10264, 0x1747efd4},
				{939, 10263, 0x1747c363},
				{940, 10262, 0x1747971d},
				{941, 10261, 0x17476b02},
				{942, 10260, 0x17473ee7},
				{943, 10259, 0x174712f7},
				{944, 10258, 0x1746e707},
				{945, 10257, 0x1746bb42},
				{946, 10256, 0x17468fa7},
				{947, 10255, 0x1746640d},
				{948, 10254, 0x1746389e},
				{949, 10253, 0x17460d59},
				{950, 10252, 0x1745e215},
				{951, 10251, 0x1745b6fc},
				{952, 10250, 0x17458be2},
				{953, 10249, 0x174560f4},
				{954, 10248, 0x1745365b},
				{955, 10247, 0x17450b97},
				{956, 10246, 0x1744e0ff},
				{957, 10245, 0x1744b666},
				{958, 10244, 0x17448bf9},
				{959, 10243, 0x174461b6},
				{960, 10242, 0x17443773},
				{961, 10241, 0x17440d5b},
				{962, 10240, 0x1743e36e},
				{963, 10239, 0x1743b981},
				{964, 10238, 0x17438fbf},
				{965, 10237, 0x174365fd},
				{966, 10236, 0x17433c66},
				{967, 10235, 0x174312cf},
				{968, 10234, 0x1742e98e},
				{969, 10233, 0x1742c04d},
				{970, 10232, 0x17429737},
				{971, 10231, 0x17426e20},
				{972, 10230, 0x1742450a},
				{973, 10229, 0x17421c1f},
				{974, 10228, 0x1741f35e},
				{975, 10227, 0x1741ca9e},
				{976, 10226, 0x1741a209},
				{977, 10225, 0x1741799e},
				{978, 10224, 0x17415134},
				{979, 10223, 0x174128c9},
				{980, 10222, 0x17410089},
				{981, 10221, 0x1740d875},
				{982, 10220, 0x1740b08b},
				{983, 10219, 0x174088a1},
				{984, 10218, 0x174060e2},
				{985, 10217, 0x17403924},
				{986, 10216, 0x17401190},
				{987, 10215, 0x173fe9fc},
				{988, 10214, 0x173fc293},
				{989, 10213, 0x173f9b2a},
				{990, 10212, 0x173f73ec},
				{991, 10211, 0x173f4cd9},
				{992, 10210, 0x173f25c6},
				{993, 10209, 0x173efede},
				{994, 10208, 0x173ed7f6},
				{995, 10207, 0x173eb138},
				{996, 10206, 0x173e8a7b},
				{997, 10205, 0x173e63e9},
				{998, 10204, 0x173e3d82},
				{999, 10203, 0x173e171a},
				{1000, 10202, 0x173df0de},
				{1001, 10201, 0x173dcaa1},
				{1002, 10200, 0x173da490},
				{1003, 10199, 0x173d7e7e},
				{1004, 10198, 0x173d5898},
				{1005, 10197, 0x173d32b1},
				{1006, 10196, 0x173d0cf6},
				{1007, 10195, 0x173ce765},
				{1008, 10194, 0x173cc1d4},
				{1009, 10193, 0x173c9c44},
				{1010, 10192, 0x173c7709},
				{1011, 10191, 0x173c51a3},
				{1012, 10190, 0x173c2c68},
				{1013, 10189, 0x173c0758},
				{1014, 10188, 0x173be248},
				{1015, 10187, 0x173bbd8e},
				{1016, 10186, 0x173b98aa},
				{1017, 10185, 0x173b73ef},
				{1018, 10184, 0x173b4f60},
				{1019, 10183, 0x173b2ad1},
				{1020, 10182, 0x173b0642},
				{1021, 10181, 0x173ae1de},
				{1022, 10180, 0x173abda5},
				{1023, 10179, 0x173a996c},
				{1024, 10178, 0x173a7532},
				{1025, 10177, 0x173a5124},
				{1026, 10176, 0x173a2d41},
				{1027, 10175, 0x173a095d},
				{1028, 10174, 0x1739e5a5},
				{1029, 10173, 0x1739c1ec},
				{1030, 10172, 0x17399e8a},
				{1031, 10171, 0x17397afc},
				{1032, 10170, 0x1739579a},
				{1033, 10169, 0x17393437},
				{1034, 10168, 0x17391100},
				{1035, 10167, 0x1738edc8},
				{1036, 10166, 0x1738cabb},
				{1037, 10165, 0x1738a7ae},
				{1038, 10164, 0x173884cd},
				{1039, 10163, 0x173861eb},
				{1040, 10162, 0x17383f34},
				{1041, 10161, 0x17381c7d},
				{1042, 10160, 0x1737f9f1},
				{1043, 10159, 0x1737d765},
				{1044, 10158, 0x1737b504},
				{1045, 10157, 0x173792ce},
				{1046, 10156, 0x17377098},
				{1047, 10155, 0x17374e8d},
				{1048, 10154, 0x17372c57},
				{1049, 10153, 0x17370a77},
				{1050, 10152, 0x1736e897},
				{1051, 10151, 0x1736c6b6},
				{1052, 10150, 0x1736a501},
				{1053, 10149, 0x1736834c},
				{1054, 10148, 0x173661c2},
				{1055, 10147, 0x17364037},
				{1056, 10146, 0x17361ed8},
				{1057, 10145, 0x1735fd79},
				{1058, 10144, 0x1735dc19},
				{1059, 10143, 0x1735bb10},
				{1060, 10142, 0x173599db},
				{1061, 10141, 0x173578fd},
				{1062, 10140, 0x1735581e},
				{1063, 10139, 0x1735373f},
				{1064, 10138, 0x1735168c},
				{1065, 10137, 0x1734f5d8},
				{1066, 10136, 0x1734d524},
				{1067, 10135, 0x1734b49c},
				{1068, 10134, 0x17349413},
				{1069, 10133, 0x173473b5},
				{1070, 10132, 0x17345382},
				{1071, 10131, 0x1734334f},
				{1072, 10130, 0x1734131c},
				{1073, 10129, 0x1733f314},
				{1074, 10128, 0x1733d30d},
				{1075, 10127, 0x1733b330},
				{1076, 10126, 0x1733937d},
				{1077, 10125, 0x173373a0},
				{1078, 10124, 0x173353ee},
				{1079, 10123, 0x17333467},
				{1080, 10122, 0x173314e0},
				{1081, 10121, 0x1732f559},
				{1082, 10120, 0x1732d5fd},
				{1083, 10119, 0x1732b6a0},
				{1084, 10118, 0x1732976f},
				{1085, 10117, 0x1732783e},
				{1086, 10116, 0x17325938},
				{1087, 10115, 0x17323a31},
				{1088, 10114, 0x17321b56},
				{1089, 10113, 0x1731fc7a},
				{1090, 10112, 0x1731dd9f},
				{1091, 10111, 0x1731beee},
				{1092, 10110, 0x1731a069},
				{1093, 10109, 0x173181e3},
				{1094, 10108, 0x1731635e},
				{1095, 10107, 0x17314503},
				{1096, 10106, 0x173126a8},
				{1097, 10105, 0x1731084e},
				{1098, 10104, 0x1730ea1e},
				{1099, 10103, 0x1730cbee},
				{1100, 10102, 0x1730adea},
				{1101, 10101, 0x17308fe5},
				{1102, 10100, 0x1730720b},
				{1103, 10099, 0x17305431},
				{1104, 10098, 0x17303657},
				{1105, 10097, 0x173018a9},
				{1106, 10096, 0x172ffb25},
				{1107, 10095, 0x172fdda1},
				{1108, 10094, 0x172fc01d},
				{1109, 10093, 0x172fa2c4},
				{1110, 10092, 0x172f856b},
				{1111, 10091, 0x172f6812},
				{1112, 10090, 0x172f4ae3},
				{1113, 10089, 0x172f2db5},
				{1114, 10088, 0x172f10b2},
				{1115, 10087, 0x172ef3af},
				{1116, 10086, 0x172ed6ac},
				{1117, 10085, 0x172eb9d4},
				{1118, 10084, 0x172e9cfb},
				{1119, 10083, 0x172e804e},
				{1120, 10082, 0x172e63a1},
				{1121, 10081, 0x172e46f3},
				{1122, 10080, 0x172e2a9c},
				{1123, 10079, 0x172e0e1a},
				{1124, 10078, 0x172df1c2},
				{1125, 10077, 0x172dd56b},
				{1126, 10076, 0x172db913},
				{1127, 10075, 0x172d9ce7},
				{1128, 10074, 0x172d80ba},
				{1129, 10073, 0x172d64b9},
				{1130, 10072, 0x172d48b7},
				{1131, 10071, 0x172d2cb5},
				{1132, 10070, 0x172d10df},
				{1133, 10069, 0x172cf508},
				{1134, 10068, 0x172cd95c},
				{1135, 10067, 0x172cbdb1},
				{1136, 10066, 0x172ca205},
				{1137, 10065, 0x172c8684},
				{1138, 10064, 0x172c6b03},
				{1139, 10063, 0x172c4fae},
				{1140, 10062, 0x172c3458},
				{1141, 10061, 0x172c1902},
				{1142, 10060, 0x172bfdac},
				{1143, 10059, 0x172be281},
				{1144, 10058, 0x172bc781},
				{1145, 10057, 0x172bac56},
				{1146, 10056, 0x172b9156},
				{1147, 10055, 0x172b7681},
				{1148, 10054, 0x172b5bac},
				{1149, 10053, 0x172b40d7},
				{1150, 10052, 0x172b2602},
				{1151, 10051, 0x172b0b58},
				{1152, 10050, 0x172af0ae},
				{1153, 10049, 0x172ad66f},
				{1154, 10048, 0x172abc1b},
				{1155, 10047, 0x172aa1c6},
				{1156, 10046, 0x172a8787},
				{1157, 10045, 0x172a6d5e},
				{1158, 10044, 0x172a534a},
				{1159, 10043, 0x172a3936},
				{1160, 10042, 0x172a1f4d},
				{1161, 10041, 0x172a054f},
				{1162, 10040, 0x1729eb7c},
				{1163, 10039, 0x1729d1a8},
				{1164, 10038, 0x1729b7ea},
				{1165, 10037, 0x17299e42},
				{1166, 10036, 0x17298499},
				{1167, 10035, 0x17296b1b},
				{1168, 10034, 0x172951b3},
				{1169, 10033, 0x17293836},
				{1170, 10032, 0x17291ee3},
				{1171, 10031, 0x17290590},
				{1172, 10030, 0x1728ec3e},
				{1173, 10029, 0x1728d316},
				{1174, 10028, 0x1728b9ee},
				{1175, 10027, 0x1728a0c6},
				{1176, 10026, 0x172887ca},
				{1177, 10025, 0x17286ecd},
				{1178, 10024, 0x172855e5},
				{1179, 10023, 0x17283d14},
				{1180, 10022, 0x17282442},
				{1181, 10021, 0x17280b85},
				{1182, 10020, 0x1727f2de},
				{1183, 10019, 0x1727da37},
				{1184, 10018, 0x1727c1bb},
				{1185, 10017, 0x1727a93f},
				{1186, 10016, 0x172790d9},
				{1187, 10015, 0x17277872},
				{1188, 10014, 0x1727600c},
				{1189, 10013, 0x172747d0},
				{1190, 10012, 0x17272f94},
				{1191, 10011, 0x1727176e},
				{1192, 10010, 0x1726ff5e},
				{1193, 10009, 0x1726e74d},
				{1194, 10008, 0x1726cf52},
				{1195, 10007, 0x1726b756},
				{1196, 10006, 0x17269f86},
				{1197, 10005, 0x172687a0},
				{1198, 10004, 0x17266fe6},
				{1199, 10003, 0x17265856},
				{1200, 10002, 0x172640b0},
				{1201, 10001, 0x1726290b},
				{1202, 10000, 0x1726117b},
				{1203, 9999, 0x1725fa01},
				{1204, 9998, 0x1725e29c},
				{1205, 9997, 0x1725cb37},
				{1206, 9996, 0x1725b3e7},
				{1207, 9995, 0x17259cad},
				{1208, 9994, 0x17258573},
				{1209, 9993, 0x17256e4f},
				{1210, 9992, 0x1725573f},
				{1211, 9991, 0x17254030},
				{1212, 9990, 0x17252937},
				{1213, 9989, 0x1725123d},
				{1214, 9988, 0x1724fb84},
				{1215, 9987, 0x1724e4a0},
				{1216, 9986, 0x1724cde6},
				{1217, 9985, 0x1724b72d},
				{1218, 9984, 0x1724a089},
				{1219, 9983, 0x172489e6},
				{1220, 9982, 0x17247357},
				{1221, 9981, 0x17245cde},
				{1222, 9980, 0x17244666},
				{1223, 9979, 0x17243002},
				{1224, 9978, 0x172419b4},
				{1225, 9977, 0x17240366},
				{1226, 9976, 0x1723ed2e},
				{1227, 9975, 0x1723d70b},
				{1228, 9974, 0x1723c0e8},
				{1229, 9973, 0x1723aadb},
				{1230, 9972, 0x172394f8},
				{1231, 9971, 0x17237f00},
				{1232, 9970, 0x17236908},
				{1233, 9969, 0x1723533b},
				{1234, 9968, 0x17233d6e},
				{1235, 9967, 0x172327a1},
				{1236, 9966, 0x172311e9},
				{1237, 9965, 0x1722fc47},
				{1238, 9964, 0x1722e6ba},
				{1239, 9963, 0x1722d12d},
				{1240, 9962, 0x1722bba1},
				{1241, 9961, 0x1722a63f},
				{1242, 9960, 0x172290dd},
				{1243, 9959, 0x17227b7b},
				{1244, 9958, 0x1722662f},
				{1245, 9957, 0x1722510d},
				{1246, 9956, 0x17223bec},
				{1247, 9955, 0x172226cb},
				{1248, 9954, 0x172211a9},
				{1249, 9953, 0x1721fc9d},
				{1250, 9952, 0x1721e7a7},
				{1251, 9951, 0x1721d2c6},
				{1252, 9950, 0x1721bde5},
				{1253, 9949, 0x1721a904},
				{1254, 9948, 0x1721944e},
				{1255, 9947, 0x17217f82},
				{1256, 9946, 0x17216ae2},
				{1257, 9945, 0x17215641},
				{1258, 9944, 0x172141a1},
				{1259, 9943, 0x17212d2b},
				{1260, 9942, 0x172118cb},
				{1261, 9941, 0x17210455},
				{1262, 9940, 0x1720eff5},
				{1263, 9939, 0x1720db95},
				{1264, 9938, 0x1720c75f},
				{1265, 9937, 0x1720b315},
				{1266, 9936, 0x17209ef5},
				{1267, 9935, 0x17208ad5},
				{1268, 9934, 0x172076b5},
				{1269, 9933, 0x172062ab},
				{1270, 9932, 0x17204eb6},
				{1271, 9931, 0x17203ac1},
				{1272, 9930, 0x172026e2},
				{1273, 9929, 0x17201302},
				{1274, 9928, 0x171fff39},
				{1275, 9927, 0x171feb84},
				{1276, 9926, 0x171fd7e5},
				{1277, 9925, 0x171fc431},
				{1278, 9924, 0x171fb0a7},
				{1279, 9923, 0x171f9d1e},
				{1280, 9922, 0x171f8994},
				{1281, 9921, 0x171f7620},
				{1282, 9920, 0x171f62ac},
				{1283, 9919, 0x171f4f4d},
				{1284, 9918, 0x171f3c04},
				{1285, 9917, 0x171f28bb},
				{1286, 9916, 0x171f1587},
				{1287, 9915, 0x171f0254},
				{1288, 9914, 0x171eef35},
				{1289, 9913, 0x171edc2d},
				{1290, 9912, 0x171ec924},
				{1291, 9911, 0x171eb631},
				{1292, 9910, 0x171ea33d},
				{1293, 9909, 0x171e9060},
				{1294, 9908, 0x171e7d82},
				{1295, 9907, 0x171e6ab9},
				{1296, 9906, 0x171e57f1},
				{1297, 9905, 0x171e453e},
				{1298, 9904, 0x171e328b},
				{1299, 9903, 0x171e1fee},
				{1300, 9902, 0x171e0d51},
				{1301, 9901, 0x171dfac9},
				{1302, 9900, 0x171de856},
				{1303, 9899, 0x171dd5e4},
				{1304, 9898, 0x171dc387},
				{1305, 9897, 0x171db12a},
				{1306, 9896, 0x171d9ee2},
				{1307, 9895, 0x171d8cb0},
				{1308, 9894, 0x171d7a68},
				{1309, 9893, 0x171d684c},
				{1310, 9892, 0x171d561a},
				{1311, 9891, 0x171d4413},
				{1312, 9890, 0x171d31f6},
				{1313, 9889, 0x171d2004},
				{1314, 9888, 0x171d0e13},
				{1315, 9887, 0x171cfc21},
				{1316, 9886, 0x171cea45},
				{1317, 9885, 0x171cd868},
				{1318, 9884, 0x171cc6a2},
				{1319, 9883, 0x171cb4f0},
				{1320, 9882, 0x171ca33f},
				{1321, 9881, 0x171c918e},
				{1322, 9880, 0x171c8007},
				{1323, 9879, 0x171c6e81},
				{1324, 9878, 0x171c5cfb},
				{1325, 9877, 0x171c4b74},
				{1326, 9876, 0x171c3a03},
				{1327, 9875, 0x171c28a8},
				{1328, 9874, 0x171c174d},
				{1329, 9873, 0x171c0607},
				{1330, 9872, 0x171bf4c1},
				{1331, 9871, 0x171be37b},
				{1332, 9870, 0x171bd24a},
				{1333, 9869, 0x171bc12f},
				{1334, 9868, 0x171bb014},
				{1335, 9867, 0x171b9f0f},
				{1336, 9866, 0x171b8e09},
				{1337, 9865, 0x171b7d19},
				{1338, 9864, 0x171b6c29},
				{1339, 9863, 0x171b5b4e},
				{1340, 9862, 0x171b4a74},
				{1341, 9861, 0x171b3999},
				{1342, 9860, 0x171b28d4},
				{1343, 9859, 0x171b1824},
				{1344, 9858, 0x171b075f},
				{1345, 9857, 0x171af6c5},
				{1346, 9856, 0x171ae62b},
				{1347, 9855, 0x171ad590},
				{1348, 9854, 0x171ac50c},
				{1349, 9853, 0x171ab49c},
				{1350, 9852, 0x171aa417},
				{1351, 9851, 0x171a93be},
				{1352, 9850, 0x171a834e},
				{1353, 9849, 0x171a731f},
				{1354, 9848, 0x171a62c6},
				{1355, 9847, 0x171a5281},
				{1356, 9846, 0x171a4252},
				{1357, 9845, 0x171a3223},
				{1358, 9844, 0x171a220a},
				{1359, 9843, 0x171a11f1},
				{1360, 9842, 0x171a01d7},
				{1361, 9841, 0x1719f1d3},
				{1362, 9840, 0x1719e1e5},
				{1363, 9839, 0x1719d1f6},
				{1364, 9838, 0x1719c208},
				{1365, 9837, 0x1719b22f},
				{1366, 9836, 0x1719a256},
				{1367, 9835, 0x17199292},
				{1368, 9834, 0x171982e4},
				{1369, 9833, 0x17197336},
				{1370, 9832, 0x17196388},
				{1371, 9831, 0x171953da},
				{1372, 9830, 0x17194441},
				{1373, 9829, 0x171934be},
				{1374, 9828, 0x1719253b},
				{1375, 9827, 0x171915b7},
				{1376, 9826, 0x1719064a},
				{1377, 9825, 0x1718f6dc},
				{1378, 9824, 0x1718e76e},
				{1379, 9823, 0x1718d82b},
				{1380, 9822, 0x1718c8d3},
				{1381, 9821, 0x1718b990},
				{1382, 9820, 0x1718aa4e},
				{1383, 9819, 0x17189b36},
				{1384, 9818, 0x17188c08},
				{1385, 9817, 0x17187cf1},
				{1386, 9816, 0x17186dd9},
				{1387, 9815, 0x17185ed6},
				{1388, 9814, 0x17184fd4},
				{1389, 9813, 0x171840d2},
				{1390, 9812, 0x171831e5},
				{1391, 9811, 0x171822f8},
				{1392, 9810, 0x17181420},
				{1393, 9809, 0x17180549},
				{1394, 9808, 0x1717f671},
				{1395, 9807, 0x1717e7af},
				{1396, 9806, 0x1717d903},
				{1397, 9805, 0x1717ca41},
				{1398, 9804, 0x1717bb94},
				{1399, 9803, 0x1717ad13},
				{1400, 9802, 0x17179e7c},
				{1401, 9801, 0x17178fe5},
				{1402, 9800, 0x17178163},
				{1403, 9799, 0x171772e2},
				{1404, 9798, 0x17176475},
				{1405, 9797, 0x17175609},
				{1406, 9796, 0x1717479d},
				{1407, 9795, 0x17173947},
				{1408, 9794, 0x17172af0},
				{1409, 9793, 0x17171caf},
				{1410, 9792, 0x17170e6e},
				{1411, 9791, 0x1717002c},
				{1412, 9790, 0x1716f201},
				{1413, 9789, 0x1716e3ea},
				{1414, 9788, 0x1716d5d4},
				{1415, 9787, 0x1716c7be},
				{1416, 9786, 0x1716b9a8},
				{1417, 9785, 0x1716aba7},
				{1418, 9784, 0x17169da6},
				{1419, 9783, 0x17168fbb},
				{1420, 9782, 0x171681cf},
				{1421, 9781, 0x171673e4},
				{1422, 9780, 0x1716660e},
				{1423, 9779, 0x17165838},
				{1424, 9778, 0x17164a62},
				{1425, 9777, 0x17163ca2},
				{1426, 9776, 0x17162ee2},
				{1427, 9775, 0x17162137},
				{1428, 9774, 0x1716138c},
				{1429, 9773, 0x171605f6},
				{1430, 9772, 0x1715f861},
				{1431, 9771, 0x1715eacb},
				{1432, 9770, 0x1715dd36},
				{1433, 9769, 0x1715cfb6},
				{1434, 9768, 0x1715c24b},
				{1435, 9767, 0x1715b4cb},
				{1436, 9766, 0x1715a761},
				{1437, 9765, 0x171599f6},
				{1438, 9764, 0x17158ca1},
				{1439, 9763, 0x17157f4c},
				{1440, 9762, 0x1715720d},
				{1441, 9761, 0x171564d8},
				{1442, 9760, 0x171557ae},
				{1443, 9759, 0x17154a8e},
				{1444, 9758, 0x17153d7a},
				{1445, 9757, 0x17153070},
				{1446, 9756, 0x17152366},
				{1447, 9755, 0x17151666},
				{1448, 9754, 0x17150972},
				{1449, 9753, 0x1714fc7e},
				{1450, 9752, 0x1714ef94},
				{1451, 9751, 0x1714e2aa},
				{1452, 9750, 0x1714d5d6},
				{1453, 9749, 0x1714c902},
				{1454, 9748, 0x1714bc38},
				{1455, 9747, 0x1714af6e},
				{1456, 9746, 0x1714a2b0},
				{1457, 9745, 0x171495fc},
				{1458, 9744, 0x17148952},
				{1459, 9743, 0x17147ca9},
				{1460, 9742, 0x17147015},
				{1461, 9741, 0x17146381},
				{1462, 9740, 0x171456f8},
				{1463, 9739, 0x17144a6f},
				{1464, 9738, 0x17143df0},
				{1465, 9737, 0x17143172},
				{1466, 9736, 0x171424fe},
				{1467, 9735, 0x17141895},
				{1468, 9734, 0x17140c37},
				{1469, 9733, 0x1713ffd9},
				{1470, 9732, 0x1713f386},
				{1471, 9731, 0x1713e73d},
				{1472, 9730, 0x1713daf4},
				{1473, 9729, 0x1713ceb6},
				{1474, 9728, 0x1713c283},
				{1475, 9727, 0x1713b665},
				{1476, 9726, 0x1713aa3c},
				{1477, 9725, 0x17139e1f},
				{1478, 9724, 0x17139201},
				{1479, 9723, 0x171385f8},
				{1480, 9722, 0x171379e5},
				{1481, 9721, 0x17136de8},
				{1482, 9720, 0x171361ea},
				{1483, 9719, 0x171355f7},
				{1484, 9718, 0x17134a0f},
				{1485, 9717, 0x17133e27},
				{1486, 9716, 0x1713324a},
				{1487, 9715, 0x1713266c},
				{1488, 9714, 0x17131aa4},
				{1489, 9713, 0x17130edc},
				{1490, 9712, 0x17130314},
				{1491, 9711, 0x1712f76d},
				{1492, 9710, 0x1712ebba},
				{1493, 9709, 0x1712e008},
				{1494, 9708, 0x1712d46b},
				{1495, 9707, 0x1712c8c3},
				{1496, 9706, 0x1712bd31},
				{1497, 9705, 0x1712b19e},
				{1498, 9704, 0x1712a617},
				{1499, 9703, 0x17129a9a},
				{1500, 9702, 0x17128f1d},
				{1501, 9701, 0x171283ab},
				{1502, 9700, 0x17127839},
				{1503, 9699, 0x17126cd2},
				{1504, 9698, 0x17126175},
				{1505, 9697, 0x17125623},
				{1506, 9696, 0x17124adc},
				{1507, 9695, 0x17123f95},
				{1508, 9694, 0x1712344e},
				{1509, 9693, 0x17122911},
				{1510, 9692, 0x17121ddf},
				{1511, 9691, 0x171212ae},
				{1512, 9690, 0x17120787},
				{1513, 9689, 0x1711fc6b},
				{1514, 9688, 0x1711f14e},
				{1515, 9687, 0x1711e63d},
				{1516, 9686, 0x1711db2b},
				{1517, 9685, 0x1711d025},
				{1518, 9684, 0x1711c529},
				{1519, 9683, 0x1711ba2d},
				{1520, 9682, 0x1711af3b},
				{1521, 9681, 0x1711a460},
				{1522, 9680, 0x17119984},
				{1523, 9679, 0x17118e9d},
				{1524, 9678, 0x171183cc},
				{1525, 9677, 0x171178fb},
				{1526, 9676, 0x17116e35},
				{1527, 9675, 0x1711636e},
				{1528, 9674, 0x171158b3},
				{1529, 9673, 0x17114e02},
				{1530, 9672, 0x17114351},
				{1531, 9671, 0x171138ab},
				{1532, 9670, 0x17112e05},
				{1533, 9669, 0x17112369},
				{1534, 9668, 0x171118d8},
				{1535, 9667, 0x17110e48},
				{1536, 9666, 0x171103c2},
				{1537, 9665, 0x1710f951},
				{1538, 9664, 0x1710eed6},
				{1539, 9663, 0x1710e466},
				{1540, 9662, 0x1710d9f5},
				{1541, 9661, 0x1710cf8f},
				{1542, 9660, 0x1710c52a},
				{1543, 9659, 0x1710bace},
				{1544, 9658, 0x1710b07e},
				{1545, 9657, 0x1710a62e},
				{1546, 9656, 0x17109be8},
				{1547, 9655, 0x171091ad},
				{1548, 9654, 0x17108773},
				{1549, 9653, 0x17107d43},
				{1550, 9652, 0x17107312},
				{1551, 9651, 0x171068ed},
				{1552, 9650, 0x17105ed2},
				{1553, 9649, 0x171054b8},
				{1554, 9648, 0x17104aa8},
				{1555, 9647, 0x17104098},
				{1556, 9646, 0x17103693},
				{1557, 9645, 0x17102c8e},
				{1558, 9644, 0x17102293},
				{1559, 9643, 0x171018a4},
				{1560, 9642, 0x17100eb4},
				{1561, 9641, 0x171004c4},
				{1562, 9640, 0x170ffaea},
				{1563, 9639, 0x170ff105},
				{1564, 9638, 0x170fe735},
				{1565, 9637, 0x170fdd66},
				{1566, 9636, 0x170fd396},
				{1567, 9635, 0x170fc9dc},
				{1568, 9634, 0x170fc022},
				{1569, 9633, 0x170fb668},
				{1570, 9632, 0x170facb9},
				{1571, 9631, 0x170fa30a},
				{1572, 9630, 0x170f9965},
				{1573, 9629, 0x170f8fcb},
				{1574, 9628, 0x170f8632},
				{1575, 9627, 0x170f7c98},
				{1576, 9626, 0x170f7309},
				{1577, 9625, 0x170f6984},
				{1578, 9624, 0x170f6000},
				{1579, 9623, 0x170f5686},
				{1580, 9622, 0x170f4d0d},
				{1581, 9621, 0x170f439e},
				{1582, 9620, 0x170f3a39},
				{1583, 9619, 0x170f30e0},
				{1584, 9618, 0x170f277c},
				{1585, 9617, 0x170f1e22},
				{1586, 9616, 0x170f14d4},
				{1587, 9615, 0x170f0b85},
				{1588, 9614, 0x170f0241},
				{1589, 9613, 0x170ef8fd},
				{1590, 9612, 0x170eefc4},
				{1591, 9611, 0x170ee68b},
				{1592, 9610, 0x170edd5c},
				{1593, 9609, 0x170ed42e},
				{1594, 9608, 0x170ecb0a},
				{1595, 9607, 0x170ec1f1},
				{1596, 9606, 0x170eb8d8},
				{1597, 9605, 0x170eafbf},
				{1598, 9604, 0x170ea6bb},
				{1599, 9603, 0x170e9db8},
				{1600, 9602, 0x170e94b4},
				{1601, 9601, 0x170e8bb0},
				{1602, 9600, 0x170e82c2},
				{1603, 9599, 0x170e79c9},
				{1604, 9598, 0x170e70db},
				{1605, 9597, 0x170e67f8},
				{1606, 9596, 0x170e5f15},
				{1607, 9595, 0x170e563c},
				{1608, 9594, 0x170e4d63},
				{1609, 9593, 0x170e4495},
				{1610, 9592, 0x170e3bc7},
				{1611, 9591, 0x170e3304},
				{1612, 9590, 0x170e2a41},
				{1613, 9589, 0x170e2193},
				{1614, 9588, 0x170e18db},
				{1615, 9587, 0x170e102d},
				{1616, 9586, 0x170e0780},
				{1617, 9585, 0x170dfedd},
				{1618, 9584, 0x170df63a},
				{1619, 9583, 0x170deda1},
				{1620, 9582, 0x170de514},
				{1621, 9581, 0x170ddc7c},
				{1622, 9580, 0x170dd3f9},
				{1623, 9579, 0x170dcb6b},
				{1624, 9578, 0x170dc2f3},
				{1625, 9577, 0x170dba70},
				{1626, 9576, 0x170db203},
				{1627, 9575, 0x170da996},
				{1628, 9574, 0x170da128},
				{1629, 9573, 0x170d98c6},
				{1630, 9572, 0x170d906e},
				{1631, 9571, 0x170d880b},
				{1632, 9570, 0x170d7fbe},
				{1633, 9569, 0x170d7767},
				{1634, 9568, 0x170d6f19},
				{1635, 9567, 0x170d66d7},
				{1636, 9566, 0x170d5e95},
				{1637, 9565, 0x170d565d},
				{1638, 9564, 0x170d4e25},
				{1639, 9563, 0x170d45ee},
				{1640, 9562, 0x170d3dc1},
				{1641, 9561, 0x170d359f},
				{1642, 9560, 0x170d2d7c},
				{1643, 9559, 0x170d255a},
				{1644, 9558, 0x170d1d4d},
				{1645, 9557, 0x170d1536},
				{1646, 9556, 0x170d0d29},
				{1647, 9555, 0x170d051d},
				{1648, 9554, 0x170cfd1b},
				{1649, 9553, 0x170cf519},
				{1650, 9552, 0x170ced21},
				{1651, 9551, 0x170ce52a},
				{1652, 9550, 0x170cdd3e},
				{1653, 9549, 0x170cd551},
				{1654, 9548, 0x170ccd65},
				{1655, 9547, 0x170cc583},
				{1656, 9546, 0x170cbdac},
				{1657, 9545, 0x170cb5ca},
				{1658, 9544, 0x170cadfe},
				{1659, 9543, 0x170ca631},
				{1660, 9542, 0x170c9e65},
				{1661, 9541, 0x170c96a3},
				{1662, 9540, 0x170c8ee2},
				{1663, 9539, 0x170c872b},
				{1664, 9538, 0x170c7f74},
				{1665, 9537, 0x170c77bd},
				{1666, 9536, 0x170c7011},
				{1667, 9535, 0x170c6865},
				{1668, 9534, 0x170c60c4},
				{1669, 9533, 0x170c5922},
				{1670, 9532, 0x170c518c},
				{1671, 9531, 0x170c49f5},
				{1672, 9530, 0x170c425e},
				{1673, 9529, 0x170c3ad2},
				{1674, 9528, 0x170c3346},
				{1675, 9527, 0x170c2bd0},
				{1676, 9526, 0x170c244f},
				{1677, 9525, 0x170c1cd8},
				{1678, 9524, 0x170c1562},
				{1679, 9523, 0x170c0deb},
				{1680, 9522, 0x170c0680},
				{1681, 9521, 0x170bff14},
				{1682, 9520, 0x170bf7b3},
				{1683, 9519, 0x170bf052},
				{1684, 9518, 0x170be8fc},
				{1685, 9517, 0x170be1a5},
				{1686, 9516, 0x170bda4f},
				{1687, 9515, 0x170bd304},
				{1688, 9514, 0x170bcbb8},
				{1689, 9513, 0x170bc477},
				{1690, 9512, 0x170bbd37},
				{1691, 9511, 0x170bb600},
				{1692, 9510, 0x170baeca},
				{1693, 9509, 0x170ba794},
				{1694, 9508, 0x170ba05e},
				{1695, 9507, 0x170b9933},
				{1696, 9506, 0x170b9212},
				{1697, 9505, 0x170b8af2},
				{1698, 9504, 0x170b83d1},
				{1699, 9503, 0x170b7cbb},
				{1700, 9502, 0x170b75a5},
				{1701, 9501, 0x170b6e90},
				{1702, 9500, 0x170b6784},
				{1703, 9499, 0x170b6079},
				{1704, 9498, 0x170b5979},
				{1705, 9497, 0x170b5279},
				{1706, 9496, 0x170b4b83},
				{1707, 9495, 0x170b4482},
				{1708, 9494, 0x170b3d98},
				{1709, 9493, 0x170b36a2},
				{1710, 9492, 0x170b2fb7},
				{1711, 9491, 0x170b28cc},
				{1712, 9490, 0x170b21ec},
				{1713, 9489, 0x170b1b0c},
				{1714, 9488, 0x170b142b},
				{1715, 9487, 0x170b0d56},
				{1716, 9486, 0x170b0680},
				{1717, 9485, 0x170affb6},
				{1718, 9484, 0x170af8eb},
				{1719, 9483, 0x170af220},
				{1720, 9482, 0x170aeb60},
				{1721, 9481, 0x170ae4a0},
				{1722, 9480, 0x170addeb},
				{1723, 9479, 0x170ad736},
				{1724, 9478, 0x170ad080},
				{1725, 9477, 0x170ac9cb},
				{1726, 9476, 0x170ac321},
				{1727, 9475, 0x170abc76},
				{1728, 9474, 0x170ab5dc},
				{1729, 9473, 0x170aaf47},
				{1730, 9472, 0x170aa8b7},
				{1731, 9471, 0x170aa227},
				{1732, 9470, 0x170a9b9d},
				{1733, 9469, 0x170a9513},
				{1734, 9468, 0x170a8e93},
				{1735, 9467, 0x170a8813},
				{1736, 9466, 0x170a819f},
				{1737, 9465, 0x170a7b24},
				{1738, 9464, 0x170a74b5},
				{1739, 9463, 0x170a6e45},
				{1740, 9462, 0x170a67d6},
				{1741, 9461, 0x170a6171},
				{1742, 9460, 0x170a5b0c},
				{1743, 9459, 0x170a54ad},
				{1744, 9458, 0x170a4e4d},
				{1745, 9457, 0x170a47f3},
				{1746, 9456, 0x170a419f},
				{1747, 9455, 0x170a3b4f},
				{1748, 9454, 0x170a3500},
				{1749, 9453, 0x170a2ebc},
				{1750, 9452, 0x170a2872},
				{1751, 9451, 0x170a2238},
				{1752, 9450, 0x170a1bf9},
				{1753, 9449, 0x170a15bf},
				{1754, 9448, 0x170a0f8a},
				{1755, 9447, 0x170a0956},
				{1756, 9446, 0x170a0327},
				{1757, 9445, 0x1709fcfd},
				{1758, 9444, 0x1709f6d3},
				{1759, 9443, 0x1709f0af},
				{1760, 9442, 0x1709ea90},
				{1761, 9441, 0x1709e471},
				{1762, 9440, 0x1709de57},
				{1763, 9439, 0x1709d843},
				{1764, 9438, 0x1709d22f},
				{1765, 9437, 0x1709cc25},
				{1766, 9436, 0x1709c616},
				{1767, 9435, 0x1709c017},
				{1768, 9434, 0x1709ba13},
				{1769, 9433, 0x1709b414},
				{1770, 9432, 0x1709ae1b},
				{1771, 9431, 0x1709a822},
				{1772, 9430, 0x1709a22d},
				{1773, 9429, 0x17099c3f},
				{1774, 9428, 0x17099650},
				{1775, 9427, 0x17099067},
				{1776, 9426, 0x17098a7d},
				{1777, 9425, 0x1709849f},
				{1778, 9424, 0x17097ebb},
				{1779, 9423, 0x170978e2},
				{1780, 9422, 0x17097308},
				{1781, 9421, 0x17096d34},
				{1782, 9420, 0x1709676b},
				{1783, 9419, 0x1709619d},
				{1784, 9418, 0x17095bd4},
				{1785, 9417, 0x17095610},
				{1786, 9416, 0x1709504c},
				{1787, 9415, 0x17094a8e},
				{1788, 9414, 0x170944cf},
				{1789, 9413, 0x17093f16},
				{1790, 9412, 0x17093963},
				{1791, 9411, 0x170933b4},
				{1792, 9410, 0x17092e06},
				{1793, 9409, 0x1709285d},
				{1794, 9408, 0x170922b4},
				{1795, 9407, 0x17091d10},
				{1796, 9406, 0x17091772},
				{1797, 9405, 0x170911d9},
				{1798, 9404, 0x17090c41},
				{1799, 9403, 0x170906ad},
				{1800, 9402, 0x1709011a},
				{1801, 9401, 0x1708fb8c},
				{1802, 9400, 0x1708f5fd},
				{1803, 9399, 0x1708f07a},
				{1804, 9398, 0x1708eaf1},
				{1805, 9397, 0x1708e573},
				{1806, 9396, 0x1708dff5},
				{1807, 9395, 0x1708da7d},
				{1808, 9394, 0x1708d504},
				{1809, 9393, 0x1708cf91},
				{1810, 9392, 0x1708ca1d},
				{1811, 9391, 0x1708c4b5},
				{1812, 9390, 0x1708bf47},
				{1813, 9389, 0x1708b9e9},
				{1814, 9388, 0x1708b486},
				{1815, 9387, 0x1708af28},
				{1816, 9386, 0x1708a9ca},
				{1817, 9385, 0x1708a472},
				{1818, 9384, 0x17089f19},
				{1819, 9383, 0x170899cc},
				{1820, 9382, 0x17089479},
				{1821, 9381, 0x17088f30},
				{1822, 9380, 0x170889e8},
				{1823, 9379, 0x170884a0},
				{1824, 9378, 0x17087f5d},
				{1825, 9377, 0x17087a1f},
				{1826, 9376, 0x170874e7},
				{1827, 9375, 0x17086faf},
				{1828, 9374, 0x17086a7c},
				{1829, 9373, 0x17086549},
				{1830, 9372, 0x1708601b},
				{1831, 9371, 0x17085af3},
				{1832, 9370, 0x170855cb},
				{1833, 9369, 0x170850a3},
				{1834, 9368, 0x17084b80},
				{1835, 9367, 0x17084663},
				{1836, 9366, 0x1708414b},
				{1837, 9365, 0x17083c2d},
				{1838, 9364, 0x1708371a},
				{1839, 9363, 0x17083208},
				{1840, 9362, 0x17082cfa},
				{1841, 9361, 0x170827ed},
				{1842, 9360, 0x170822e5},
				{1843, 9359, 0x17081de8},
				{1844, 9358, 0x170818e5},
				{1845, 9357, 0x170813e3},
				{1846, 9356, 0x17080ee6},
				{1847, 9355, 0x170809ee},
				{1848, 9354, 0x170804fb},
				{1849, 9353, 0x17080003},
				{1850, 9352, 0x1707fb16},
				{1851, 9351, 0x1707f629},
				{1852, 9350, 0x1707f141},
				{1853, 9349, 0x1707ec5a},
				{1854, 9348, 0x1707e777},
				{1855, 9347, 0x1707e295},
				{1856, 9346, 0x1707ddb8},
				{1857, 9345, 0x1707d8e0},
				{1858, 9344, 0x1707d409},
				{1859, 9343, 0x1707cf36},
				{1860, 9342, 0x1707ca64},
				{1861, 9341, 0x1707c597},
				{1862, 9340, 0x1707c0ca},
				{1863, 9339, 0x1707bc03},
				{1864, 9338, 0x1707b740},
				{1865, 9337, 0x1707b27e},
				{1866, 9336, 0x1707adbc},
				{1867, 9335, 0x1707a8ff},
				{1868, 9334, 0x1707a448},
				{1869, 9333, 0x17079f90},
				{1870, 9332, 0x17079ade},
				{1871, 9331, 0x1707962c},
				{1872, 9330, 0x1707917f},
				{1873, 9329, 0x17078cd3},
				{1874, 9328, 0x17078831},
				{1875, 9327, 0x17078389},
				{1876, 9326, 0x17077ee7},
				{1877, 9325, 0x17077a4b},
				{1878, 9324, 0x170775ae},
				{1879, 9323, 0x17077112},
				{1880, 9322, 0x17076c7a},
				{1881, 9321, 0x170767e8},
				{1882, 9320, 0x17076357},
				{1883, 9319, 0x17075eca},
				{1884, 9318, 0x17075a3e},
				{1885, 9317, 0x170755b6},
				{1886, 9316, 0x1707512f},
				{1887, 9315, 0x17074cad},
				{1888, 9314, 0x1707482c},
				{1889, 9313, 0x170743b5},
				{1890, 9312, 0x17073f38},
				{1891, 9311, 0x17073ac1},
				{1892, 9310, 0x1707364f},
				{1893, 9309, 0x170731de},
				{1894, 9308, 0x17072d6c},
				{1895, 9307, 0x17072900},
				{1896, 9306, 0x17072499},
				{1897, 9305, 0x17072032},
				{1898, 9304, 0x17071bcb},
				{1899, 9303, 0x17071769},
				{1900, 9302, 0x1707130d},
				{1901, 9301, 0x17070eb1},
				{1902, 9300, 0x17070a55},
				{1903, 9299, 0x170705fe},
				{1904, 9298, 0x170701ac},
				{1905, 9297, 0x1706fd60},
				{1906, 9296, 0x1706f90f},
				{1907, 9295, 0x1706f4c2},
				{1908, 9294, 0x1706f07c},
				{1909, 9293, 0x1706ec35},
				{1910, 9292, 0x1706e7ee},
				{1911, 9291, 0x1706e3ad},
				{1912, 9290, 0x1706df71},
				{1913, 9289, 0x1706db35},
				{1914, 9288, 0x1706d6f9},
				{1915, 9287, 0x1706d2c2},
				{1916, 9286, 0x1706ce91},
				{1917, 9285, 0x1706ca60},
				{1918, 9284, 0x1706c62e},
				{1919, 9283, 0x1706c202},
				{1920, 9282, 0x1706bde1},
				{1921, 9281, 0x1706b9bb},
				{1922, 9280, 0x1706b594},
				{1923, 9279, 0x1706b173},
				{1924, 9278, 0x1706ad52},
				{1925, 9277, 0x1706a936},
				{1926, 9276, 0x1706a51a},
				{1927, 9275, 0x1706a104},
				{1928, 9274, 0x17069ced},
				{1929, 9273, 0x170698dc},
				{1930, 9272, 0x170694cb},
				{1931, 9271, 0x170690bf},
				{1932, 9270, 0x17068cb3},
				{1933, 9269, 0x170688a8},
				{1934, 9268, 0x170684a1},
				{1935, 9267, 0x170680a6},
				{1936, 9266, 0x17067ca5},
				{1937, 9265, 0x170678a4},
				{1938, 9264, 0x170674a8},
				{1939, 9263, 0x170670ac},
				{1940, 9262, 0x17066cb6},
				{1941, 9261, 0x170668c5},
				{1942, 9260, 0x170664cf},
				{1943, 9259, 0x170660de},
				{1944, 9258, 0x17065cf3},
				{1945, 9257, 0x17065907},
				{1946, 9256, 0x17065521},
				{1947, 9255, 0x1706513b},
				{1948, 9254, 0x17064d55},
				{1949, 9253, 0x17064974},
				{1950, 9252, 0x17064593},
				{1951, 9251, 0x170641bd},
				{1952, 9250, 0x17063de2},
				{1953, 9249, 0x17063a06},
				{1954, 9248, 0x17063635},
				{1955, 9247, 0x1706325f},
				{1956, 9246, 0x17062e8f},
				{1957, 9245, 0x17062abe},
				{1958, 9244, 0x170626f3},
				{1959, 9243, 0x17062327},
				{1960, 9242, 0x17061f61},
				{1961, 9241, 0x17061b9b},
				{1962, 9240, 0x170617db},
				{1963, 9239, 0x1706141a},
				{1964, 9238, 0x1706105a},
				{1965, 9237, 0x17060c9e},
				{1966, 9236, 0x170608e9},
				{1967, 9235, 0x1706052d},
				{1968, 9234, 0x1706017d},
				{1969, 9233, 0x1705fdc7},
				{1970, 9232, 0x1705fa16},
				{1971, 9231, 0x1705f666},
				{1972, 9230, 0x1705f2bb},
				{1973, 9229, 0x1705ef10},
				{1974, 9228, 0x1705eb6a},
				{1975, 9227, 0x1705e7c4},
				{1976, 9226, 0x1705e41e},
				{1977, 9225, 0x1705e07e},
				{1978, 9224, 0x1705dcde},
				{1979, 9223, 0x1705d943},
				{1980, 9222, 0x1705d5a8},
				{1981, 9221, 0x1705d212},
				{1982, 9220, 0x1705ce7c},
				{1983, 9219, 0x1705cae7},
				{1984, 9218, 0x1705c751},
				{1985, 9217, 0x1705c3c6},
				{1986, 9216, 0x1705c036},
				{1987, 9215, 0x1705bcab},
				{1988, 9214, 0x1705b920},
				{1989, 9213, 0x1705b59a},
				{1990, 9212, 0x1705b215},
				{1991, 9211, 0x1705ae8f},
				{1992, 9210, 0x1705ab0f},
				{1993, 9209, 0x1705a78f},
				{1994, 9208, 0x1705a40e},
				{1995, 9207, 0x1705a094},
				{1996, 9206, 0x17059d1e},
				{1997, 9205, 0x170599a9},
				{1998, 9204, 0x17059633},
				{1999, 9203, 0x170592c3},
				{2000, 9202, 0x17058f4e},
				{2001, 9201, 0x17058be3},
				{2002, 9200, 0x17058873},
				{2003, 9199, 0x17058508},
				{2004, 9198, 0x170581a3},
				{2005, 9197, 0x17057e38},
				{2006, 9196, 0x17057ad3},
				{2007, 9195, 0x17057773},
				{2008, 9194, 0x17057413},
				{2009, 9193, 0x170570b3},
				{2010, 9192, 0x17056d53},
				{2011, 9191, 0x170569f8},
				{2012, 9190, 0x170566a3},
				{2013, 9189, 0x1705634e},
				{2014, 9188, 0x17055ff8},
				{2015, 9187, 0x17055ca8},
				{2016, 9186, 0x1705595b},
				{2017, 9185, 0x17055611},
				{2018, 9184, 0x170552c9},
				{2019, 9183, 0x17054f81},
				{2020, 9182, 0x17054c3e},
				{2021, 9181, 0x170548fc},
				{2022, 9180, 0x170545b9},
				{2023, 9179, 0x1705427c},
				{2024, 9178, 0x17053f3f},
				{2025, 9177, 0x17053c05},
				{2026, 9176, 0x170538cd},
				{2027, 9175, 0x17053595},
				{2028, 9174, 0x17053263},
				{2029, 9173, 0x17052f31},
				{2030, 9172, 0x17052bfe},
				{2031, 9171, 0x170528d1},
				{2032, 9170, 0x170525a1},
				{2033, 9169, 0x17052277},
				{2034, 9168, 0x17051f4d},
				{2035, 9167, 0x17051c25},
				{2036, 9166, 0x17051900},
				{2037, 9165, 0x170515db},
				{2038, 9164, 0x170512b9},
				{2039, 9163, 0x17050f99},
				{2040, 9162, 0x17050c7c},
				{2041, 9161, 0x1705095f},
				{2042, 9160, 0x17050645},
				{2043, 9159, 0x17050330},
				{2044, 9158, 0x17050019},
				{2045, 9157, 0x1704fd04},
				{2046, 9156, 0x1704f9f2},
				{2047, 9155, 0x1704f6e0},
				{2048, 9154, 0x1704f3d0},
				{2049, 9153, 0x1704f0c3},
				{2050, 9152, 0x1704edb6},
				{2051, 9151, 0x1704eaac},
				{2052, 9150, 0x1704e7a5},
				{2053, 9149, 0x1704e4a0},
				{2054, 9148, 0x1704e19b},
				{2055, 9147, 0x1704de99},
				{2056, 9146, 0x1704db97},
				{2057, 9145, 0x1704d898},
				{2058, 9144, 0x1704d5a0},
				{2059, 9143, 0x1704d2a4},
				{2060, 9142, 0x1704cfaa},
				{2061, 9141, 0x1704ccb2},
				{2062, 9140, 0x1704c9be},
				{2063, 9139, 0x1704c6c9},
				{2064, 9138, 0x1704c3d7},
				{2065, 9137, 0x1704c0e8},
				{2066, 9136, 0x1704bdf8},
				{2067, 9135, 0x1704bb0c},
				{2068, 9134, 0x1704b81f},
				{2069, 9133, 0x1704b535},
				{2070, 9132, 0x1704b24e},
				{2071, 9131, 0x1704af69},
				{2072, 9130, 0x1704ac85},
				{2073, 9129, 0x1704a9a3},
				{2074, 9128, 0x1704a6c6},
				{2075, 9127, 0x1704a3e7},
				{2076, 9126, 0x1704a10b},
				{2077, 9125, 0x17049e2e},
				{2078, 9124, 0x17049b54},
				{2079, 9123, 0x1704987d},
				{2080, 9122, 0x170495a9},
				{2081, 9121, 0x170492d4},
				{2082, 9120, 0x17049000},
				{2083, 9119, 0x17048d31},
				{2084, 9118, 0x17048a61},
				{2085, 9117, 0x17048792},
				{2086, 9116, 0x170484c9},
				{2087, 9115, 0x170481ff},
				{2088, 9114, 0x17047f35},
				{2089, 9113, 0x17047c71},
				{2090, 9112, 0x170479ac},
				{2091, 9111, 0x170476eb},
				{2092, 9110, 0x17047429},
				{2093, 9109, 0x1704716a},
				{2094, 9108, 0x17046eab},
				{2095, 9107, 0x17046bef},
				{2096, 9106, 0x17046935},
				{2097, 9105, 0x1704667b},
				{2098, 9104, 0x170463c4},
				{2099, 9103, 0x1704610d},
				{2100, 9102, 0x17045e59},
				{2101, 9101, 0x17045ba8},
				{2102, 9100, 0x170458f6},
				{2103, 9099, 0x17045647},
				{2104, 9098, 0x1704539e},
				{2105, 9097, 0x170450f1},
				{2106, 9096, 0x17044e48},
				{2107, 9095, 0x17044b9e},
				{2108, 9094, 0x170448f7},
				{2109, 9093, 0x17044653},
				{2110, 9092, 0x170443af},
				{2111, 9091, 0x1704410e},
				{2112, 9090, 0x17043e6f},
				{2113, 9089, 0x17043bd0},
				{2114, 9088, 0x17043931},
				{2115, 9087, 0x17043698},
				{2116, 9086, 0x170433fe},
				{2117, 9085, 0x17043165},
				{2118, 9084, 0x17042ece},
				{2119, 9083, 0x17042c3a},
				{2120, 9082, 0x170429a9},
				{2121, 9081, 0x17042717},
				{2122, 9080, 0x17042489},
				{2123, 9079, 0x170421fa},
				{2124, 9078, 0x17041f6b},
				{2125, 9077, 0x17041cdf},
				{2126, 9076, 0x17041a56},
				{2127, 9075, 0x170417cf},
				{2128, 9074, 0x17041548},
				{2129, 9073, 0x170412c2},
				{2130, 9072, 0x1704103e},
				{2131, 9071, 0x17040dbc},
				{2132, 9070, 0x17040b3e},
				{2133, 9069, 0x170408bd},
				{2134, 9068, 0x17040641},
				{2135, 9067, 0x170403c7},
				{2136, 9066, 0x1704014e},
				{2137, 9065, 0x1703fed5},
				{2138, 9064, 0x1703fc5e},
				{2139, 9063, 0x1703f9e8},
				{2140, 9062, 0x1703f774},
				{2141, 9061, 0x1703f500},
				{2142, 9060, 0x1703f28f},
				{2143, 9059, 0x1703f020},
				{2144, 9058, 0x1703edb2},
				{2145, 9057, 0x1703eb46},
				{2146, 9056, 0x1703e8da},
				{2147, 9055, 0x1703e671},
				{2148, 9054, 0x1703e40b},
				{2149, 9053, 0x1703e1a4},
				{2150, 9052, 0x1703df40},
				{2151, 9051, 0x1703dcdd},
				{2152, 9050, 0x1703da7c},
				{2153, 9049, 0x1703d81a},
				{2154, 9048, 0x1703d5bc},
				{2155, 9047, 0x1703d35e},
				{2156, 9046, 0x1703d102},
				{2157, 9045, 0x1703cea9},
				{2158, 9044, 0x1703cc50},
				{2159, 9043, 0x1703c9f7},
				{2160, 9042, 0x1703c7a0},
				{2161, 9041, 0x1703c54d},
				{2162, 9040, 0x1703c2f9},
				{2163, 9039, 0x1703c0a5},
				{2164, 9038, 0x1703be57},
				{2165, 9037, 0x1703bc06},
				{2166, 9036, 0x1703b9bd},
				{2167, 9035, 0x1703b771},
				{2168, 9034, 0x1703b526},
				{2169, 9033, 0x1703b2dd},
				{2170, 9032, 0x1703b094},
				{2171, 9031, 0x1703ae4e},
				{2172, 9030, 0x1703ac0a},
				{2173, 9029, 0x1703a9c7},
				{2174, 9028, 0x1703a783},
				{2175, 9027, 0x1703a542},
				{2176, 9026, 0x1703a304},
				{2177, 9025, 0x1703a0c6},
				{2178, 9024, 0x17039e8a},
				{2179, 9023, 0x17039c4f},
				{2180, 9022, 0x17039a13},
				{2181, 9021, 0x170397dd},
				{2182, 9020, 0x170395a7},
				{2183, 9019, 0x17039371},
				{2184, 9018, 0x1703913d},
				{2185, 9017, 0x17038f0a},
				{2186, 9016, 0x17038cd9},
				{2187, 9015, 0x17038aa8},
				{2188, 9014, 0x1703887a},
				{2189, 9013, 0x1703864c},
				{2190, 9012, 0x17038420},
				{2191, 9011, 0x170381f5},
				{2192, 9010, 0x17037fcc},
				{2193, 9009, 0x17037da6},
				{2194, 9008, 0x17037b7d},
				{2195, 9007, 0x1703795a},
				{2196, 9006, 0x17037737},
				{2197, 9005, 0x17037516},
				{2198, 9004, 0x170372f3},
				{2199, 9003, 0x170370d5},
				{2200, 9002, 0x17036eb4},
				{2201, 9001, 0x17036c99},
				{2202, 9000, 0x17036a7b},
				{2203, 8999, 0x17036862},
				{2204, 8998, 0x17036647},
				{2205, 8997, 0x1703642e},
				{2206, 8996, 0x17036218},
				{2207, 8995, 0x17036002},
				{2208, 8994, 0x17035def},
				{2209, 8993, 0x17035bdb},
				{2210, 8992, 0x170359cb},
				{2211, 8991, 0x170357ba},
				{2212, 8990, 0x170355ac},
				{2213, 8989, 0x1703539e},
				{2214, 8988, 0x17035193},
				{2215, 8987, 0x17034f88},
				{2216, 8986, 0x17034d7d},
				{2217, 8985, 0x17034b74},
				{2218, 8984, 0x1703496e},
				{2219, 8983, 0x17034766},
				{2220, 8982, 0x17034563},
				{2221, 8981, 0x1703435f},
				{2222, 8980, 0x1703415c},
				{2223, 8979, 0x17033f5c},
				{2224, 8978, 0x17033d5b},
				{2225, 8977, 0x17033b5b},
				{2226, 8976, 0x17033960},
				{2227, 8975, 0x17033764},
				{2228, 8974, 0x17033569},
				{2229, 8973, 0x1703336e},
				{2230, 8972, 0x17033176},
				{2231, 8971, 0x17032f80},
				{2232, 8970, 0x17032d8a},
				{2233, 8969, 0x17032b94},
				{2234, 8968, 0x170329a1},
				{2235, 8967, 0x170327ae},
				{2236, 8966, 0x170325bb},
				{2237, 8965, 0x170323cb},
				{2238, 8964, 0x170321dd},
				{2239, 8963, 0x17031fef},
				{2240, 8962, 0x17031e02},
				{2241, 8961, 0x17031c17},
				{2242, 8960, 0x17031a2e},
				{2243, 8959, 0x17031846},
				{2244, 8958, 0x1703165e},
				{2245, 8957, 0x17031478},
				{2246, 8956, 0x17031292},
				{2247, 8955, 0x170310ad},
				{2248, 8954, 0x17030eca},
				{2249, 8953, 0x17030ce7},
				{2250, 8952, 0x17030b06},
				{2251, 8951, 0x17030926},
				{2252, 8950, 0x17030748},
				{2253, 8949, 0x1703056b},
				{2254, 8948, 0x1703038d},
				{2255, 8947, 0x170301b2},
				{2256, 8946, 0x1702ffd7},
				{2257, 8945, 0x1702fdff},
				{2258, 8944, 0x1702fc2a},
				{2259, 8943, 0x1702fa51},
				{2260, 8942, 0x1702f87c},
				{2261, 8941, 0x1702f6a9},
				{2262, 8940, 0x1702f4d3},
				{2263, 8939, 0x1702f303},
				{2264, 8938, 0x1702f130},
				{2265, 8937, 0x1702ef60},
				{2266, 8936, 0x1702ed93},
				{2267, 8935, 0x1702ebc2},
				{2268, 8934, 0x1702e9f7},
				{2269, 8933, 0x1702e82a},
				{2270, 8932, 0x1702e65f},
				{2271, 8931, 0x1702e497},
				{2272, 8930, 0x1702e2cc},
				{2273, 8929, 0x1702e107},
				{2274, 8928, 0x1702df41},
				{2275, 8927, 0x1702dd7c},
				{2276, 8926, 0x1702dbb6},
				{2277, 8925, 0x1702d9f4},
				{2278, 8924, 0x1702d831},
				{2279, 8923, 0x1702d671},
				{2280, 8922, 0x1702d4b1},
				{2281, 8921, 0x1702d2f0},
				{2282, 8920, 0x1702d133},
				{2283, 8919, 0x1702cf76},
				{2284, 8918, 0x1702cdbb},
				{2285, 8917, 0x1702cbfd},
				{2286, 8916, 0x1702ca45},
				{2287, 8915, 0x1702c88b},
				{2288, 8914, 0x1702c6d5},
				{2289, 8913, 0x1702c520},
				{2290, 8912, 0x1702c368},
				{2291, 8911, 0x1702c1b3},
				{2292, 8910, 0x1702c000},
				{2293, 8909, 0x1702be4d},
				{2294, 8908, 0x1702bc9a},
				{2295, 8907, 0x1702baea},
				{2296, 8906, 0x1702b93a},
				{2297, 8905, 0x1702b78a},
				{2298, 8904, 0x1702b5dd},
				{2299, 8903, 0x1702b430},
				{2300, 8902, 0x1702b283},
				{2301, 8901, 0x1702b0d8},
				{2302, 8900, 0x1702af2d},
				{2303, 8899, 0x1702ad88},
				{2304, 8898, 0x1702abe4},
				{2305, 8897, 0x1702aa3f},
				{2306, 8896, 0x1702a89b},
				{2307, 8895, 0x1702a6f8},
				{2308, 8894, 0x1702a557},
				{2309, 8893, 0x1702a3b6},
				{2310, 8892, 0x1702a216},
				{2311, 8891, 0x1702a077},
				{2312, 8890, 0x17029eda},
				{2313, 8889, 0x17029d3d},
				{2314, 8888, 0x17029ba1},
				{2315, 8887, 0x17029a05},
				{2316, 8886, 0x1702986c},
				{2317, 8885, 0x170296d3},
				{2318, 8884, 0x1702953b},
				{2319, 8883, 0x170293a6},
				{2320, 8882, 0x1702920f},
				{2321, 8881, 0x1702907a},
				{2322, 8880, 0x17028ee5},
				{2323, 8879, 0x17028d52},
				{2324, 8878, 0x17028bc0},
				{2325, 8877, 0x17028a2d},
				{2326, 8876, 0x1702889e},
				{2327, 8875, 0x1702870e},
				{2328, 8874, 0x1702857f},
				{2329, 8873, 0x170283f1},
				{2330, 8872, 0x17028265},
				{2331, 8871, 0x170280d9},
				{2332, 8870, 0x17027f4e},
				{2333, 8869, 0x17027dc3},
				{2334, 8868, 0x17027c3d},
				{2335, 8867, 0x17027ab4},
				{2336, 8866, 0x1702792d},
				{2337, 8865, 0x170277a7},
				{2338, 8864, 0x17027622},
				{2339, 8863, 0x1702749d},
				{2340, 8862, 0x17027319},
				{2341, 8861, 0x17027197},
				{2342, 8860, 0x17027016},
				{2343, 8859, 0x17026e95},
				{2344, 8858, 0x17026d15},
				{2345, 8857, 0x17026b95},
				{2346, 8856, 0x17026a17},
				{2347, 8855, 0x1702689a},
				{2348, 8854, 0x1702671e},
				{2349, 8853, 0x170265a2},
				{2350, 8852, 0x17026429},
				{2351, 8851, 0x170262b0},
				{2352, 8850, 0x17026137},
				{2353, 8849, 0x17025fc0},
				{2354, 8848, 0x17025e48},
				{2355, 8847, 0x17025cd2},
				{2356, 8846, 0x17025b5d},
				{2357, 8845, 0x170259e9},
				{2358, 8844, 0x17025876},
				{2359, 8843, 0x17025703},
				{2360, 8842, 0x17025592},
				{2361, 8841, 0x17025421},
				{2362, 8840, 0x170252b2},
				{2363, 8839, 0x17025142},
				{2364, 8838, 0x17024fd4},
				{2365, 8837, 0x17024e68},
				{2366, 8836, 0x17024cfd},
				{2367, 8835, 0x17024b91},
				{2368, 8834, 0x17024a26},
				{2369, 8833, 0x170248bd},
				{2370, 8832, 0x17024754},
				{2371, 8831, 0x170245eb},
				{2372, 8830, 0x17024485},
				{2373, 8829, 0x1702431f},
				{2374, 8828, 0x170241b8},
				{2375, 8827, 0x17024055},
				{2376, 8826, 0x17023ef0},
				{2377, 8825, 0x17023d8e},
				{2378, 8824, 0x17023c2c},
				{2379, 8823, 0x17023acb},
				{2380, 8822, 0x1702396b},
				{2381, 8821, 0x1702380c},
				{2382, 8820, 0x170236ae},
				{2383, 8819, 0x1702354f},
				{2384, 8818, 0x170233f3},
				{2385, 8817, 0x17023297},
				{2386, 8816, 0x1702313c},
				{2387, 8815, 0x17022fe1},
				{2388, 8814, 0x17022e87},
				{2389, 8813, 0x17022d2f},
				{2390, 8812, 0x17022bd7},
				{2391, 8811, 0x17022a80},
				{2392, 8810, 0x1702292a},
				{2393, 8809, 0x170227d3},
				{2394, 8808, 0x17022680},
				{2395, 8807, 0x1702252b},
				{2396, 8806, 0x170223da},
				{2397, 8805, 0x17022288},
				{2398, 8804, 0x17022136},
				{2399, 8803, 0x17021fe6},
				{2400, 8802, 0x17021e96},
				{2401, 8801, 0x17021d47},
				{2402, 8800, 0x17021bf9},
				{2403, 8799, 0x17021aac},
				{2404, 8798, 0x1702195f},
				{2405, 8797, 0x17021814},
				{2406, 8796, 0x170216c8},
				{2407, 8795, 0x1702157e},
				{2408, 8794, 0x17021436},
				{2409, 8793, 0x170212ed},
				{2410, 8792, 0x170211a6},
				{2411, 8791, 0x17021060},
				{2412, 8790, 0x17020f1a},
				{2413, 8789, 0x17020dd5},
				{2414, 8788, 0x17020c90},
				{2415, 8787, 0x17020b4c},
				{2416, 8786, 0x17020a0a},
				{2417, 8785, 0x170208c8},
				{2418, 8784, 0x17020786},
				{2419, 8783, 0x17020645},
				{2420, 8782, 0x17020506},
				{2421, 8781, 0x170203c6},
				{2422, 8780, 0x17020288},
				{2423, 8779, 0x1702014b},
				{2424, 8778, 0x1702000e},
				{2425, 8777, 0x1701fed1},
				{2426, 8776, 0x1701fd97},
				{2427, 8775, 0x1701fc5d},
				{2428, 8774, 0x1701fb23},
				{2429, 8773, 0x1701f9eb},
				{2430, 8772, 0x1701f8b2},
				{2431, 8771, 0x1701f77b},
				{2432, 8770, 0x1701f645},
				{2433, 8769, 0x1701f50f},
				{2434, 8768, 0x1701f3d9},
				{2435, 8767, 0x1701f2a5},
				{2436, 8766, 0x1701f171},
				{2437, 8765, 0x1701f03e},
				{2438, 8764, 0x1701ef0c},
				{2439, 8763, 0x1701eddc},
				{2440, 8762, 0x1701ecaa},
				{2441, 8761, 0x1701eb7b},
				{2442, 8760, 0x1701ea4d},
				{2443, 8759, 0x1701e91f},
				{2444, 8758, 0x1701e7f1},
				{2445, 8757, 0x1701e6c3},
				{2446, 8756, 0x1701e598},
				{2447, 8755, 0x1701e46c},
				{2448, 8754, 0x1701e342},
				{2449, 8753, 0x1701e218},
				{2450, 8752, 0x1701e0ee},
				{2451, 8751, 0x1701dfc6},
				{2452, 8750, 0x1701de9d},
				{2453, 8749, 0x1701dd76},
				{2454, 8748, 0x1701dc50},
				{2455, 8747, 0x1701db2a},
				{2456, 8746, 0x1701da06},
				{2457, 8745, 0x1701d8e3},
				{2458, 8744, 0x1701d7be},
				{2459, 8743, 0x1701d69c},
				{2460, 8742, 0x1701d579},
				{2461, 8741, 0x1701d459},
				{2462, 8740, 0x1701d337},
				{2463, 8739, 0x1701d218},
				{2464, 8738, 0x1701d0f9},
				{2465, 8737, 0x1701cfda},
				{2466, 8736, 0x1701cebc},
				{2467, 8735, 0x1701cd9e},
				{2468, 8734, 0x1701cc82},
				{2469, 8733, 0x1701cb65},
				{2470, 8732, 0x1701ca4a},
				{2471, 8731, 0x1701c931},
				{2472, 8730, 0x1701c818},
				{2473, 8729, 0x1701c6fe},
				{2474, 8728, 0x1701c5e6},
				{2475, 8727, 0x1701c4cf},
				{2476, 8726, 0x1701c3b8},
				{2477, 8725, 0x1701c2a1},
				{2478, 8724, 0x1701c18b},
				{2479, 8723, 0x1701c077},
				{2480, 8722, 0x1701bf62},
				{2481, 8721, 0x1701be4e},
				{2482, 8720, 0x1701bd3b},
				{2483, 8719, 0x1701bc29},
				{2484, 8718, 0x1701bb18},
				{2485, 8717, 0x1701ba06},
				{2486, 8716, 0x1701b8f6},
				{2487, 8715, 0x1701b7e7},
				{2488, 8714, 0x1701b6d9},
				{2489, 8713, 0x1701b5ca},
				{2490, 8712, 0x1701b4bc},
				{2491, 8711, 0x1701b3b0},
				{2492, 8710, 0x1701b2a4},
				{2493, 8709, 0x1701b197},
				{2494, 8708, 0x1701b08c},
				{2495, 8707, 0x1701af81},
				{2496, 8706, 0x1701ae78},
				{2497, 8705, 0x1701ad6f},
				{2498, 8704, 0x1701ac67},
				{2499, 8703, 0x1701ab5f},
				{2500, 8702, 0x1701aa58},
				{2501, 8701, 0x1701a951},
				{2502, 8700, 0x1701a84b},
				{2503, 8699, 0x1701a747},
				{2504, 8698, 0x1701a641},
				{2505, 8697, 0x1701a53e},
				{2506, 8696, 0x1701a43a},
				{2507, 8695, 0x1701a337},
				{2508, 8694, 0x1701a236},
				{2509, 8693, 0x1701a134},
				{2510, 8692, 0x1701a032},
				{2511, 8691, 0x17019f32},
				{2512, 8690, 0x17019e33},
				{2513, 8689, 0x17019d34},
				{2514, 8688, 0x17019c35},
				{2515, 8687, 0x17019b38},
				{2516, 8686, 0x17019a3a},
				{2517, 8685, 0x1701993e},
				{2518, 8684, 0x17019843},
				{2519, 8683, 0x17019747},
				{2520, 8682, 0x1701964c},
				{2521, 8681, 0x17019553},
				{2522, 8680, 0x17019459},
				{2523, 8679, 0x17019360},
				{2524, 8678, 0x17019267},
				{2525, 8677, 0x1701916f},
				{2526, 8676, 0x17019078},
				{2527, 8675, 0x17018f81},
				{2528, 8674, 0x17018e8c},
				{2529, 8673, 0x17018d96},
				{2530, 8672, 0x17018ca1},
				{2531, 8671, 0x17018bad},
				{2532, 8670, 0x17018aba},
				{2533, 8669, 0x170189c6},
				{2534, 8668, 0x170188d4},
				{2535, 8667, 0x170187e3},
				{2536, 8666, 0x170186f1},
				{2537, 8665, 0x17018601},
				{2538, 8664, 0x17018510},
				{2539, 8663, 0x17018421},
				{2540, 8662, 0x17018332},
				{2541, 8661, 0x17018243},
				{2542, 8660, 0x17018154},
				{2543, 8659, 0x17018068},
				{2544, 8658, 0x17017f7b},
				{2545, 8657, 0x17017e8f},
				{2546, 8656, 0x17017da2},
				{2547, 8655, 0x17017cb8},
				{2548, 8654, 0x17017bcd},
				{2549, 8653, 0x17017ae3},
				{2550, 8652, 0x170179fa},
				{2551, 8651, 0x17017912},
				{2552, 8650, 0x17017829},
				{2553, 8649, 0x17017742},
				{2554, 8648, 0x1701765a},
				{2555, 8647, 0x17017573},
				{2556, 8646, 0x1701748d},
				{2557, 8645, 0x170173a7},
				{2558, 8644, 0x170172c3},
				{2559, 8643, 0x170171dd},
				{2560, 8642, 0x170170f9},
				{2561, 8641, 0x17017016},
				{2562, 8640, 0x17016f32},
				{2563, 8639, 0x17016e51},
				{2564, 8638, 0x17016d70},
				{2565, 8637, 0x17016c8e},
				{2566, 8636, 0x17016bad},
				{2567, 8635, 0x17016acd},
				{2568, 8634, 0x170169ed},
				{2569, 8633, 0x1701690e},
				{2570, 8632, 0x1701682f},
				{2571, 8631, 0x17016751},
				{2572, 8630, 0x17016673},
				{2573, 8629, 0x17016596},
				{2574, 8628, 0x170164b8},
				{2575, 8627, 0x170163dc},
				{2576, 8626, 0x17016300},
				{2577, 8625, 0x17016226},
				{2578, 8624, 0x1701614b},
				{2579, 8623, 0x17016070},
				{2580, 8622, 0x17015f97},
				{2581, 8621, 0x17015ebe},
				{2582, 8620, 0x17015de6},
				{2583, 8619, 0x17015d0e},
				{2584, 8618, 0x17015c36},
				{2585, 8617, 0x17015b5e},
				{2586, 8616, 0x17015a87},
				{2587, 8615, 0x170159b2},
				{2588, 8614, 0x170158db},
				{2589, 8613, 0x17015806},
				{2590, 8612, 0x17015732},
				{2591, 8611, 0x1701565e},
				{2592, 8610, 0x1701558c},
				{2593, 8609, 0x170154ba},
				{2594, 8608, 0x170153e9},
				{2595, 8607, 0x17015318},
				{2596, 8606, 0x17015247},
				{2597, 8605, 0x17015177},
				{2598, 8604, 0x170150a7},
				{2599, 8603, 0x17014fd9},
				{2600, 8602, 0x17014f0a},
				{2601, 8601, 0x17014e3c},
				{2602, 8600, 0x17014d6e},
				{2603, 8599, 0x17014ca0},
				{2604, 8598, 0x17014bd4},
				{2605, 8597, 0x17014b07},
				{2606, 8596, 0x17014a3c},
				{2607, 8595, 0x17014970},
				{2608, 8594, 0x170148a5},
				{2609, 8593, 0x170147db},
				{2610, 8592, 0x17014712},
				{2611, 8591, 0x17014648},
				{2612, 8590, 0x17014580},
				{2613, 8589, 0x170144b7},
				{2614, 8588, 0x170143ef},
				{2615, 8587, 0x17014327},
				{2616, 8586, 0x17014260},
				{2617, 8585, 0x1701419a},
				{2618, 8584, 0x170140d3},
				{2619, 8583, 0x1701400e},
				{2620, 8582, 0x17013f49},
				{2621, 8581, 0x17013e83},
				{2622, 8580, 0x17013dc0},
				{2623, 8579, 0x17013cfc},
				{2624, 8578, 0x17013c38},
				{2625, 8577, 0x17013b75},
				{2626, 8576, 0x17013ab3},
				{2627, 8575, 0x170139f2},
				{2628, 8574, 0x17013930},
				{2629, 8573, 0x1701386e},
				{2630, 8572, 0x170137ae},
				{2631, 8571, 0x170136ee},
				{2632, 8570, 0x1701362e},
				{2633, 8569, 0x1701356f},
				{2634, 8568, 0x170134b0},
				{2635, 8567, 0x170133f2},
				{2636, 8566, 0x17013334},
				{2637, 8565, 0x17013276},
				{2638, 8564, 0x170131b9},
				{2639, 8563, 0x170130fd},
				{2640, 8562, 0x17013041},
				{2641, 8561, 0x17012f86},
				{2642, 8560, 0x17012ecb},
				{2643, 8559, 0x17012e0f},
				{2644, 8558, 0x17012d56},
				{2645, 8557, 0x17012c9b},
				{2646, 8556, 0x17012be2},
				{2647, 8555, 0x17012b29},
				{2648, 8554, 0x17012a70},
				{2649, 8553, 0x170129b9},
				{2650, 8552, 0x17012901},
				{2651, 8551, 0x17012849},
				{2652, 8550, 0x17012793},
				{2653, 8549, 0x170126dc},
				{2654, 8548, 0x17012626},
				{2655, 8547, 0x17012570},
				{2656, 8546, 0x170124bc},
				{2657, 8545, 0x17012408},
				{2658, 8544, 0x17012353},
				{2659, 8543, 0x1701229f},
				{2660, 8542, 0x170121ec},
				{2661, 8541, 0x17012139},
				{2662, 8540, 0x17012087},
				{2663, 8539, 0x17011fd5},
				{2664, 8538, 0x17011f23},
				{2665, 8537, 0x17011e72},
				{2666, 8536, 0x17011dc1},
				{2667, 8535, 0x17011d11},
				{2668, 8534, 0x17011c61},
				{2669, 8533, 0x17011bb1},
				{2670, 8532, 0x17011b02},
				{2671, 8531, 0x17011a54},
				{2672, 8530, 0x170119a6},
				{2673, 8529, 0x170118f8},
				{2674, 8528, 0x1701184a},
				{2675, 8527, 0x1701179e},
				{2676, 8526, 0x170116f1},
				{2677, 8525, 0x17011645},
				{2678, 8524, 0x17011599},
				{2679, 8523, 0x170114ee},
				{2680, 8522, 0x17011443},
				{2681, 8521, 0x17011398},
				{2682, 8520, 0x170112ee},
				{2683, 8519, 0x17011244},
				{2684, 8518, 0x1701119b},
				{2685, 8517, 0x170110f2},
				{2686, 8516, 0x1701104a},
				{2687, 8515, 0x17010fa3},
				{2688, 8514, 0x17010efb},
				{2689, 8513, 0x17010e53},
				{2690, 8512, 0x17010dad},
				{2691, 8511, 0x17010d06},
				{2692, 8510, 0x17010c60},
				{2693, 8509, 0x17010bbb},
				{2694, 8508, 0x17010b16},
				{2695, 8507, 0x17010a70},
				{2696, 8506, 0x170109cc},
				{2697, 8505, 0x17010928},
				{2698, 8504, 0x17010884},
				{2699, 8503, 0x170107e1},
				{2700, 8502, 0x1701073e},
				{2701, 8501, 0x1701069c},
				{2702, 8500, 0x170105f9},
				{2703, 8499, 0x17010558},
				{2704, 8498, 0x170104b7},
				{2705, 8497, 0x17010416},
				{2706, 8496, 0x17010376},
				{2707, 8495, 0x170102d6},
				{2708, 8494, 0x17010236},
				{2709, 8493, 0x17010197},
				{2710, 8492, 0x170100f8},
				{2711, 8491, 0x17010059},
				{2712, 8490, 0x1700ffbb},
				{2713, 8489, 0x1700ff1d},
				{2714, 8488, 0x1700fe80},
				{2715, 8487, 0x1700fde3},
				{2716, 8486, 0x1700fd46},
				{2717, 8485, 0x1700fcaa},
				{2718, 8484, 0x1700fc0e},
				{2719, 8483, 0x1700fb73},
				{2720, 8482, 0x1700fad8},
				{2721, 8481, 0x1700fa3d},
				{2722, 8480, 0x1700f9a3},
				{2723, 8479, 0x1700f909},
				{2724, 8478, 0x1700f86f},
				{2725, 8477, 0x1700f7d6},
				{2726, 8476, 0x1700f73d},
				{2727, 8475, 0x1700f6a5},
				{2728, 8474, 0x1700f60c},
				{2729, 8473, 0x1700f575},
				{2730, 8472, 0x1700f4dd},
				{2731, 8471, 0x1700f446},
				{2732, 8470, 0x1700f3af},
				{2733, 8469, 0x1700f31a},
				{2734, 8468, 0x1700f284},
				{2735, 8467, 0x1700f1ee},
				{2736, 8466, 0x1700f159},
				{2737, 8465, 0x1700f0c5},
				{2738, 8464, 0x1700f030},
				{2739, 8463, 0x1700ef9b},
				{2740, 8462, 0x1700ef08},
				{2741, 8461, 0x1700ee74},
				{2742, 8460, 0x1700ede1},
				{2743, 8459, 0x1700ed4f},
				{2744, 8458, 0x1700ecbc},
				{2745, 8457, 0x1700ec2b},
				{2746, 8456, 0x1700eb99},
				{2747, 8455, 0x1700eb08},
				{2748, 8454, 0x1700ea77},
				{2749, 8453, 0x1700e9e7},
				{2750, 8452, 0x1700e957},
				{2751, 8451, 0x1700e8c7},
				{2752, 8450, 0x1700e837},
				{2753, 8449, 0x1700e7a8},
				{2754, 8448, 0x1700e719},
				{2755, 8447, 0x1700e68b},
				{2756, 8446, 0x1700e5fd},
				{2757, 8445, 0x1700e56f},
				{2758, 8444, 0x1700e4e2},
				{2759, 8443, 0x1700e454},
				{2760, 8442, 0x1700e3c8},
				{2761, 8441, 0x1700e33b},
				{2762, 8440, 0x1700e2b0},
				{2763, 8439, 0x1700e223},
				{2764, 8438, 0x1700e199},
				{2765, 8437, 0x1700e10e},
				{2766, 8436, 0x1700e084},
				{2767, 8435, 0x1700dff9},
				{2768, 8434, 0x1700df6f},
				{2769, 8433, 0x1700dee5},
				{2770, 8432, 0x1700de5c},
				{2771, 8431, 0x1700ddd3},
				{2772, 8430, 0x1700dd4a},
				{2773, 8429, 0x1700dcc2},
				{2774, 8428, 0x1700dc3a},
				{2775, 8427, 0x1700dbb2},
				{2776, 8426, 0x1700db2b},
				{2777, 8425, 0x1700daa4},
				{2778, 8424, 0x1700da1d},
				{2779, 8423, 0x1700d997},
				{2780, 8422, 0x1700d911},
				{2781, 8421, 0x1700d88c},
				{2782, 8420, 0x1700d806},
				{2783, 8419, 0x1700d781},
				{2784, 8418, 0x1700d6fd},
				{2785, 8417, 0x1700d678},
				{2786, 8416, 0x1700d5f4},
				{2787, 8415, 0x1700d570},
				{2788, 8414, 0x1700d4ed},
				{2789, 8413, 0x1700d46a},
				{2790, 8412, 0x1700d3e7},
				{2791, 8411, 0x1700d364},
				{2792, 8410, 0x1700d2e2},
				{2793, 8409, 0x1700d260},
				{2794, 8408, 0x1700d1de},
				{2795, 8407, 0x1700d15e},
				{2796, 8406, 0x1700d0dd},
				{2797, 8405, 0x1700d05c},
				{2798, 8404, 0x1700cfdc},
				{2799, 8403, 0x1700cf5c},
				{2800, 8402, 0x1700cedc},
				{2801, 8401, 0x1700ce5d},
				{2802, 8400, 0x1700cddd},
				{2803, 8399, 0x1700cd5f},
				{2804, 8398, 0x1700cce1},
				{2805, 8397, 0x1700cc62},
				{2806, 8396, 0x1700cbe4},
				{2807, 8395, 0x1700cb67},
				{2808, 8394, 0x1700cae9},
				{2809, 8393, 0x1700ca6d},
				{2810, 8392, 0x1700c9f1},
				{2811, 8391, 0x1700c974},
				{2812, 8390, 0x1700c8f8},
				{2813, 8389, 0x1700c87c},
				{2814, 8388, 0x1700c801},
				{2815, 8387, 0x1700c785},
				{2816, 8386, 0x1700c70b},
				{2817, 8385, 0x1700c690},
				{2818, 8384, 0x1700c616},
				{2819, 8383, 0x1700c59c},
				{2820, 8382, 0x1700c522},
				{2821, 8381, 0x1700c4a9},
				{2822, 8380, 0x1700c42f},
				{2823, 8379, 0x1700c3b7},
				{2824, 8378, 0x1700c33e},
				{2825, 8377, 0x1700c2c7},
				{2826, 8376, 0x1700c24f},
				{2827, 8375, 0x1700c1d7},
				{2828, 8374, 0x1700c160},
				{2829, 8373, 0x1700c0e8},
				{2830, 8372, 0x1700c072},
				{2831, 8371, 0x1700bffb},
				{2832, 8370, 0x1700bf85},
				{2833, 8369, 0x1700bf0f},
				{2834, 8368, 0x1700be99},
				{2835, 8367, 0x1700be23},
				{2836, 8366, 0x1700bdaf},
				{2837, 8365, 0x1700bd39},
				{2838, 8364, 0x1700bcc4},
				{2839, 8363, 0x1700bc50},
				{2840, 8362, 0x1700bbdc},
				{2841, 8361, 0x1700bb69},
				{2842, 8360, 0x1700baf6},
				{2843, 8359, 0x1700ba82},
				{2844, 8358, 0x1700ba0f},
				{2845, 8357, 0x1700b99d},
				{2846, 8356, 0x1700b92a},
				{2847, 8355, 0x1700b8b8},
				{2848, 8354, 0x1700b846},
				{2849, 8353, 0x1700b7d5},
				{2850, 8352, 0x1700b763},
				{2851, 8351, 0x1700b6f2},
				{2852, 8350, 0x1700b681},
				{2853, 8349, 0x1700b611},
				{2854, 8348, 0x1700b5a0},
				{2855, 8347, 0x1700b530},
				{2856, 8346, 0x1700b4c1},
				{2857, 8345, 0x1700b451},
				{2858, 8344, 0x1700b3e2},
				{2859, 8343, 0x1700b373},
				{2860, 8342, 0x1700b304},
				{2861, 8341, 0x1700b296},
				{2862, 8340, 0x1700b228},
				{2863, 8339, 0x1700b1ba},
				{2864, 8338, 0x1700b14c},
				{2865, 8337, 0x1700b0de},
				{2866, 8336, 0x1700b071},
				{2867, 8335, 0x1700b004},
				{2868, 8334, 0x1700af97},
				{2869, 8333, 0x1700af2a},
				{2870, 8332, 0x1700aebe},
				{2871, 8331, 0x1700ae53},
				{2872, 8330, 0x1700ade7},
				{2873, 8329, 0x1700ad7c},
				{2874, 8328, 0x1700ad10},
				{2875, 8327, 0x1700aca5},
				{2876, 8326, 0x1700ac3a},
				{2877, 8325, 0x1700abd0},
				{2878, 8324, 0x1700ab66},
				{2879, 8323, 0x1700aafd},
				{2880, 8322, 0x1700aa94},
				{2881, 8321, 0x1700aa2b},
				{2882, 8320, 0x1700a9c2},
				{2883, 8319, 0x1700a959},
				{2884, 8318, 0x1700a8f1},
				{2885, 8317, 0x1700a889},
				{2886, 8316, 0x1700a821},
				{2887, 8315, 0x1700a7bb},
				{2888, 8314, 0x1700a753},
				{2889, 8313, 0x1700a6ec},
				{2890, 8312, 0x1700a686},
				{2891, 8311, 0x1700a61f},
				{2892, 8310, 0x1700a5b9},
				{2893, 8309, 0x1700a553},
				{2894, 8308, 0x1700a4ed},
				{2895, 8307, 0x1700a487},
				{2896, 8306, 0x1700a422},
				{2897, 8305, 0x1700a3bd},
				{2898, 8304, 0x1700a358},
				{2899, 8303, 0x1700a2f3},
				{2900, 8302, 0x1700a28f},
				{2901, 8301, 0x1700a22b},
				{2902, 8300, 0x1700a1c7},
				{2903, 8299, 0x1700a164},
				{2904, 8298, 0x1700a100},
				{2905, 8297, 0x1700a09d},
				{2906, 8296, 0x1700a03a},
				{2907, 8295, 0x17009fd7},
				{2908, 8294, 0x17009f75},
				{2909, 8293, 0x17009f13},
				{2910, 8292, 0x17009eb1},
				{2911, 8291, 0x17009e4f},
				{2912, 8290, 0x17009ded},
				{2913, 8289, 0x17009d8c},
				{2914, 8288, 0x17009d2b},
				{2915, 8287, 0x17009cca},
				{2916, 8286, 0x17009c69},
				{2917, 8285, 0x17009c09},
				{2918, 8284, 0x17009ba9},
				{2919, 8283, 0x17009b49},
				{2920, 8282, 0x17009ae9},
				{2921, 8281, 0x17009a8a},
				{2922, 8280, 0x17009a2a},
				{2923, 8279, 0x170099cb},
				{2924, 8278, 0x1700996c},
				{2925, 8277, 0x1700990e},
				{2926, 8276, 0x170098af},
				{2927, 8275, 0x17009851},
				{2928, 8274, 0x170097f3},
				{2929, 8273, 0x17009795},
				{2930, 8272, 0x17009738},
				{2931, 8271, 0x170096db},
				{2932, 8270, 0x1700967e},
				{2933, 8269, 0x17009621},
				{2934, 8268, 0x170095c4},
				{2935, 8267, 0x17009568},
				{2936, 8266, 0x1700950c},
				{2937, 8265, 0x170094b0},
				{2938, 8264, 0x17009454},
				{2939, 8263, 0x170093f9},
				{2940, 8262, 0x1700939d},
				{2941, 8261, 0x17009342},
				{2942, 8260, 0x170092e7},
				{2943, 8259, 0x1700928d},
				{2944, 8258, 0x17009232},
				{2945, 8257, 0x170091d8},
				{2946, 8256, 0x1700917e},
				{2947, 8255, 0x17009124},
				{2948, 8254, 0x170090cb},
				{2949, 8253, 0x17009072},
				{2950, 8252, 0x17009019},
				{2951, 8251, 0x17008fc0},
				{2952, 8250, 0x17008f67},
				{2953, 8249, 0x17008f0f},
				{2954, 8248, 0x17008eb6},
				{2955, 8247, 0x17008e5e},
				{2956, 8246, 0x17008e06},
				{2957, 8245, 0x17008dae},
				{2958, 8244, 0x17008d57},
				{2959, 8243, 0x17008d00},
				{2960, 8242, 0x17008ca9},
				{2961, 8241, 0x17008c52},
				{2962, 8240, 0x17008bfb},
				{2963, 8239, 0x17008ba5},
				{2964, 8238, 0x17008b4f},
				{2965, 8237, 0x17008af9},
				{2966, 8236, 0x17008aa3},
				{2967, 8235, 0x17008a4e},
				{2968, 8234, 0x170089f9},
				{2969, 8233, 0x170089a3},
				{2970, 8232, 0x1700894e},
				{2971, 8231, 0x170088fa},
				{2972, 8230, 0x170088a5},
				{2973, 8229, 0x17008851},
				{2974, 8228, 0x170087fc},
				{2975, 8227, 0x170087a9},
				{2976, 8226, 0x17008755},
				{2977, 8225, 0x17008701},
				{2978, 8224, 0x170086ae},
				{2979, 8223, 0x1700865b},
				{2980, 8222, 0x17008608},
				{2981, 8221, 0x170085b5},
				{2982, 8220, 0x17008563},
				{2983, 8219, 0x17008511},
				{2984, 8218, 0x170084bf},
				{2985, 8217, 0x1700846c},
				{2986, 8216, 0x1700841b},
				{2987, 8215, 0x170083c9},
				{2988, 8214, 0x17008378},
				{2989, 8213, 0x17008327},
				{2990, 8212, 0x170082d6},
				{2991, 8211, 0x17008285},
				{2992, 8210, 0x17008235},
				{2993, 8209, 0x170081e4},
				{2994, 8208, 0x17008194},
				{2995, 8207, 0x17008144},
				{2996, 8206, 0x170080f5},
				{2997, 8205, 0x170080a5},
				{2998, 8204, 0x17008056},
				{2999, 8203, 0x17008006},
				{3000, 8202, 0x167fb822},
				{3001, 8201, 0x167f6950},
				{3002, 8200, 0x167f1a7e},
				{3003, 8199, 0x167ecc02},
				{3004, 8198, 0x167e7ddc},
				{3005, 8197, 0x167e2fb6},
				{3006, 8196, 0x167de1e5},
				{3007, 8195, 0x167d946b},
				{3008, 8194, 0x167d46f1},
				{3009, 8193, 0x167cf9cc},
				{3010, 8192, 0x167caca7},
				{3011, 8191, 0x167c5fd9},
				{3012, 8190, 0x167c130a},
				{3013, 8189, 0x167bc691},
				{3014, 8188, 0x167b7a6e},
				{3015, 8187, 0x167b2df5},
				{3016, 8186, 0x167ae228},
				{3017, 8185, 0x167a965b},
				{3018, 8184, 0x167a4a8e},
				{3019, 8183, 0x1679ff6d},
				{3020, 8182, 0x1679b3f6},
				{3021, 8181, 0x167968d4},
				{3022, 8180, 0x16791e09},
				{3023, 8179, 0x1678d33d},
				{3024, 8178, 0x167888c8},
				{3025, 8177, 0x16783efe},
				{3026, 8176, 0x1677f4de},
				{3027, 8175, 0x1677aabe},
				{3028, 8174, 0x167760f4},
				{3029, 8173, 0x16771780},
				{3030, 8172, 0x1676ce0c},
				{3031, 8171, 0x16768498},
				{3032, 8170, 0x16763b7a},
				{3033, 8169, 0x1675f2b1},
				{3034, 8168, 0x1675a9e9},
				{3035, 8167, 0x16756177},
				{3036, 8166, 0x16751904},
				{3037, 8165, 0x1674d0e8},
				{3038, 8164, 0x167488cb},
				{3039, 8163, 0x16744104},
				{3040, 8162, 0x1673f993},
				{3041, 8161, 0x1673b222},
				{3042, 8160, 0x16736ab2},
				{3043, 8159, 0x16732397},
				{3044, 8158, 0x1672dcd1},
				{3045, 8157, 0x1672960c},
				{3046, 8156, 0x16724f47},
				{3047, 8155, 0x167208d8},
				{3048, 8154, 0x1671c2be},
				{3049, 8153, 0x16717c4f},
				{3050, 8152, 0x1671368b},
				{3051, 8151, 0x1670f0c8},
				{3052, 8150, 0x1670ab04},
				{3053, 8149, 0x16706597},
				{3054, 8148, 0x1670207f},
				{3055, 8147, 0x166fdbbd},
				{3056, 8146, 0x166f96a5},
				{3057, 8145, 0x166f51e3},
				{3058, 8144, 0x166f0d77},
				{3059, 8143, 0x166ec90b},
				{3060, 8142, 0x166e849f},
				{3061, 8141, 0x166e4088},
				{3062, 8140, 0x166dfcc8},
				{3063, 8139, 0x166db908},
				{3064, 8138, 0x166d7547},
				{3065, 8137, 0x166d31dd},
				{3066, 8136, 0x166ceec8},
				{3067, 8135, 0x166cabb4},
				{3068, 8134, 0x166c689f},
				{3069, 8133, 0x166c25e0},
				{3070, 8132, 0x166be321},
				{3071, 8131, 0x166ba10e},
				{3072, 8130, 0x166b5ea5},
				{3073, 8129, 0x166b1c92},
				{3074, 8128, 0x166ada7f},
				{3075, 8127, 0x166a98c2},
				{3076, 8126, 0x166a575b},
				{3077, 8125, 0x166a159d},
				{3078, 8124, 0x1669d436},
				{3079, 8123, 0x16699325},
				{3080, 8122, 0x16695213},
				{3081, 8121, 0x16691157},
				{3082, 8120, 0x1668d09c},
				{3083, 8119, 0x16688fe0},
				{3084, 8118, 0x16684f7b},
				{3085, 8117, 0x16680f6b},
				{3086, 8116, 0x1667cfb1},
				{3087, 8115, 0x16678fa1},
				{3088, 8114, 0x16674fe7},
				{3089, 8113, 0x1667102d},
				{3090, 8112, 0x1666d0c8},
				{3091, 8111, 0x16669164},
				{3092, 8110, 0x16665256},
				{3093, 8109, 0x16661348},
				{3094, 8108, 0x1665d48f},
				{3095, 8107, 0x166595d7},
				{3096, 8106, 0x1665571e},
				{3097, 8105, 0x166518bc},
				{3098, 8104, 0x1664da59},
				{3099, 8103, 0x16649c4d},
				{3100, 8102, 0x16645e96},
				{3101, 8101, 0x166420df},
				{3102, 8100, 0x1663e328},
				{3103, 8099, 0x1663a5c7},
				{3104, 8098, 0x16636866},
				{3105, 8097, 0x16632b05},
				{3106, 8096, 0x1662edfa},
				{3107, 8095, 0x1662b0ef},
				{3108, 8094, 0x1662743a},
				{3109, 8093, 0x16623784},
				{3110, 8092, 0x1661fb25},
				{3111, 8091, 0x1661bec6},
				{3112, 8090, 0x166182bc},
				{3113, 8089, 0x166146b3},
				{3114, 8088, 0x16610aa9},
				{3115, 8087, 0x1660cef5},
				{3116, 8086, 0x16609342},
				{3117, 8085, 0x166057e4},
				{3118, 8084, 0x16601c86},
				{3119, 8083, 0x165fe17e},
				{3120, 8082, 0x165fa676},
				{3121, 8081, 0x165f6b6e},
				{3122, 8080, 0x165f30bc},
				{3123, 8079, 0x165ef60a},
				{3124, 8078, 0x165ebb58},
				{3125, 8077, 0x165e80fc},
				{3126, 8076, 0x165e46a0},
				{3127, 8075, 0x165e0c99},
				{3128, 8074, 0x165dd293},
				{3129, 8073, 0x165d98e2},
				{3130, 8072, 0x165d5f32},
				{3131, 8071, 0x165d2581},
				{3132, 8070, 0x165cec7d},
				{3133, 8069, 0x165cb322},
				{3134, 8068, 0x165c7a1d},
				{3135, 8067, 0x165c4119},
				{3136, 8066, 0x165c0814},
				{3137, 8065, 0x165bcf65},
				{3138, 8064, 0x165b96b6},
				{3139, 8063, 0x165b5e5d},
				{3140, 8062, 0x165b2604},
				{3141, 8061, 0x165aedab},
				{3142, 8060, 0x165ab5a8},
				{3143, 8059, 0x165a7da5},
				{3144, 8058, 0x165a45a1},
				{3145, 8057, 0x165a0df4},
				{3146, 8056, 0x1659d647},
				{3147, 8055, 0x16599f45},
				{3148, 8054, 0x165967ee},
				{3149, 8053, 0x16593096},
				{3150, 8052, 0x1658f995},
				{3151, 8051, 0x1658c293},
				{3152, 8050, 0x16588be8},
				{3153, 8049, 0x1658553c},
				{3154, 8048, 0x16581e90},
				{3155, 8047, 0x1657e83a},
				{3156, 8046, 0x1657b1e4},
				{3157, 8045, 0x16577b8f},
				{3158, 8044, 0x1657458f},
				{3159, 8043, 0x16570f8f},
				{3160, 8042, 0x1656d9e4},
				{3161, 8041, 0x1656a43a},
				{3162, 8040, 0x16566e90},
				{3163, 8039, 0x1656393c},
				{3164, 8038, 0x165603e8},
				{3165, 8037, 0x1655cee9},
				{3166, 8036, 0x16559a16},
				{3167, 8035, 0x1655656d},
				{3168, 8034, 0x165530f0},
				{3169, 8033, 0x1654fc72},
				{3170, 8032, 0x1654c84a},
				{3171, 8031, 0x16549423},
				{3172, 8030, 0x16545ffb},
				{3173, 8029, 0x16542c29},
				{3174, 8028, 0x1653f857},
				{3175, 8027, 0x1653c4b0},
				{3176, 8026, 0x16539109},
				{3177, 8025, 0x16535d8d},
				{3178, 8024, 0x16532a92},
				{3179, 8023, 0x1652f741},
				{3180, 8022, 0x1652c446},
				{3181, 8021, 0x1652914a},
				{3182, 8020, 0x16525e7a},
				{3183, 8019, 0x16522baa},
				{3184, 8018, 0x1651f904},
				{3185, 8017, 0x1651c68a},
				{3186, 8016, 0x1651943b},
				{3187, 8015, 0x165161eb},
				{3188, 8014, 0x16512fc7},
				{3189, 8013, 0x1650fdcd},
				{3190, 8012, 0x1650cbfe},
				{3191, 8011, 0x16509a2f},
				{3192, 8010, 0x1650688c},
				{3193, 8009, 0x16503713},
				{3194, 8008, 0x165005c5},
				{3195, 8007, 0x164fd477},
				{3196, 8006, 0x164fa329},
				{3197, 8005, 0x164f7206},
				{3198, 8004, 0x164f410e},
				{3199, 8003, 0x164f1041},
				{3200, 8002, 0x164edf9f},
				{3201, 8001, 0x164eaefd},
				{3202, 8000, 0x164e7e5b},
				{3203, 7999, 0x164e4e0e},
				{3204, 7998, 0x164e1dc2},
				{3205, 7997, 0x164deda1},
				{3206, 7996, 0x164dbd7f},
				{3207, 7995, 0x164d8d89},
				{3208, 7994, 0x164d5dbd},
				{3209, 7993, 0x164d2e48},
				{3210, 7992, 0x164cfea7},
				{3211, 7991, 0x164ccf31},
				{3212, 7990, 0x164c9fbc},
				{3213, 7989, 0x164c709c},
				{3214, 7988, 0x164c4151},
				{3215, 7987, 0x164c125c},
				{3216, 7986, 0x164be367},
				{3217, 7985, 0x164bb49d},
				{3218, 7984, 0x164b85d4},
				{3219, 7983, 0x164b575f},
				{3220, 7982, 0x164b28c0},
				{3221, 7981, 0x164afa77},
				{3222, 7980, 0x164acc2e},
				{3223, 7979, 0x164a9e10},
				{3224, 7978, 0x164a7048},
				{3225, 7977, 0x164a4254},
				{3226, 7976, 0x164a1461},
				{3227, 7975, 0x1649e6c4},
				{3228, 7974, 0x1649b926},
				{3229, 7973, 0x16498bb4},
				{3230, 7972, 0x16495e41},
				{3231, 7971, 0x164930fa},
				{3232, 7970, 0x164903dd},
				{3233, 7969, 0x1648d6c1},
				{3234, 7968, 0x1648a9cf},
				{3235, 7967, 0x16487d08},
				{3236, 7966, 0x16485041},
				{3237, 7965, 0x164823a6},
				{3238, 7964, 0x1647f70a},
				{3239, 7963, 0x1647cac4},
				{3240, 7962, 0x16479e7e},
				{3241, 7961, 0x16477263},
				{3242, 7960, 0x16474648},
				{3243, 7959, 0x16471a2d},
				{3244, 7958, 0x1646ee68},
				{3245, 7957, 0x1646c2a3},
				{3246, 7956, 0x164696dd},
				{3247, 7955, 0x16466b43},
				{3248, 7954, 0x16463fd4},
				{3249, 7953, 0x1646148f},
				{3250, 7952, 0x1645e94b},
				{3251, 7951, 0x1645be07},
				{3252, 7950, 0x16459318},
				{3253, 7949, 0x1645682a},
				{3254, 7948, 0x16453d3b},
				{3255, 7947, 0x164512a3},
				{3256, 7946, 0x1644e80a},
				{3257, 7945, 0x1644bd9c},
				{3258, 7944, 0x1644932f},
				{3259, 7943, 0x164468c1},
				{3260, 7942, 0x16443e7e},
				{3261, 7941, 0x16441466},
				{3262, 7940, 0x1643ea79},
				{3263, 7939, 0x1643c08c},
				{3264, 7938, 0x164396a0},
				{3265, 7937, 0x16436cde},
				{3266, 7936, 0x16434346},
				{3267, 7935, 0x164319da},
				{3268, 7934, 0x1642f06e},
				{3269, 7933, 0x1642c702},
				{3270, 7932, 0x16429e17},
				{3271, 7931, 0x164274d6},
				{3272, 7930, 0x16424bea},
				{3273, 7929, 0x164222ff},
				{3274, 7928, 0x1641fa14},
				{3275, 7927, 0x1641d17e},
				{3276, 7926, 0x1641a8be},
				{3277, 7925, 0x16418053},
				{3278, 7924, 0x164157e9},
				{3279, 7923, 0x16412f7e},
				{3280, 7922, 0x1641073f},
				{3281, 7921, 0x1640df2a},
				{3282, 7920, 0x1640b715},
				{3283, 7919, 0x16408f2c},
				{3284, 7918, 0x1640676d},
				{3285, 7917, 0x16403fd9},
				{3286, 7916, 0x1640181a},
				{3287, 7915, 0x163ff086},
				{3288, 7914, 0x163fc91d},
				{3289, 7913, 0x163fa1df},
				{3290, 7912, 0x163f7a76},
				{3291, 7911, 0x163f5363},
				{3292, 7910, 0x163f2c50},
				{3293, 7909, 0x163f053d},
				{3294, 7908, 0x163ede80},
				{3295, 7907, 0x163eb798},
				{3296, 7906, 0x163e9106},
				{3297, 7905, 0x163e6a48},
				{3298, 7904, 0x163e43e1},
				{3299, 7903, 0x163e1d7a},
				{3300, 7902, 0x163df712},
				{3301, 7901, 0x163dd101},
				{3302, 7900, 0x163daaef},
				{3303, 7899, 0x163d84de},
				{3304, 7898, 0x163d5ef7},
				{3305, 7897, 0x163d3911},
				{3306, 7896, 0x163d1355},
				{3307, 7895, 0x163ced9a},
				{3308, 7894, 0x163cc809},
				{3309, 7893, 0x163ca2a3},
				{3310, 7892, 0x163c7d3d},
				{3311, 7891, 0x163c57d8},
				{3312, 7890, 0x163c329d},
				{3313, 7889, 0x163c0d8d},
				{3314, 7888, 0x163be87d},
				{3315, 7887, 0x163bc398},
				{3316, 7886, 0x163b9ede},
				{3317, 7885, 0x163b7a24},
				{3318, 7884, 0x163b556a},
				{3319, 7883, 0x163b30db},
				{3320, 7882, 0x163b0c4c},
				{3321, 7881, 0x163ae7e8},
				{3322, 7880, 0x163ac3ae},
				{3323, 7879, 0x163a9f75},
				{3324, 7878, 0x163a7b3c},
				{3325, 7877, 0x163a572e},
				{3326, 7876, 0x163a334a},
				{3327, 7875, 0x163a0f67},
				{3328, 7874, 0x1639eb83},
				{3329, 7873, 0x1639c7cb},
				{3330, 7872, 0x1639a43d},
				{3331, 7871, 0x163980db},
				{3332, 7870, 0x16395d78},
				{3333, 7869, 0x16393a16},
				{3334, 7868, 0x163916de},
				{3335, 7867, 0x1638f3a6},
				{3336, 7866, 0x1638d09a},
				{3337, 7865, 0x1638ad8d},
				{3338, 7864, 0x16388a80},
				{3339, 7863, 0x163867c9},
				{3340, 7862, 0x163844e8},
				{3341, 7861, 0x1638225c},
				{3342, 7860, 0x1637ffa5},
				{3343, 7859, 0x1637dd44},
				{3344, 7858, 0x1637bab8},
				{3345, 7857, 0x16379882},
				{3346, 7856, 0x16377621},
				{3347, 7855, 0x16375441},
				{3348, 7854, 0x1637320b},
				{3349, 7853, 0x16371000},
				{3350, 7852, 0x1636ee1f},
				{3351, 7851, 0x1636cc3f},
				{3352, 7850, 0x1636aa8a},
				{3353, 7849, 0x163688d5},
				{3354, 7848, 0x1636674a},
				{3355, 7847, 0x163645c0},
				{3356, 7846, 0x16362461},
				{3357, 7845, 0x16360301},
				{3358, 7844, 0x1635e1cd},
				{3359, 7843, 0x1635c098},
				{3360, 7842, 0x16359f64},
				{3361, 7841, 0x16357e5a},
				{3362, 7840, 0x16355da7},
				{3363, 7839, 0x16353cc8},
				{3364, 7838, 0x16351be9},
				{3365, 7837, 0x1634fb36},
				{3366, 7836, 0x1634da82},
				{3367, 7835, 0x1634b9f9},
				{3368, 7834, 0x1634999c},
				{3369, 7833, 0x1634793e},
				{3370, 7832, 0x163458e0},
				{3371, 7831, 0x163438ad},
				{3372, 7830, 0x1634187a},
				{3373, 7829, 0x1633f872},
				{3374, 7828, 0x1633d86a},
				{3375, 7827, 0x1633b862},
				{3376, 7826, 0x16339885},
				{3377, 7825, 0x163378d3},
				{3378, 7824, 0x1633594c},
				{3379, 7823, 0x1633399a},
				{3380, 7822, 0x16331a13},
				{3381, 7821, 0x1632fa8c},
				{3382, 7820, 0x1632db30},
				{3383, 7819, 0x1632bbd3},
				{3384, 7818, 0x16329ca2},
				{3385, 7817, 0x16327d71},
				{3386, 7816, 0x16325e6a},
				{3387, 7815, 0x16323f64},
				{3388, 7814, 0x1632205e},
				{3389, 7813, 0x16320182},
				{3390, 7812, 0x1631e2d2},
				{3391, 7811, 0x1631c3f6},
				{3392, 7810, 0x1631a571},
				{3393, 7809, 0x163186eb},
				{3394, 7808, 0x16316866},
				{3395, 7807, 0x16314a0b},
				{3396, 7806, 0x16312bb0},
				{3397, 7805, 0x16310d56},
				{3398, 7804, 0x1630ef26},
				{3399, 7803, 0x1630d0f6},
				{3400, 7802, 0x1630b2f2},
				{3401, 7801, 0x163094ed},
				{3402, 7800, 0x16307713},
				{3403, 7799, 0x16305939},
				{3404, 7798, 0x16303b5f},
				{3405, 7797, 0x16301db1},
				{3406, 7796, 0x16300002},
				{3407, 7795, 0x162fe27e},
				{3408, 7794, 0x162fc525},
				{3409, 7793, 0x162fa7a1},
				{3410, 7792, 0x162f8a48},
				{3411, 7791, 0x162f6cef},
				{3412, 7790, 0x162f4fc0},
				{3413, 7789, 0x162f3292},
				{3414, 7788, 0x162f158f},
				{3415, 7787, 0x162ef88c},
				{3416, 7786, 0x162edb89},
				{3417, 7785, 0x162ebeb1},
				{3418, 7784, 0x162ea1d8},
				{3419, 7783, 0x162e8500},
				{3420, 7782, 0x162e6853},
				{3421, 7781, 0x162e4bd0},
				{3422, 7780, 0x162e2f4e},
				{3423, 7779, 0x162e12cc},
				{3424, 7778, 0x162df674},
				{3425, 7777, 0x162dda1d},
				{3426, 7776, 0x162dbdc5},
				{3427, 7775, 0x162da199},
				{3428, 7774, 0x162d856c},
				{3429, 7773, 0x162d696b},
				{3430, 7772, 0x162d4d69},
				{3431, 7771, 0x162d3168},
				{3432, 7770, 0x162d1591},
				{3433, 7769, 0x162cf9ba},
				{3434, 7768, 0x162cdde4},
				{3435, 7767, 0x162cc238},
				{3436, 7766, 0x162ca68c},
				{3437, 7765, 0x162c8b0b},
				{3438, 7764, 0x162c6f8b},
				{3439, 7763, 0x162c5435},
				{3440, 7762, 0x162c38df},
				{3441, 7761, 0x162c1d89},
				{3442, 7760, 0x162c0233},
				{3443, 7759, 0x162be708},
				{3444, 7758, 0x162bcc08},
				{3445, 7757, 0x162bb0dd},
				{3446, 7756, 0x162b95dd},
				{3447, 7755, 0x162b7add},
				{3448, 7754, 0x162b6008},
				{3449, 7753, 0x162b4533},
				{3450, 7752, 0x162b2a89},
				{3451, 7751, 0x162b0fb4},
				{3452, 7750, 0x162af535},
				{3453, 7749, 0x162adab6},
				{3454, 7748, 0x162ac077},
				{3455, 7747, 0x162aa622},
				{3456, 7746, 0x162a8bf9},
				{3457, 7745, 0x162a71ba},
				{3458, 7744, 0x162a57a6},
				{3459, 7743, 0x162a3d93},
				{3460, 7742, 0x162a2394},
				{3461, 7741, 0x162a09ab},
				{3462, 7740, 0x1629efc2},
				{3463, 7739, 0x1629d5ef},
				{3464, 7738, 0x1629bc31},
				{3465, 7737, 0x1629a288},
				{3466, 7736, 0x162988e0},
				{3467, 7735, 0x16296f4d},
				{3468, 7734, 0x162955cf},
				{3469, 7733, 0x16293c51},
				{3470, 7732, 0x16292314},
				{3471, 7731, 0x162909c2},
				{3472, 7730, 0x1628f06f},
				{3473, 7729, 0x1628d747},
				{3474, 7728, 0x1628be1f},
				{3475, 7727, 0x1628a4f8},
				{3476, 7726, 0x16288bfb},
				{3477, 7725, 0x162872fe},
				{3478, 7724, 0x16285a17},
				{3479, 7723, 0x1628412f},
				{3480, 7722, 0x1628285d},
				{3481, 7721, 0x16280fa1},
				{3482, 7720, 0x1627f6fa},
				{3483, 7719, 0x1627de53},
				{3484, 7718, 0x1627c5c2},
				{3485, 7717, 0x1627ad5b},
				{3486, 7716, 0x162794df},
				{3487, 7715, 0x16277c79},
				{3488, 7714, 0x16276427},
				{3489, 7713, 0x16274bd6},
				{3490, 7712, 0x1627339b},
				{3491, 7711, 0x16271b75},
				{3492, 7710, 0x1627034e},
				{3493, 7709, 0x1626eb53},
				{3494, 7708, 0x1626d343},
				{3495, 7707, 0x1626bb5d},
				{3496, 7706, 0x1626a377},
				{3497, 7705, 0x16268ba7},
				{3498, 7704, 0x162673d6},
				{3499, 7703, 0x16265c1c},
				{3500, 7702, 0x162644a1},
				{3501, 7701, 0x16262cfc},
				{3502, 7700, 0x1626156c},
				{3503, 7699, 0x1625fdf1},
				{3504, 7698, 0x1625e68d},
				{3505, 7697, 0x1625cf28},
				{3506, 7696, 0x1625b7d8},
				{3507, 7695, 0x1625a089},
				{3508, 7694, 0x1625894f},
				{3509, 7693, 0x1625722a},
				{3510, 7692, 0x16255b05},
				{3511, 7691, 0x1625440c},
				{3512, 7690, 0x16252cfd},
				{3513, 7689, 0x16251618},
				{3514, 7688, 0x1624ff34},
				{3515, 7687, 0x1624e866},
				{3516, 7686, 0x1624d1ac},
				{3517, 7685, 0x1624baf3},
				{3518, 7684, 0x1624a44f},
				{3519, 7683, 0x16248dac},
				{3520, 7682, 0x1624771d},
				{3521, 7681, 0x162460a4},
				{3522, 7680, 0x16244a2c},
				{3523, 7679, 0x162433c8},
				{3524, 7678, 0x16241d65},
				{3525, 7677, 0x1624072c},
				{3526, 7676, 0x1623f0de},
				{3527, 7675, 0x1623dabc},
				{3528, 7674, 0x1623c499},
				{3529, 7673, 0x1623ae8b},
				{3530, 7672, 0x1623987e},
				{3531, 7671, 0x1623829b},
				{3532, 7670, 0x16236cb8},
				{3533, 7669, 0x162356d6},
				{3534, 7668, 0x16234109},
				{3535, 7667, 0x16232b3c},
				{3536, 7666, 0x16231599},
				{3537, 7665, 0x1622ffe2},
				{3538, 7664, 0x1622ea55},
				{3539, 7663, 0x1622d4c8},
				{3540, 7662, 0x1622bf3c},
				{3541, 7661, 0x1622a9c4},
				{3542, 7660, 0x16229463},
				{3543, 7659, 0x16227f16},
				{3544, 7658, 0x162269ca},
				{3545, 7657, 0x1622547e},
				{3546, 7656, 0x16223f72},
				{3547, 7655, 0x16222a50},
				{3548, 7654, 0x1622152f},
				{3549, 7653, 0x16220023},
				{3550, 7652, 0x1621eb2d},
				{3551, 7651, 0x1621d636},
				{3552, 7650, 0x1621c155},
				{3553, 7649, 0x1621ac8a},
				{3554, 7648, 0x162197be},
				{3555, 7647, 0x16218308},
				{3556, 7646, 0x16216e52},
				{3557, 7645, 0x162159b1},
				{3558, 7644, 0x16214511},
				{3559, 7643, 0x16213086},
				{3560, 7642, 0x16211c10},
				{3561, 7641, 0x1621079a},
				{3562, 7640, 0x1620f350},
				{3563, 7639, 0x1620df05},
				{3564, 7638, 0x1620caba},
				{3565, 7637, 0x1620b685},
				{3566, 7636, 0x1620a250},
				{3567, 7635, 0x16208e30},
				{3568, 7634, 0x16207a10},
				{3569, 7633, 0x16206606},
				{3570, 7632, 0x162051fb},
				{3571, 7631, 0x16203e1c},
				{3572, 7630, 0x16202a27},
				{3573, 7629, 0x1620165d},
				{3574, 7628, 0x1620027e},
				{3575, 7627, 0x161feec9},
				{3576, 7626, 0x161fdb15},
				{3577, 7625, 0x161fc78b},
				{3578, 7624, 0x161fb3ec},
				{3579, 7623, 0x161fa04d},
				{3580, 7622, 0x161f8cd9},
				{3581, 7621, 0x161f7965},
				{3582, 7620, 0x161f65f1},
				{3583, 7619, 0x161f5292},
				{3584, 7618, 0x161f3f34},
				{3585, 7617, 0x161f2c00},
				{3586, 7616, 0x161f18b7},
				{3587, 7615, 0x161f0583},
				{3588, 7614, 0x161ef265},
				{3589, 7613, 0x161edf5c},
				{3590, 7612, 0x161ecc3e},
				{3591, 7611, 0x161eb94b},
				{3592, 7610, 0x161ea66d},
				{3593, 7609, 0x161e937a},
				{3594, 7608, 0x161e809c},
				{3595, 7607, 0x161e6dd4},
				{3596, 7606, 0x161e5b0b},
				{3597, 7605, 0x161e4858},
				{3598, 7604, 0x161e35a6},
				{3599, 7603, 0x161e2308},
				{3600, 7602, 0x161e106b},
				{3601, 7601, 0x161dfde3},
				{3602, 7600, 0x161deb70},
				{3603, 7599, 0x161dd8fe},
				{3604, 7598, 0x161dc68b},
				{3605, 7597, 0x161db42e},
				{3606, 7596, 0x161da1e7},
				{3607, 7595, 0x161d8f9f},
				{3608, 7594, 0x161d7d6d},
				{3609, 7593, 0x161d6b51},
				{3610, 7592, 0x161d591e},
				{3611, 7591, 0x161d4717},
				{3612, 7590, 0x161d34fb},
				{3613, 7589, 0x161d2309},
				{3614, 7588, 0x161d1102},
				{3615, 7587, 0x161cff26},
				{3616, 7586, 0x161ced34},
				{3617, 7585, 0x161cdb6d},
				{3618, 7584, 0x161cc9a6},
				{3619, 7583, 0x161cb7e0},
				{3620, 7582, 0x161ca62e},
				{3621, 7581, 0x161c947d},
				{3622, 7580, 0x161c82e1},
				{3623, 7579, 0x161c7170},
				{3624, 7578, 0x161c5fea},
				{3625, 7577, 0x161c4e64},
				{3626, 7576, 0x161c3cf3},
				{3627, 7575, 0x161c2b97},
				{3628, 7574, 0x161c1a3c},
				{3629, 7573, 0x161c08e1},
				{3630, 7572, 0x161bf79b},
				{3631, 7571, 0x161be655},
				{3632, 7570, 0x161bd524},
				{3633, 7569, 0x161bc409},
				{3634, 7568, 0x161bb2ee},
				{3635, 7567, 0x161ba1d3},
				{3636, 7566, 0x161b90cd},
				{3637, 7565, 0x161b7fdd},
				{3638, 7564, 0x161b6f03},
				{3639, 7563, 0x161b5e13},
				{3640, 7562, 0x161b4d38},
				{3641, 7561, 0x161b3c5d},
				{3642, 7560, 0x161b2b98},
				{3643, 7559, 0x161b1ae9},
				{3644, 7558, 0x161b0a39},
				{3645, 7557, 0x161af989},
				{3646, 7556, 0x161ae8ef},
				{3647, 7555, 0x161ad855},
				{3648, 7554, 0x161ac7d0},
				{3649, 7553, 0x161ab74b},
				{3650, 7552, 0x161aa6dc},
				{3651, 7551, 0x161a966d},
				{3652, 7550, 0x161a8613},
				{3653, 7549, 0x161a75b9},
				{3654, 7548, 0x161a658a},
				{3655, 7547, 0x161a5546},
				{3656, 7546, 0x161a4501},
				{3657, 7545, 0x161a34d2},
				{3658, 7544, 0x161a24b9},
				{3659, 7543, 0x161a149f},
				{3660, 7542, 0x161a0486},
				{3661, 7541, 0x1619f482},
				{3662, 7540, 0x1619e494},
				{3663, 7539, 0x1619d490},
				{3664, 7538, 0x1619c4b7},
				{3665, 7537, 0x1619b4c8},
				{3666, 7536, 0x1619a504},
				{3667, 7535, 0x1619952b},
				{3668, 7534, 0x16198568},
				{3669, 7533, 0x161975cf},
				{3670, 7532, 0x16196621},
				{3671, 7531, 0x16195673},
				{3672, 7530, 0x161946da},
				{3673, 7529, 0x16193757},
				{3674, 7528, 0x161927bf},
				{3675, 7527, 0x16191851},
				{3676, 7526, 0x161908ce},
				{3677, 7525, 0x1618f960},
				{3678, 7524, 0x1618ea08},
				{3679, 7523, 0x1618daaf},
				{3680, 7522, 0x1618cb57},
				{3681, 7521, 0x1618bc14},
				{3682, 7520, 0x1618ace7},
				{3683, 7519, 0x16189da4},
				{3684, 7518, 0x16188e8c},
				{3685, 7517, 0x16187f75},
				{3686, 7516, 0x1618705d},
				{3687, 7515, 0x16186145},
				{3688, 7514, 0x16185243},
				{3689, 7513, 0x16184356},
				{3690, 7512, 0x16183453},
				{3691, 7511, 0x1618257c},
				{3692, 7510, 0x1618168f},
				{3693, 7509, 0x161807b7},
				{3694, 7508, 0x1617f8f5},
				{3695, 7507, 0x1617ea33},
				{3696, 7506, 0x1617db71},
				{3697, 7505, 0x1617ccaf},
				{3698, 7504, 0x1617be18},
				{3699, 7503, 0x1617af6c},
				{3700, 7502, 0x1617a0ea},
				{3701, 7501, 0x16179253},
				{3702, 7500, 0x161783d2},
				{3703, 7499, 0x16177550},
				{3704, 7498, 0x161766e4},
				{3705, 7497, 0x16175878},
				{3706, 7496, 0x16174a0c},
				{3707, 7495, 0x16173bb5},
				{3708, 7494, 0x16172d5e},
				{3709, 7493, 0x16171f08},
				{3710, 7492, 0x161710c7},
				{3711, 7491, 0x1617029b},
				{3712, 7490, 0x1616f45a},
				{3713, 7489, 0x1616e643},
				{3714, 7488, 0x1616d818},
				{3715, 7487, 0x1616ca17},
				{3716, 7486, 0x1616bc01},
				{3717, 7485, 0x1616ae00},
				{3718, 7484, 0x16169fff},
				{3719, 7483, 0x16169214},
				{3720, 7482, 0x16168413},
				{3721, 7481, 0x1616763d},
				{3722, 7480, 0x16166852},
				{3723, 7479, 0x16165a7c},
				{3724, 7478, 0x16164cbb},
				{3725, 7477, 0x16163efb},
				{3726, 7476, 0x1616313b},
				{3727, 7475, 0x1616237a},
				{3728, 7474, 0x161615cf},
				{3729, 7473, 0x1616083a},
				{3730, 7472, 0x1615faa4},
				{3731, 7471, 0x1615ed0f},
				{3732, 7470, 0x1615df7a},
				{3733, 7469, 0x1615d1fa},
				{3734, 7468, 0x1615c47a},
				{3735, 7467, 0x1615b70f},
				{3736, 7466, 0x1615a9a4},
				{3737, 7465, 0x16159c3a},
				{3738, 7464, 0x16158ee5},
				{3739, 7463, 0x16158190},
				{3740, 7462, 0x1615743b},
				{3741, 7461, 0x16156711},
				{3742, 7460, 0x161559e6},
				{3743, 7459, 0x16154cbc},
				{3744, 7458, 0x16153fa8},
				{3745, 7457, 0x16153293},
				{3746, 7456, 0x16152594},
				{3747, 7455, 0x16151895},
				{3748, 7454, 0x16150b95},
				{3749, 7453, 0x1614fea1},
				{3750, 7452, 0x1614f1b7},
				{3751, 7451, 0x1614e4d8},
				{3752, 7450, 0x1614d7f9},
				{3753, 7449, 0x1614cb25},
				{3754, 7448, 0x1614be51},
				{3755, 7447, 0x1614b192},
				{3756, 7446, 0x1614a4d3},
				{3757, 7445, 0x1614981f},
				{3758, 7444, 0x16148b6b},
				{3759, 7443, 0x16147ec2},
				{3760, 7442, 0x16147223},
				{3761, 7441, 0x1614659a},
				{3762, 7440, 0x16145906},
				{3763, 7439, 0x16144c7d},
				{3764, 7438, 0x16143ffe},
				{3765, 7437, 0x1614338b},
				{3766, 7436, 0x16142717},
				{3767, 7435, 0x16141aae},
				{3768, 7434, 0x16140e45},
				{3769, 7433, 0x161401e7},
				{3770, 7432, 0x1613f593},
				{3771, 7431, 0x1613e94b},
				{3772, 7430, 0x1613dd02},
				{3773, 7429, 0x1613d0c4},
				{3774, 7428, 0x1613c491},
				{3775, 7427, 0x1613b85d},
				{3776, 7426, 0x1613ac40},
				{3777, 7425, 0x1613a022},
				{3778, 7424, 0x16139404},
				{3779, 7423, 0x161387f1},
				{3780, 7422, 0x16137be9},
				{3781, 7421, 0x16136feb},
				{3782, 7420, 0x161363ed},
				{3783, 7419, 0x161357fa},
				{3784, 7418, 0x16134c08},
				{3785, 7417, 0x1613401f},
				{3786, 7416, 0x16133442},
				{3787, 7415, 0x16132865},
				{3788, 7414, 0x16131c9d},
				{3789, 7413, 0x161310ca},
				{3790, 7412, 0x1613050d},
				{3791, 7411, 0x1612f950},
				{3792, 7410, 0x1612eda8},
				{3793, 7409, 0x1612e200},
				{3794, 7408, 0x1612d658},
				{3795, 7407, 0x1612cabb},
				{3796, 7406, 0x1612bf1e},
				{3797, 7405, 0x1612b38c},
				{3798, 7404, 0x1612a804},
				{3799, 7403, 0x16129c7d},
				{3800, 7402, 0x16129100},
				{3801, 7401, 0x1612858e},
				{3802, 7400, 0x16127a27},
				{3803, 7399, 0x16126ebf},
				{3804, 7398, 0x16126358},
				{3805, 7397, 0x16125806},
				{3806, 7396, 0x16124cb4},
				{3807, 7395, 0x16124178},
				{3808, 7394, 0x16123631},
				{3809, 7393, 0x16122af4},
				{3810, 7392, 0x16121fb8},
				{3811, 7391, 0x16121486},
				{3812, 7390, 0x1612095f},
				{3813, 7389, 0x1611fe43},
				{3814, 7388, 0x1611f327},
				{3815, 7387, 0x1611e815},
				{3816, 7386, 0x1611dd04},
				{3817, 7385, 0x1611d1fd},
				{3818, 7384, 0x1611c701},
				{3819, 7383, 0x1611bc05},
				{3820, 7382, 0x1611b114},
				{3821, 7381, 0x1611a622},
				{3822, 7380, 0x16119b51},
				{3823, 7379, 0x16119075},
				{3824, 7378, 0x1611859a},
				{3825, 7377, 0x16117ac8},
				{3826, 7376, 0x16117002},
				{3827, 7375, 0x1611653c},
				{3828, 7374, 0x16115a80},
				{3829, 7373, 0x16114fc5},
				{3830, 7372, 0x16114514},
				{3831, 7371, 0x16113a6d},
				{3832, 7370, 0x16112fd2},
				{3833, 7369, 0x1611252c},
				{3834, 7368, 0x16111a9b},
				{3835, 7367, 0x1611100a},
				{3836, 7366, 0x16110585},
				{3837, 7365, 0x1610faff},
				{3838, 7364, 0x1610f099},
				{3839, 7363, 0x1610e61e},
				{3840, 7362, 0x1610dbad},
				{3841, 7361, 0x1610d147},
				{3842, 7360, 0x1610c6ec},
				{3843, 7359, 0x1610bc91},
				{3844, 7358, 0x1610b236},
				{3845, 7357, 0x1610a7e6},
				{3846, 7356, 0x16109da0},
				{3847, 7355, 0x16109366},
				{3848, 7354, 0x1610892b},
				{3849, 7353, 0x16107ef0},
				{3850, 7352, 0x161074c0},
				{3851, 7351, 0x16106a9a},
				{3852, 7350, 0x16106075},
				{3853, 7349, 0x16105665},
				{3854, 7348, 0x16104c55},
				{3855, 7347, 0x16104245},
				{3856, 7346, 0x16103840},
				{3857, 7345, 0x16102e3b},
				{3858, 7344, 0x16102441},
				{3859, 7343, 0x16101a46},
				{3860, 7342, 0x16101056},
				{3861, 7341, 0x16100671},
				{3862, 7340, 0x160ffc8d},
				{3863, 7339, 0x160ff2a8},
				{3864, 7338, 0x160fe8d8},
				{3865, 7337, 0x160fdf09},
				{3866, 7336, 0x160fd539},
				{3867, 7335, 0x160fcb74},
				{3868, 7334, 0x160fc1c5},
				{3869, 7333, 0x160fb80b},
				{3870, 7332, 0x160fae51},
				{3871, 7331, 0x160fa4ac},
				{3872, 7330, 0x160f9afd},
				{3873, 7329, 0x160f9163},
				{3874, 7328, 0x160f87c9},
				{3875, 7327, 0x160f7e30},
				{3876, 7326, 0x160f74a0},
				{3877, 7325, 0x160f6b1c},
				{3878, 7324, 0x160f6198},
				{3879, 7323, 0x160f581e},
				{3880, 7322, 0x160f4ea4},
				{3881, 7321, 0x160f4536},
				{3882, 7320, 0x160f3bc7},
				{3883, 7319, 0x160f3262},
				{3884, 7318, 0x160f2909},
				{3885, 7317, 0x160f1fb0},
				{3886, 7316, 0x160f1661},
				{3887, 7315, 0x160f0d12},
				{3888, 7314, 0x160f03ce},
				{3889, 7313, 0x160efa8a},
				{3890, 7312, 0x160ef146},
				{3891, 7311, 0x160ee818},
				{3892, 7310, 0x160edede},
				{3893, 7309, 0x160ed5bb},
				{3894, 7308, 0x160ecc97},
				{3895, 7307, 0x160ec373},
				{3896, 7306, 0x160eba5a},
				{3897, 7305, 0x160eb141},
				{3898, 7304, 0x160ea833},
				{3899, 7303, 0x160e9f3a},
				{3900, 7302, 0x160e9636},
				{3901, 7301, 0x160e8d33},
				{3902, 7300, 0x160e843a},
				{3903, 7299, 0x160e7b4c},
				{3904, 7298, 0x160e725e},
				{3905, 7297, 0x160e6970},
				{3906, 7296, 0x160e608c},
				{3907, 7295, 0x160e57b4},
				{3908, 7294, 0x160e4edb},
				{3909, 7293, 0x160e460d},
				{3910, 7292, 0x160e3d3f},
				{3911, 7291, 0x160e347c},
				{3912, 7290, 0x160e2bb9},
				{3913, 7289, 0x160e2300},
				{3914, 7288, 0x160e1a53},
				{3915, 7287, 0x160e11a5},
				{3916, 7286, 0x160e08f7},
				{3917, 7285, 0x160e0049},
				{3918, 7284, 0x160df7b1},
				{3919, 7283, 0x160def0e},
				{3920, 7282, 0x160de681},
				{3921, 7281, 0x160ddde8},
				{3922, 7280, 0x160dd566},
				{3923, 7279, 0x160dccd8},
				{3924, 7278, 0x160dc455},
				{3925, 7277, 0x160dbbdd},
				{3926, 7276, 0x160db365},
				{3927, 7275, 0x160daaf8},
				{3928, 7274, 0x160da28b},
				{3929, 7273, 0x160d9a28},
				{3930, 7272, 0x160d91d0},
				{3931, 7271, 0x160d8978},
				{3932, 7270, 0x160d8121},
				{3933, 7269, 0x160d78c9},
				{3934, 7268, 0x160d707c},
				{3935, 7267, 0x160d6839},
				{3936, 7266, 0x160d5ff7},
				{3937, 7265, 0x160d57b4},
				{3938, 7264, 0x160d4f7d},
				{3939, 7263, 0x160d4750},
				{3940, 7262, 0x160d3f23},
				{3941, 7261, 0x160d36f6},
				{3942, 7260, 0x160d2ed4},
				{3943, 7259, 0x160d26b2},
				{3944, 7258, 0x160d1e9a},
				{3945, 7257, 0x160d168d},
				{3946, 7256, 0x160d0e81},
				{3947, 7255, 0x160d0674},
				{3948, 7254, 0x160cfe72},
				{3949, 7253, 0x160cf670},
				{3950, 7252, 0x160cee79},
				{3951, 7251, 0x160ce682},
				{3952, 7250, 0x160cde8a},
				{3953, 7249, 0x160cd69e},
				{3954, 7248, 0x160ccebc},
				{3955, 7247, 0x160cc6cf},
				{3956, 7246, 0x160cbef8},
				{3957, 7245, 0x160cb721},
				{3958, 7244, 0x160caf4a},
				{3959, 7243, 0x160ca773},
				{3960, 7242, 0x160c9fb2},
				{3961, 7241, 0x160c97f0},
				{3962, 7240, 0x160c902f},
				{3963, 7239, 0x160c886d},
				{3964, 7238, 0x160c80b6},
				{3965, 7237, 0x160c790a},
				{3966, 7236, 0x160c7153},
				{3967, 7235, 0x160c69b2},
				{3968, 7234, 0x160c6206},
				{3969, 7233, 0x160c5a64},
				{3970, 7232, 0x160c52ce},
				{3971, 7231, 0x160c4b37},
				{3972, 7230, 0x160c43a0},
				{3973, 7229, 0x160c3c14},
				{3974, 7228, 0x160c3488},
				{3975, 7227, 0x160c2d07},
				{3976, 7226, 0x160c2591},
				{3977, 7225, 0x160c1e10},
				{3978, 7224, 0x160c1699},
				{3979, 7223, 0x160c0f2d},
				{3980, 7222, 0x160c07c2},
				{3981, 7221, 0x160c0056},
				{3982, 7220, 0x160bf8ea},
				{3983, 7219, 0x160bf189},
				{3984, 7218, 0x160bea33},
				{3985, 7217, 0x160be2dd},
				{3986, 7216, 0x160bdb86},
				{3987, 7215, 0x160bd43b},
				{3988, 7214, 0x160bccef},
				{3989, 7213, 0x160bc5a4},
				{3990, 7212, 0x160bbe63},
				{3991, 7211, 0x160bb738},
				{3992, 7210, 0x160baff7},
				{3993, 7209, 0x160ba8c1},
				{3994, 7208, 0x160ba196},
				{3995, 7207, 0x160b9a6a},
				{3996, 7206, 0x160b933f},
				{3997, 7205, 0x160b8c1e},
				{3998, 7204, 0x160b84fe},
				{3999, 7203, 0x160b7de8},
				{4000, 7202, 0x160b76d2},
				{4001, 7201, 0x160b6fbc},
				{4002, 7200, 0x160b68b1},
				{4003, 7199, 0x160b61a6},
				{4004, 7198, 0x160b5a9b},
				{4005, 7197, 0x160b539a},
				{4006, 7196, 0x160b4ca5},
				{4007, 7195, 0x160b45af},
				{4008, 7194, 0x160b3eb9},
				{4009, 7193, 0x160b37ce},
				{4010, 7192, 0x160b30d9},
				{4011, 7191, 0x160b29f9},
				{4012, 7190, 0x160b230e},
				{4013, 7189, 0x160b1c2d},
				{4014, 7188, 0x160b1558},
				{4015, 7187, 0x160b0e78},
				{4016, 7186, 0x160b07a2},
				{4017, 7185, 0x160b00d8},
				{4018, 7184, 0x160afa0d},
				{4019, 7183, 0x160af342},
				{4020, 7182, 0x160aec82},
				{4021, 7181, 0x160ae5c2},
				{4022, 7180, 0x160adf0d},
				{4023, 7179, 0x160ad84d},
				{4024, 7178, 0x160ad198},
				{4025, 7177, 0x160acaed},
				{4026, 7176, 0x160ac438},
				{4027, 7175, 0x160abd98},
				{4028, 7174, 0x160ab6f8},
				{4029, 7173, 0x160ab05e},
				{4030, 7172, 0x160aa9ce},
				{4031, 7171, 0x160aa33e},
				{4032, 7170, 0x160a9cb4},
				{4033, 7169, 0x160a962a},
				{4034, 7168, 0x160a8fa5},
				{4035, 7167, 0x160a8925},
				{4036, 7166, 0x160a82ab},
				{4037, 7165, 0x160a7c3b},
				{4038, 7164, 0x160a75c6},
				{4039, 7163, 0x160a6f57},
				{4040, 7162, 0x160a68e7},
				{4041, 7161, 0x160a6283},
				{4042, 7160, 0x160a5c1e},
				{4043, 7159, 0x160a55b9},
				{4044, 7158, 0x160a4f5f},
				{4045, 7157, 0x160a4905},
				{4046, 7156, 0x160a42b0},
				{4047, 7155, 0x160a3c5c},
				{4048, 7154, 0x160a3612},
				{4049, 7153, 0x160a2fc3},
				{4050, 7152, 0x160a297e},
				{4051, 7151, 0x160a2339},
				{4052, 7150, 0x160a1d00},
				{4053, 7149, 0x160a16cb},
				{4054, 7148, 0x160a1091},
				{4055, 7147, 0x160a0a5d},
				{4056, 7146, 0x160a042e},
				{4057, 7145, 0x1609fe04},
				{4058, 7144, 0x1609f7da},
				{4059, 7143, 0x1609f1b6},
				{4060, 7142, 0x1609eb91},
				{4061, 7141, 0x1609e578},
				{4062, 7140, 0x1609df5e},
				{4063, 7139, 0x1609d945},
				{4064, 7138, 0x1609d336},
				{4065, 7137, 0x1609cd27},
				{4066, 7136, 0x1609c718},
				{4067, 7135, 0x1609c114},
				{4068, 7134, 0x1609bb15},
				{4069, 7133, 0x1609b516},
				{4070, 7132, 0x1609af17},
				{4071, 7131, 0x1609a91e},
				{4072, 7130, 0x1609a32a},
				{4073, 7129, 0x16099d3b},
				{4074, 7128, 0x1609974c},
				{4075, 7127, 0x16099163},
				{4076, 7126, 0x16098b7a},
				{4077, 7125, 0x16098596},
				{4078, 7124, 0x16097fb7},
				{4079, 7123, 0x160979de},
				{4080, 7122, 0x16097405},
				{4081, 7121, 0x16096e2b},
				{4082, 7120, 0x1609685d},
				{4083, 7119, 0x16096294},
				{4084, 7118, 0x16095ccb},
				{4085, 7117, 0x16095707},
				{4086, 7116, 0x16095143},
				{4087, 7115, 0x16094b7f},
				{4088, 7114, 0x160945c6},
				{4089, 7113, 0x1609400d},
				{4090, 7112, 0x16093a59},
				{4091, 7111, 0x160934a6},
				{4092, 7110, 0x16092ef7},
				{4093, 7109, 0x1609294e},
				{4094, 7108, 0x160923a6},
				{4095, 7107, 0x16091e02},
				{4096, 7106, 0x1609185e},
				{4097, 7105, 0x160912c6},
				{4098, 7104, 0x16090d27},
				{4099, 7103, 0x16090799},
				{4100, 7102, 0x16090206},
				{4101, 7101, 0x1608fc78},
				{4102, 7100, 0x1608f6ef},
				{4103, 7099, 0x1608f166},
				{4104, 7098, 0x1608ebdd},
				{4105, 7097, 0x1608e65f},
				{4106, 7096, 0x1608e0e1},
				{4107, 7095, 0x1608db63},
				{4108, 7094, 0x1608d5eb},
				{4109, 7093, 0x1608d077},
				{4110, 7092, 0x1608cb0a},
				{4111, 7091, 0x1608c59c},
				{4112, 7090, 0x1608c02e},
				{4113, 7089, 0x1608bacb},
				{4114, 7088, 0x1608b56d},
				{4115, 7087, 0x1608b00a},
				{4116, 7086, 0x1608aab1},
				{4117, 7085, 0x1608a553},
				{4118, 7084, 0x1608a000},
				{4119, 7083, 0x16089aad},
				{4120, 7082, 0x1608955a},
				{4121, 7081, 0x16089012},
				{4122, 7080, 0x16088ac9},
				{4123, 7079, 0x16088581},
				{4124, 7078, 0x1608803e},
				{4125, 7077, 0x16087b00},
				{4126, 7076, 0x160875c3},
				{4127, 7075, 0x1608708b},
				{4128, 7074, 0x16086b58},
				{4129, 7073, 0x1608662a},
				{4130, 7072, 0x160860f7},
				{4131, 7071, 0x16085bcf},
				{4132, 7070, 0x160856a7},
				{4133, 7069, 0x1608517f},
				{4134, 7068, 0x16084c5c},
				{4135, 7067, 0x1608473f},
				{4136, 7066, 0x16084221},
				{4137, 7065, 0x16083d09},
				{4138, 7064, 0x160837f1},
				{4139, 7063, 0x160832de},
				{4140, 7062, 0x16082dd1},
				{4141, 7061, 0x160828c4},
				{4142, 7060, 0x160823bc},
				{4143, 7059, 0x16081eb4},
				{4144, 7058, 0x160819b1},
				{4145, 7057, 0x160814b9},
				{4146, 7056, 0x16080fbc},
				{4147, 7055, 0x16080ac4},
				{4148, 7054, 0x160805cd},
				{4149, 7053, 0x160800da},
				{4150, 7052, 0x1607fbe8},
				{4151, 7051, 0x1607f6fa},
				{4152, 7050, 0x1607f213},
				{4153, 7049, 0x1607ed2b},
				{4154, 7048, 0x1607e849},
				{4155, 7047, 0x1607e366},
				{4156, 7046, 0x1607de89},
				{4157, 7045, 0x1607d9ac},
				{4158, 7044, 0x1607d4d5},
				{4159, 7043, 0x1607d002},
				{4160, 7042, 0x1607cb35},
				{4161, 7041, 0x1607c663},
				{4162, 7040, 0x1607c196},
				{4163, 7039, 0x1607bccf},
				{4164, 7038, 0x1607b807},
				{4165, 7037, 0x1607b345},
				{4166, 7036, 0x1607ae88},
				{4167, 7035, 0x1607a9cb},
				{4168, 7034, 0x1607a50e},
				{4169, 7033, 0x1607a057},
				{4170, 7032, 0x16079ba5},
				{4171, 7031, 0x160796f3},
				{4172, 7030, 0x16079246},
				{4173, 7029, 0x16078d99},
				{4174, 7028, 0x160788f2},
				{4175, 7027, 0x16078450},
				{4176, 7026, 0x16077fae},
				{4177, 7025, 0x16077b0c},
				{4178, 7024, 0x1607766f},
				{4179, 7023, 0x160771d8},
				{4180, 7022, 0x16076d41},
				{4181, 7021, 0x160768aa},
				{4182, 7020, 0x16076418},
				{4183, 7019, 0x16075f8b},
				{4184, 7018, 0x16075aff},
				{4185, 7017, 0x16075678},
				{4186, 7016, 0x160751f0},
				{4187, 7015, 0x16074d6f},
				{4188, 7014, 0x160748ed},
				{4189, 7013, 0x16074470},
				{4190, 7012, 0x16073ff4},
				{4191, 7011, 0x16073b82},
				{4192, 7010, 0x1607370b},
				{4193, 7009, 0x1607329a},
				{4194, 7008, 0x16072e28},
				{4195, 7007, 0x160729bc},
				{4196, 7006, 0x16072555},
				{4197, 7005, 0x160720ee},
				{4198, 7004, 0x16071c87},
				{4199, 7003, 0x16071825},
				{4200, 7002, 0x160713c3},
				{4201, 7001, 0x16070f67},
				{4202, 7000, 0x16070b10},
				{4203, 6999, 0x160706ba},
				{4204, 6998, 0x16070263},
				{4205, 6997, 0x1606fe11},
				{4206, 6996, 0x1606f9c5},
				{4207, 6995, 0x1606f579},
				{4208, 6994, 0x1606f132},
				{4209, 6993, 0x1606eceb},
				{4210, 6992, 0x1606e8a5},
				{4211, 6991, 0x1606e463},
				{4212, 6990, 0x1606e027},
				{4213, 6989, 0x1606dbeb},
				{4214, 6988, 0x1606d7af},
				{4215, 6987, 0x1606d379},
				{4216, 6986, 0x1606cf42},
				{4217, 6985, 0x1606cb11},
				{4218, 6984, 0x1606c6e5},
				{4219, 6983, 0x1606c2b3},
				{4220, 6982, 0x1606be8d},
				{4221, 6981, 0x1606ba6c},
				{4222, 6980, 0x1606b645},
				{4223, 6979, 0x1606b224},
				{4224, 6978, 0x1606ae03},
				{4225, 6977, 0x1606a9e7},
				{4226, 6976, 0x1606a5cb},
				{4227, 6975, 0x1606a1af},
				{4228, 6974, 0x16069d9e},
				{4229, 6973, 0x16069988},
				{4230, 6972, 0x16069577},
				{4231, 6971, 0x1606916b},
				{4232, 6970, 0x16068d5f},
				{4233, 6969, 0x16068953},
				{4234, 6968, 0x1606854d},
				{4235, 6967, 0x1606814c},
				{4236, 6966, 0x16067d4b},
				{4237, 6965, 0x1606794f},
				{4238, 6964, 0x16067554},
				{4239, 6963, 0x16067158},
				{4240, 6962, 0x16066d62},
				{4241, 6961, 0x1606696c},
				{4242, 6960, 0x1606657b},
				{4243, 6959, 0x1606618a},
				{4244, 6958, 0x16065d99},
				{4245, 6957, 0x160659ad},
				{4246, 6956, 0x160655c7},
				{4247, 6955, 0x160651e1},
				{4248, 6954, 0x16064dfb},
				{4249, 6953, 0x16064a1a},
				{4250, 6952, 0x16064639},
				{4251, 6951, 0x1606425e},
				{4252, 6950, 0x16063e83},
				{4253, 6949, 0x16063aac},
				{4254, 6948, 0x160636d6},
				{4255, 6947, 0x16063300},
				{4256, 6946, 0x16062f30},
				{4257, 6945, 0x16062b64},
				{4258, 6944, 0x16062794},
				{4259, 6943, 0x160623c8},
				{4260, 6942, 0x16062002},
				{4261, 6941, 0x16061c3c},
				{4262, 6940, 0x1606187c},
				{4263, 6939, 0x160614b6},
				{4264, 6938, 0x160610fb},
				{4265, 6937, 0x16060d3a},
				{4266, 6936, 0x16060984},
				{4267, 6935, 0x160605ce},
				{4268, 6934, 0x16060218},
				{4269, 6933, 0x1605fe63},
				{4270, 6932, 0x1605fab2},
				{4271, 6931, 0x1605f702},
				{4272, 6930, 0x1605f356},
				{4273, 6929, 0x1605efab},
				{4274, 6928, 0x1605ec06},
				{4275, 6927, 0x1605e860},
				{4276, 6926, 0x1605e4ba},
				{4277, 6925, 0x1605e11a},
				{4278, 6924, 0x1605dd79},
				{4279, 6923, 0x1605d9d9},
				{4280, 6922, 0x1605d63e},
				{4281, 6921, 0x1605d2a8},
				{4282, 6920, 0x1605cf0d},
				{4283, 6919, 0x1605cb7d},
				{4284, 6918, 0x1605c7ec},
				{4285, 6917, 0x1605c45c},
				{4286, 6916, 0x1605c0cc},
				{4287, 6915, 0x1605bd41},
				{4288, 6914, 0x1605b9b6},
				{4289, 6913, 0x1605b630},
				{4290, 6912, 0x1605b2ab},
				{4291, 6911, 0x1605af25},
				{4292, 6910, 0x1605aba5},
				{4293, 6909, 0x1605a825},
				{4294, 6908, 0x1605a4a5},
				{4295, 6907, 0x1605a12a},
				{4296, 6906, 0x16059daf},
				{4297, 6905, 0x16059a3a},
				{4298, 6904, 0x160596ca},
				{4299, 6903, 0x16059354},
				{4300, 6902, 0x16058fe4},
				{4301, 6901, 0x16058c74},
				{4302, 6900, 0x16058904},
				{4303, 6899, 0x16058599},
				{4304, 6898, 0x16058234},
				{4305, 6897, 0x16057ec9},
				{4306, 6896, 0x16057b64},
				{4307, 6895, 0x16057804},
				{4308, 6894, 0x160574a4},
				{4309, 6893, 0x16057144},
				{4310, 6892, 0x16056de4},
				{4311, 6891, 0x16056a89},
				{4312, 6890, 0x1605672e},
				{4313, 6889, 0x160563de},
				{4314, 6888, 0x16056084},
				{4315, 6887, 0x16055d34},
				{4316, 6886, 0x160559e7},
				{4317, 6885, 0x1605569c},
				{4318, 6884, 0x16055354},
				{4319, 6883, 0x1605500c},
				{4320, 6882, 0x16054cca},
				{4321, 6881, 0x16054987},
				{4322, 6880, 0x16054645},
				{4323, 6879, 0x16054305},
				{4324, 6878, 0x16053fc8},
				{4325, 6877, 0x16053c8e},
				{4326, 6876, 0x16053956},
				{4327, 6875, 0x1605361e},
				{4328, 6874, 0x160532e9},
				{4329, 6873, 0x16052fb9},
				{4330, 6872, 0x16052c87},
				{4331, 6871, 0x16052957},
				{4332, 6870, 0x1605262a},
				{4333, 6869, 0x160522fd},
				{4334, 6868, 0x16051fd6},
				{4335, 6867, 0x16051cab},
				{4336, 6866, 0x16051986},
				{4337, 6865, 0x16051661},
				{4338, 6864, 0x1605133f},
				{4339, 6863, 0x1605101f},
				{4340, 6862, 0x16050d00},
				{4341, 6861, 0x160509e3},
				{4342, 6860, 0x160506c9},
				{4343, 6859, 0x160503b1},
				{4344, 6858, 0x1605009c},
				{4345, 6857, 0x1604fd87},
				{4346, 6856, 0x1604fa75},
				{4347, 6855, 0x1604f763},
				{4348, 6854, 0x1604f453},
				{4349, 6853, 0x1604f144},
				{4350, 6852, 0x1604ee3a},
				{4351, 6851, 0x1604eb30},
				{4352, 6850, 0x1604e826},
				{4353, 6849, 0x1604e521},
				{4354, 6848, 0x1604e21c},
				{4355, 6847, 0x1604df1a},
				{4356, 6846, 0x1604dc18},
				{4357, 6845, 0x1604d918},
				{4358, 6844, 0x1604d61c},
				{4359, 6843, 0x1604d322},
				{4360, 6842, 0x1604d02a},
				{4361, 6841, 0x1604cd30},
				{4362, 6840, 0x1604ca3c},
				{4363, 6839, 0x1604c747},
				{4364, 6838, 0x1604c455},
				{4365, 6837, 0x1604c163},
				{4366, 6836, 0x1604be74},
				{4367, 6835, 0x1604bb87},
				{4368, 6834, 0x1604b89d},
				{4369, 6833, 0x1604b5b3},
				{4370, 6832, 0x1604b2c9},
				{4371, 6831, 0x1604afe5},
				{4372, 6830, 0x1604ad00},
				{4373, 6829, 0x1604aa1e},
				{4374, 6828, 0x1604a73c},
				{4375, 6827, 0x1604a460},
				{4376, 6826, 0x1604a183},
				{4377, 6825, 0x16049ea9},
				{4378, 6824, 0x16049bd0},
				{4379, 6823, 0x160498f6},
				{4380, 6822, 0x16049621},
				{4381, 6821, 0x1604934d},
				{4382, 6820, 0x16049078},
				{4383, 6819, 0x16048da7},
				{4384, 6818, 0x16048ad7},
				{4385, 6817, 0x1604880b},
				{4386, 6816, 0x1604853f},
				{4387, 6815, 0x16048275},
				{4388, 6814, 0x16047fab},
				{4389, 6813, 0x16047ce4},
				{4390, 6812, 0x16047a22},
				{4391, 6811, 0x16047761},
				{4392, 6810, 0x1604749f},
				{4393, 6809, 0x160471dd},
				{4394, 6808, 0x16046f1e},
				{4395, 6807, 0x16046c62},
				{4396, 6806, 0x160469a8},
				{4397, 6805, 0x160466ef},
				{4398, 6804, 0x16046438},
				{4399, 6803, 0x16046181},
				{4400, 6802, 0x16045ecd},
				{4401, 6801, 0x16045c1b},
				{4402, 6800, 0x16045969},
				{4403, 6799, 0x160456ba},
				{4404, 6798, 0x1604540c},
				{4405, 6797, 0x16045165},
				{4406, 6796, 0x16044eb9},
				{4407, 6795, 0x16044c12},
				{4408, 6794, 0x1604496b},
				{4409, 6793, 0x160446c4},
				{4410, 6792, 0x16044420},
				{4411, 6791, 0x1604417e},
				{4412, 6790, 0x16043edd},
				{4413, 6789, 0x16043c3e},
				{4414, 6788, 0x160439a2},
				{4415, 6787, 0x16043706},
				{4416, 6786, 0x1604346c},
				{4417, 6785, 0x160431d3},
				{4418, 6784, 0x16042f3c},
				{4419, 6783, 0x16042ca8},
				{4420, 6782, 0x16042a14},
				{4421, 6781, 0x16042785},
				{4422, 6780, 0x160424f4},
				{4423, 6779, 0x16042265},
				{4424, 6778, 0x16041fd9},
				{4425, 6777, 0x16041d4d},
				{4426, 6776, 0x16041ac4},
				{4427, 6775, 0x1604183a},
				{4428, 6774, 0x160415b4},
				{4429, 6773, 0x1604132d},
				{4430, 6772, 0x160410a9},
				{4431, 6771, 0x16040e28},
				{4432, 6770, 0x16040ba7},
				{4433, 6769, 0x16040928},
				{4434, 6768, 0x160406a9},
				{4435, 6767, 0x1604042d},
				{4436, 6766, 0x160401b7},
				{4437, 6765, 0x1603ff3e},
				{4438, 6764, 0x1603fcc7},
				{4439, 6763, 0x1603fa50},
				{4440, 6762, 0x1603f7dd},
				{4441, 6761, 0x1603f569},
				{4442, 6760, 0x1603f2f8},
				{4443, 6759, 0x1603f089},
				{4444, 6758, 0x1603ee1b},
				{4445, 6757, 0x1603ebaf},
				{4446, 6756, 0x1603e943},
				{4447, 6755, 0x1603e6da},
				{4448, 6754, 0x1603e471},
				{4449, 6753, 0x1603e20a},
				{4450, 6752, 0x1603dfa4},
				{4451, 6751, 0x1603dd43},
				{4452, 6750, 0x1603dae1},
				{4453, 6749, 0x1603d880},
				{4454, 6748, 0x1603d622},
				{4455, 6747, 0x1603d3c4},
				{4456, 6746, 0x1603d168},
				{4457, 6745, 0x1603cf0c},
				{4458, 6744, 0x1603ccb3},
				{4459, 6743, 0x1603ca5a},
				{4460, 6742, 0x1603c804},
				{4461, 6741, 0x1603c5b0},
				{4462, 6740, 0x1603c35c},
				{4463, 6739, 0x1603c109},
				{4464, 6738, 0x1603beb8},
				{4465, 6737, 0x1603bc69},
				{4466, 6736, 0x1603ba1b},
				{4467, 6735, 0x1603b7d2},
				{4468, 6734, 0x1603b586},
				{4469, 6733, 0x1603b33e},
				{4470, 6732, 0x1603b0f5},
				{4471, 6731, 0x1603aeae},
				{4472, 6730, 0x1603ac6b},
				{4473, 6729, 0x1603aa27},
				{4474, 6728, 0x1603a7e4},
				{4475, 6727, 0x1603a5a3},
				{4476, 6726, 0x1603a364},
				{4477, 6725, 0x1603a126},
				{4478, 6724, 0x16039ee8},
				{4479, 6723, 0x16039cad},
				{4480, 6722, 0x16039a74},
				{4481, 6721, 0x1603983b},
				{4482, 6720, 0x16039605},
				{4483, 6719, 0x160393d1},
				{4484, 6718, 0x1603919b},
				{4485, 6717, 0x16038f68},
				{4486, 6716, 0x16038d37},
				{4487, 6715, 0x16038b06},
				{4488, 6714, 0x160388d8},
				{4489, 6713, 0x160386aa},
				{4490, 6712, 0x1603847e},
				{4491, 6711, 0x16038253},
				{4492, 6710, 0x1603802a},
				{4493, 6709, 0x16037e01},
				{4494, 6708, 0x16037bd9},
				{4495, 6707, 0x160379b5},
				{4496, 6706, 0x1603778f},
				{4497, 6705, 0x1603756f},
				{4498, 6704, 0x1603734e},
				{4499, 6703, 0x1603712d},
				{4500, 6702, 0x16036f0f},
				{4501, 6701, 0x16036cf1},
				{4502, 6700, 0x16036ad6},
				{4503, 6699, 0x160368ba},
				{4504, 6698, 0x160366a2},
				{4505, 6697, 0x16036489},
				{4506, 6696, 0x16036271},
				{4507, 6695, 0x1603605b},
				{4508, 6694, 0x16035e47},
				{4509, 6693, 0x16035c34},
				{4510, 6692, 0x16035a23},
				{4511, 6691, 0x16035813},
				{4512, 6690, 0x16035602},
				{4513, 6689, 0x160353f7},
				{4514, 6688, 0x160351e9},
				{4515, 6687, 0x16034fde},
				{4516, 6686, 0x16034dd5},
				{4517, 6685, 0x16034bcd},
				{4518, 6684, 0x160349c4},
				{4519, 6683, 0x160347be},
				{4520, 6682, 0x160345b8},
				{4521, 6681, 0x160343b5},
				{4522, 6680, 0x160341b2},
				{4523, 6679, 0x16033faf},
				{4524, 6678, 0x16033db1},
				{4525, 6677, 0x16033bb1},
				{4526, 6676, 0x160339b3},
				{4527, 6675, 0x160337b8},
				{4528, 6674, 0x160335bd},
				{4529, 6673, 0x160333c4},
				{4530, 6672, 0x160331cc},
				{4531, 6671, 0x16032fd3},
				{4532, 6670, 0x16032ddd},
				{4533, 6669, 0x16032be8},
				{4534, 6668, 0x160329f5},
				{4535, 6667, 0x16032802},
				{4536, 6666, 0x1603260e},
				{4537, 6665, 0x1603241e},
				{4538, 6664, 0x16032230},
				{4539, 6663, 0x16032043},
				{4540, 6662, 0x16031e55},
				{4541, 6661, 0x16031c6a},
				{4542, 6660, 0x16031a7f},
				{4543, 6659, 0x16031896},
				{4544, 6658, 0x160316ae},
				{4545, 6657, 0x160314c8},
				{4546, 6656, 0x160312e3},
				{4547, 6655, 0x160310fd},
				{4548, 6654, 0x16030f1a},
				{4549, 6653, 0x16030d37},
				{4550, 6652, 0x16030b57},
				{4551, 6651, 0x16030977},
				{4552, 6650, 0x16030796},
				{4553, 6649, 0x160305b9},
				{4554, 6648, 0x160303de},
				{4555, 6647, 0x16030203},
				{4556, 6646, 0x16030028},
				{4557, 6645, 0x1602fe4d},
				{4558, 6644, 0x1602fc75},
				{4559, 6643, 0x1602faa2},
				{4560, 6642, 0x1602f8ca},
				{4561, 6641, 0x1602f6f7},
				{4562, 6640, 0x1602f521},
				{4563, 6639, 0x1602f351},
				{4564, 6638, 0x1602f17e},
				{4565, 6637, 0x1602efae},
				{4566, 6636, 0x1602edde},
				{4567, 6635, 0x1602ec10},
				{4568, 6634, 0x1602ea43},
				{4569, 6633, 0x1602e878},
				{4570, 6632, 0x1602e6ad},
				{4571, 6631, 0x1602e4e2},
				{4572, 6630, 0x1602e31a},
				{4573, 6629, 0x1602e152},
				{4574, 6628, 0x1602df8c},
				{4575, 6627, 0x1602ddc7},
				{4576, 6626, 0x1602dc01},
				{4577, 6625, 0x1602da3f},
				{4578, 6624, 0x1602d87c},
				{4579, 6623, 0x1602d6bc},
				{4580, 6622, 0x1602d4fc},
				{4581, 6621, 0x1602d33c},
				{4582, 6620, 0x1602d17e},
				{4583, 6619, 0x1602cfc1},
				{4584, 6618, 0x1602ce03},
				{4585, 6617, 0x1602cc49},
				{4586, 6616, 0x1602ca8e},
				{4587, 6615, 0x1602c8d6},
				{4588, 6614, 0x1602c71e},
				{4589, 6613, 0x1602c568},
				{4590, 6612, 0x1602c3b0},
				{4591, 6611, 0x1602c1fe},
				{4592, 6610, 0x1602c048},
				{4593, 6609, 0x1602be96},
				{4594, 6608, 0x1602bce3},
				{4595, 6607, 0x1602bb33},
				{4596, 6606, 0x1602b983},
				{4597, 6605, 0x1602b7d3},
				{4598, 6604, 0x1602b626},
				{4599, 6603, 0x1602b478},
				{4600, 6602, 0x1602b2cb},
				{4601, 6601, 0x1602b120},
				{4602, 6600, 0x1602af76},
				{4603, 6599, 0x1602adce},
				{4604, 6598, 0x1602ac28},
				{4605, 6597, 0x1602aa86},
				{4606, 6596, 0x1602a8e2},
				{4607, 6595, 0x1602a73e},
				{4608, 6594, 0x1602a59d},
				{4609, 6593, 0x1602a3fc},
				{4610, 6592, 0x1602a25c},
				{4611, 6591, 0x1602a0bd},
				{4612, 6590, 0x16029f1f},
				{4613, 6589, 0x16029d81},
				{4614, 6588, 0x16029be5},
				{4615, 6587, 0x16029a4a},
				{4616, 6586, 0x160298b0},
				{4617, 6585, 0x16029717},
				{4618, 6584, 0x1602957e},
				{4619, 6583, 0x160293e8},
				{4620, 6582, 0x16029252},
				{4621, 6581, 0x160290bd},
				{4622, 6580, 0x16028f28},
				{4623, 6579, 0x16028d96},
				{4624, 6578, 0x16028c03},
				{4625, 6577, 0x16028a71},
				{4626, 6576, 0x160288e1},
				{4627, 6575, 0x16028751},
				{4628, 6574, 0x160285c1},
				{4629, 6573, 0x16028434},
				{4630, 6572, 0x160282a7},
				{4631, 6571, 0x1602811b},
				{4632, 6570, 0x16027f8f},
				{4633, 6569, 0x16027e05},
				{4634, 6568, 0x16027c7c},
				{4635, 6567, 0x16027af5},
				{4636, 6566, 0x1602796e},
				{4637, 6565, 0x160277e7},
				{4638, 6564, 0x16027662},
				{4639, 6563, 0x160274dd},
				{4640, 6562, 0x16027359},
				{4641, 6561, 0x160271d7},
				{4642, 6560, 0x16027055},
				{4643, 6559, 0x16026ed4},
				{4644, 6558, 0x16026d54},
				{4645, 6557, 0x16026bd6},
				{4646, 6556, 0x16026a57},
				{4647, 6555, 0x160268d9},
				{4648, 6554, 0x1602675d},
				{4649, 6553, 0x160265e1},
				{4650, 6552, 0x16026467},
				{4651, 6551, 0x160262ee},
				{4652, 6550, 0x16026176},
				{4653, 6549, 0x16025ffd},
				{4654, 6548, 0x16025e87},
				{4655, 6547, 0x16025d11},
				{4656, 6546, 0x16025b9c},
				{4657, 6545, 0x16025a27},
				{4658, 6544, 0x160258b3},
				{4659, 6543, 0x16025741},
				{4660, 6542, 0x160255cf},
				{4661, 6541, 0x1602545e},
				{4662, 6540, 0x160252ee},
				{4663, 6539, 0x16025180},
				{4664, 6538, 0x16025012},
				{4665, 6537, 0x16024ea3},
				{4666, 6536, 0x16024d39},
				{4667, 6535, 0x16024bce},
				{4668, 6534, 0x16024a62},
				{4669, 6533, 0x160248f9},
				{4670, 6532, 0x16024790},
				{4671, 6531, 0x16024627},
				{4672, 6530, 0x160244c0},
				{4673, 6529, 0x1602435a},
				{4674, 6528, 0x160241f5},
				{4675, 6527, 0x16024090},
				{4676, 6526, 0x16023f2c},
				{4677, 6525, 0x16023dc9},
				{4678, 6524, 0x16023c67},
				{4679, 6523, 0x16023b06},
				{4680, 6522, 0x160239a5},
				{4681, 6521, 0x16023847},
				{4682, 6520, 0x160236e9},
				{4683, 6519, 0x1602358a},
				{4684, 6518, 0x1602342e},
				{4685, 6517, 0x160232d1},
				{4686, 6516, 0x16023175},
				{4687, 6515, 0x1602301b},
				{4688, 6514, 0x16022ec1},
				{4689, 6513, 0x16022d68},
				{4690, 6512, 0x16022c11},
				{4691, 6511, 0x16022ab9},
				{4692, 6510, 0x16022962},
				{4693, 6509, 0x1602280d},
				{4694, 6508, 0x160226b8},
				{4695, 6507, 0x16022564},
				{4696, 6506, 0x16022411},
				{4697, 6505, 0x160222c1},
				{4698, 6504, 0x1602216f},
				{4699, 6503, 0x1602201e},
				{4700, 6502, 0x16021ece},
				{4701, 6501, 0x16021d7f},
				{4702, 6500, 0x16021c31},
				{4703, 6499, 0x16021ae3},
				{4704, 6498, 0x16021996},
				{4705, 6497, 0x1602184b},
				{4706, 6496, 0x160216ff},
				{4707, 6495, 0x160215b5},
				{4708, 6494, 0x1602146d},
				{4709, 6493, 0x16021324},
				{4710, 6492, 0x160211db},
				{4711, 6491, 0x16021095},
				{4712, 6490, 0x16020f51},
				{4713, 6489, 0x16020e0b},
				{4714, 6488, 0x16020cc6},
				{4715, 6487, 0x16020b83},
				{4716, 6486, 0x16020a3f},
				{4717, 6485, 0x160208fd},
				{4718, 6484, 0x160207bb},
				{4719, 6483, 0x1602067b},
				{4720, 6482, 0x1602053a},
				{4721, 6481, 0x160203fb},
				{4722, 6480, 0x160202bd},
				{4723, 6479, 0x1602017f},
				{4724, 6478, 0x16020042},
				{4725, 6477, 0x1601ff07},
				{4726, 6476, 0x1601fdca},
				{4727, 6475, 0x1601fc90},
				{4728, 6474, 0x1601fb58},
				{4729, 6473, 0x1601fa1f},
				{4730, 6472, 0x1601f8e7},
				{4731, 6471, 0x1601f7af},
				{4732, 6470, 0x1601f678},
				{4733, 6469, 0x1601f542},
				{4734, 6468, 0x1601f40e},
				{4735, 6467, 0x1601f2d9},
				{4736, 6466, 0x1601f1a4},
				{4737, 6465, 0x1601f071},
				{4738, 6464, 0x1601ef3f},
				{4739, 6463, 0x1601ee0d},
				{4740, 6462, 0x1601ecdd},
				{4741, 6461, 0x1601ebae},
				{4742, 6460, 0x1601ea7e},
				{4743, 6459, 0x1601e951},
				{4744, 6458, 0x1601e823},
				{4745, 6457, 0x1601e6f6},
				{4746, 6456, 0x1601e5ca},
				{4747, 6455, 0x1601e49f},
				{4748, 6454, 0x1601e373},
				{4749, 6453, 0x1601e24a},
				{4750, 6452, 0x1601e120},
				{4751, 6451, 0x1601dff7},
				{4752, 6450, 0x1601decf},
				{4753, 6449, 0x1601dda8},
				{4754, 6448, 0x1601dc80},
				{4755, 6447, 0x1601db5b},
				{4756, 6446, 0x1601da36},
				{4757, 6445, 0x1601d912},
				{4758, 6444, 0x1601d7f0},
				{4759, 6443, 0x1601d6cd},
				{4760, 6442, 0x1601d5aa},
				{4761, 6441, 0x1601d489},
				{4762, 6440, 0x1601d367},
				{4763, 6439, 0x1601d247},
				{4764, 6438, 0x1601d128},
				{4765, 6437, 0x1601d009},
				{4766, 6436, 0x1601ceeb},
				{4767, 6435, 0x1601cdcf},
				{4768, 6434, 0x1601ccb1},
				{4769, 6433, 0x1601cb96},
				{4770, 6432, 0x1601ca7b},
				{4771, 6431, 0x1601c960},
				{4772, 6430, 0x1601c846},
				{4773, 6429, 0x1601c72c},
				{4774, 6428, 0x1601c615},
				{4775, 6427, 0x1601c4fd},
				{4776, 6426, 0x1601c3e6},
				{4777, 6425, 0x1601c2d0},
				{4778, 6424, 0x1601c1ba},
				{4779, 6423, 0x1601c0a4},
				{4780, 6422, 0x1601bf90},
				{4781, 6421, 0x1601be7d},
				{4782, 6420, 0x1601bd6a},
				{4783, 6419, 0x1601bc57},
				{4784, 6418, 0x1601bb45},
				{4785, 6417, 0x1601ba34},
				{4786, 6416, 0x1601b923},
				{4787, 6415, 0x1601b814},
				{4788, 6414, 0x1601b705},
				{4789, 6413, 0x1601b5f8},
				{4790, 6412, 0x1601b4ea},
				{4791, 6411, 0x1601b3dc},
				{4792, 6410, 0x1601b2d0},
				{4793, 6409, 0x1601b1c4},
				{4794, 6408, 0x1601b0b9},
				{4795, 6407, 0x1601afaf},
				{4796, 6406, 0x1601aea4},
				{4797, 6405, 0x1601ad9c},
				{4798, 6404, 0x1601ac92},
				{4799, 6403, 0x1601ab8b},
				{4800, 6402, 0x1601aa83},
				{4801, 6401, 0x1601a97d},
				{4802, 6400, 0x1601a876},
				{4803, 6399, 0x1601a771},
				{4804, 6398, 0x1601a66e},
				{4805, 6397, 0x1601a569},
				{4806, 6396, 0x1601a465},
				{4807, 6395, 0x1601a362},
				{4808, 6394, 0x1601a261},
				{4809, 6393, 0x1601a15f},
				{4810, 6392, 0x1601a05d},
				{4811, 6391, 0x16019f5d},
				{4812, 6390, 0x16019e5d},
				{4813, 6389, 0x16019d5e},
				{4814, 6388, 0x16019c60},
				{4815, 6387, 0x16019b61},
				{4816, 6386, 0x16019a64},
				{4817, 6385, 0x16019968},
				{4818, 6384, 0x1601986b},
				{4819, 6383, 0x16019771},
				{4820, 6382, 0x16019676},
				{4821, 6381, 0x1601957c},
				{4822, 6380, 0x16019483},
				{4823, 6379, 0x16019389},
				{4824, 6378, 0x16019291},
				{4825, 6377, 0x16019199},
				{4826, 6376, 0x160190a2},
				{4827, 6375, 0x16018fab},
				{4828, 6374, 0x16018eb4},
				{4829, 6373, 0x16018dbf},
				{4830, 6372, 0x16018cca},
				{4831, 6371, 0x16018bd5},
				{4832, 6370, 0x16018ae2},
				{4833, 6369, 0x160189ee},
				{4834, 6368, 0x160188fc},
				{4835, 6367, 0x1601880b},
				{4836, 6366, 0x16018719},
				{4837, 6365, 0x16018628},
				{4838, 6364, 0x16018538},
				{4839, 6363, 0x16018449},
				{4840, 6362, 0x16018359},
				{4841, 6361, 0x1601826a},
				{4842, 6360, 0x1601817d},
				{4843, 6359, 0x1601808f},
				{4844, 6358, 0x16017fa2},
				{4845, 6357, 0x16017eb6},
				{4846, 6356, 0x16017dc9},
				{4847, 6355, 0x16017cdf},
				{4848, 6354, 0x16017bf4},
				{4849, 6353, 0x16017b0a},
				{4850, 6352, 0x16017a21},
				{4851, 6351, 0x16017939},
				{4852, 6350, 0x1601784f},
				{4853, 6349, 0x16017767},
				{4854, 6348, 0x16017681},
				{4855, 6347, 0x1601759a},
				{4856, 6346, 0x160174b3},
				{4857, 6345, 0x160173ce},
				{4858, 6344, 0x160172e8},
				{4859, 6343, 0x16017204},
				{4860, 6342, 0x16017120},
				{4861, 6341, 0x1601703c},
				{4862, 6340, 0x16016f59},
				{4863, 6339, 0x16016e77},
				{4864, 6338, 0x16016d94},
				{4865, 6337, 0x16016cb2},
				{4866, 6336, 0x16016bd2},
				{4867, 6335, 0x16016af2},
				{4868, 6334, 0x16016a12},
				{4869, 6333, 0x16016932},
				{4870, 6332, 0x16016853},
				{4871, 6331, 0x16016775},
				{4872, 6330, 0x16016697},
				{4873, 6329, 0x160165ba},
				{4874, 6328, 0x160164dd},
				{4875, 6327, 0x16016401},
				{4876, 6326, 0x16016325},
				{4877, 6325, 0x1601624a},
				{4878, 6324, 0x1601616f},
				{4879, 6323, 0x16016095},
				{4880, 6322, 0x16015fbb},
				{4881, 6321, 0x16015ee2},
				{4882, 6320, 0x16015e0a},
				{4883, 6319, 0x16015d32},
				{4884, 6318, 0x16015c5a},
				{4885, 6317, 0x16015b82},
				{4886, 6316, 0x16015aab},
				{4887, 6315, 0x160159d5},
				{4888, 6314, 0x160158ff},
				{4889, 6313, 0x1601582a},
				{4890, 6312, 0x16015755},
				{4891, 6311, 0x16015682},
				{4892, 6310, 0x160155af},
				{4893, 6309, 0x160154dd},
				{4894, 6308, 0x1601540b},
				{4895, 6307, 0x1601533a},
				{4896, 6306, 0x1601526a},
				{4897, 6305, 0x1601519a},
				{4898, 6304, 0x160150ca},
				{4899, 6303, 0x16014ffb},
				{4900, 6302, 0x16014f2c},
				{4901, 6301, 0x16014e5e},
				{4902, 6300, 0x16014d90},
				{4903, 6299, 0x16014cc3},
				{4904, 6298, 0x16014bf6},
				{4905, 6297, 0x16014b29},
				{4906, 6296, 0x16014a5d},
				{4907, 6295, 0x16014992},
				{4908, 6294, 0x160148c7},
				{4909, 6293, 0x160147fc},
				{4910, 6292, 0x16014732},
				{4911, 6291, 0x16014669},
				{4912, 6290, 0x160145a1},
				{4913, 6289, 0x160144d8},
				{4914, 6288, 0x16014410},
				{4915, 6287, 0x16014348},
				{4916, 6286, 0x16014281},
				{4917, 6285, 0x160141bb},
				{4918, 6284, 0x160140f5},
				{4919, 6283, 0x1601402f},
				{4920, 6282, 0x16013f6a},
				{4921, 6281, 0x16013ea4},
				{4922, 6280, 0x16013de0},
				{4923, 6279, 0x16013d1c},
				{4924, 6278, 0x16013c59},
				{4925, 6277, 0x16013b95},
				{4926, 6276, 0x16013ad3},
				{4927, 6275, 0x16013a12},
				{4928, 6274, 0x16013950},
				{4929, 6273, 0x1601388f},
				{4930, 6272, 0x160137ce},
				{4931, 6271, 0x1601370e},
				{4932, 6270, 0x1601364e},
				{4933, 6269, 0x1601358f},
				{4934, 6268, 0x160134d0},
				{4935, 6267, 0x16013412},
				{4936, 6266, 0x16013354},
				{4937, 6265, 0x16013296},
				{4938, 6264, 0x160131d9},
				{4939, 6263, 0x1601311c},
				{4940, 6262, 0x16013060},
				{4941, 6261, 0x16012fa4},
				{4942, 6260, 0x16012ee9},
				{4943, 6259, 0x16012e2f},
				{4944, 6258, 0x16012d74},
				{4945, 6257, 0x16012cbb},
				{4946, 6256, 0x16012c01},
				{4947, 6255, 0x16012b48},
				{4948, 6254, 0x16012a8f},
				{4949, 6253, 0x160129d7},
				{4950, 6252, 0x1601291f},
				{4951, 6251, 0x16012868},
				{4952, 6250, 0x160127b1},
				{4953, 6249, 0x160126fa},
				{4954, 6248, 0x16012644},
				{4955, 6247, 0x1601258f},
				{4956, 6246, 0x160124da},
				{4957, 6245, 0x16012425},
				{4958, 6244, 0x16012371},
				{4959, 6243, 0x160122be},
				{4960, 6242, 0x1601220a},
				{4961, 6241, 0x16012157},
				{4962, 6240, 0x160120a4},
				{4963, 6239, 0x16011ff2},
				{4964, 6238, 0x16011f41},
				{4965, 6237, 0x16011e90},
				{4966, 6236, 0x16011ddf},
				{4967, 6235, 0x16011d2e},
				{4968, 6234, 0x16011c7e},
				{4969, 6233, 0x16011bcf},
				{4970, 6232, 0x16011b20},
				{4971, 6231, 0x16011a70},
				{4972, 6230, 0x160119c2},
				{4973, 6229, 0x16011915},
				{4974, 6228, 0x16011867},
				{4975, 6227, 0x160117ba},
				{4976, 6226, 0x1601170e},
				{4977, 6225, 0x16011661},
				{4978, 6224, 0x160115b6},
				{4979, 6223, 0x1601150b},
				{4980, 6222, 0x16011460},
				{4981, 6221, 0x160113b4},
				{4982, 6220, 0x1601130b},
				{4983, 6219, 0x16011261},
				{4984, 6218, 0x160111b7},
				{4985, 6217, 0x1601110e},
				{4986, 6216, 0x16011066},
				{4987, 6215, 0x16010fbe},
				{4988, 6214, 0x16010f16},
				{4989, 6213, 0x16010e6f},
				{4990, 6212, 0x16010dc8},
				{4991, 6211, 0x16010d22},
				{4992, 6210, 0x16010c7c},
				{4993, 6209, 0x16010bd6},
				{4994, 6208, 0x16010b31},
				{4995, 6207, 0x16010a8c},
				{4996, 6206, 0x160109e8},
				{4997, 6205, 0x16010943},
				{4998, 6204, 0x160108a0},
				{4999, 6203, 0x160107fc},
				{5000, 6202, 0x16010759},
				{5001, 6201, 0x160106b7},
				{5002, 6200, 0x16010614},
				{5003, 6199, 0x16010573},
				{5004, 6198, 0x160104d2},
				{5005, 6197, 0x16010431},
				{5006, 6196, 0x16010391},
				{5007, 6195, 0x160102f1},
				{5008, 6194, 0x16010250},
				{5009, 6193, 0x160101b1},
				{5010, 6192, 0x16010112},
				{5011, 6191, 0x16010073},
				{5012, 6190, 0x1600ffd5},
				{5013, 6189, 0x1600ff37},
				{5014, 6188, 0x1600fe9a},
				{5015, 6187, 0x1600fdfd},
				{5016, 6186, 0x1600fd60},
				{5017, 6185, 0x1600fcc4},
				{5018, 6184, 0x1600fc28},
				{5019, 6183, 0x1600fb8d},
				{5020, 6182, 0x1600faf2},
				{5021, 6181, 0x1600fa57},
				{5022, 6180, 0x1600f9bd},
				{5023, 6179, 0x1600f922},
				{5024, 6178, 0x1600f889},
				{5025, 6177, 0x1600f7ef},
				{5026, 6176, 0x1600f756},
				{5027, 6175, 0x1600f6be},
				{5028, 6174, 0x1600f626},
				{5029, 6173, 0x1600f58e},
				{5030, 6172, 0x1600f4f7},
				{5031, 6171, 0x1600f460},
				{5032, 6170, 0x1600f3c9},
				{5033, 6169, 0x1600f332},
				{5034, 6168, 0x1600f29d},
				{5035, 6167, 0x1600f207},
				{5036, 6166, 0x1600f172},
				{5037, 6165, 0x1600f0dd},
				{5038, 6164, 0x1600f048},
				{5039, 6163, 0x1600efb4},
				{5040, 6162, 0x1600ef21},
				{5041, 6161, 0x1600ee8d},
				{5042, 6160, 0x1600edfa},
				{5043, 6159, 0x1600ed67},
				{5044, 6158, 0x1600ecd5},
				{5045, 6157, 0x1600ec43},
				{5046, 6156, 0x1600ebb1},
				{5047, 6155, 0x1600eb20},
				{5048, 6154, 0x1600ea8f},
				{5049, 6153, 0x1600e9ff},
				{5050, 6152, 0x1600e96e},
				{5051, 6151, 0x1600e8df},
				{5052, 6150, 0x1600e84f},
				{5053, 6149, 0x1600e7c0},
				{5054, 6148, 0x1600e731},
				{5055, 6147, 0x1600e6a3},
				{5056, 6146, 0x1600e614},
				{5057, 6145, 0x1600e587},
				{5058, 6144, 0x1600e4f9},
				{5059, 6143, 0x1600e46c},
				{5060, 6142, 0x1600e3e0},
				{5061, 6141, 0x1600e353},
				{5062, 6140, 0x1600e2c6},
				{5063, 6139, 0x1600e23b},
				{5064, 6138, 0x1600e1af},
				{5065, 6137, 0x1600e125},
				{5066, 6136, 0x1600e09a},
				{5067, 6135, 0x1600e010},
				{5068, 6134, 0x1600df86},
				{5069, 6133, 0x1600defc},
				{5070, 6132, 0x1600de73},
				{5071, 6131, 0x1600dde9},
				{5072, 6130, 0x1600dd61},
				{5073, 6129, 0x1600dcd8},
				{5074, 6128, 0x1600dc50},
				{5075, 6127, 0x1600dbc9},
				{5076, 6126, 0x1600db41},
				{5077, 6125, 0x1600daba},
				{5078, 6124, 0x1600da34},
				{5079, 6123, 0x1600d9ad},
				{5080, 6122, 0x1600d928},
				{5081, 6121, 0x1600d8a2},
				{5082, 6120, 0x1600d81c},
				{5083, 6119, 0x1600d797},
				{5084, 6118, 0x1600d713},
				{5085, 6117, 0x1600d68e},
				{5086, 6116, 0x1600d60a},
				{5087, 6115, 0x1600d586},
				{5088, 6114, 0x1600d503},
				{5089, 6113, 0x1600d47f},
				{5090, 6112, 0x1600d3fc},
				{5091, 6111, 0x1600d37a},
				{5092, 6110, 0x1600d2f8},
				{5093, 6109, 0x1600d275},
				{5094, 6108, 0x1600d1f4},
				{5095, 6107, 0x1600d172},
				{5096, 6106, 0x1600d0f2},
				{5097, 6105, 0x1600d072},
				{5098, 6104, 0x1600cff1},
				{5099, 6103, 0x1600cf71},
				{5100, 6102, 0x1600cef1},
				{5101, 6101, 0x1600ce72},
				{5102, 6100, 0x1600cdf3},
				{5103, 6099, 0x1600cd74},
				{5104, 6098, 0x1600ccf5},
				{5105, 6097, 0x1600cc77},
				{5106, 6096, 0x1600cbf9},
				{5107, 6095, 0x1600cb7c},
				{5108, 6094, 0x1600cafe},
				{5109, 6093, 0x1600ca81},
				{5110, 6092, 0x1600ca05},
				{5111, 6091, 0x1600c989},
				{5112, 6090, 0x1600c90c},
				{5113, 6089, 0x1600c891},
				{5114, 6088, 0x1600c816},
				{5115, 6087, 0x1600c79a},
				{5116, 6086, 0x1600c71f},
				{5117, 6085, 0x1600c6a5},
				{5118, 6084, 0x1600c62a},
				{5119, 6083, 0x1600c5b1},
				{5120, 6082, 0x1600c536},
				{5121, 6081, 0x1600c4bd},
				{5122, 6080, 0x1600c444},
				{5123, 6079, 0x1600c3cb},
				{5124, 6078, 0x1600c352},
				{5125, 6077, 0x1600c2da},
				{5126, 6076, 0x1600c263},
				{5127, 6075, 0x1600c1eb},
				{5128, 6074, 0x1600c173},
				{5129, 6073, 0x1600c0fc},
				{5130, 6072, 0x1600c085},
				{5131, 6071, 0x1600c00e},
				{5132, 6070, 0x1600bf98},
				{5133, 6069, 0x1600bf22},
				{5134, 6068, 0x1600bead},
				{5135, 6067, 0x1600be37},
				{5136, 6066, 0x1600bdc2},
				{5137, 6065, 0x1600bd4d},
				{5138, 6064, 0x1600bcd8},
				{5139, 6063, 0x1600bc64},
				{5140, 6062, 0x1600bbf0},
				{5141, 6061, 0x1600bb7c},
				{5142, 6060, 0x1600bb09},
				{5143, 6059, 0x1600ba96},
				{5144, 6058, 0x1600ba22},
				{5145, 6057, 0x1600b9b0},
				{5146, 6056, 0x1600b93d},
				{5147, 6055, 0x1600b8cb},
				{5148, 6054, 0x1600b859},
				{5149, 6053, 0x1600b7e7},
				{5150, 6052, 0x1600b776},
				{5151, 6051, 0x1600b705},
				{5152, 6050, 0x1600b694},
				{5153, 6049, 0x1600b623},
				{5154, 6048, 0x1600b5b3},
				{5155, 6047, 0x1600b543},
				{5156, 6046, 0x1600b4d3},
				{5157, 6045, 0x1600b464},
				{5158, 6044, 0x1600b3f5},
				{5159, 6043, 0x1600b385},
				{5160, 6042, 0x1600b316},
				{5161, 6041, 0x1600b2a8},
				{5162, 6040, 0x1600b23a},
				{5163, 6039, 0x1600b1cc},
				{5164, 6038, 0x1600b15e},
				{5165, 6037, 0x1600b0f0},
				{5166, 6036, 0x1600b083},
				{5167, 6035, 0x1600b016},
				{5168, 6034, 0x1600afa9},
				{5169, 6033, 0x1600af3c},
				{5170, 6032, 0x1600aed0},
				{5171, 6031, 0x1600ae64},
				{5172, 6030, 0x1600adf9},
				{5173, 6029, 0x1600ad8e},
				{5174, 6028, 0x1600ad22},
				{5175, 6027, 0x1600acb7},
				{5176, 6026, 0x1600ac4c},
				{5177, 6025, 0x1600abe2},
				{5178, 6024, 0x1600ab78},
				{5179, 6023, 0x1600ab0e},
				{5180, 6022, 0x1600aaa5},
				{5181, 6021, 0x1600aa3c},
				{5182, 6020, 0x1600a9d4},
				{5183, 6019, 0x1600a96b},
				{5184, 6018, 0x1600a903},
				{5185, 6017, 0x1600a89b},
				{5186, 6016, 0x1600a833},
				{5187, 6015, 0x1600a7cb},
				{5188, 6014, 0x1600a765},
				{5189, 6013, 0x1600a6fd},
				{5190, 6012, 0x1600a697},
				{5191, 6011, 0x1600a630},
				{5192, 6010, 0x1600a5ca},
				{5193, 6009, 0x1600a564},
				{5194, 6008, 0x1600a4fe},
				{5195, 6007, 0x1600a498},
				{5196, 6006, 0x1600a433},
				{5197, 6005, 0x1600a3ce},
				{5198, 6004, 0x1600a369},
				{5199, 6003, 0x1600a304},
				{5200, 6002, 0x1600a2a0},
				{5201, 6001, 0x1600a23c},
				{5202, 6000, 0x1600a1d8},
				{5203, 5999, 0x1600a174},
				{5204, 5998, 0x1600a111},
				{5205, 5997, 0x1600a0ae},
				{5206, 5996, 0x1600a04b},
				{5207, 5995, 0x16009fe8},
				{5208, 5994, 0x16009f85},
				{5209, 5993, 0x16009f23},
				{5210, 5992, 0x16009ec1},
				{5211, 5991, 0x16009e5f},
				{5212, 5990, 0x16009dfd},
				{5213, 5989, 0x16009d9c},
				{5214, 5988, 0x16009d3b},
				{5215, 5987, 0x16009cda},
				{5216, 5986, 0x16009c79},
				{5217, 5985, 0x16009c19},
				{5218, 5984, 0x16009bb9},
				{5219, 5983, 0x16009b59},
				{5220, 5982, 0x16009af9},
				{5221, 5981, 0x16009a99},
				{5222, 5980, 0x16009a3a},
				{5223, 5979, 0x160099db},
				{5224, 5978, 0x1600997c},
				{5225, 5977, 0x1600991d},
				{5226, 5976, 0x160098bf},
				{5227, 5975, 0x16009861},
				{5228, 5974, 0x16009803},
				{5229, 5973, 0x160097a5},
				{5230, 5972, 0x16009748},
				{5231, 5971, 0x160096ea},
				{5232, 5970, 0x1600968d},
				{5233, 5969, 0x16009630},
				{5234, 5968, 0x160095d4},
				{5235, 5967, 0x16009577},
				{5236, 5966, 0x1600951b},
				{5237, 5965, 0x160094bf},
				{5238, 5964, 0x16009464},
				{5239, 5963, 0x16009408},
				{5240, 5962, 0x160093ac},
				{5241, 5961, 0x16009351},
				{5242, 5960, 0x160092f6},
				{5243, 5959, 0x1600929c},
				{5244, 5958, 0x16009241},
				{5245, 5957, 0x160091e7},
				{5246, 5956, 0x1600918d},
				{5247, 5955, 0x16009133},
				{5248, 5954, 0x160090da},
				{5249, 5953, 0x16009080},
				{5250, 5952, 0x16009027},
				{5251, 5951, 0x16008fce},
				{5252, 5950, 0x16008f76},
				{5253, 5949, 0x16008f1d},
				{5254, 5948, 0x16008ec5},
				{5255, 5947, 0x16008e6d},
				{5256, 5946, 0x16008e15},
				{5257, 5945, 0x16008dbd},
				{5258, 5944, 0x16008d66},
				{5259, 5943, 0x16008d0e},
				{5260, 5942, 0x16008cb7},
				{5261, 5941, 0x16008c60},
				{5262, 5940, 0x16008c0a},
				{5263, 5939, 0x16008bb3},
				{5264, 5938, 0x16008b5d},
				{5265, 5937, 0x16008b08},
				{5266, 5936, 0x16008ab2},
				{5267, 5935, 0x16008a5c},
				{5268, 5934, 0x16008a07},
				{5269, 5933, 0x160089b1},
				{5270, 5932, 0x1600895d},
				{5271, 5931, 0x16008908},
				{5272, 5930, 0x160088b3},
				{5273, 5929, 0x1600885f},
				{5274, 5928, 0x1600880a},
				{5275, 5927, 0x160087b7},
				{5276, 5926, 0x16008763},
				{5277, 5925, 0x1600870f},
				{5278, 5924, 0x160086bc},
				{5279, 5923, 0x16008669},
				{5280, 5922, 0x16008616},
				{5281, 5921, 0x160085c3},
				{5282, 5920, 0x16008571},
				{5283, 5919, 0x1600851e},
				{5284, 5918, 0x160084cc},
				{5285, 5917, 0x1600847a},
				{5286, 5916, 0x16008428},
				{5287, 5915, 0x160083d7},
				{5288, 5914, 0x16008386},
				{5289, 5913, 0x16008334},
				{5290, 5912, 0x160082e3},
				{5291, 5911, 0x16008292},
				{5292, 5910, 0x16008242},
				{5293, 5909, 0x160081f1},
				{5294, 5908, 0x160081a1},
				{5295, 5907, 0x16008152},
				{5296, 5906, 0x16008102},
				{5297, 5905, 0x160080b2},
				{5298, 5904, 0x16008063},
				{5299, 5903, 0x16008014},
				{5300, 5902, 0x157fc536},
				{5301, 5901, 0x157f7665},
				{5302, 5900, 0x157f2793},
				{5303, 5899, 0x157ed917},
				{5304, 5898, 0x157e8af1},
				{5305, 5897, 0x157e3cca},
				{5306, 5896, 0x157deefa},
				{5307, 5895, 0x157da12a},
				{5308, 5894, 0x157d53af},
				{5309, 5893, 0x157d068b},
				{5310, 5892, 0x157cb9bc},
				{5311, 5891, 0x157c6c98},
				{5312, 5890, 0x157c1fc9},
				{5313, 5889, 0x157bd350},
				{5314, 5888, 0x157b86d7},
				{5315, 5887, 0x157b3ab4},
				{5316, 5886, 0x157aeee7},
				{5317, 5885, 0x157aa31a},
				{5318, 5884, 0x157a574d},
				{5319, 5883, 0x157a0bd6},
				{5320, 5882, 0x1579c0b4},
				{5321, 5881, 0x15797593},
				{5322, 5880, 0x15792a72},
				{5323, 5879, 0x1578dffc},
				{5324, 5878, 0x15789530},
				{5325, 5877, 0x15784b11},
				{5326, 5876, 0x157800f1},
				{5327, 5875, 0x1577b727},
				{5328, 5874, 0x15776d5d},
				{5329, 5873, 0x15772393},
				{5330, 5872, 0x1576da1f},
				{5331, 5871, 0x15769101},
				{5332, 5870, 0x157647e3},
				{5333, 5869, 0x1575fec4},
				{5334, 5868, 0x1575b5fc},
				{5335, 5867, 0x15756d8a},
				{5336, 5866, 0x15752517},
				{5337, 5865, 0x1574dcfb},
				{5338, 5864, 0x157494de},
				{5339, 5863, 0x15744d17},
				{5340, 5862, 0x15740551},
				{5341, 5861, 0x1573be35},
				{5342, 5860, 0x157376c5},
				{5343, 5859, 0x15732faa},
				{5344, 5858, 0x1572e88f},
				{5345, 5857, 0x1572a1c9},
				{5346, 5856, 0x15725b04},
				{5347, 5855, 0x15721495},
				{5348, 5854, 0x1571ce26},
				{5349, 5853, 0x1571880c},
				{5350, 5852, 0x15714249},
				{5351, 5851, 0x1570fc85},
				{5352, 5850, 0x1570b6c2},
				{5353, 5849, 0x15707154},
				{5354, 5848, 0x15702be6},
				{5355, 5847, 0x156fe6ce},
				{5356, 5846, 0x156fa262},
				{5357, 5845, 0x156f5d4a},
				{5358, 5844, 0x156f18de},
				{5359, 5843, 0x156ed472},
				{5360, 5842, 0x156e9006},
				{5361, 5841, 0x156e4bf0},
				{5362, 5840, 0x156e082f},
				{5363, 5839, 0x156dc419},
				{5364, 5838, 0x156d80af},
				{5365, 5837, 0x156d3d44},
				{5366, 5836, 0x156cf9da},
				{5367, 5835, 0x156cb6c5},
				{5368, 5834, 0x156c73b0},
				{5369, 5833, 0x156c30f2},
				{5370, 5832, 0x156bee33},
				{5371, 5831, 0x156babca},
				{5372, 5830, 0x156b69b7},
				{5373, 5829, 0x156b27a4},
				{5374, 5828, 0x156ae591},
				{5375, 5827, 0x156aa3d3},
				{5376, 5826, 0x156a6216},
				{5377, 5825, 0x156a20af},
				{5378, 5824, 0x1569df48},
				{5379, 5823, 0x15699de0},
				{5380, 5822, 0x15695d25},
				{5381, 5821, 0x15691c13},
				{5382, 5820, 0x1568db57},
				{5383, 5819, 0x15689af2},
				{5384, 5818, 0x15685a36},
				{5385, 5817, 0x15681a26},
				{5386, 5816, 0x1567da16},
				{5387, 5815, 0x15679a5c},
				{5388, 5814, 0x15675aa2},
				{5389, 5813, 0x15671ae8},
				{5390, 5812, 0x1566db84},
				{5391, 5811, 0x15669c20},
				{5392, 5810, 0x15665cbc},
				{5393, 5809, 0x15661dae},
				{5394, 5808, 0x1565def5},
				{5395, 5807, 0x1565a03d},
				{5396, 5806, 0x15656184},
				{5397, 5805, 0x15652322},
				{5398, 5804, 0x1564e4bf},
				{5399, 5803, 0x1564a6b2},
				{5400, 5802, 0x156468a6},
				{5401, 5801, 0x15642aef},
				{5402, 5800, 0x1563ed38},
				{5403, 5799, 0x1563afd7},
				{5404, 5798, 0x15637276},
				{5405, 5797, 0x1563356b},
				{5406, 5796, 0x1562f80a},
				{5407, 5795, 0x1562bb55},
				{5408, 5794, 0x15627e4a},
				{5409, 5793, 0x156241ea},
				{5410, 5792, 0x15620535},
				{5411, 5791, 0x1561c8d5},
				{5412, 5790, 0x15618ccc},
				{5413, 5789, 0x1561506d},
				{5414, 5788, 0x156114b9},
				{5415, 5787, 0x1560d8af},
				{5416, 5786, 0x15609cfc},
				{5417, 5785, 0x1560619e},
				{5418, 5784, 0x15602696},
				{5419, 5783, 0x155feb38},
				{5420, 5782, 0x155fb030},
				{5421, 5781, 0x155f7528},
				{5422, 5780, 0x155f3a76},
				{5423, 5779, 0x155effc4},
				{5424, 5778, 0x155ec512},
				{5425, 5777, 0x155e8ab6},
				{5426, 5776, 0x155e505a},
				{5427, 5775, 0x155e1653},
				{5428, 5774, 0x155ddc4d},
				{5429, 5773, 0x155da29c},
				{5430, 5772, 0x155d68ec},
				{5431, 5771, 0x155d2f3c},
				{5432, 5770, 0x155cf5e1},
				{5433, 5769, 0x155cbcdc},
				{5434, 5768, 0x155c8382},
				{5435, 5767, 0x155c4a7d},
				{5436, 5766, 0x155c1178},
				{5437, 5765, 0x155bd8c9},
				{5438, 5764, 0x155ba01a},
				{5439, 5763, 0x155b67c1},
				{5440, 5762, 0x155b2f12},
				{5441, 5761, 0x155af70f},
				{5442, 5760, 0x155abeb6},
				{5443, 5759, 0x155a8709},
				{5444, 5758, 0x155a4f06},
				{5445, 5757, 0x155a1758},
				{5446, 5756, 0x1559dfab},
				{5447, 5755, 0x1559a854},
				{5448, 5754, 0x155970fc},
				{5449, 5753, 0x155939fb},
				{5450, 5752, 0x155902a3},
				{5451, 5751, 0x1558cbf7},
				{5452, 5750, 0x155894f6},
				{5453, 5749, 0x15585e4a},
				{5454, 5748, 0x1558279e},
				{5455, 5747, 0x1557f149},
				{5456, 5746, 0x1557baf3},
				{5457, 5745, 0x1557849d},
				{5458, 5744, 0x15574e9d},
				{5459, 5743, 0x1557189d},
				{5460, 5742, 0x1556e2f3},
				{5461, 5741, 0x1556acf3},
				{5462, 5740, 0x1556779e},
				{5463, 5739, 0x155641f4},
				{5464, 5738, 0x15560cf6},
				{5465, 5737, 0x1555d7a2},
				{5466, 5736, 0x1555a2f9},
				{5467, 5735, 0x15556e51},
				{5468, 5734, 0x155539a8},
				{5469, 5733, 0x15550555},
				{5470, 5732, 0x1554d103},
				{5471, 5731, 0x15549cb0},
				{5472, 5730, 0x155468b3},
				{5473, 5729, 0x155434b6},
				{5474, 5728, 0x155400e5},
				{5475, 5727, 0x1553cd3e},
				{5476, 5726, 0x15539997},
				{5477, 5725, 0x1553661b},
				{5478, 5724, 0x155332ca},
				{5479, 5723, 0x1552ffce},
				{5480, 5722, 0x1552cca8},
				{5481, 5721, 0x155299ad},
				{5482, 5720, 0x155266dd},
				{5483, 5719, 0x15523437},
				{5484, 5718, 0x15520192},
				{5485, 5717, 0x1551ceed},
				{5486, 5716, 0x15519c9d},
				{5487, 5715, 0x15516a4e},
				{5488, 5714, 0x15513829},
				{5489, 5713, 0x1551062f},
				{5490, 5712, 0x1550d436},
				{5491, 5711, 0x1550a267},
				{5492, 5710, 0x155070c3},
				{5493, 5709, 0x15503f20},
				{5494, 5708, 0x15500da7},
				{5495, 5707, 0x154fdcaf},
				{5496, 5706, 0x154fab61},
				{5497, 5705, 0x154f7a3e},
				{5498, 5704, 0x154f4946},
				{5499, 5703, 0x154f1879},
				{5500, 5702, 0x154ee7ac},
				{5501, 5701, 0x154eb709},
				{5502, 5700, 0x154e8667},
				{5503, 5699, 0x154e561b},
				{5504, 5698, 0x154e25cf},
				{5505, 5697, 0x154df582},
				{5506, 5696, 0x154dc58c},
				{5507, 5695, 0x154d9595},
				{5508, 5694, 0x154d65ca},
				{5509, 5693, 0x154d35fe},
				{5510, 5692, 0x154d0689},
				{5511, 5691, 0x154cd713},
				{5512, 5690, 0x154ca79d},
				{5513, 5689, 0x154c7853},
				{5514, 5688, 0x154c4933},
				{5515, 5687, 0x154c1a3e},
				{5516, 5686, 0x154beb49},
				{5517, 5685, 0x154bbc54},
				{5518, 5684, 0x154b8db5},
				{5519, 5683, 0x154b5f16},
				{5520, 5682, 0x154b30a2},
				{5521, 5681, 0x154b022e},
				{5522, 5680, 0x154ad3e5},
				{5523, 5679, 0x154aa5c7},
				{5524, 5678, 0x154a77a9},
				{5525, 5677, 0x154a49e0},
				{5526, 5676, 0x154a1c18},
				{5527, 5675, 0x1549ee50},
				{5528, 5674, 0x1549c0b2},
				{5529, 5673, 0x15499340},
				{5530, 5672, 0x154965cd},
				{5531, 5671, 0x15493886},
				{5532, 5670, 0x15490b69},
				{5533, 5669, 0x1548de4d},
				{5534, 5668, 0x1548b15b},
				{5535, 5667, 0x15488469},
				{5536, 5666, 0x154857a2},
				{5537, 5665, 0x15482b07},
				{5538, 5664, 0x1547fe6b},
				{5539, 5663, 0x1547d1fa},
				{5540, 5662, 0x1547a5b4},
				{5541, 5661, 0x154779c4},
				{5542, 5660, 0x15474da9},
				{5543, 5659, 0x1547218e},
				{5544, 5658, 0x1546f59e},
				{5545, 5657, 0x1546c9d9},
				{5546, 5656, 0x15469e3e},
				{5547, 5655, 0x154672a4},
				{5548, 5654, 0x1546470a},
				{5549, 5653, 0x15461bc6},
				{5550, 5652, 0x1545f081},
				{5551, 5651, 0x1545c53d},
				{5552, 5650, 0x15459a23},
				{5553, 5649, 0x15456f35},
				{5554, 5648, 0x15454471},
				{5555, 5647, 0x154519ae},
				{5556, 5646, 0x1544ef15},
				{5557, 5645, 0x1544c4a7},
				{5558, 5644, 0x15449a3a},
				{5559, 5643, 0x15446fcc},
				{5560, 5642, 0x15444589},
				{5561, 5641, 0x15441b71},
				{5562, 5640, 0x1543f15a},
				{5563, 5639, 0x1543c76d},
				{5564, 5638, 0x15439dab},
				{5565, 5637, 0x154373e9},
				{5566, 5636, 0x15434a52},
				{5567, 5635, 0x154320bb},
				{5568, 5634, 0x1542f74e},
				{5569, 5633, 0x1542cde2},
				{5570, 5632, 0x1542a4cc},
				{5571, 5631, 0x15427bb6},
				{5572, 5630, 0x154252cb},
				{5573, 5629, 0x154229df},
				{5574, 5628, 0x154200f4},
				{5575, 5627, 0x1541d833},
				{5576, 5626, 0x1541af9e},
				{5577, 5625, 0x15418709},
				{5578, 5624, 0x15415e9e},
				{5579, 5623, 0x15413634},
				{5580, 5622, 0x15410df4},
				{5581, 5621, 0x1540e5df},
				{5582, 5620, 0x1540bdcb},
				{5583, 5619, 0x154095e1},
				{5584, 5618, 0x15406df7},
				{5585, 5617, 0x15404638},
				{5586, 5616, 0x15401ea4},
				{5587, 5615, 0x153ff73c},
				{5588, 5614, 0x153fcfa8},
				{5589, 5613, 0x153fa86a},
				{5590, 5612, 0x153f8101},
				{5591, 5611, 0x153f59ee},
				{5592, 5610, 0x153f32db},
				{5593, 5609, 0x153f0bc8},
				{5594, 5608, 0x153ee4df},
				{5595, 5607, 0x153ebe22},
				{5596, 5606, 0x153e9765},
				{5597, 5605, 0x153e70d3},
				{5598, 5604, 0x153e4a40},
				{5599, 5603, 0x153e23d9},
				{5600, 5602, 0x153dfd72},
				{5601, 5601, 0x153dd735},
				{5602, 5600, 0x153db14f},
				{5603, 5599, 0x153d8b3d},
				{5604, 5598, 0x153d652c},
				{5605, 5597, 0x153d3f70},
				{5606, 5596, 0x153d198a},
				{5607, 5595, 0x153cf3f9},
				{5608, 5594, 0x153cce3d},
				{5609, 5593, 0x153ca8d8},
				{5610, 5592, 0x153c8372},
				{5611, 5591, 0x153c5e0c},
				{5612, 5590, 0x153c38d1},
				{5613, 5589, 0x153c13c1},
				{5614, 5588, 0x153beeb1},
				{5615, 5587, 0x153bc9a2},
				{5616, 5586, 0x153ba4e8},
				{5617, 5585, 0x153b802d},
				{5618, 5584, 0x153b5b9e},
				{5619, 5583, 0x153b36e4},
				{5620, 5582, 0x153b1280},
				{5621, 5581, 0x153aedf1},
				{5622, 5580, 0x153ac9b8},
				{5623, 5579, 0x153aa57f},
				{5624, 5578, 0x153a8145},
				{5625, 5577, 0x153a5d37},
				{5626, 5576, 0x153a3929},
				{5627, 5575, 0x153a1545},
				{5628, 5574, 0x1539f18d},
				{5629, 5573, 0x1539cdd5},
				{5630, 5572, 0x1539aa1c},
				{5631, 5571, 0x1539868f},
				{5632, 5570, 0x1539632c},
				{5633, 5569, 0x15393ff4},
				{5634, 5568, 0x15391cbd},
				{5635, 5567, 0x1538f985},
				{5636, 5566, 0x1538d64d},
				{5637, 5565, 0x1538b36c},
				{5638, 5564, 0x1538905f},
				{5639, 5563, 0x15386d7d},
				{5640, 5562, 0x15384ac6},
				{5641, 5561, 0x1538280f},
				{5642, 5560, 0x15380584},
				{5643, 5559, 0x1537e2f8},
				{5644, 5558, 0x1537c097},
				{5645, 5557, 0x15379e36},
				{5646, 5556, 0x15377bd5},
				{5647, 5555, 0x153759ca},
				{5648, 5554, 0x153737be},
				{5649, 5553, 0x153715b3},
				{5650, 5552, 0x1536f3d3},
				{5651, 5551, 0x1536d1f3},
				{5652, 5550, 0x1536b03e},
				{5653, 5549, 0x15368e88},
				{5654, 5548, 0x15366cd3},
				{5655, 5547, 0x15364b49},
				{5656, 5546, 0x153629e9},
				{5657, 5545, 0x1536088a},
				{5658, 5544, 0x1535e756},
				{5659, 5543, 0x1535c621},
				{5660, 5542, 0x1535a4ed},
				{5661, 5541, 0x153583e3},
				{5662, 5540, 0x153562da},
				{5663, 5539, 0x15354226},
				{5664, 5538, 0x15352172},
				{5665, 5537, 0x153500bf},
				{5666, 5536, 0x1534e00b},
				{5667, 5535, 0x1534bf82},
				{5668, 5534, 0x15349ef9},
				{5669, 5533, 0x15347e9c},
				{5670, 5532, 0x15345e3e},
				{5671, 5531, 0x15343e0b},
				{5672, 5530, 0x15341dd8},
				{5673, 5529, 0x1533fda5},
				{5674, 5528, 0x1533dd9d},
				{5675, 5527, 0x1533bdc0},
				{5676, 5526, 0x15339de3},
				{5677, 5525, 0x15337e06},
				{5678, 5524, 0x15335e54},
				{5679, 5523, 0x15333ef8},
				{5680, 5522, 0x15331f46},
				{5681, 5521, 0x1532ffea},
				{5682, 5520, 0x1532e062},
				{5683, 5519, 0x1532c131},
				{5684, 5518, 0x1532a1d5},
				{5685, 5517, 0x153282a4},
				{5686, 5516, 0x1532639d},
				{5687, 5515, 0x15324497},
				{5688, 5514, 0x15322591},
				{5689, 5513, 0x153206b5},
				{5690, 5512, 0x1531e7da},
				{5691, 5511, 0x1531c929},
				{5692, 5510, 0x1531aa79},
				{5693, 5509, 0x15318bf3},
				{5694, 5508, 0x15316d99},
				{5695, 5507, 0x15314f13},
				{5696, 5506, 0x153130b8},
				{5697, 5505, 0x1531125e},
				{5698, 5504, 0x1530f42e},
				{5699, 5503, 0x1530d5fe},
				{5700, 5502, 0x1530b7fa},
				{5701, 5501, 0x153099f5},
				{5702, 5500, 0x15307bf0},
				{5703, 5499, 0x15305e16},
				{5704, 5498, 0x15304067},
				{5705, 5497, 0x1530228e},
				{5706, 5496, 0x153004df},
				{5707, 5495, 0x152fe75b},
				{5708, 5494, 0x152fc9d7},
				{5709, 5493, 0x152fac7e},
				{5710, 5492, 0x152f8f25},
				{5711, 5491, 0x152f71cc},
				{5712, 5490, 0x152f549d},
				{5713, 5489, 0x152f376f},
				{5714, 5488, 0x152f1a6c},
				{5715, 5487, 0x152efd3e},
				{5716, 5486, 0x152ee066},
				{5717, 5485, 0x152ec363},
				{5718, 5484, 0x152ea68a},
				{5719, 5483, 0x152e89dd},
				{5720, 5482, 0x152e6d30},
				{5721, 5481, 0x152e5082},
				{5722, 5480, 0x152e3400},
				{5723, 5479, 0x152e177e},
				{5724, 5478, 0x152dfb26},
				{5725, 5477, 0x152ddecf},
				{5726, 5476, 0x152dc277},
				{5727, 5475, 0x152da64b},
				{5728, 5474, 0x152d8a1e},
				{5729, 5473, 0x152d6e1d},
				{5730, 5472, 0x152d521b},
				{5731, 5471, 0x152d361a},
				{5732, 5470, 0x152d1a18},
				{5733, 5469, 0x152cfe6c},
				{5734, 5468, 0x152ce296},
				{5735, 5467, 0x152cc6ea},
				{5736, 5466, 0x152cab3e},
				{5737, 5465, 0x152c8f92},
				{5738, 5464, 0x152c7412},
				{5739, 5463, 0x152c58bc},
				{5740, 5462, 0x152c3d66},
				{5741, 5461, 0x152c2210},
				{5742, 5460, 0x152c06e5},
				{5743, 5459, 0x152beb8f},
				{5744, 5458, 0x152bd08f},
				{5745, 5457, 0x152bb564},
				{5746, 5456, 0x152b9a64},
				{5747, 5455, 0x152b7f64},
				{5748, 5454, 0x152b648f},
				{5749, 5453, 0x152b49ba},
				{5750, 5452, 0x152b2ee5},
				{5751, 5451, 0x152b143b},
				{5752, 5450, 0x152af991},
				{5753, 5449, 0x152adf12},
				{5754, 5448, 0x152ac4be},
				{5755, 5447, 0x152aaa94},
				{5756, 5446, 0x152a9055},
				{5757, 5445, 0x152a7617},
				{5758, 5444, 0x152a5c03},
				{5759, 5443, 0x152a41ef},
				{5760, 5442, 0x152a27f0},
				{5761, 5441, 0x152a0df2},
				{5762, 5440, 0x1529f41f},
				{5763, 5439, 0x1529da4b},
				{5764, 5438, 0x1529c08d},
				{5765, 5437, 0x1529a6cf},
				{5766, 5436, 0x15298d27},
				{5767, 5435, 0x15297393},
				{5768, 5434, 0x15295a16},
				{5769, 5433, 0x15294098},
				{5770, 5432, 0x15292730},
				{5771, 5431, 0x15290df3},
				{5772, 5430, 0x1528f4b6},
				{5773, 5429, 0x1528db78},
				{5774, 5428, 0x1528c251},
				{5775, 5427, 0x1528a929},
				{5776, 5426, 0x15289017},
				{5777, 5425, 0x1528771a},
				{5778, 5424, 0x15285e33},
				{5779, 5423, 0x1528454b},
				{5780, 5422, 0x15282c79},
				{5781, 5421, 0x152813bd},
				{5782, 5420, 0x1527fb16},
				{5783, 5419, 0x1527e26f},
				{5784, 5418, 0x1527c9dd},
				{5785, 5417, 0x1527b14c},
				{5786, 5416, 0x152798fb},
				{5787, 5415, 0x15278094},
				{5788, 5414, 0x1527682e},
				{5789, 5413, 0x15274fdd},
				{5790, 5412, 0x152737a1},
				{5791, 5411, 0x15271f7b},
				{5792, 5410, 0x15270755},
				{5793, 5409, 0x1526ef44},
				{5794, 5408, 0x1526d749},
				{5795, 5407, 0x1526bf4e},
				{5796, 5406, 0x1526a768},
				{5797, 5405, 0x15268f98},
				{5798, 5404, 0x152677c7},
				{5799, 5403, 0x15266022},
				{5800, 5402, 0x15264867},
				{5801, 5401, 0x152630ed},
				{5802, 5400, 0x1526195d},
				{5803, 5399, 0x152601e2},
				{5804, 5398, 0x1525ea68},
				{5805, 5397, 0x1525d303},
				{5806, 5396, 0x1525bbb3},
				{5807, 5395, 0x1525a464},
				{5808, 5394, 0x15258d2a},
				{5809, 5393, 0x15257605},
				{5810, 5392, 0x15255ee1},
				{5811, 5391, 0x152547d2},
				{5812, 5390, 0x152530d8},
				{5813, 5389, 0x152519de},
				{5814, 5388, 0x152502fa},
				{5815, 5387, 0x1524ec2c},
				{5816, 5386, 0x1524d55d},
				{5817, 5385, 0x1524beb9},
				{5818, 5384, 0x1524a815},
				{5819, 5383, 0x15249172},
				{5820, 5382, 0x15247ae3},
				{5821, 5381, 0x15246455},
				{5822, 5380, 0x15244df2},
				{5823, 5379, 0x15243779},
				{5824, 5378, 0x1524212b},
				{5825, 5377, 0x15240add},
				{5826, 5376, 0x1523f4a4},
				{5827, 5375, 0x1523de6c},
				{5828, 5374, 0x1523c849},
				{5829, 5373, 0x1523b23c},
				{5830, 5372, 0x15239c2e},
				{5831, 5371, 0x15238636},
				{5832, 5370, 0x15237053},
				{5833, 5369, 0x15235a71},
				{5834, 5368, 0x152344a4},
				{5835, 5367, 0x15232eec},
				{5836, 5366, 0x15231934},
				{5837, 5365, 0x1523037d},
				{5838, 5364, 0x1522eddb},
				{5839, 5363, 0x1522d84e},
				{5840, 5362, 0x1522c2d7},
				{5841, 5361, 0x1522ad5f},
				{5842, 5360, 0x152297fe},
				{5843, 5359, 0x1522829c},
				{5844, 5358, 0x15226d4f},
				{5845, 5357, 0x15225819},
				{5846, 5356, 0x152242e2},
				{5847, 5355, 0x15222dd6},
				{5848, 5354, 0x152218b4},
				{5849, 5353, 0x152203a9},
				{5850, 5352, 0x1521eeb2},
				{5851, 5351, 0x1521d9bc},
				{5852, 5350, 0x1521c4db},
				{5853, 5349, 0x1521affa},
				{5854, 5348, 0x15219b2e},
				{5855, 5347, 0x15218678},
				{5856, 5346, 0x152171c2},
				{5857, 5345, 0x15215d21},
				{5858, 5344, 0x15214881},
				{5859, 5343, 0x152133f6},
				{5860, 5342, 0x15211f80},
				{5861, 5341, 0x15210b0a},
				{5862, 5340, 0x1520f6aa},
				{5863, 5339, 0x1520e25f},
				{5864, 5338, 0x1520ce15},
				{5865, 5337, 0x1520b9df},
				{5866, 5336, 0x1520a5aa},
				{5867, 5335, 0x1520918a},
				{5868, 5334, 0x15207d6b},
				{5869, 5333, 0x15206960},
				{5870, 5332, 0x15205556},
				{5871, 5331, 0x15204161},
				{5872, 5330, 0x15202d82},
				{5873, 5329, 0x152019a2},
				{5874, 5328, 0x152005d8},
				{5875, 5327, 0x151ff20e},
				{5876, 5326, 0x151fde5a},
				{5877, 5325, 0x151fcabb},
				{5878, 5324, 0x151fb731},
				{5879, 5323, 0x151fa392},
				{5880, 5322, 0x151f9009},
				{5881, 5321, 0x151f7c95},
				{5882, 5320, 0x151f6921},
				{5883, 5319, 0x151f55c2},
				{5884, 5318, 0x151f4279},
				{5885, 5317, 0x151f2f30},
				{5886, 5316, 0x151f1be7},
				{5887, 5315, 0x151f08b3},
				{5888, 5314, 0x151ef595},
				{5889, 5313, 0x151ee277},
				{5890, 5312, 0x151ecf6e},
				{5891, 5311, 0x151ebc7b},
				{5892, 5310, 0x151ea972},
				{5893, 5309, 0x151e96aa},
				{5894, 5308, 0x151e83cc},
				{5895, 5307, 0x151e70ee},
				{5896, 5306, 0x151e5e26},
				{5897, 5305, 0x151e4b73},
				{5898, 5304, 0x151e38c0},
				{5899, 5303, 0x151e2622},
				{5900, 5302, 0x151e1385},
				{5901, 5301, 0x151e00fd},
				{5902, 5300, 0x151dee75},
				{5903, 5299, 0x151ddc03},
				{5904, 5298, 0x151dc9a6},
				{5905, 5297, 0x151db749},
				{5906, 5296, 0x151da4ec},
				{5907, 5295, 0x151d92a4},
				{5908, 5294, 0x151d8072},
				{5909, 5293, 0x151d6e55},
				{5910, 5292, 0x151d5c23},
				{5911, 5291, 0x151d4a07},
				{5912, 5290, 0x151d37ff},
				{5913, 5289, 0x151d25f8},
				{5914, 5288, 0x151d1407},
				{5915, 5287, 0x151d0215},
				{5916, 5286, 0x151cf039},
				{5917, 5285, 0x151cde5d},
				{5918, 5284, 0x151ccc96},
				{5919, 5283, 0x151cbacf},
				{5920, 5282, 0x151ca91e},
				{5921, 5281, 0x151c976c},
				{5922, 5280, 0x151c85d1},
				{5923, 5279, 0x151c744a},
				{5924, 5278, 0x151c62c4},
				{5925, 5277, 0x151c5153},
				{5926, 5276, 0x151c3fe2},
				{5927, 5275, 0x151c2e71},
				{5928, 5274, 0x151c1d16},
				{5929, 5273, 0x151c0bba},
				{5930, 5272, 0x151bfa74},
				{5931, 5271, 0x151be944},
				{5932, 5270, 0x151bd813},
				{5933, 5269, 0x151bc6e3},
				{5934, 5268, 0x151bb5c8},
				{5935, 5267, 0x151ba4ad},
				{5936, 5266, 0x151b93a7},
				{5937, 5265, 0x151b82b7},
				{5938, 5264, 0x151b71b2},
				{5939, 5263, 0x151b60ec},
				{5940, 5262, 0x151b5012},
				{5941, 5261, 0x151b3f37},
				{5942, 5260, 0x151b2e72},
				{5943, 5259, 0x151b1dad},
				{5944, 5258, 0x151b0cfd},
				{5945, 5257, 0x151afc4e},
				{5946, 5256, 0x151aebb3},
				{5947, 5255, 0x151adb19},
				{5948, 5254, 0x151aca94},
				{5949, 5253, 0x151aba10},
				{5950, 5252, 0x151aa9a0},
				{5951, 5251, 0x151a9931},
				{5952, 5250, 0x151a88c2},
				{5953, 5249, 0x151a787d},
				{5954, 5248, 0x151a6823},
				{5955, 5247, 0x151a57f4},
				{5956, 5246, 0x151a47c6},
				{5957, 5245, 0x151a3781},
				{5958, 5244, 0x151a2768},
				{5959, 5243, 0x151a174e},
				{5960, 5242, 0x151a0735},
				{5961, 5241, 0x1519f731},
				{5962, 5240, 0x1519e72d},
				{5963, 5239, 0x1519d73e},
				{5964, 5238, 0x1519c750},
				{5965, 5237, 0x1519b777},
				{5966, 5236, 0x1519a79e},
				{5967, 5235, 0x151997da},
				{5968, 5234, 0x15198817},
				{5969, 5233, 0x15197853},
				{5970, 5232, 0x151968bb},
				{5971, 5231, 0x1519590c},
				{5972, 5230, 0x15194974},
				{5973, 5229, 0x151939db},
				{5974, 5228, 0x15192a58},
				{5975, 5227, 0x15191ad5},
				{5976, 5226, 0x15190b67},
				{5977, 5225, 0x1518fbf9},
				{5978, 5224, 0x1518eca1},
				{5979, 5223, 0x1518dd33},
				{5980, 5222, 0x1518cdf1},
				{5981, 5221, 0x1518beae},
				{5982, 5220, 0x1518af6b},
				{5983, 5219, 0x1518a03e},
				{5984, 5218, 0x15189110},
				{5985, 5217, 0x151881f9},
				{5986, 5216, 0x151872e1},
				{5987, 5215, 0x151863c9},
				{5988, 5214, 0x151854c6},
				{5989, 5213, 0x151845c4},
				{5990, 5212, 0x151836d7},
				{5991, 5211, 0x151827ea},
				{5992, 5210, 0x15181913},
				{5993, 5209, 0x15180a3b},
				{5994, 5208, 0x1517fb64},
				{5995, 5207, 0x1517eca2},
				{5996, 5206, 0x1517dde0},
				{5997, 5205, 0x1517cf33},
				{5998, 5204, 0x1517c087},
				{5999, 5203, 0x1517b1da},
				{6000, 5202, 0x1517a343},
				{6001, 5201, 0x151794c2},
				{6002, 5200, 0x15178640},
				{6003, 5199, 0x151777bf},
				{6004, 5198, 0x1517693d},
				{6005, 5197, 0x15175ad1},
				{6006, 5196, 0x15174c65},
				{6007, 5195, 0x15173e0e},
				{6008, 5194, 0x15172fb7},
				{6009, 5193, 0x15172176},
				{6010, 5192, 0x15171335},
				{6011, 5191, 0x151704f4},
				{6012, 5190, 0x1516f6c8},
				{6013, 5189, 0x1516e89c},
				{6014, 5188, 0x1516da71},
				{6015, 5187, 0x1516cc5a},
				{6016, 5186, 0x1516be5a},
				{6017, 5185, 0x1516b059},
				{6018, 5184, 0x1516a258},
				{6019, 5183, 0x15169457},
				{6020, 5182, 0x1516866c},
				{6021, 5181, 0x15167881},
				{6022, 5180, 0x15166aab},
				{6023, 5179, 0x15165cd5},
				{6024, 5178, 0x15164eff},
				{6025, 5177, 0x1516413f},
				{6026, 5176, 0x1516337e},
				{6027, 5175, 0x151625d3},
				{6028, 5174, 0x15161813},
				{6029, 5173, 0x15160a7d},
				{6030, 5172, 0x1515fcd3},
				{6031, 5171, 0x1515ef53},
				{6032, 5170, 0x1515e1bd},
				{6033, 5169, 0x1515d43d},
				{6034, 5168, 0x1515c6bd},
				{6035, 5167, 0x1515b953},
				{6036, 5166, 0x1515abd3},
				{6037, 5165, 0x15159e7e},
				{6038, 5164, 0x15159113},
				{6039, 5163, 0x151583be},
				{6040, 5162, 0x15157674},
				{6041, 5161, 0x1515693f},
				{6042, 5160, 0x15155c15},
				{6043, 5159, 0x15154ef5},
				{6044, 5158, 0x151541d6},
				{6045, 5157, 0x151534c1},
				{6046, 5156, 0x151527b7},
				{6047, 5155, 0x15151ac3},
				{6048, 5154, 0x15150dc3},
				{6049, 5153, 0x151500cf},
				{6050, 5152, 0x1514f3db},
				{6051, 5151, 0x1514e6fc},
				{6052, 5150, 0x1514da1d},
				{6053, 5149, 0x1514cd48},
				{6054, 5148, 0x1514c074},
				{6055, 5147, 0x1514b3aa},
				{6056, 5146, 0x1514a6ec},
				{6057, 5145, 0x15149a38},
				{6058, 5144, 0x15148d84},
				{6059, 5143, 0x151480e5},
				{6060, 5142, 0x1514743c},
				{6061, 5141, 0x151467a8},
				{6062, 5140, 0x15145b1f},
				{6063, 5139, 0x15144e95},
				{6064, 5138, 0x15144217},
				{6065, 5137, 0x15143599},
				{6066, 5136, 0x15142925},
				{6067, 5135, 0x15141cbc},
				{6068, 5134, 0x15141053},
				{6069, 5133, 0x151403f5},
				{6070, 5132, 0x1513f7a1},
				{6071, 5131, 0x1513eb59},
				{6072, 5130, 0x1513df10},
				{6073, 5129, 0x1513d2d2},
				{6074, 5128, 0x1513c694},
				{6075, 5127, 0x1513ba61},
				{6076, 5126, 0x1513ae38},
				{6077, 5125, 0x1513a21a},
				{6078, 5124, 0x15139607},
				{6079, 5123, 0x151389f4},
				{6080, 5122, 0x15137dec},
				{6081, 5121, 0x151371e3},
				{6082, 5120, 0x151365e6},
				{6083, 5119, 0x151359f3},
				{6084, 5118, 0x15134e00},
				{6085, 5117, 0x15134218},
				{6086, 5116, 0x1513363a},
				{6087, 5115, 0x15132a68},
				{6088, 5114, 0x15131e95},
				{6089, 5113, 0x151312c2},
				{6090, 5112, 0x15130705},
				{6091, 5111, 0x1512fb48},
				{6092, 5110, 0x1512ef8b},
				{6093, 5109, 0x1512e3ee},
				{6094, 5108, 0x1512d846},
				{6095, 5107, 0x1512cca9},
				{6096, 5106, 0x1512c10c},
				{6097, 5105, 0x1512b57a},
				{6098, 5104, 0x1512a9f2},
				{6099, 5103, 0x15129e6b},
				{6100, 5102, 0x151292ee},
				{6101, 5101, 0x1512877c},
				{6102, 5100, 0x15127c0a},
				{6103, 5099, 0x151270a2},
				{6104, 5098, 0x15126546},
				{6105, 5097, 0x151259e9},
				{6106, 5096, 0x15124e97},
				{6107, 5095, 0x15124345},
				{6108, 5094, 0x15123814},
				{6109, 5093, 0x15122ccc},
				{6110, 5092, 0x1512219b},
				{6111, 5091, 0x15121669},
				{6112, 5090, 0x15120b42},
				{6113, 5089, 0x1512001b},
				{6114, 5088, 0x1511f4ff},
				{6115, 5087, 0x1511e9ed},
				{6116, 5086, 0x1511dedc},
				{6117, 5085, 0x1511d3d5},
				{6118, 5084, 0x1511c8ce},
				{6119, 5083, 0x1511bddd},
				{6120, 5082, 0x1511b2e1},
				{6121, 5081, 0x1511a7fb},
				{6122, 5080, 0x15119d14},
				{6123, 5079, 0x15119238},
				{6124, 5078, 0x15118767},
				{6125, 5077, 0x15117c96},
				{6126, 5076, 0x151171c5},
				{6127, 5075, 0x15116709},
				{6128, 5074, 0x15115c43},
				{6129, 5073, 0x15115192},
				{6130, 5072, 0x151146e1},
				{6131, 5071, 0x15113c3b},
				{6132, 5070, 0x15113195},
				{6133, 5069, 0x151126f9},
				{6134, 5068, 0x15111c5e},
				{6135, 5067, 0x151111cd},
				{6136, 5066, 0x15110747},
				{6137, 5065, 0x1510fcc1},
				{6138, 5064, 0x1510f246},
				{6139, 5063, 0x1510e7e0},
				{6140, 5062, 0x1510dd70},
				{6141, 5061, 0x1510d2ff},
				{6142, 5060, 0x1510c8a4},
				{6143, 5059, 0x1510be49},
				{6144, 5058, 0x1510b3ee},
				{6145, 5057, 0x1510a99e},
				{6146, 5056, 0x15109f58},
				{6147, 5055, 0x15109513},
				{6148, 5054, 0x15108ad8},
				{6149, 5053, 0x151080a8},
				{6150, 5052, 0x15107678},
				{6151, 5051, 0x15106c48},
				{6152, 5050, 0x1510622d},
				{6153, 5049, 0x15105808},
				{6154, 5048, 0x15104e03},
				{6155, 5047, 0x151043f3},
				{6156, 5046, 0x151039e3},
				{6157, 5045, 0x15102fe8},
				{6158, 5044, 0x151025e3},
				{6159, 5043, 0x15101bf3},
				{6160, 5042, 0x15101204},
				{6161, 5041, 0x15100814},
				{6162, 5040, 0x150ffe2f},
				{6163, 5039, 0x150ff455},
				{6164, 5038, 0x150fea7b},
				{6165, 5037, 0x150fe0ab},
				{6166, 5036, 0x150fd6dc},
				{6167, 5035, 0x150fcd17},
				{6168, 5034, 0x150fc352},
				{6169, 5033, 0x150fb998},
				{6170, 5032, 0x150faff3},
				{6171, 5031, 0x150fa644},
				{6172, 5030, 0x150f9ca0},
				{6173, 5029, 0x150f92fb},
				{6174, 5028, 0x150f8961},
				{6175, 5027, 0x150f7fc7},
				{6176, 5026, 0x150f7638},
				{6177, 5025, 0x150f6cb4},
				{6178, 5024, 0x150f6330},
				{6179, 5023, 0x150f59ab},
				{6180, 5022, 0x150f503c},
				{6181, 5021, 0x150f46c3},
				{6182, 5020, 0x150f3d5e},
				{6183, 5019, 0x150f33f0},
				{6184, 5018, 0x150f2a96},
				{6185, 5017, 0x150f2147},
				{6186, 5016, 0x150f17ee},
				{6187, 5015, 0x150f0e9f},
				{6188, 5014, 0x150f0551},
				{6189, 5013, 0x150efc17},
				{6190, 5012, 0x150ef2d3},
				{6191, 5011, 0x150ee99a},
				{6192, 5010, 0x150ee06c},
				{6193, 5009, 0x150ed73d},
				{6194, 5008, 0x150ece19},
				{6195, 5007, 0x150ec4f6},
				{6196, 5006, 0x150ebbdc},
				{6197, 5005, 0x150eb2c3},
				{6198, 5004, 0x150ea9b5},
				{6199, 5003, 0x150ea0b2},
				{6200, 5002, 0x150e97ae},
				{6201, 5001, 0x150e8eb5},
				{6202, 5000, 0x150e85bc},
				{6203, 4999, 0x150e7cc3},
				{6204, 4998, 0x150e73d5},
				{6205, 4997, 0x150e6af2},
				{6206, 4996, 0x150e620f},
				{6207, 4995, 0x150e592b},
				{6208, 4994, 0x150e5053},
				{6209, 4993, 0x150e4785},
				{6210, 4992, 0x150e3eb7},
				{6211, 4991, 0x150e35f4},
				{6212, 4990, 0x150e2d30},
				{6213, 4989, 0x150e246d},
				{6214, 4988, 0x150e1bb5},
				{6215, 4987, 0x150e1307},
				{6216, 4986, 0x150e0a64},
				{6217, 4985, 0x150e01c1},
				{6218, 4984, 0x150df91e},
				{6219, 4983, 0x150df086},
				{6220, 4982, 0x150de7ee},
				{6221, 4981, 0x150ddf55},
				{6222, 4980, 0x150dd6c8},
				{6223, 4979, 0x150dce45},
				{6224, 4978, 0x150dc5c2},
				{6225, 4977, 0x150dbd4a},
				{6226, 4976, 0x150db4d2},
				{6227, 4975, 0x150dac65},
				{6228, 4974, 0x150da3f8},
				{6229, 4973, 0x150d9b8a},
				{6230, 4972, 0x150d9328},
				{6231, 4971, 0x150d8adb},
				{6232, 4970, 0x150d8283},
				{6233, 4969, 0x150d7a2b},
				{6234, 4968, 0x150d71de},
				{6235, 4967, 0x150d699b},
				{6236, 4966, 0x150d6159},
				{6237, 4965, 0x150d5917},
				{6238, 4964, 0x150d50df},
				{6239, 4963, 0x150d48a7},
				{6240, 4962, 0x150d407a},
				{6241, 4961, 0x150d3858},
				{6242, 4960, 0x150d302b},
				{6243, 4959, 0x150d2814},
				{6244, 4958, 0x150d1ff2},
				{6245, 4957, 0x150d17e5},
				{6246, 4956, 0x150d0fd8},
				{6247, 4955, 0x150d07cb},
				{6248, 4954, 0x150cffca},
				{6249, 4953, 0x150cf7c8},
				{6250, 4952, 0x150cefc6},
				{6251, 4951, 0x150ce7ce},
				{6252, 4950, 0x150cdfe2},
				{6253, 4949, 0x150cd7f5},
				{6254, 4948, 0x150cd009},
				{6255, 4947, 0x150cc827},
				{6256, 4946, 0x150cc045},
				{6257, 4945, 0x150cb86e},
				{6258, 4944, 0x150cb097},
				{6259, 4943, 0x150ca8c0},
				{6260, 4942, 0x150ca0fe},
				{6261, 4941, 0x150c9932},
				{6262, 4940, 0x150c917b},
				{6263, 4939, 0x150c89ba},
				{6264, 4938, 0x150c8203},
				{6265, 4937, 0x150c7a4c},
				{6266, 4936, 0x150c72a0},
				{6267, 4935, 0x150c6af4},
				{6268, 4934, 0x150c6352},
				{6269, 4933, 0x150c5bb1},
				{6270, 4932, 0x150c5410},
				{6271, 4931, 0x150c4c79},
				{6272, 4930, 0x150c44e2},
				{6273, 4929, 0x150c3d56},
				{6274, 4928, 0x150c35ca},
				{6275, 4927, 0x150c2e49},
				{6276, 4926, 0x150c26c8},
				{6277, 4925, 0x150c1f52},
				{6278, 4924, 0x150c17db},
				{6279, 4923, 0x150c1065},
				{6280, 4922, 0x150c08f9},
				{6281, 4921, 0x150c018d},
				{6282, 4920, 0x150bfa2c},
				{6283, 4919, 0x150bf2cb},
				{6284, 4918, 0x150beb6a},
				{6285, 4917, 0x150be414},
				{6286, 4916, 0x150bdcbe},
				{6287, 4915, 0x150bd572},
				{6288, 4914, 0x150bce27},
				{6289, 4913, 0x150bc6db},
				{6290, 4912, 0x150bbf9a},
				{6291, 4911, 0x150bb859},
				{6292, 4910, 0x150bb12e},
				{6293, 4909, 0x150ba9f8},
				{6294, 4908, 0x150ba2cd},
				{6295, 4907, 0x150b9b97},
				{6296, 4906, 0x150b9476},
				{6297, 4905, 0x150b8d4b},
				{6298, 4904, 0x150b862a},
				{6299, 4903, 0x150b7f14},
				{6300, 4902, 0x150b77fe},
				{6301, 4901, 0x150b70e9},
				{6302, 4900, 0x150b69dd},
				{6303, 4899, 0x150b62d2},
				{6304, 4898, 0x150b5bc7},
				{6305, 4897, 0x150b54c7},
				{6306, 4896, 0x150b4dc6},
				{6307, 4895, 0x150b46d1},
				{6308, 4894, 0x150b3fe6},
				{6309, 4893, 0x150b38f0},
				{6310, 4892, 0x150b3205},
				{6311, 4891, 0x150b2b1a},
				{6312, 4890, 0x150b243a},
				{6313, 4889, 0x150b1d5a},
				{6314, 4888, 0x150b167a},
				{6315, 4887, 0x150b0f9a},
				{6316, 4886, 0x150b08cf},
				{6317, 4885, 0x150b01f9},
				{6318, 4884, 0x150afb2f},
				{6319, 4883, 0x150af464},
				{6320, 4882, 0x150aed99},
				{6321, 4881, 0x150ae6d9},
				{6322, 4880, 0x150ae024},
				{6323, 4879, 0x150ad96f},
				{6324, 4878, 0x150ad2b9},
				{6325, 4877, 0x150acc04},
				{6326, 4876, 0x150ac55a},
				{6327, 4875, 0x150abeaf},
				{6328, 4874, 0x150ab80f},
				{6329, 4873, 0x150ab17a},
				{6330, 4872, 0x150aaae5},
				{6331, 4871, 0x150aa455},
				{6332, 4870, 0x150a9dcb},
				{6333, 4869, 0x150a9741},
				{6334, 4868, 0x150a90bc},
				{6335, 4867, 0x150a8a3c},
				{6336, 4866, 0x150a83c2},
				{6337, 4865, 0x150a7d48},
				{6338, 4864, 0x150a76d8},
				{6339, 4863, 0x150a7069},
				{6340, 4862, 0x150a69f9},
				{6341, 4861, 0x150a6394},
				{6342, 4860, 0x150a5d2a},
				{6343, 4859, 0x150a56cb},
				{6344, 4858, 0x150a506b},
				{6345, 4857, 0x150a4a11},
				{6346, 4856, 0x150a43bd},
				{6347, 4855, 0x150a3d68},
				{6348, 4854, 0x150a371e},
				{6349, 4853, 0x150a30cf},
				{6350, 4852, 0x150a2a8a},
				{6351, 4851, 0x150a2446},
				{6352, 4850, 0x150a1e06},
				{6353, 4849, 0x150a17cd},
				{6354, 4848, 0x150a1198},
				{6355, 4847, 0x150a0b64},
				{6356, 4846, 0x150a0535},
				{6357, 4845, 0x1509ff0b},
				{6358, 4844, 0x1509f8e1},
				{6359, 4843, 0x1509f2b7},
				{6360, 4842, 0x1509ec98},
				{6361, 4841, 0x1509e679},
				{6362, 4840, 0x1509e060},
				{6363, 4839, 0x1509da4b},
				{6364, 4838, 0x1509d437},
				{6365, 4837, 0x1509ce28},
				{6366, 4836, 0x1509c819},
				{6367, 4835, 0x1509c215},
				{6368, 4834, 0x1509bc11},
				{6369, 4833, 0x1509b618},
				{6370, 4832, 0x1509b019},
				{6371, 4831, 0x1509aa1f},
				{6372, 4830, 0x1509a42b},
				{6373, 4829, 0x15099e37},
				{6374, 4828, 0x15099849},
				{6375, 4827, 0x1509925f},
				{6376, 4826, 0x15098c76},
				{6377, 4825, 0x15098692},
				{6378, 4824, 0x150980b3},
				{6379, 4823, 0x15097ad5},
				{6380, 4822, 0x150974fb},
				{6381, 4821, 0x15096f28},
				{6382, 4820, 0x15096954},
				{6383, 4819, 0x15096385},
				{6384, 4818, 0x15095dc1},
				{6385, 4817, 0x150957f8},
				{6386, 4816, 0x15095234},
				{6387, 4815, 0x15094c76},
				{6388, 4814, 0x150946b8},
				{6389, 4813, 0x150940ff},
				{6390, 4812, 0x15093b4b},
				{6391, 4811, 0x15093597},
				{6392, 4810, 0x15092fe9},
				{6393, 4809, 0x15092a40},
				{6394, 4808, 0x15092497},
				{6395, 4807, 0x15091ef3},
				{6396, 4806, 0x15091950},
				{6397, 4805, 0x150913b2},
				{6398, 4804, 0x15090e19},
				{6399, 4803, 0x15090880},
				{6400, 4802, 0x150902f7},
				{6401, 4801, 0x1508fd64},
				{6402, 4800, 0x1508f7db},
				{6403, 4799, 0x1508f252},
				{6404, 4798, 0x1508ecca},
				{6405, 4797, 0x1508e746},
				{6406, 4796, 0x1508e1c8},
				{6407, 4795, 0x1508dc50},
				{6408, 4794, 0x1508d6d7},
				{6409, 4793, 0x1508d15e},
				{6410, 4792, 0x1508cbf0},
				{6411, 4791, 0x1508c682},
				{6412, 4790, 0x1508c114},
				{6413, 4789, 0x1508bbb1},
				{6414, 4788, 0x1508b64e},
				{6415, 4787, 0x1508b0f0},
				{6416, 4786, 0x1508ab92},
				{6417, 4785, 0x1508a63a},
				{6418, 4784, 0x1508a0e2},
				{6419, 4783, 0x15089b8f},
				{6420, 4782, 0x15089641},
				{6421, 4781, 0x150890f3},
				{6422, 4780, 0x15088bab},
				{6423, 4779, 0x15088662},
				{6424, 4778, 0x1508811f},
				{6425, 4777, 0x15087be2},
				{6426, 4776, 0x150876a4},
				{6427, 4775, 0x15087167},
				{6428, 4774, 0x15086c34},
				{6429, 4773, 0x15086701},
				{6430, 4772, 0x150861d9},
				{6431, 4771, 0x15085cab},
				{6432, 4770, 0x15085783},
				{6433, 4769, 0x1508525b},
				{6434, 4768, 0x15084d38},
				{6435, 4767, 0x15084815},
				{6436, 4766, 0x150842fd},
				{6437, 4765, 0x15083de0},
				{6438, 4764, 0x150838cd},
				{6439, 4763, 0x150833ba},
				{6440, 4762, 0x15082ea8},
				{6441, 4761, 0x1508299a},
				{6442, 4760, 0x15082493},
				{6443, 4759, 0x15081f8b},
				{6444, 4758, 0x15081a88},
				{6445, 4757, 0x1508158b},
				{6446, 4756, 0x15081093},
				{6447, 4755, 0x15080b96},
				{6448, 4754, 0x1508069e},
				{6449, 4753, 0x150801ab},
				{6450, 4752, 0x1507fcb9},
				{6451, 4751, 0x1507f7cc},
				{6452, 4750, 0x1507f2e4},
				{6453, 4749, 0x1507edfc},
				{6454, 4748, 0x1507e915},
				{6455, 4747, 0x1507e438},
				{6456, 4746, 0x1507df55},
				{6457, 4745, 0x1507da7d},
				{6458, 4744, 0x1507d5a0},
				{6459, 4743, 0x1507d0ce},
				{6460, 4742, 0x1507cbfc},
				{6461, 4741, 0x1507c72f},
				{6462, 4740, 0x1507c267},
				{6463, 4739, 0x1507bd9a},
				{6464, 4738, 0x1507b8d3},
				{6465, 4737, 0x1507b411},
				{6466, 4736, 0x1507af4f},
				{6467, 4735, 0x1507aa92},
				{6468, 4734, 0x1507a5da},
				{6469, 4733, 0x1507a123},
				{6470, 4732, 0x15079c6b},
				{6471, 4731, 0x150797b9},
				{6472, 4730, 0x1507930d},
				{6473, 4729, 0x15078e60},
				{6474, 4728, 0x150789b9},
				{6475, 4727, 0x15078511},
				{6476, 4726, 0x15078075},
				{6477, 4725, 0x15077bd3},
				{6478, 4724, 0x15077736},
				{6479, 4723, 0x15077299},
				{6480, 4722, 0x15076e02},
				{6481, 4721, 0x15076970},
				{6482, 4720, 0x150764de},
				{6483, 4719, 0x1507604c},
				{6484, 4718, 0x15075bc0},
				{6485, 4717, 0x15075739},
				{6486, 4716, 0x150752b2},
				{6487, 4715, 0x15074e30},
				{6488, 4714, 0x150749ae},
				{6489, 4713, 0x1507452c},
				{6490, 4712, 0x150740b5},
				{6491, 4711, 0x15073c39},
				{6492, 4710, 0x150737cc},
				{6493, 4709, 0x15073355},
				{6494, 4708, 0x15072ee9},
				{6495, 4707, 0x15072a77},
				{6496, 4706, 0x15072610},
				{6497, 4705, 0x150721a9},
				{6498, 4704, 0x15071d42},
				{6499, 4703, 0x150718e1},
				{6500, 4702, 0x1507147f},
				{6501, 4701, 0x15071023},
				{6502, 4700, 0x15070bc7},
				{6503, 4699, 0x15070770},
				{6504, 4698, 0x1507031f},
				{6505, 4697, 0x1506fec8},
				{6506, 4696, 0x1506fa7c},
				{6507, 4695, 0x1506f635},
				{6508, 4694, 0x1506f1e9},
				{6509, 4693, 0x1506eda2},
				{6510, 4692, 0x1506e95b},
				{6511, 4691, 0x1506e51a},
				{6512, 4690, 0x1506e0d8},
				{6513, 4689, 0x1506dc9c},
				{6514, 4688, 0x1506d866},
				{6515, 4687, 0x1506d42a},
				{6516, 4686, 0x1506cff8},
				{6517, 4685, 0x1506cbc7},
				{6518, 4684, 0x1506c796},
				{6519, 4683, 0x1506c36a},
				{6520, 4682, 0x1506bf3e},
				{6521, 4681, 0x1506bb17},
				{6522, 4680, 0x1506b6f6},
				{6523, 4679, 0x1506b2d0},
				{6524, 4678, 0x1506aeb4},
				{6525, 4677, 0x1506aa93},
				{6526, 4676, 0x1506a677},
				{6527, 4675, 0x1506a260},
				{6528, 4674, 0x15069e4a},
				{6529, 4673, 0x15069a39},
				{6530, 4672, 0x15069628},
				{6531, 4671, 0x15069217},
				{6532, 4670, 0x15068e0b},
				{6533, 4669, 0x15068a05},
				{6534, 4668, 0x150685f9},
				{6535, 4667, 0x150681f8},
				{6536, 4666, 0x15067df7},
				{6537, 4665, 0x150679f6},
				{6538, 4664, 0x150675fa},
				{6539, 4663, 0x15067204},
				{6540, 4662, 0x15066e08},
				{6541, 4661, 0x15066a12},
				{6542, 4660, 0x15066621},
				{6543, 4659, 0x15066230},
				{6544, 4658, 0x15065e3f},
				{6545, 4657, 0x15065a54},
				{6546, 4656, 0x1506566e},
				{6547, 4655, 0x15065287},
				{6548, 4654, 0x15064ea1},
				{6549, 4653, 0x15064abb},
				{6550, 4652, 0x150646e0},
				{6551, 4651, 0x150642ff},
				{6552, 4650, 0x15063f24},
				{6553, 4649, 0x15063b53},
				{6554, 4648, 0x15063777},
				{6555, 4647, 0x150633a7},
				{6556, 4646, 0x15062fd6},
				{6557, 4645, 0x15062c05},
				{6558, 4644, 0x1506283a},
				{6559, 4643, 0x1506246f},
				{6560, 4642, 0x150620a3},
				{6561, 4641, 0x15061cdd},
				{6562, 4640, 0x15061917},
				{6563, 4639, 0x15061557},
				{6564, 4638, 0x15061196},
				{6565, 4637, 0x15060ddb},
				{6566, 4636, 0x15060a20},
				{6567, 4635, 0x1506066a},
				{6568, 4634, 0x150602b4},
				{6569, 4633, 0x1505ff04},
				{6570, 4632, 0x1505fb53},
				{6571, 4631, 0x1505f7a3},
				{6572, 4630, 0x1505f3f2},
				{6573, 4629, 0x1505f047},
				{6574, 4628, 0x1505eca1},
				{6575, 4627, 0x1505e8fb},
				{6576, 4626, 0x1505e556},
				{6577, 4625, 0x1505e1b5},
				{6578, 4624, 0x1505de15},
				{6579, 4623, 0x1505da74},
				{6580, 4622, 0x1505d6d9},
				{6581, 4621, 0x1505d33e},
				{6582, 4620, 0x1505cfa9},
				{6583, 4619, 0x1505cc13},
				{6584, 4618, 0x1505c883},
				{6585, 4617, 0x1505c4f2},
				{6586, 4616, 0x1505c162},
				{6587, 4615, 0x1505bdd7},
				{6588, 4614, 0x1505ba4c},
				{6589, 4613, 0x1505b6c7},
				{6590, 4612, 0x1505b33c},
				{6591, 4611, 0x1505afbc},
				{6592, 4610, 0x1505ac36},
				{6593, 4609, 0x1505a8b6},
				{6594, 4608, 0x1505a53b},
				{6595, 4607, 0x1505a1c0},
				{6596, 4606, 0x15059e45},
				{6597, 4605, 0x15059acb},
				{6598, 4604, 0x15059755},
				{6599, 4603, 0x150593e5},
				{6600, 4602, 0x15059075},
				{6601, 4601, 0x15058d05},
				{6602, 4600, 0x1505899a},
				{6603, 4599, 0x1505862f},
				{6604, 4598, 0x150582c5},
				{6605, 4597, 0x15057f5a},
				{6606, 4596, 0x15057bf4},
				{6607, 4595, 0x15057894},
				{6608, 4594, 0x1505752f},
				{6609, 4593, 0x150571cf},
				{6610, 4592, 0x15056e74},
				{6611, 4591, 0x15056b1a},
				{6612, 4590, 0x150567bf},
				{6613, 4589, 0x15056465},
				{6614, 4588, 0x15056115},
				{6615, 4587, 0x15055dc2},
				{6616, 4586, 0x15055a75},
				{6617, 4585, 0x1505572a},
				{6618, 4584, 0x150553e0},
				{6619, 4583, 0x1505509b},
				{6620, 4582, 0x15054d55},
				{6621, 4581, 0x15054a10},
				{6622, 4580, 0x150546d0},
				{6623, 4579, 0x15054391},
				{6624, 4578, 0x15054054},
				{6625, 4577, 0x15053d19},
				{6626, 4576, 0x150539df},
				{6627, 4575, 0x150536a7},
				{6628, 4574, 0x15053372},
				{6629, 4573, 0x1505303d},
				{6630, 4572, 0x15052d10},
				{6631, 4571, 0x150529e0},
				{6632, 4570, 0x150526b0},
				{6633, 4569, 0x15052386},
				{6634, 4568, 0x1505205c},
				{6635, 4567, 0x15051d34},
				{6636, 4566, 0x15051a0c},
				{6637, 4565, 0x150516e7},
				{6638, 4564, 0x150513c5},
				{6639, 4563, 0x150510a6},
				{6640, 4562, 0x15050d86},
				{6641, 4561, 0x15050a69},
				{6642, 4560, 0x1505074f},
				{6643, 4559, 0x15050435},
				{6644, 4558, 0x1505011d},
				{6645, 4557, 0x1504fe0b},
				{6646, 4556, 0x1504faf6},
				{6647, 4555, 0x1504f7e6},
				{6648, 4554, 0x1504f4d4},
				{6649, 4553, 0x1504f1c7},
				{6650, 4552, 0x1504eebb},
				{6651, 4551, 0x1504ebb1},
				{6652, 4550, 0x1504e8a9},
				{6653, 4549, 0x1504e5a2},
				{6654, 4548, 0x1504e29d},
				{6655, 4547, 0x1504df98},
				{6656, 4546, 0x1504dc99},
				{6657, 4545, 0x1504d999},
				{6658, 4544, 0x1504d69a},
				{6659, 4543, 0x1504d3a0},
				{6660, 4542, 0x1504d0a8},
				{6661, 4541, 0x1504cdb1},
				{6662, 4540, 0x1504caba},
				{6663, 4539, 0x1504c7c5},
				{6664, 4538, 0x1504c4d3},
				{6665, 4537, 0x1504c1e1},
				{6666, 4536, 0x1504bef2},
				{6667, 4535, 0x1504bc05},
				{6668, 4534, 0x1504b919},
				{6669, 4533, 0x1504b62f},
				{6670, 4532, 0x1504b347},
				{6671, 4531, 0x1504b060},
				{6672, 4530, 0x1504ad7c},
				{6673, 4529, 0x1504aa9a},
				{6674, 4528, 0x1504a7b8},
				{6675, 4527, 0x1504a4d9},
				{6676, 4526, 0x1504a1ff},
				{6677, 4525, 0x15049f22},
				{6678, 4524, 0x15049c48},
				{6679, 4523, 0x1504996f},
				{6680, 4522, 0x1504969a},
				{6681, 4521, 0x150493c3},
				{6682, 4520, 0x150490f1},
				{6683, 4519, 0x15048e1f},
				{6684, 4518, 0x15048b50},
				{6685, 4517, 0x15048881},
				{6686, 4516, 0x150485b5},
				{6687, 4515, 0x150482eb},
				{6688, 4514, 0x15048021},
				{6689, 4513, 0x15047d5a},
				{6690, 4512, 0x15047a96},
				{6691, 4511, 0x150477d7},
				{6692, 4510, 0x15047512},
				{6693, 4509, 0x15047253},
				{6694, 4508, 0x15046f94},
				{6695, 4507, 0x15046cd8},
				{6696, 4506, 0x15046a1c},
				{6697, 4505, 0x15046762},
				{6698, 4504, 0x150464ab},
				{6699, 4503, 0x150461f4},
				{6700, 4502, 0x15045f40},
				{6701, 4501, 0x15045c8e},
				{6702, 4500, 0x150459dd},
				{6703, 4499, 0x1504572e},
				{6704, 4498, 0x1504547f},
				{6705, 4497, 0x150451d3},
				{6706, 4496, 0x15044f2c},
				{6707, 4495, 0x15044c82},
				{6708, 4494, 0x150449db},
				{6709, 4493, 0x15044735},
				{6710, 4492, 0x15044490},
				{6711, 4491, 0x150441ef},
				{6712, 4490, 0x15043f4d},
				{6713, 4489, 0x15043caf},
				{6714, 4488, 0x15043a10},
				{6715, 4487, 0x15043776},
				{6716, 4486, 0x150434da},
				{6717, 4485, 0x15043244},
				{6718, 4484, 0x15042faa},
				{6719, 4483, 0x15042d16},
				{6720, 4482, 0x15042a82},
				{6721, 4481, 0x150427f1},
				{6722, 4480, 0x15042562},
				{6723, 4479, 0x150422d3},
				{6724, 4478, 0x15042044},
				{6725, 4477, 0x15041db8},
				{6726, 4476, 0x15041b2f},
				{6727, 4475, 0x150418a6},
				{6728, 4474, 0x1504161f},
				{6729, 4473, 0x15041398},
				{6730, 4472, 0x15041114},
				{6731, 4471, 0x15040e93},
				{6732, 4470, 0x15040c12},
				{6733, 4469, 0x15040993},
				{6734, 4468, 0x15040715},
				{6735, 4467, 0x15040499},
				{6736, 4466, 0x1504021d},
				{6737, 4465, 0x1503ffa6},
				{6738, 4464, 0x1503fd30},
				{6739, 4463, 0x1503fab9},
				{6740, 4462, 0x1503f845},
				{6741, 4461, 0x1503f5d1},
				{6742, 4460, 0x1503f360},
				{6743, 4459, 0x1503f0ef},
				{6744, 4458, 0x1503ee80},
				{6745, 4457, 0x1503ec15},
				{6746, 4456, 0x1503e9a9},
				{6747, 4455, 0x1503e740},
				{6748, 4454, 0x1503e4d7},
				{6749, 4453, 0x1503e270},
				{6750, 4452, 0x1503e00a},
				{6751, 4451, 0x1503dda6},
				{6752, 4450, 0x1503db45},
				{6753, 4449, 0x1503d8e6},
				{6754, 4448, 0x1503d685},
				{6755, 4447, 0x1503d427},
				{6756, 4446, 0x1503d1cb},
				{6757, 4445, 0x1503cf6f},
				{6758, 4444, 0x1503cd16},
				{6759, 4443, 0x1503cabd},
				{6760, 4442, 0x1503c867},
				{6761, 4441, 0x1503c613},
				{6762, 4440, 0x1503c3c0},
				{6763, 4439, 0x1503c16c},
				{6764, 4438, 0x1503bf1b},
				{6765, 4437, 0x1503bccd},
				{6766, 4436, 0x1503ba7e},
				{6767, 4435, 0x1503b833},
				{6768, 4434, 0x1503b5ea},
				{6769, 4433, 0x1503b39e},
				{6770, 4432, 0x1503b158},
				{6771, 4431, 0x1503af0f},
				{6772, 4430, 0x1503accb},
				{6773, 4429, 0x1503aa88},
				{6774, 4428, 0x1503a844},
				{6775, 4427, 0x1503a603},
				{6776, 4426, 0x1503a3c2},
				{6777, 4425, 0x1503a184},
				{6778, 4424, 0x15039f49},
				{6779, 4423, 0x15039d0d},
				{6780, 4422, 0x15039ad2},
				{6781, 4421, 0x15039899},
				{6782, 4420, 0x15039663},
				{6783, 4419, 0x1503942f},
				{6784, 4418, 0x150391f9},
				{6785, 4417, 0x15038fc5},
				{6786, 4416, 0x15038d95},
				{6787, 4415, 0x15038b64},
				{6788, 4414, 0x15038933},
				{6789, 4413, 0x15038708},
				{6790, 4412, 0x150384da},
				{6791, 4411, 0x150382ae},
				{6792, 4410, 0x15038085},
				{6793, 4409, 0x15037e5d},
				{6794, 4408, 0x15037c37},
				{6795, 4407, 0x15037a10},
				{6796, 4406, 0x150377ea},
				{6797, 4405, 0x150375ca},
				{6798, 4404, 0x150373a6},
				{6799, 4403, 0x15037188},
				{6800, 4402, 0x15036f6a},
				{6801, 4401, 0x15036d4c},
				{6802, 4400, 0x15036b2e},
				{6803, 4399, 0x15036913},
				{6804, 4398, 0x150366fa},
				{6805, 4397, 0x150364e2},
				{6806, 4396, 0x150362cc},
				{6807, 4395, 0x150360b6},
				{6808, 4394, 0x15035ea0},
				{6809, 4393, 0x15035c8d},
				{6810, 4392, 0x15035a79},
				{6811, 4391, 0x15035869},
				{6812, 4390, 0x1503565b},
				{6813, 4389, 0x1503544d},
				{6814, 4388, 0x15035242},
				{6815, 4387, 0x15035036},
				{6816, 4386, 0x15034e2b},
				{6817, 4385, 0x15034c23},
				{6818, 4384, 0x15034a1a},
				{6819, 4383, 0x15034814},
				{6820, 4382, 0x1503460e},
				{6821, 4381, 0x1503440b},
				{6822, 4380, 0x15034208},
				{6823, 4379, 0x15034005},
				{6824, 4378, 0x15033e04},
				{6825, 4377, 0x15033c06},
				{6826, 4376, 0x15033a09},
				{6827, 4375, 0x1503380b},
				{6828, 4374, 0x15033610},
				{6829, 4373, 0x15033417},
				{6830, 4372, 0x1503321f},
				{6831, 4371, 0x15033026},
				{6832, 4370, 0x15032e31},
				{6833, 4369, 0x15032c3b},
				{6834, 4368, 0x15032a48},
				{6835, 4367, 0x15032855},
				{6836, 4366, 0x15032662},
				{6837, 4365, 0x15032471},
				{6838, 4364, 0x15032281},
				{6839, 4363, 0x15032093},
				{6840, 4362, 0x15031ea5},
				{6841, 4361, 0x15031cba},
				{6842, 4360, 0x15031acf},
				{6843, 4359, 0x150318e7},
				{6844, 4358, 0x150316ff},
				{6845, 4357, 0x15031519},
				{6846, 4356, 0x15031333},
				{6847, 4355, 0x1503114e},
				{6848, 4354, 0x15030f6b},
				{6849, 4353, 0x15030d88},
				{6850, 4352, 0x15030ba7},
				{6851, 4351, 0x150309c7},
				{6852, 4350, 0x150307e7},
				{6853, 4349, 0x15030609},
				{6854, 4348, 0x1503042c},
				{6855, 4347, 0x15030251},
				{6856, 4346, 0x15030076},
				{6857, 4345, 0x1502fe9d},
				{6858, 4344, 0x1502fcc5},
				{6859, 4343, 0x1502faed},
				{6860, 4342, 0x1502f91a},
				{6861, 4341, 0x1502f745},
				{6862, 4340, 0x1502f56f},
				{6863, 4339, 0x1502f39f},
				{6864, 4338, 0x1502f1cc},
				{6865, 4337, 0x1502effc},
				{6866, 4336, 0x1502ee2b},
				{6867, 4335, 0x1502ec5e},
				{6868, 4334, 0x1502ea90},
				{6869, 4333, 0x1502e8c3},
				{6870, 4332, 0x1502e6f8},
				{6871, 4331, 0x1502e52d},
				{6872, 4330, 0x1502e365},
				{6873, 4329, 0x1502e19d},
				{6874, 4328, 0x1502dfd7},
				{6875, 4327, 0x1502de12},
				{6876, 4326, 0x1502dc4f},
				{6877, 4325, 0x1502da8a},
				{6878, 4324, 0x1502d8c7},
				{6879, 4323, 0x1502d707},
				{6880, 4322, 0x1502d547},
				{6881, 4321, 0x1502d387},
				{6882, 4320, 0x1502d1c7},
				{6883, 4319, 0x1502d009},
				{6884, 4318, 0x1502ce4e},
				{6885, 4317, 0x1502cc91},
				{6886, 4316, 0x1502cad9},
				{6887, 4315, 0x1502c91e},
				{6888, 4314, 0x1502c766},
				{6889, 4313, 0x1502c5ae},
				{6890, 4312, 0x1502c3f9},
				{6891, 4311, 0x1502c246},
				{6892, 4310, 0x1502c091},
				{6893, 4309, 0x1502bede},
				{6894, 4308, 0x1502bd2b},
				{6895, 4307, 0x1502bb79},
				{6896, 4306, 0x1502b9c9},
				{6897, 4305, 0x1502b81b},
				{6898, 4304, 0x1502b66b},
				{6899, 4303, 0x1502b4be},
				{6900, 4302, 0x1502b313},
				{6901, 4301, 0x1502b166},
				{6902, 4300, 0x1502afbb},
				{6903, 4299, 0x1502ae15},
				{6904, 4298, 0x1502ac6e},
				{6905, 4297, 0x1502aac9},
				{6906, 4296, 0x1502a928},
				{6907, 4295, 0x1502a784},
				{6908, 4294, 0x1502a5e3},
				{6909, 4293, 0x1502a441},
				{6910, 4292, 0x1502a2a1},
				{6911, 4291, 0x1502a102},
				{6912, 4290, 0x15029f63},
				{6913, 4289, 0x15029dc6},
				{6914, 4288, 0x15029c2a},
				{6915, 4287, 0x15029a8e},
				{6916, 4286, 0x150298f5},
				{6917, 4285, 0x1502975a},
				{6918, 4284, 0x150295c2},
				{6919, 4283, 0x1502942b},
				{6920, 4282, 0x15029294},
				{6921, 4281, 0x15029100},
				{6922, 4280, 0x15028f6c},
				{6923, 4279, 0x15028dd9},
				{6924, 4278, 0x15028c46},
				{6925, 4277, 0x15028ab4},
				{6926, 4276, 0x15028922},
				{6927, 4275, 0x15028793},
				{6928, 4274, 0x15028604},
				{6929, 4273, 0x15028476},
				{6930, 4272, 0x150282e9},
				{6931, 4271, 0x1502815d},
				{6932, 4270, 0x15027fd1},
				{6933, 4269, 0x15027e47},
				{6934, 4268, 0x15027cbe},
				{6935, 4267, 0x15027b36},
				{6936, 4266, 0x150279ae},
				{6937, 4265, 0x15027829},
				{6938, 4264, 0x150276a3},
				{6939, 4263, 0x1502751f},
				{6940, 4262, 0x1502739a},
				{6941, 4261, 0x15027217},
				{6942, 4260, 0x15027095},
				{6943, 4259, 0x15026f14},
				{6944, 4258, 0x15026d94},
				{6945, 4257, 0x15026c15},
				{6946, 4256, 0x15026a96},
				{6947, 4255, 0x15026919},
				{6948, 4254, 0x1502679c},
				{6949, 4253, 0x15026621},
				{6950, 4252, 0x150264a6},
				{6951, 4251, 0x1502632c},
				{6952, 4250, 0x150261b4},
				{6953, 4249, 0x1502603d},
				{6954, 4248, 0x15025ec5},
				{6955, 4247, 0x15025d4f},
				{6956, 4246, 0x15025bda},
				{6957, 4245, 0x15025a65},
				{6958, 4244, 0x150258f1},
				{6959, 4243, 0x1502577f},
				{6960, 4242, 0x1502560c},
				{6961, 4241, 0x1502549b},
				{6962, 4240, 0x1502532c},
				{6963, 4239, 0x150251bc},
				{6964, 4238, 0x1502504e},
				{6965, 4237, 0x15024ee1},
				{6966, 4236, 0x15024d74},
				{6967, 4235, 0x15024c0a},
				{6968, 4234, 0x15024a9e},
				{6969, 4233, 0x15024935},
				{6970, 4232, 0x150247cb},
				{6971, 4231, 0x15024664},
				{6972, 4230, 0x150244fc},
				{6973, 4229, 0x15024396},
				{6974, 4228, 0x15024230},
				{6975, 4227, 0x150240cb},
				{6976, 4226, 0x15023f67},
				{6977, 4225, 0x15023e04},
				{6978, 4224, 0x15023ca2},
				{6979, 4223, 0x15023b3f},
				{6980, 4222, 0x150239e0},
				{6981, 4221, 0x15023880},
				{6982, 4220, 0x15023721},
				{6983, 4219, 0x150235c4},
				{6984, 4218, 0x15023467},
				{6985, 4217, 0x1502330a},
				{6986, 4216, 0x150231af},
				{6987, 4215, 0x15023055},
				{6988, 4214, 0x15022efb},
				{6989, 4213, 0x15022da2},
				{6990, 4212, 0x15022c49},
				{6991, 4211, 0x15022af2},
				{6992, 4210, 0x1502299c},
				{6993, 4209, 0x15022845},
				{6994, 4208, 0x150226f1},
				{6995, 4207, 0x1502259d},
				{6996, 4206, 0x1502244a},
				{6997, 4205, 0x150222f8},
				{6998, 4204, 0x150221a7},
				{6999, 4203, 0x15022056},
				{7000, 4202, 0x15021f05},
				{7001, 4201, 0x15021db7},
				{7002, 4200, 0x15021c68},
				{7003, 4199, 0x15021b1b},
				{7004, 4198, 0x150219cf},
				{7005, 4197, 0x15021882},
				{7006, 4196, 0x15021736},
				{7007, 4195, 0x150215ec},
				{7008, 4194, 0x150214a4},
				{7009, 4193, 0x1502135b},
				{7010, 4192, 0x15021212},
				{7011, 4191, 0x150210cb},
				{7012, 4190, 0x15020f85},
				{7013, 4189, 0x15020e40},
				{7014, 4188, 0x15020cfc},
				{7015, 4187, 0x15020bb8},
				{7016, 4186, 0x15020a75},
				{7017, 4185, 0x15020933},
				{7018, 4184, 0x150207f1},
				{7019, 4183, 0x150206b0},
				{7020, 4182, 0x15020570},
				{7021, 4181, 0x15020430},
				{7022, 4180, 0x150202f2},
				{7023, 4179, 0x150201b5},
				{7024, 4178, 0x15020077},
				{7025, 4177, 0x1501ff3b},
				{7026, 4176, 0x1501fe00},
				{7027, 4175, 0x1501fcc5},
				{7028, 4174, 0x1501fb8b},
				{7029, 4173, 0x1501fa54},
				{7030, 4172, 0x1501f91b},
				{7031, 4171, 0x1501f7e2},
				{7032, 4170, 0x1501f6ac},
				{7033, 4169, 0x1501f576},
				{7034, 4168, 0x1501f441},
				{7035, 4167, 0x1501f30c},
				{7036, 4166, 0x1501f1d7},
				{7037, 4165, 0x1501f0a6},
				{7038, 4164, 0x1501ef72},
				{7039, 4163, 0x1501ee40},
				{7040, 4162, 0x1501ed10},
				{7041, 4161, 0x1501ebdf},
				{7042, 4160, 0x1501eab0},
				{7043, 4159, 0x1501e982},
				{7044, 4158, 0x1501e856},
				{7045, 4157, 0x1501e728},
				{7046, 4156, 0x1501e5fb},
				{7047, 4155, 0x1501e4d0},
				{7048, 4154, 0x1501e3a5},
				{7049, 4153, 0x1501e27b},
				{7050, 4152, 0x1501e151},
				{7051, 4151, 0x1501e029},
				{7052, 4150, 0x1501df00},
				{7053, 4149, 0x1501ddd9},
				{7054, 4148, 0x1501dcb2},
				{7055, 4147, 0x1501db8c},
				{7056, 4146, 0x1501da66},
				{7057, 4145, 0x1501d942},
				{7058, 4144, 0x1501d81f},
				{7059, 4143, 0x1501d6fd},
				{7060, 4142, 0x1501d5da},
				{7061, 4141, 0x1501d4b8},
				{7062, 4140, 0x1501d398},
				{7063, 4139, 0x1501d277},
				{7064, 4138, 0x1501d158},
				{7065, 4137, 0x1501d039},
				{7066, 4136, 0x1501cf1b},
				{7067, 4135, 0x1501cdfd},
				{7068, 4134, 0x1501cce1},
				{7069, 4133, 0x1501cbc5},
				{7070, 4132, 0x1501caaa},
				{7071, 4131, 0x1501c98e},
				{7072, 4130, 0x1501c875},
				{7073, 4129, 0x1501c75b},
				{7074, 4128, 0x1501c643},
				{7075, 4127, 0x1501c52c},
				{7076, 4126, 0x1501c414},
				{7077, 4125, 0x1501c2fd},
				{7078, 4124, 0x1501c1e8},
				{7079, 4123, 0x1501c0d3},
				{7080, 4122, 0x1501bfbe},
				{7081, 4121, 0x1501beab},
				{7082, 4120, 0x1501bd98},
				{7083, 4119, 0x1501bc85},
				{7084, 4118, 0x1501bb73},
				{7085, 4117, 0x1501ba61},
				{7086, 4116, 0x1501b951},
				{7087, 4115, 0x1501b840},
				{7088, 4114, 0x1501b731},
				{7089, 4113, 0x1501b624},
				{7090, 4112, 0x1501b516},
				{7091, 4111, 0x1501b40a},
				{7092, 4110, 0x1501b2fc},
				{7093, 4109, 0x1501b1f1},
				{7094, 4108, 0x1501b0e5},
				{7095, 4107, 0x1501afdb},
				{7096, 4106, 0x1501aed0},
				{7097, 4105, 0x1501adc8},
				{7098, 4104, 0x1501acbe},
				{7099, 4103, 0x1501abb6},
				{7100, 4102, 0x1501aaaf},
				{7101, 4101, 0x1501a9a8},
				{7102, 4100, 0x1501a8a2},
				{7103, 4099, 0x1501a79d},
				{7104, 4098, 0x1501a697},
				{7105, 4097, 0x1501a594},
				{7106, 4096, 0x1501a491},
				{7107, 4095, 0x1501a38e},
				{7108, 4094, 0x1501a28b},
				{7109, 4093, 0x1501a18a},
				{7110, 4092, 0x1501a088},
				{7111, 4091, 0x15019f88},
				{7112, 4090, 0x15019e88},
				{7113, 4089, 0x15019d89},
				{7114, 4088, 0x15019c8a},
				{7115, 4087, 0x15019b8c},
				{7116, 4086, 0x15019a8f},
				{7117, 4085, 0x15019991},
				{7118, 4084, 0x15019895},
				{7119, 4083, 0x1501979a},
				{7120, 4082, 0x1501969f},
				{7121, 4081, 0x150195a6},
				{7122, 4080, 0x150194ac},
				{7123, 4079, 0x150193b3},
				{7124, 4078, 0x150192ba},
				{7125, 4077, 0x150191c2},
				{7126, 4076, 0x150190ca},
				{7127, 4075, 0x15018fd3},
				{7128, 4074, 0x15018ede},
				{7129, 4073, 0x15018de8},
				{7130, 4072, 0x15018cf3},
				{7131, 4071, 0x15018bff},
				{7132, 4070, 0x15018b0a},
				{7133, 4069, 0x15018a18},
				{7134, 4068, 0x15018925},
				{7135, 4067, 0x15018832},
				{7136, 4066, 0x15018742},
				{7137, 4065, 0x15018650},
				{7138, 4064, 0x15018560},
				{7139, 4063, 0x15018470},
				{7140, 4062, 0x15018381},
				{7141, 4061, 0x15018292},
				{7142, 4060, 0x150181a5},
				{7143, 4059, 0x150180b6},
				{7144, 4058, 0x15017fca},
				{7145, 4057, 0x15017ede},
				{7146, 4056, 0x15017df2},
				{7147, 4055, 0x15017d06},
				{7148, 4054, 0x15017c1b},
				{7149, 4053, 0x15017b31},
				{7150, 4052, 0x15017a47},
				{7151, 4051, 0x1501795e},
				{7152, 4050, 0x15017876},
				{7153, 4049, 0x1501778e},
				{7154, 4048, 0x150176a7},
				{7155, 4047, 0x150175bf},
				{7156, 4046, 0x150174da},
				{7157, 4045, 0x150173f3},
				{7158, 4044, 0x1501730e},
				{7159, 4043, 0x1501722a},
				{7160, 4042, 0x15017146},
				{7161, 4041, 0x15017062},
				{7162, 4040, 0x15016f7f},
				{7163, 4039, 0x15016e9c},
				{7164, 4038, 0x15016db9},
				{7165, 4037, 0x15016cd8},
				{7166, 4036, 0x15016bf7},
				{7167, 4035, 0x15016b18},
				{7168, 4034, 0x15016a38},
				{7169, 4033, 0x15016958},
				{7170, 4032, 0x15016879},
				{7171, 4031, 0x1501679a},
				{7172, 4030, 0x150166bd},
				{7173, 4029, 0x150165e0},
				{7174, 4028, 0x15016502},
				{7175, 4027, 0x15016426},
				{7176, 4026, 0x1501634a},
				{7177, 4025, 0x1501626e},
				{7178, 4024, 0x15016193},
				{7179, 4023, 0x150160b9},
				{7180, 4022, 0x15015fdf},
				{7181, 4021, 0x15015f06},
				{7182, 4020, 0x15015e2e},
				{7183, 4019, 0x15015d55},
				{7184, 4018, 0x15015c7d},
				{7185, 4017, 0x15015ba6},
				{7186, 4016, 0x15015acf},
				{7187, 4015, 0x150159f9},
				{7188, 4014, 0x15015922},
				{7189, 4013, 0x1501584d},
				{7190, 4012, 0x15015778},
				{7191, 4011, 0x150156a5},
				{7192, 4010, 0x150155d2},
				{7193, 4009, 0x15015500},
				{7194, 4008, 0x1501542e},
				{7195, 4007, 0x1501535d},
				{7196, 4006, 0x1501528c},
				{7197, 4005, 0x150151bc},
				{7198, 4004, 0x150150ed},
				{7199, 4003, 0x1501501d},
				{7200, 4002, 0x15014f4f},
				{7201, 4001, 0x15014e80},
				{7202, 4000, 0x15014db2},
				{7203, 3999, 0x15014ce5},
				{7204, 3998, 0x15014c18},
				{7205, 3997, 0x15014b4c},
				{7206, 3996, 0x15014a80},
				{7207, 3995, 0x150149b4},
				{7208, 3994, 0x150148e9},
				{7209, 3993, 0x1501481f},
				{7210, 3992, 0x15014754},
				{7211, 3991, 0x1501468b},
				{7212, 3990, 0x150145c2},
				{7213, 3989, 0x150144fa},
				{7214, 3988, 0x15014432},
				{7215, 3987, 0x1501436a},
				{7216, 3986, 0x150142a3},
				{7217, 3985, 0x150141dc},
				{7218, 3984, 0x15014115},
				{7219, 3983, 0x15014050},
				{7220, 3982, 0x15013f8a},
				{7221, 3981, 0x15013ec5},
				{7222, 3980, 0x15013e01},
				{7223, 3979, 0x15013d3d},
				{7224, 3978, 0x15013c79},
				{7225, 3977, 0x15013bb6},
				{7226, 3976, 0x15013af3},
				{7227, 3975, 0x15013a31},
				{7228, 3974, 0x15013970},
				{7229, 3973, 0x150138af},
				{7230, 3972, 0x150137ee},
				{7231, 3971, 0x1501372e},
				{7232, 3970, 0x1501366e},
				{7233, 3969, 0x150135af},
				{7234, 3968, 0x150134f0},
				{7235, 3967, 0x15013431},
				{7236, 3966, 0x15013373},
				{7237, 3965, 0x150132b5},
				{7238, 3964, 0x150131f8},
				{7239, 3963, 0x1501313c},
				{7240, 3962, 0x1501307f},
				{7241, 3961, 0x15012fc3},
				{7242, 3960, 0x15012f08},
				{7243, 3959, 0x15012e4e},
				{7244, 3958, 0x15012d93},
				{7245, 3957, 0x15012cda},
				{7246, 3956, 0x15012c20},
				{7247, 3955, 0x15012b67},
				{7248, 3954, 0x15012aae},
				{7249, 3953, 0x150129f6},
				{7250, 3952, 0x1501293e},
				{7251, 3951, 0x15012886},
				{7252, 3950, 0x150127d0},
				{7253, 3949, 0x15012719},
				{7254, 3948, 0x15012663},
				{7255, 3947, 0x150125ad},
				{7256, 3946, 0x150124f8},
				{7257, 3945, 0x15012443},
				{7258, 3944, 0x1501238f},
				{7259, 3943, 0x150122dc},
				{7260, 3942, 0x15012228},
				{7261, 3941, 0x15012175},
				{7262, 3940, 0x150120c2},
				{7263, 3939, 0x15012010},
				{7264, 3938, 0x15011f5e},
				{7265, 3937, 0x15011ead},
				{7266, 3936, 0x15011dfc},
				{7267, 3935, 0x15011d4c},
				{7268, 3934, 0x15011c9b},
				{7269, 3933, 0x15011beb},
				{7270, 3932, 0x15011b3c},
				{7271, 3931, 0x15011a8e},
				{7272, 3930, 0x150119e0},
				{7273, 3929, 0x15011931},
				{7274, 3928, 0x15011885},
				{7275, 3927, 0x150117d7},
				{7276, 3926, 0x1501172b},
				{7277, 3925, 0x1501167e},
				{7278, 3924, 0x150115d2},
				{7279, 3923, 0x15011527},
				{7280, 3922, 0x1501147c},
				{7281, 3921, 0x150113d1},
				{7282, 3920, 0x15011327},
				{7283, 3919, 0x1501127d},
				{7284, 3918, 0x150111d3},
				{7285, 3917, 0x1501112a},
				{7286, 3916, 0x15011082},
				{7287, 3915, 0x15010fda},
				{7288, 3914, 0x15010f32},
				{7289, 3913, 0x15010e8b},
				{7290, 3912, 0x15010de4},
				{7291, 3911, 0x15010d3e},
				{7292, 3910, 0x15010c97},
				{7293, 3909, 0x15010bf2},
				{7294, 3908, 0x15010b4d},
				{7295, 3907, 0x15010aa7},
				{7296, 3906, 0x15010a02},
				{7297, 3905, 0x1501095f},
				{7298, 3904, 0x150108ba},
				{7299, 3903, 0x15010817},
				{7300, 3902, 0x15010774},
				{7301, 3901, 0x150106d2},
				{7302, 3900, 0x15010630},
				{7303, 3899, 0x1501058d},
				{7304, 3898, 0x150104ec},
				{7305, 3897, 0x1501044c},
				{7306, 3896, 0x150103ab},
				{7307, 3895, 0x1501030b},
				{7308, 3894, 0x1501026b},
				{7309, 3893, 0x150101cb},
				{7310, 3892, 0x1501012d},
				{7311, 3891, 0x1501008e},
				{7312, 3890, 0x1500fff0},
				{7313, 3889, 0x1500ff52},
				{7314, 3888, 0x1500feb4},
				{7315, 3887, 0x1500fe17},
				{7316, 3886, 0x1500fd7a},
				{7317, 3885, 0x1500fcde},
				{7318, 3884, 0x1500fc42},
				{7319, 3883, 0x1500fba6},
				{7320, 3882, 0x1500fb0b},
				{7321, 3881, 0x1500fa71},
				{7322, 3880, 0x1500f9d6},
				{7323, 3879, 0x1500f93c},
				{7324, 3878, 0x1500f8a2},
				{7325, 3877, 0x1500f809},
				{7326, 3876, 0x1500f770},
				{7327, 3875, 0x1500f6d8},
				{7328, 3874, 0x1500f63f},
				{7329, 3873, 0x1500f5a7},
				{7330, 3872, 0x1500f50f},
				{7331, 3871, 0x1500f478},
				{7332, 3870, 0x1500f3e2},
				{7333, 3869, 0x1500f34b},
				{7334, 3868, 0x1500f2b5},
				{7335, 3867, 0x1500f220},
				{7336, 3866, 0x1500f18b},
				{7337, 3865, 0x1500f0f6},
				{7338, 3864, 0x1500f061},
				{7339, 3863, 0x1500efcd},
				{7340, 3862, 0x1500ef39},
				{7341, 3861, 0x1500eea6},
				{7342, 3860, 0x1500ee12},
				{7343, 3859, 0x1500ed7f},
				{7344, 3858, 0x1500eced},
				{7345, 3857, 0x1500ec5b},
				{7346, 3856, 0x1500ebc9},
				{7347, 3855, 0x1500eb38},
				{7348, 3854, 0x1500eaa7},
				{7349, 3853, 0x1500ea16},
				{7350, 3852, 0x1500e986},
				{7351, 3851, 0x1500e8f7},
				{7352, 3850, 0x1500e867},
				{7353, 3849, 0x1500e7d8},
				{7354, 3848, 0x1500e749},
				{7355, 3847, 0x1500e6ba},
				{7356, 3846, 0x1500e62c},
				{7357, 3845, 0x1500e59e},
				{7358, 3844, 0x1500e511},
				{7359, 3843, 0x1500e483},
				{7360, 3842, 0x1500e3f7},
				{7361, 3841, 0x1500e36a},
				{7362, 3840, 0x1500e2de},
				{7363, 3839, 0x1500e252},
				{7364, 3838, 0x1500e1c7},
				{7365, 3837, 0x1500e13b},
				{7366, 3836, 0x1500e0b1},
				{7367, 3835, 0x1500e027},
				{7368, 3834, 0x1500df9d},
				{7369, 3833, 0x1500df13},
				{7370, 3832, 0x1500de8a},
				{7371, 3831, 0x1500de00},
				{7372, 3830, 0x1500dd77},
				{7373, 3829, 0x1500dcef},
				{7374, 3828, 0x1500dc67},
				{7375, 3827, 0x1500dbe0},
				{7376, 3826, 0x1500db58},
				{7377, 3825, 0x1500dad1},
				{7378, 3824, 0x1500da4a},
				{7379, 3823, 0x1500d9c4},
				{7380, 3822, 0x1500d93d},
				{7381, 3821, 0x1500d8b8},
				{7382, 3820, 0x1500d832},
				{7383, 3819, 0x1500d7ae},
				{7384, 3818, 0x1500d729},
				{7385, 3817, 0x1500d6a4},
				{7386, 3816, 0x1500d620},
				{7387, 3815, 0x1500d59c},
				{7388, 3814, 0x1500d518},
				{7389, 3813, 0x1500d495},
				{7390, 3812, 0x1500d412},
				{7391, 3811, 0x1500d390},
				{7392, 3810, 0x1500d30e},
				{7393, 3809, 0x1500d28c},
				{7394, 3808, 0x1500d209},
				{7395, 3807, 0x1500d189},
				{7396, 3806, 0x1500d107},
				{7397, 3805, 0x1500d087},
				{7398, 3804, 0x1500d007},
				{7399, 3803, 0x1500cf87},
				{7400, 3802, 0x1500cf07},
				{7401, 3801, 0x1500ce87},
				{7402, 3800, 0x1500ce08},
				{7403, 3799, 0x1500cd89},
				{7404, 3798, 0x1500cd0b},
				{7405, 3797, 0x1500cc8c},
				{7406, 3796, 0x1500cc0f},
				{7407, 3795, 0x1500cb90},
				{7408, 3794, 0x1500cb13},
				{7409, 3793, 0x1500ca96},
				{7410, 3792, 0x1500ca19},
				{7411, 3791, 0x1500c99d},
				{7412, 3790, 0x1500c921},
				{7413, 3789, 0x1500c8a6},
				{7414, 3788, 0x1500c82a},
				{7415, 3787, 0x1500c7af},
				{7416, 3786, 0x1500c734},
				{7417, 3785, 0x1500c6b9},
				{7418, 3784, 0x1500c63f},
				{7419, 3783, 0x1500c5c5},
				{7420, 3782, 0x1500c54b},
				{7421, 3781, 0x1500c4d1},
				{7422, 3780, 0x1500c458},
				{7423, 3779, 0x1500c3df},
				{7424, 3778, 0x1500c366},
				{7425, 3777, 0x1500c2ee},
				{7426, 3776, 0x1500c276},
				{7427, 3775, 0x1500c1fe},
				{7428, 3774, 0x1500c187},
				{7429, 3773, 0x1500c110},
				{7430, 3772, 0x1500c099},
				{7431, 3771, 0x1500c022},
				{7432, 3770, 0x1500bfac},
				{7433, 3769, 0x1500bf36},
				{7434, 3768, 0x1500bec0},
				{7435, 3767, 0x1500be4b},
				{7436, 3766, 0x1500bdd5},
				{7437, 3765, 0x1500bd60},
				{7438, 3764, 0x1500bceb},
				{7439, 3763, 0x1500bc77},
				{7440, 3762, 0x1500bc03},
				{7441, 3761, 0x1500bb8f},
				{7442, 3760, 0x1500bb1c},
				{7443, 3759, 0x1500baa8},
				{7444, 3758, 0x1500ba36},
				{7445, 3757, 0x1500b9c3},
				{7446, 3756, 0x1500b950},
				{7447, 3755, 0x1500b8de},
				{7448, 3754, 0x1500b86c},
				{7449, 3753, 0x1500b7fa},
				{7450, 3752, 0x1500b789},
				{7451, 3751, 0x1500b717},
				{7452, 3750, 0x1500b6a7},
				{7453, 3749, 0x1500b636},
				{7454, 3748, 0x1500b5c5},
				{7455, 3747, 0x1500b555},
				{7456, 3746, 0x1500b4e5},
				{7457, 3745, 0x1500b476},
				{7458, 3744, 0x1500b407},
				{7459, 3743, 0x1500b398},
				{7460, 3742, 0x1500b329},
				{7461, 3741, 0x1500b2bb},
				{7462, 3740, 0x1500b24c},
				{7463, 3739, 0x1500b1de},
				{7464, 3738, 0x1500b170},
				{7465, 3737, 0x1500b103},
				{7466, 3736, 0x1500b095},
				{7467, 3735, 0x1500b028},
				{7468, 3734, 0x1500afbb},
				{7469, 3733, 0x1500af4e},
				{7470, 3732, 0x1500aee2},
				{7471, 3731, 0x1500ae76},
				{7472, 3730, 0x1500ae0a},
				{7473, 3729, 0x1500ad9f},
				{7474, 3728, 0x1500ad34},
				{7475, 3727, 0x1500acc9},
				{7476, 3726, 0x1500ac5e},
				{7477, 3725, 0x1500abf3},
				{7478, 3724, 0x1500ab8a},
				{7479, 3723, 0x1500ab20},
				{7480, 3722, 0x1500aab7},
				{7481, 3721, 0x1500aa4e},
				{7482, 3720, 0x1500a9e5},
				{7483, 3719, 0x1500a97c},
				{7484, 3718, 0x1500a914},
				{7485, 3717, 0x1500a8ac},
				{7486, 3716, 0x1500a844},
				{7487, 3715, 0x1500a7dd},
				{7488, 3714, 0x1500a775},
				{7489, 3713, 0x1500a70f},
				{7490, 3712, 0x1500a6a8},
				{7491, 3711, 0x1500a641},
				{7492, 3710, 0x1500a5db},
				{7493, 3709, 0x1500a575},
				{7494, 3708, 0x1500a50f},
				{7495, 3707, 0x1500a4a9},
				{7496, 3706, 0x1500a444},
				{7497, 3705, 0x1500a3df},
				{7498, 3704, 0x1500a37a},
				{7499, 3703, 0x1500a315},
				{7500, 3702, 0x1500a2b1},
				{7501, 3701, 0x1500a24c},
				{7502, 3700, 0x1500a1e8},
				{7503, 3699, 0x1500a185},
				{7504, 3698, 0x1500a121},
				{7505, 3697, 0x1500a0be},
				{7506, 3696, 0x1500a05b},
				{7507, 3695, 0x15009ff8},
				{7508, 3694, 0x15009f96},
				{7509, 3693, 0x15009f33},
				{7510, 3692, 0x15009ed1},
				{7511, 3691, 0x15009e6f},
				{7512, 3690, 0x15009e0e},
				{7513, 3689, 0x15009dac},
				{7514, 3688, 0x15009d4b},
				{7515, 3687, 0x15009cea},
				{7516, 3686, 0x15009c89},
				{7517, 3685, 0x15009c29},
				{7518, 3684, 0x15009bc9},
				{7519, 3683, 0x15009b68},
				{7520, 3682, 0x15009b09},
				{7521, 3681, 0x15009aaa},
				{7522, 3680, 0x15009a4a},
				{7523, 3679, 0x150099eb},
				{7524, 3678, 0x1500998c},
				{7525, 3677, 0x1500992d},
				{7526, 3676, 0x150098cf},
				{7527, 3675, 0x15009870},
				{7528, 3674, 0x15009812},
				{7529, 3673, 0x150097b5},
				{7530, 3672, 0x15009757},
				{7531, 3671, 0x150096fa},
				{7532, 3670, 0x1500969c},
				{7533, 3669, 0x15009640},
				{7534, 3668, 0x150095e3},
				{7535, 3667, 0x15009587},
				{7536, 3666, 0x1500952b},
				{7537, 3665, 0x150094cf},
				{7538, 3664, 0x15009473},
				{7539, 3663, 0x15009417},
				{7540, 3662, 0x150093bc},
				{7541, 3661, 0x15009361},
				{7542, 3660, 0x15009306},
				{7543, 3659, 0x150092ab},
				{7544, 3658, 0x15009250},
				{7545, 3657, 0x150091f6},
				{7546, 3656, 0x1500919c},
				{7547, 3655, 0x15009142},
				{7548, 3654, 0x150090e8},
				{7549, 3653, 0x1500908f},
				{7550, 3652, 0x15009036},
				{7551, 3651, 0x15008fdd},
				{7552, 3650, 0x15008f85},
				{7553, 3649, 0x15008f2c},
				{7554, 3648, 0x15008ed4},
				{7555, 3647, 0x15008e7b},
				{7556, 3646, 0x15008e23},
				{7557, 3645, 0x15008dcc},
				{7558, 3644, 0x15008d74},
				{7559, 3643, 0x15008d1d},
				{7560, 3642, 0x15008cc6},
				{7561, 3641, 0x15008c6f},
				{7562, 3640, 0x15008c18},
				{7563, 3639, 0x15008bc2},
				{7564, 3638, 0x15008b6b},
				{7565, 3637, 0x15008b15},
				{7566, 3636, 0x15008ac0},
				{7567, 3635, 0x15008a6a},
				{7568, 3634, 0x15008a15},
				{7569, 3633, 0x150089bf},
				{7570, 3632, 0x1500896b},
				{7571, 3631, 0x15008916},
				{7572, 3630, 0x150088c1},
				{7573, 3629, 0x1500886d},
				{7574, 3628, 0x15008819},
				{7575, 3627, 0x150087c4},
				{7576, 3626, 0x15008771},
				{7577, 3625, 0x1500871d},
				{7578, 3624, 0x150086ca},
				{7579, 3623, 0x15008677},
				{7580, 3622, 0x15008623},
				{7581, 3621, 0x150085d1},
				{7582, 3620, 0x1500857e},
				{7583, 3619, 0x1500852c},
				{7584, 3618, 0x150084da},
				{7585, 3617, 0x15008488},
				{7586, 3616, 0x15008436},
				{7587, 3615, 0x150083e5},
				{7588, 3614, 0x15008393},
				{7589, 3613, 0x15008342},
				{7590, 3612, 0x150082f1},
				{7591, 3611, 0x150082a0},
				{7592, 3610, 0x1500824f},
				{7593, 3609, 0x150081ff},
				{7594, 3608, 0x150081af},
				{7595, 3607, 0x1500815f},
				{7596, 3606, 0x1500810f},
				{7597, 3605, 0x150080c0},
				{7598, 3604, 0x15008070},
				{7599, 3603, 0x15008021},
				{7600, 3602, 0x147fd24b},
				{7601, 3601, 0x147f8379},
				{7602, 3600, 0x147f34a7},
				{7603, 3599, 0x147ee62b},
				{7604, 3598, 0x147e9805},
				{7605, 3597, 0x147e49df},
				{7606, 3596, 0x147dfc0f},
				{7607, 3595, 0x147dae3e},
				{7608, 3594, 0x147d60c4},
				{7609, 3593, 0x147d134a},
				{7610, 3592, 0x147cc625},
				{7611, 3591, 0x147c7956},
				{7612, 3590, 0x147c2cde},
				{7613, 3589, 0x147be00f},
				{7614, 3588, 0x147b9396},
				{7615, 3587, 0x147b4773},
				{7616, 3586, 0x147afb50},
				{7617, 3585, 0x147aaf83},
				{7618, 3584, 0x147a640c},
				{7619, 3583, 0x147a183f},
				{7620, 3582, 0x1479cd1d},
				{7621, 3581, 0x147981fc},
				{7622, 3580, 0x14793730},
				{7623, 3579, 0x1478ec65},
				{7624, 3578, 0x1478a199},
				{7625, 3577, 0x14785724},
				{7626, 3576, 0x14780d04},
				{7627, 3575, 0x1477c33a},
				{7628, 3574, 0x14777970},
				{7629, 3573, 0x14772ffc},
				{7630, 3572, 0x1476e688},
				{7631, 3571, 0x14769d14},
				{7632, 3570, 0x147653f6},
				{7633, 3569, 0x14760ad8},
				{7634, 3568, 0x1475c20f},
				{7635, 3567, 0x1475799d},
				{7636, 3566, 0x1475312a},
				{7637, 3565, 0x1474e90e},
				{7638, 3564, 0x1474a0f1},
				{7639, 3563, 0x147458d4},
				{7640, 3562, 0x1474110e},
				{7641, 3561, 0x1473c99d},
				{7642, 3560, 0x14738282},
				{7643, 3559, 0x14733b67},
				{7644, 3558, 0x1472f44c},
				{7645, 3557, 0x1472ad87},
				{7646, 3556, 0x147266c1},
				{7647, 3555, 0x14722052},
				{7648, 3554, 0x1471d9e3},
				{7649, 3553, 0x147193c9},
				{7650, 3552, 0x14714db0},
				{7651, 3551, 0x147107ed},
				{7652, 3550, 0x1470c27f},
				{7653, 3549, 0x14707cbb},
				{7654, 3548, 0x147037a3},
				{7655, 3547, 0x146ff28c},
				{7656, 3546, 0x146fad74},
				{7657, 3545, 0x146f68b2},
				{7658, 3544, 0x146f2446},
				{7659, 3543, 0x146edfda},
				{7660, 3542, 0x146e9b6d},
				{7661, 3541, 0x146e5757},
				{7662, 3540, 0x146e1341},
				{7663, 3539, 0x146dcf81},
				{7664, 3538, 0x146d8bc0},
				{7665, 3537, 0x146d4856},
				{7666, 3536, 0x146d04eb},
				{7667, 3535, 0x146cc1d6},
				{7668, 3534, 0x146c7ec2},
				{7669, 3533, 0x146c3c03},
				{7670, 3532, 0x146bf944},
				{7671, 3531, 0x146bb6db},
				{7672, 3530, 0x146b7472},
				{7673, 3529, 0x146b32b5},
				{7674, 3528, 0x146af0a2},
				{7675, 3527, 0x146aaee5},
				{7676, 3526, 0x146a6d28},
				{7677, 3525, 0x146a2b6a},
				{7678, 3524, 0x1469ea03},
				{7679, 3523, 0x1469a8f2},
				{7680, 3522, 0x146967e0},
				{7681, 3521, 0x146926cf},
				{7682, 3520, 0x1468e613},
				{7683, 3519, 0x1468a5ad},
				{7684, 3518, 0x146864f2},
				{7685, 3517, 0x146824e2},
				{7686, 3516, 0x1467e47c},
				{7687, 3515, 0x1467a4c2},
				{7688, 3514, 0x14676508},
				{7689, 3513, 0x1467254e},
				{7690, 3512, 0x1466e5ea},
				{7691, 3511, 0x1466a686},
				{7692, 3510, 0x14666777},
				{7693, 3509, 0x14662869},
				{7694, 3508, 0x1465e95b},
				{7695, 3507, 0x1465aaa2},
				{7696, 3506, 0x14656bea},
				{7697, 3505, 0x14652d87},
				{7698, 3504, 0x1464ef25},
				{7699, 3503, 0x1464b118},
				{7700, 3502, 0x1464730b},
				{7701, 3501, 0x14643555},
				{7702, 3500, 0x1463f79e},
				{7703, 3499, 0x1463b9e7},
				{7704, 3498, 0x14637cdc},
				{7705, 3497, 0x14633f7b},
				{7706, 3496, 0x14630270},
				{7707, 3495, 0x1462c565},
				{7708, 3494, 0x146288af},
				{7709, 3493, 0x14624bfa},
				{7710, 3492, 0x14620f45},
				{7711, 3491, 0x1461d2e5},
				{7712, 3490, 0x14619686},
				{7713, 3489, 0x14615a7c},
				{7714, 3488, 0x14611e73},
				{7715, 3487, 0x1460e2bf},
				{7716, 3486, 0x1460a70c},
				{7717, 3485, 0x14606bae},
				{7718, 3484, 0x14602ffa},
				{7719, 3483, 0x145ff548},
				{7720, 3482, 0x145fb9ea},
				{7721, 3481, 0x145f7ee2},
				{7722, 3480, 0x145f4430},
				{7723, 3479, 0x145f097e},
				{7724, 3478, 0x145ececc},
				{7725, 3477, 0x145e9470},
				{7726, 3476, 0x145e5a14},
				{7727, 3475, 0x145e200d},
				{7728, 3474, 0x145de607},
				{7729, 3473, 0x145dac01},
				{7730, 3472, 0x145d7250},
				{7731, 3471, 0x145d38f6},
				{7732, 3470, 0x145cff45},
				{7733, 3469, 0x145cc5ea},
				{7734, 3468, 0x145c8ce6},
				{7735, 3467, 0x145c53e1},
				{7736, 3466, 0x145c1b32},
				{7737, 3465, 0x145be22d},
				{7738, 3464, 0x145ba97e},
				{7739, 3463, 0x145b7125},
				{7740, 3462, 0x145b3876},
				{7741, 3461, 0x145b0073},
				{7742, 3460, 0x145ac81a},
				{7743, 3459, 0x145a9017},
				{7744, 3458, 0x145a586a},
				{7745, 3457, 0x145a2067},
				{7746, 3456, 0x1459e90f},
				{7747, 3455, 0x1459b162},
				{7748, 3454, 0x14597a0a},
				{7749, 3453, 0x145942b3},
				{7750, 3452, 0x14590c07},
				{7751, 3451, 0x1458d506},
				{7752, 3450, 0x14589e04},
				{7753, 3449, 0x14586758},
				{7754, 3448, 0x145830ad},
				{7755, 3447, 0x1457fa57},
				{7756, 3446, 0x1457c401},
				{7757, 3445, 0x14578dab},
				{7758, 3444, 0x145757ab},
				{7759, 3443, 0x145721ab},
				{7760, 3442, 0x1456ebab},
				{7761, 3441, 0x1456b601},
				{7762, 3440, 0x14568057},
				{7763, 3439, 0x14564b03},
				{7764, 3438, 0x145615ae},
				{7765, 3437, 0x1455e05a},
				{7766, 3436, 0x1455abb2},
				{7767, 3435, 0x14557709},
				{7768, 3434, 0x14554261},
				{7769, 3433, 0x14550e0e},
				{7770, 3432, 0x1454d9bb},
				{7771, 3431, 0x1454a569},
				{7772, 3430, 0x1454716c},
				{7773, 3429, 0x14543d6f},
				{7774, 3428, 0x1454099d},
				{7775, 3427, 0x1453d5cb},
				{7776, 3426, 0x1453a24f},
				{7777, 3425, 0x14536ed3},
				{7778, 3424, 0x14533b57},
				{7779, 3423, 0x14530831},
				{7780, 3422, 0x1452d536},
				{7781, 3421, 0x1452a23b},
				{7782, 3420, 0x14526f6a},
				{7783, 3419, 0x14523c9a},
				{7784, 3418, 0x145209f5},
				{7785, 3417, 0x1451d77a},
				{7786, 3416, 0x1451a500},
				{7787, 3415, 0x145172b0},
				{7788, 3414, 0x1451408c},
				{7789, 3413, 0x14510e67},
				{7790, 3412, 0x1450dc98},
				{7791, 3411, 0x1450aaca},
				{7792, 3410, 0x145078fb},
				{7793, 3409, 0x14504757},
				{7794, 3408, 0x145015de},
				{7795, 3407, 0x144fe491},
				{7796, 3406, 0x144fb398},
				{7797, 3405, 0x144f8276},
				{7798, 3404, 0x144f517d},
				{7799, 3403, 0x144f2085},
				{7800, 3402, 0x144eefb8},
				{7801, 3401, 0x144ebf16},
				{7802, 3400, 0x144e8e9f},
				{7803, 3399, 0x144e5e28},
				{7804, 3398, 0x144e2ddb},
				{7805, 3397, 0x144dfd8f},
				{7806, 3396, 0x144dcd99},
				{7807, 3395, 0x144d9d77},
				{7808, 3394, 0x144d6dac},
				{7809, 3393, 0x144d3de0},
				{7810, 3392, 0x144d0e40},
				{7811, 3391, 0x144cdef5},
				{7812, 3390, 0x144caf7f},
				{7813, 3389, 0x144c8034},
				{7814, 3388, 0x144c5115},
				{7815, 3387, 0x144c21f5},
				{7816, 3386, 0x144bf300},
				{7817, 3385, 0x144bc436},
				{7818, 3384, 0x144b956c},
				{7819, 3383, 0x144b66cd},
				{7820, 3382, 0x144b3859},
				{7821, 3381, 0x144b09e5},
				{7822, 3380, 0x144adb9c},
				{7823, 3379, 0x144aad7e},
				{7824, 3378, 0x144a7f60},
				{7825, 3377, 0x144a516c},
				{7826, 3376, 0x144a23a4},
				{7827, 3375, 0x1449f607},
				{7828, 3374, 0x1449c869},
				{7829, 3373, 0x14499acc},
				{7830, 3372, 0x14496d59},
				{7831, 3371, 0x14494012},
				{7832, 3370, 0x144912ca},
				{7833, 3369, 0x1448e5d8},
				{7834, 3368, 0x1448b8bc},
				{7835, 3367, 0x14488bf5},
				{7836, 3366, 0x14485f2e},
				{7837, 3365, 0x14483268},
				{7838, 3364, 0x144805f7},
				{7839, 3363, 0x1447d986},
				{7840, 3362, 0x1447ad15},
				{7841, 3361, 0x144780cf},
				{7842, 3360, 0x144754df},
				{7843, 3359, 0x144728ef},
				{7844, 3358, 0x1446fcff},
				{7845, 3357, 0x1446d13a},
				{7846, 3356, 0x1446a574},
				{7847, 3355, 0x144679da},
				{7848, 3354, 0x14464e6b},
				{7849, 3353, 0x144622fc},
				{7850, 3352, 0x1445f7b7},
				{7851, 3351, 0x1445cc73},
				{7852, 3350, 0x1445a159},
				{7853, 3349, 0x1445766b},
				{7854, 3348, 0x14454b7c},
				{7855, 3347, 0x144520b9},
				{7856, 3346, 0x1444f620},
				{7857, 3345, 0x1444cbb3},
				{7858, 3344, 0x1444a145},
				{7859, 3343, 0x144476d7},
				{7860, 3342, 0x14444c94},
				{7861, 3341, 0x1444227d},
				{7862, 3340, 0x1443f865},
				{7863, 3339, 0x1443ce78},
				{7864, 3338, 0x1443a48b},
				{7865, 3337, 0x14437ac9},
				{7866, 3336, 0x14435132},
				{7867, 3335, 0x1443279b},
				{7868, 3334, 0x1442fe2f},
				{7869, 3333, 0x1442d4c3},
				{7870, 3332, 0x1442ab81},
				{7871, 3331, 0x1442826b},
				{7872, 3330, 0x14425980},
				{7873, 3329, 0x14423094},
				{7874, 3328, 0x144207d4},
				{7875, 3327, 0x1441df14},
				{7876, 3326, 0x1441b653},
				{7877, 3325, 0x14418dbe},
				{7878, 3324, 0x14416553},
				{7879, 3323, 0x14413ce9},
				{7880, 3322, 0x144114a9},
				{7881, 3321, 0x1440ec95},
				{7882, 3320, 0x1440c480},
				{7883, 3319, 0x14409c96},
				{7884, 3318, 0x144074ac},
				{7885, 3317, 0x14404cee},
				{7886, 3316, 0x1440252f},
				{7887, 3315, 0x143ffd9b},
				{7888, 3314, 0x143fd65d},
				{7889, 3313, 0x143faef4},
				{7890, 3312, 0x143f878b},
				{7891, 3311, 0x143f6078},
				{7892, 3310, 0x143f393a},
				{7893, 3309, 0x143f1252},
				{7894, 3308, 0x143eeb6a},
				{7895, 3307, 0x143ec482},
				{7896, 3306, 0x143e9dc4},
				{7897, 3305, 0x143e7732},
				{7898, 3304, 0x143e50a0},
				{7899, 3303, 0x143e2a38},
				{7900, 3302, 0x143e03d1},
				{7901, 3301, 0x143ddd95},
				{7902, 3300, 0x143db783},
				{7903, 3299, 0x143d919d},
				{7904, 3298, 0x143d6b8b},
				{7905, 3297, 0x143d45a5},
				{7906, 3296, 0x143d1fe9},
				{7907, 3295, 0x143cfa2d},
				{7908, 3294, 0x143cd49d},
				{7909, 3293, 0x143caf0c},
				{7910, 3292, 0x143c89a6},
				{7911, 3291, 0x143c6441},
				{7912, 3290, 0x143c3f06},
				{7913, 3289, 0x143c19f6},
				{7914, 3288, 0x143bf4e6},
				{7915, 3287, 0x143bcfd6},
				{7916, 3286, 0x143baaf1},
				{7917, 3285, 0x143b8637},
				{7918, 3284, 0x143b61a8},
				{7919, 3283, 0x143b3d19},
				{7920, 3282, 0x143b188a},
				{7921, 3281, 0x143af426},
				{7922, 3280, 0x143acfc1},
				{7923, 3279, 0x143aab88},
				{7924, 3278, 0x143a874f},
				{7925, 3277, 0x143a6341},
				{7926, 3276, 0x143a3f32},
				{7927, 3275, 0x143a1b4f},
				{7928, 3274, 0x1439f76c},
				{7929, 3273, 0x1439d3b3},
				{7930, 3272, 0x1439b026},
				{7931, 3271, 0x14398c98},
				{7932, 3270, 0x1439690b},
				{7933, 3269, 0x143945a8},
				{7934, 3268, 0x1439229b},
				{7935, 3267, 0x1438ff64},
				{7936, 3266, 0x1438dc2c},
				{7937, 3265, 0x1438b91f},
				{7938, 3264, 0x1438963e},
				{7939, 3263, 0x1438735c},
				{7940, 3262, 0x1438507a},
				{7941, 3261, 0x14382dee},
				{7942, 3260, 0x14380b37},
				{7943, 3259, 0x1437e8ab},
				{7944, 3258, 0x1437c64a},
				{7945, 3257, 0x1437a3e9},
				{7946, 3256, 0x14378188},
				{7947, 3255, 0x14375f52},
				{7948, 3254, 0x14373d47},
				{7949, 3253, 0x14371b67},
				{7950, 3252, 0x1436f987},
				{7951, 3251, 0x1436d7a7},
				{7952, 3250, 0x1436b5c6},
				{7953, 3249, 0x14369411},
				{7954, 3248, 0x14367287},
				{7955, 3247, 0x143650fc},
				{7956, 3246, 0x14362f72},
				{7957, 3245, 0x14360e13},
				{7958, 3244, 0x1435ecde},
				{7959, 3243, 0x1435cbaa},
				{7960, 3242, 0x1435aa75},
				{7961, 3241, 0x1435896c},
				{7962, 3240, 0x14356862},
				{7963, 3239, 0x14354784},
				{7964, 3238, 0x143526d0},
				{7965, 3237, 0x1435061c},
				{7966, 3236, 0x1434e569},
				{7967, 3235, 0x1434c4e0},
				{7968, 3234, 0x1434a457},
				{7969, 3233, 0x143483f9},
				{7970, 3232, 0x1434639c},
				{7971, 3231, 0x14344369},
				{7972, 3230, 0x14342336},
				{7973, 3229, 0x14340303},
				{7974, 3228, 0x1433e2fb},
				{7975, 3227, 0x1433c31e},
				{7976, 3226, 0x1433a341},
				{7977, 3225, 0x14338364},
				{7978, 3224, 0x143363b2},
				{7979, 3223, 0x14334400},
				{7980, 3222, 0x143324a4},
				{7981, 3221, 0x1433051c},
				{7982, 3220, 0x1432e5c0},
				{7983, 3219, 0x1432c664},
				{7984, 3218, 0x1432a708},
				{7985, 3217, 0x143287d7},
				{7986, 3216, 0x143268d0},
				{7987, 3215, 0x1432499f},
				{7988, 3214, 0x14322ac3},
				{7989, 3213, 0x14320be8},
				{7990, 3212, 0x1431ed0d},
				{7991, 3211, 0x1431ce31},
				{7992, 3210, 0x1431afac},
				{7993, 3209, 0x143190fb},
				{7994, 3208, 0x14317276},
				{7995, 3207, 0x1431541b},
				{7996, 3206, 0x143135c0},
				{7997, 3205, 0x14311766},
				{7998, 3204, 0x1430f936},
				{7999, 3203, 0x1430db06},
				{8000, 3202, 0x1430bd02},
				{8001, 3201, 0x14309efd},
				{8002, 3200, 0x143080f8},
				{8003, 3199, 0x1430631e},
				{8004, 3198, 0x14304544},
				{8005, 3197, 0x14302795},
				{8006, 3196, 0x143009e7},
				{8007, 3195, 0x142fec38},
				{8008, 3194, 0x142fceb4},
				{8009, 3193, 0x142fb15b},
				{8010, 3192, 0x142f9402},
				{8011, 3191, 0x142f76a9},
				{8012, 3190, 0x142f597a},
				{8013, 3189, 0x142f3c4c},
				{8014, 3188, 0x142f1f1e},
				{8015, 3187, 0x142f021b},
				{8016, 3186, 0x142ee518},
				{8017, 3185, 0x142ec840},
				{8018, 3184, 0x142eab67},
				{8019, 3183, 0x142e8eba},
				{8020, 3182, 0x142e71e2},
				{8021, 3181, 0x142e5560},
				{8022, 3180, 0x142e38b2},
				{8023, 3179, 0x142e1c30},
				{8024, 3178, 0x142dffd8},
				{8025, 3177, 0x142de381},
				{8026, 3176, 0x142dc754},
				{8027, 3175, 0x142daafd},
				{8028, 3174, 0x142d8ed0},
				{8029, 3173, 0x142d72cf},
				{8030, 3172, 0x142d56a2},
				{8031, 3171, 0x142d3acc},
				{8032, 3170, 0x142d1eca},
				{8033, 3169, 0x142d02f3},
				{8034, 3168, 0x142ce71d},
				{8035, 3167, 0x142ccb71},
				{8036, 3166, 0x142cafc5},
				{8037, 3165, 0x142c9445},
				{8038, 3164, 0x142c78c4},
				{8039, 3163, 0x142c5d43},
				{8040, 3162, 0x142c41ed},
				{8041, 3161, 0x142c2697},
				{8042, 3160, 0x142c0b6c},
				{8043, 3159, 0x142bf017},
				{8044, 3158, 0x142bd4ec},
				{8045, 3157, 0x142bb9ec},
				{8046, 3156, 0x142b9eec},
				{8047, 3155, 0x142b83ec},
				{8048, 3154, 0x142b6917},
				{8049, 3153, 0x142b4e41},
				{8050, 3152, 0x142b336c},
				{8051, 3151, 0x142b18c2},
				{8052, 3150, 0x142afe18},
				{8053, 3149, 0x142ae383},
				{8054, 3148, 0x142ac91a},
				{8055, 3147, 0x142aaec5},
				{8056, 3146, 0x142a94b2},
				{8057, 3145, 0x142a7a73},
				{8058, 3144, 0x142a605f},
				{8059, 3143, 0x142a464b},
				{8060, 3142, 0x142a2c37},
				{8061, 3141, 0x142a124e},
				{8062, 3140, 0x1429f865},
				{8063, 3139, 0x1429de92},
				{8064, 3138, 0x1429c4d4},
				{8065, 3137, 0x1429ab16},
				{8066, 3136, 0x1429916d},
				{8067, 3135, 0x142977da},
				{8068, 3134, 0x14295e47},
				{8069, 3133, 0x142944df},
				{8070, 3132, 0x14292b61},
				{8071, 3131, 0x1429120f},
				{8072, 3130, 0x1428f8e7},
				{8073, 3129, 0x1428dfaa},
				{8074, 3128, 0x1428c682},
				{8075, 3127, 0x1428ad5a},
				{8076, 3126, 0x14289448},
				{8077, 3125, 0x14287b4b},
				{8078, 3124, 0x14286264},
				{8079, 3123, 0x1428497c},
				{8080, 3122, 0x142830ab},
				{8081, 3121, 0x142817d9},
				{8082, 3120, 0x1427ff32},
				{8083, 3119, 0x1427e68b},
				{8084, 3118, 0x1427cde4},
				{8085, 3117, 0x1427b568},
				{8086, 3116, 0x14279cec},
				{8087, 3115, 0x1427849b},
				{8088, 3114, 0x14276c4a},
				{8089, 3113, 0x142753f9},
				{8090, 3112, 0x14273bbd},
				{8091, 3111, 0x14272381},
				{8092, 3110, 0x14270b5b},
				{8093, 3109, 0x1426f34a},
				{8094, 3108, 0x1426db4f},
				{8095, 3107, 0x1426c354},
				{8096, 3106, 0x1426ab6e},
				{8097, 3105, 0x14269388},
				{8098, 3104, 0x14267bce},
				{8099, 3103, 0x14266413},
				{8100, 3102, 0x14264c58},
				{8101, 3101, 0x142634c8},
				{8102, 3100, 0x14261d23},
				{8103, 3099, 0x142605be},
				{8104, 3098, 0x1425ee59},
				{8105, 3097, 0x1425d6f4},
				{8106, 3096, 0x1425bf8f},
				{8107, 3095, 0x1425a83f},
				{8108, 3094, 0x14259105},
				{8109, 3093, 0x142579e1},
				{8110, 3092, 0x142562bc},
				{8111, 3091, 0x14254bad},
				{8112, 3090, 0x142534b3},
				{8113, 3089, 0x14251dba},
				{8114, 3088, 0x142506d6},
				{8115, 3087, 0x1424eff2},
				{8116, 3086, 0x1424d923},
				{8117, 3085, 0x1424c26a},
				{8118, 3084, 0x1424abdb},
				{8119, 3083, 0x14249538},
				{8120, 3082, 0x14247ea9},
				{8121, 3081, 0x1424681b},
				{8122, 3080, 0x142451a2},
				{8123, 3079, 0x14243b3f},
				{8124, 3078, 0x142424db},
				{8125, 3077, 0x14240e8d},
				{8126, 3076, 0x1423f855},
				{8127, 3075, 0x1423e21d},
				{8128, 3074, 0x1423cbfa},
				{8129, 3073, 0x1423b5d7},
				{8130, 3072, 0x14239fdf},
				{8131, 3071, 0x142389d1},
				{8132, 3070, 0x142373ee},
				{8133, 3069, 0x14235e21},
				{8134, 3068, 0x14234854},
				{8135, 3067, 0x14233287},
				{8136, 3066, 0x14231ccf},
				{8137, 3065, 0x14230718},
				{8138, 3064, 0x1422f176},
				{8139, 3063, 0x1422dbe9},
				{8140, 3062, 0x1422c672},
				{8141, 3061, 0x1422b0fa},
				{8142, 3060, 0x14229b83},
				{8143, 3059, 0x14228621},
				{8144, 3058, 0x142270d5},
				{8145, 3057, 0x14225b9e},
				{8146, 3056, 0x14224667},
				{8147, 3055, 0x14223146},
				{8148, 3054, 0x14221c25},
				{8149, 3053, 0x1422072e},
				{8150, 3052, 0x1421f222},
				{8151, 3051, 0x1421dd2c},
				{8152, 3050, 0x1421c84b},
				{8153, 3049, 0x1421b36a},
				{8154, 3048, 0x14219e9e},
				{8155, 3047, 0x142189e8},
				{8156, 3046, 0x14217532},
				{8157, 3045, 0x14216092},
				{8158, 3044, 0x14214bf1},
				{8159, 3043, 0x14213766},
				{8160, 3042, 0x142122db},
				{8161, 3041, 0x14210e7a},
				{8162, 3040, 0x1420fa05},
				{8163, 3039, 0x1420e5ba},
				{8164, 3038, 0x1420d16f},
				{8165, 3037, 0x1420bd3a},
				{8166, 3036, 0x1420a905},
				{8167, 3035, 0x142094e5},
				{8168, 3034, 0x142080c5},
				{8169, 3033, 0x14206cbb},
				{8170, 3032, 0x142058b1},
				{8171, 3031, 0x142044bc},
				{8172, 3030, 0x142030c7},
				{8173, 3029, 0x14201ce7},
				{8174, 3028, 0x1420091e},
				{8175, 3027, 0x141ff554},
				{8176, 3026, 0x141fe19f},
				{8177, 3025, 0x141fce00},
				{8178, 3024, 0x141fba61},
				{8179, 3023, 0x141fa6d8},
				{8180, 3022, 0x141f934e},
				{8181, 3021, 0x141f7fda},
				{8182, 3020, 0x141f6c66},
				{8183, 3019, 0x141f5907},
				{8184, 3018, 0x141f45a9},
				{8185, 3017, 0x141f3260},
				{8186, 3016, 0x141f1f2c},
				{8187, 3015, 0x141f0bf8},
				{8188, 3014, 0x141ef8c5},
				{8189, 3013, 0x141ee5a6},
				{8190, 3012, 0x141ed29e},
				{8191, 3011, 0x141ebf95},
				{8192, 3010, 0x141eaca2},
				{8193, 3009, 0x141e99ae},
				{8194, 3008, 0x141e86d0},
				{8195, 3007, 0x141e741e},
				{8196, 3006, 0x141e6155},
				{8197, 3005, 0x141e4e8d},
				{8198, 3004, 0x141e3bda},
				{8199, 3003, 0x141e293d},
				{8200, 3002, 0x141e169f},
				{8201, 3001, 0x141e0417},
				{8202, 3000, 0x141df18f},
				{8203, 2999, 0x141ddf1d},
				{8204, 2998, 0x141dccaa},
				{8205, 2997, 0x141dba4d},
				{8206, 2996, 0x141da806},
				{8207, 2995, 0x141d95a9},
				{8208, 2994, 0x141d8377},
				{8209, 2993, 0x141d7145},
				{8210, 2992, 0x141d5f28},
				{8211, 2991, 0x141d4d0b},
				{8212, 2990, 0x141d3b04},
				{8213, 2989, 0x141d28fd},
				{8214, 2988, 0x141d170b},
				{8215, 2987, 0x141d051a},
				{8216, 2986, 0x141cf328},
				{8217, 2985, 0x141ce161},
				{8218, 2984, 0x141ccf85},
				{8219, 2983, 0x141cbdd4},
				{8220, 2982, 0x141cac0d},
				{8221, 2981, 0x141c9a71},
				{8222, 2980, 0x141c88c0},
				{8223, 2979, 0x141c773a},
				{8224, 2978, 0x141c659e},
				{8225, 2977, 0x141c5442},
				{8226, 2976, 0x141c42bc},
				{8227, 2975, 0x141c3161},
				{8228, 2974, 0x141c1ff0},
				{8229, 2973, 0x141c0eaa},
				{8230, 2972, 0x141bfd64},
				{8231, 2971, 0x141bec1e},
				{8232, 2970, 0x141bdaed},
				{8233, 2969, 0x141bc9bd},
				{8234, 2968, 0x141bb8a2},
				{8235, 2967, 0x141ba787},
				{8236, 2966, 0x141b9681},
				{8237, 2965, 0x141b857c},
				{8238, 2964, 0x141b748b},
				{8239, 2963, 0x141b639b},
				{8240, 2962, 0x141b52c1},
				{8241, 2961, 0x141b41fc},
				{8242, 2960, 0x141b3136},
				{8243, 2959, 0x141b2071},
				{8244, 2958, 0x141b0fc2},
				{8245, 2957, 0x141aff12},
				{8246, 2956, 0x141aee78},
				{8247, 2955, 0x141adddd},
				{8248, 2954, 0x141acd59},
				{8249, 2953, 0x141abcd4},
				{8250, 2952, 0x141aac4f},
				{8251, 2951, 0x141a9be0},
				{8252, 2950, 0x141a8b86},
				{8253, 2949, 0x141a7b2c},
				{8254, 2948, 0x141a6ad2},
				{8255, 2947, 0x141a5a8e},
				{8256, 2946, 0x141a4a74},
				{8257, 2945, 0x141a3a46},
				{8258, 2944, 0x141a2a17},
				{8259, 2943, 0x141a19fd},
				{8260, 2942, 0x141a09e4},
				{8261, 2941, 0x1419f9e0},
				{8262, 2940, 0x1419e9dc},
				{8263, 2939, 0x1419d9ed},
				{8264, 2938, 0x1419c9ff},
				{8265, 2937, 0x1419ba26},
				{8266, 2936, 0x1419aa4d},
				{8267, 2935, 0x14199a74},
				{8268, 2934, 0x14198ab0},
				{8269, 2933, 0x14197aed},
				{8270, 2932, 0x14196b3e},
				{8271, 2931, 0x14195ba6},
				{8272, 2930, 0x14194c0d},
				{8273, 2929, 0x14193c75},
				{8274, 2928, 0x14192cf1},
				{8275, 2927, 0x14191d6e},
				{8276, 2926, 0x14190e00},
				{8277, 2925, 0x1418fe93},
				{8278, 2924, 0x1418ef25},
				{8279, 2923, 0x1418dfcd},
				{8280, 2922, 0x1418d075},
				{8281, 2921, 0x1418c132},
				{8282, 2920, 0x1418b1ef},
				{8283, 2919, 0x1418a2c2},
				{8284, 2918, 0x14189394},
				{8285, 2917, 0x14188467},
				{8286, 2916, 0x1418754f},
				{8287, 2915, 0x1418664d},
				{8288, 2914, 0x1418574a},
				{8289, 2913, 0x14184848},
				{8290, 2912, 0x1418395b},
				{8291, 2911, 0x14182a6e},
				{8292, 2910, 0x14181b81},
				{8293, 2909, 0x14180caa},
				{8294, 2908, 0x1417fdd2},
				{8295, 2907, 0x1417ef10},
				{8296, 2906, 0x1417e04e},
				{8297, 2905, 0x1417d1a2},
				{8298, 2904, 0x1417c2f5},
				{8299, 2903, 0x1417b449},
				{8300, 2902, 0x1417a5b2},
				{8301, 2901, 0x1417971b},
				{8302, 2900, 0x141788af},
				{8303, 2899, 0x14177a2d},
				{8304, 2898, 0x14176bac},
				{8305, 2897, 0x14175d3f},
				{8306, 2896, 0x14174ed3},
				{8307, 2895, 0x1417407d},
				{8308, 2894, 0x14173226},
				{8309, 2893, 0x141723cf},
				{8310, 2892, 0x1417158e},
				{8311, 2891, 0x1417074d},
				{8312, 2890, 0x1416f921},
				{8313, 2889, 0x1416eaf5},
				{8314, 2888, 0x1416dcca},
				{8315, 2887, 0x1416ceb4},
				{8316, 2886, 0x1416c09d},
				{8317, 2885, 0x1416b2b2},
				{8318, 2884, 0x1416a4b1},
				{8319, 2883, 0x141696b0},
				{8320, 2882, 0x141688c5},
				{8321, 2881, 0x14167ada},
				{8322, 2880, 0x14166cee},
				{8323, 2879, 0x14165f18},
				{8324, 2878, 0x14165158},
				{8325, 2877, 0x14164382},
				{8326, 2876, 0x141635c2},
				{8327, 2875, 0x14162817},
				{8328, 2874, 0x14161a6c},
				{8329, 2873, 0x14160cc1},
				{8330, 2872, 0x1415ff16},
				{8331, 2871, 0x1415f181},
				{8332, 2870, 0x1415e3eb},
				{8333, 2869, 0x1415d681},
				{8334, 2868, 0x1415c901},
				{8335, 2867, 0x1415bb81},
				{8336, 2866, 0x1415ae16},
				{8337, 2865, 0x1415a0ac},
				{8338, 2864, 0x14159357},
				{8339, 2863, 0x141585ec},
				{8340, 2862, 0x141578ac},
				{8341, 2861, 0x14156b6d},
				{8342, 2860, 0x14155e43},
				{8343, 2859, 0x14155123},
				{8344, 2858, 0x14154404},
				{8345, 2857, 0x141536ef},
				{8346, 2856, 0x141529e5},
				{8347, 2855, 0x14151cdb},
				{8348, 2854, 0x14150fe7},
				{8349, 2853, 0x141502f2},
				{8350, 2852, 0x1414f609},
				{8351, 2851, 0x1414e91f},
				{8352, 2850, 0x1414dc40},
				{8353, 2849, 0x1414cf6c},
				{8354, 2848, 0x1414c297},
				{8355, 2847, 0x1414b5ce},
				{8356, 2846, 0x1414a90f},
				{8357, 2845, 0x14149c5b},
				{8358, 2844, 0x14148fa7},
				{8359, 2843, 0x141482fe},
				{8360, 2842, 0x14147654},
				{8361, 2841, 0x141469c0},
				{8362, 2840, 0x14145d2c},
				{8363, 2839, 0x141450ae},
				{8364, 2838, 0x14144430},
				{8365, 2837, 0x141437b1},
				{8366, 2836, 0x14142b3e},
				{8367, 2835, 0x14141eca},
				{8368, 2834, 0x14141261},
				{8369, 2833, 0x14140603},
				{8370, 2832, 0x1413f9af},
				{8371, 2831, 0x1413ed5c},
				{8372, 2830, 0x1413e113},
				{8373, 2829, 0x1413d4d5},
				{8374, 2828, 0x1413c8a2},
				{8375, 2827, 0x1413bc6f},
				{8376, 2826, 0x1413b046},
				{8377, 2825, 0x1413a41d},
				{8378, 2824, 0x14139800},
				{8379, 2823, 0x14138bf7},
				{8380, 2822, 0x14137fef},
				{8381, 2821, 0x141373e7},
				{8382, 2820, 0x141367e9},
				{8383, 2819, 0x14135bf6},
				{8384, 2818, 0x14135003},
				{8385, 2817, 0x1413441b},
				{8386, 2816, 0x14133833},
				{8387, 2815, 0x14132c60},
				{8388, 2814, 0x14132083},
				{8389, 2813, 0x141314bb},
				{8390, 2812, 0x141308f3},
				{8391, 2811, 0x1412fd36},
				{8392, 2810, 0x1412f183},
				{8393, 2809, 0x1412e5d1},
				{8394, 2808, 0x1412da34},
				{8395, 2807, 0x1412ce97},
				{8396, 2806, 0x1412c2fa},
				{8397, 2805, 0x1412b767},
				{8398, 2804, 0x1412abe0},
				{8399, 2803, 0x1412a058},
				{8400, 2802, 0x141294db},
				{8401, 2801, 0x1412895f},
				{8402, 2800, 0x14127ded},
				{8403, 2799, 0x14127285},
				{8404, 2798, 0x14126729},
				{8405, 2797, 0x14125bcc},
				{8406, 2796, 0x1412507a},
				{8407, 2795, 0x14124528},
				{8408, 2794, 0x141239e1},
				{8409, 2793, 0x14122eaf},
				{8410, 2792, 0x14122373},
				{8411, 2791, 0x14121841},
				{8412, 2790, 0x14120d1a},
				{8413, 2789, 0x141201f3},
				{8414, 2788, 0x1411f6d7},
				{8415, 2787, 0x1411ebc6},
				{8416, 2786, 0x1411e0b4},
				{8417, 2785, 0x1411d5ad},
				{8418, 2784, 0x1411caa7},
				{8419, 2783, 0x1411bfab},
				{8420, 2782, 0x1411b4b9},
				{8421, 2781, 0x1411a9c8},
				{8422, 2780, 0x14119ee2},
				{8423, 2779, 0x14119406},
				{8424, 2778, 0x1411892a},
				{8425, 2777, 0x14117e63},
				{8426, 2776, 0x14117392},
				{8427, 2775, 0x141168cc},
				{8428, 2774, 0x14115e10},
				{8429, 2773, 0x14115355},
				{8430, 2772, 0x141148a4},
				{8431, 2771, 0x14113dfe},
				{8432, 2770, 0x14113358},
				{8433, 2769, 0x141128bc},
				{8434, 2768, 0x14111e21},
				{8435, 2767, 0x14111390},
				{8436, 2766, 0x1411090a},
				{8437, 2765, 0x1410fe84},
				{8438, 2764, 0x1410f409},
				{8439, 2763, 0x1410e98e},
				{8440, 2762, 0x1410df28},
				{8441, 2761, 0x1410d4c2},
				{8442, 2760, 0x1410ca5c},
				{8443, 2759, 0x1410c001},
				{8444, 2758, 0x1410b5a6},
				{8445, 2757, 0x1410ab56},
				{8446, 2756, 0x1410a110},
				{8447, 2755, 0x141096cb},
				{8448, 2754, 0x14108c90},
				{8449, 2753, 0x14108255},
				{8450, 2752, 0x14107825},
				{8451, 2751, 0x14106e00},
				{8452, 2750, 0x141063da},
				{8453, 2749, 0x141059c0},
				{8454, 2748, 0x14104fa5},
				{8455, 2747, 0x141045a0},
				{8456, 2746, 0x14103b90},
				{8457, 2745, 0x1410318b},
				{8458, 2744, 0x14102790},
				{8459, 2743, 0x14101d96},
				{8460, 2742, 0x141013a6},
				{8461, 2741, 0x141009b7},
				{8462, 2740, 0x140fffd2},
				{8463, 2739, 0x140ff5f7},
				{8464, 2738, 0x140fec1d},
				{8465, 2737, 0x140fe24e},
				{8466, 2736, 0x140fd87e},
				{8467, 2735, 0x140fceb9},
				{8468, 2734, 0x140fc4f5},
				{8469, 2733, 0x140fbb3b},
				{8470, 2732, 0x140fb181},
				{8471, 2731, 0x140fa7e7},
				{8472, 2730, 0x140f9e37},
				{8473, 2729, 0x140f9493},
				{8474, 2728, 0x140f8af9},
				{8475, 2727, 0x140f815f},
				{8476, 2726, 0x140f77d0},
				{8477, 2725, 0x140f6e4c},
				{8478, 2724, 0x140f64c7},
				{8479, 2723, 0x140f5b43},
				{8480, 2722, 0x140f51c9},
				{8481, 2721, 0x140f485a},
				{8482, 2720, 0x140f3eec},
				{8483, 2719, 0x140f3587},
				{8484, 2718, 0x140f2c23},
				{8485, 2717, 0x140f22ca},
				{8486, 2716, 0x140f197b},
				{8487, 2715, 0x140f102c},
				{8488, 2714, 0x140f06de},
				{8489, 2713, 0x140efd9a},
				{8490, 2712, 0x140ef460},
				{8491, 2711, 0x140eeb27},
				{8492, 2710, 0x140ee1ee},
				{8493, 2709, 0x140ed8ca},
				{8494, 2708, 0x140ecf9c},
				{8495, 2707, 0x140ec678},
				{8496, 2706, 0x140ebd5f},
				{8497, 2705, 0x140eb446},
				{8498, 2704, 0x140eab38},
				{8499, 2703, 0x140ea229},
				{8500, 2702, 0x140e9926},
				{8501, 2701, 0x140e9038},
				{8502, 2700, 0x140e8734},
				{8503, 2699, 0x140e7e46},
				{8504, 2698, 0x140e7558},
				{8505, 2697, 0x140e6c6a},
				{8506, 2696, 0x140e6386},
				{8507, 2695, 0x140e5aae},
				{8508, 2694, 0x140e51ca},
				{8509, 2693, 0x140e48fc},
				{8510, 2692, 0x140e402e},
				{8511, 2691, 0x140e3760},
				{8512, 2690, 0x140e2ea8},
				{8513, 2689, 0x140e25e5},
				{8514, 2688, 0x140e1d2c},
				{8515, 2687, 0x140e147f},
				{8516, 2686, 0x140e0bd1},
				{8517, 2685, 0x140e032e},
				{8518, 2684, 0x140dfa8b},
				{8519, 2683, 0x140df1f3},
				{8520, 2682, 0x140de95b},
				{8521, 2681, 0x140de0c2},
				{8522, 2680, 0x140dd835},
				{8523, 2679, 0x140dcfb2},
				{8524, 2678, 0x140dc72f},
				{8525, 2677, 0x140dbeb7},
				{8526, 2676, 0x140db63f},
				{8527, 2675, 0x140dadc7},
				{8528, 2674, 0x140da55a},
				{8529, 2673, 0x140d9cf7},
				{8530, 2672, 0x140d9495},
				{8531, 2671, 0x140d8c32},
				{8532, 2670, 0x140d83e5},
				{8533, 2669, 0x140d7b8d},
				{8534, 2668, 0x140d7340},
				{8535, 2667, 0x140d6afe},
				{8536, 2666, 0x140d62b0},
				{8537, 2665, 0x140d5a79},
				{8538, 2664, 0x140d5241},
				{8539, 2663, 0x140d4a09},
				{8540, 2662, 0x140d41dd},
				{8541, 2661, 0x140d39b0},
				{8542, 2660, 0x140d318d},
				{8543, 2659, 0x140d296b},
				{8544, 2658, 0x140d2154},
				{8545, 2657, 0x140d193c},
				{8546, 2656, 0x140d1125},
				{8547, 2655, 0x140d0923},
				{8548, 2654, 0x140d0121},
				{8549, 2653, 0x140cf91f},
				{8550, 2652, 0x140cf11d},
				{8551, 2651, 0x140ce926},
				{8552, 2650, 0x140ce12e},
				{8553, 2649, 0x140cd942},
				{8554, 2648, 0x140cd155},
				{8555, 2647, 0x140cc974},
				{8556, 2646, 0x140cc192},
				{8557, 2645, 0x140cb9bb},
				{8558, 2644, 0x140cb1e4},
				{8559, 2643, 0x140caa17},
				{8560, 2642, 0x140ca240},
				{8561, 2641, 0x140c9a7f},
				{8562, 2640, 0x140c92bd},
				{8563, 2639, 0x140c8b06},
				{8564, 2638, 0x140c8350},
				{8565, 2637, 0x140c7b99},
				{8566, 2636, 0x140c73e2},
				{8567, 2635, 0x140c6c40},
				{8568, 2634, 0x140c6494},
				{8569, 2633, 0x140c5cf3},
				{8570, 2632, 0x140c5551},
				{8571, 2631, 0x140c4dbb},
				{8572, 2630, 0x140c4624},
				{8573, 2629, 0x140c3e98},
				{8574, 2628, 0x140c370c},
				{8575, 2627, 0x140c2f8b},
				{8576, 2626, 0x140c280a},
				{8577, 2625, 0x140c2089},
				{8578, 2624, 0x140c191d},
				{8579, 2623, 0x140c11a7},
				{8580, 2622, 0x140c0a3b},
				{8581, 2621, 0x140c02cf},
				{8582, 2620, 0x140bfb63},
				{8583, 2619, 0x140bf402},
				{8584, 2618, 0x140beca1},
				{8585, 2617, 0x140be54b},
				{8586, 2616, 0x140bddf5},
				{8587, 2615, 0x140bd6a9},
				{8588, 2614, 0x140bcf5e},
				{8589, 2613, 0x140bc812},
				{8590, 2612, 0x140bc0d2},
				{8591, 2611, 0x140bb991},
				{8592, 2610, 0x140bb25b},
				{8593, 2609, 0x140bab2f},
				{8594, 2608, 0x140ba3f9},
				{8595, 2607, 0x140b9cce},
				{8596, 2606, 0x140b95a3},
				{8597, 2605, 0x140b8e82},
				{8598, 2604, 0x140b8761},
				{8599, 2603, 0x140b8041},
				{8600, 2602, 0x140b792b},
				{8601, 2601, 0x140b7215},
				{8602, 2600, 0x140b6b0a},
				{8603, 2599, 0x140b63ff},
				{8604, 2598, 0x140b5cf4},
				{8605, 2597, 0x140b55f3},
				{8606, 2596, 0x140b4ef3},
				{8607, 2595, 0x140b47fd},
				{8608, 2594, 0x140b4108},
				{8609, 2593, 0x140b3a1d},
				{8610, 2592, 0x140b3327},
				{8611, 2591, 0x140b2c47},
				{8612, 2590, 0x140b255c},
				{8613, 2589, 0x140b1e7c},
				{8614, 2588, 0x140b179c},
				{8615, 2587, 0x140b10c6},
				{8616, 2586, 0x140b09f1},
				{8617, 2585, 0x140b031b},
				{8618, 2584, 0x140afc50},
				{8619, 2583, 0x140af586},
				{8620, 2582, 0x140aeebb},
				{8621, 2581, 0x140ae7fb},
				{8622, 2580, 0x140ae13b},
				{8623, 2579, 0x140ada86},
				{8624, 2578, 0x140ad3db},
				{8625, 2577, 0x140acd26},
				{8626, 2576, 0x140ac67b},
				{8627, 2575, 0x140abfd1},
				{8628, 2574, 0x140ab92c},
				{8629, 2573, 0x140ab291},
				{8630, 2572, 0x140aabfc},
				{8631, 2571, 0x140aa56c},
				{8632, 2570, 0x140a9ee2},
				{8633, 2569, 0x140a9858},
				{8634, 2568, 0x140a91d3},
				{8635, 2567, 0x140a8b53},
				{8636, 2566, 0x140a84d4},
				{8637, 2565, 0x140a7e59},
				{8638, 2564, 0x140a77e4},
				{8639, 2563, 0x140a717a},
				{8640, 2562, 0x140a6b0b},
				{8641, 2561, 0x140a64a1},
				{8642, 2560, 0x140a5e3c},
				{8643, 2559, 0x140a57dd},
				{8644, 2558, 0x140a517d},
				{8645, 2557, 0x140a4b23},
				{8646, 2556, 0x140a44c9},
				{8647, 2555, 0x140a3e7a},
				{8648, 2554, 0x140a382a},
				{8649, 2553, 0x140a31db},
				{8650, 2552, 0x140a2b97},
				{8651, 2551, 0x140a2552},
				{8652, 2550, 0x140a1f13},
				{8653, 2549, 0x140a18d4},
				{8654, 2548, 0x140a129a},
				{8655, 2547, 0x140a0c70},
				{8656, 2546, 0x140a063c},
				{8657, 2545, 0x140a0012},
				{8658, 2544, 0x1409f9e8},
				{8659, 2543, 0x1409f3be},
				{8660, 2542, 0x1409ed9f},
				{8661, 2541, 0x1409e780},
				{8662, 2540, 0x1409e167},
				{8663, 2539, 0x1409db4d},
				{8664, 2538, 0x1409d539},
				{8665, 2537, 0x1409cf2a},
				{8666, 2536, 0x1409c91b},
				{8667, 2535, 0x1409c317},
				{8668, 2534, 0x1409bd0d},
				{8669, 2533, 0x1409b70e},
				{8670, 2532, 0x1409b115},
				{8671, 2531, 0x1409ab1c},
				{8672, 2530, 0x1409a528},
				{8673, 2529, 0x14099f33},
				{8674, 2528, 0x14099945},
				{8675, 2527, 0x1409935b},
				{8676, 2526, 0x14098d72},
				{8677, 2525, 0x1409878e},
				{8678, 2524, 0x140981aa},
				{8679, 2523, 0x14097bd1},
				{8680, 2522, 0x140975f8},
				{8681, 2521, 0x1409701e},
				{8682, 2520, 0x14096a4b},
				{8683, 2519, 0x1409647c},
				{8684, 2518, 0x14095eb3},
				{8685, 2517, 0x140958ef},
				{8686, 2516, 0x1409532b},
				{8687, 2515, 0x14094d6d},
				{8688, 2514, 0x140947ae},
				{8689, 2513, 0x140941f5},
				{8690, 2512, 0x14093c3c},
				{8691, 2511, 0x14093689},
				{8692, 2510, 0x140930da},
				{8693, 2509, 0x14092b31},
				{8694, 2508, 0x14092589},
				{8695, 2507, 0x14091fe0},
				{8696, 2506, 0x14091a41},
				{8697, 2505, 0x140914a3},
				{8698, 2504, 0x14090f05},
				{8699, 2503, 0x14090971},
				{8700, 2502, 0x140903de},
				{8701, 2501, 0x1408fe50},
				{8702, 2500, 0x1408f8c7},
				{8703, 2499, 0x1408f339},
				{8704, 2498, 0x1408edb6},
				{8705, 2497, 0x1408e832},
				{8706, 2496, 0x1408e2b4},
				{8707, 2495, 0x1408dd36},
				{8708, 2494, 0x1408d7be},
				{8709, 2493, 0x1408d24a},
				{8710, 2492, 0x1408ccd7},
				{8711, 2491, 0x1408c769},
				{8712, 2490, 0x1408c1fb},
				{8713, 2489, 0x1408bc98},
				{8714, 2488, 0x1408b730},
				{8715, 2487, 0x1408b1d2},
				{8716, 2486, 0x1408ac79},
				{8717, 2485, 0x1408a71b},
				{8718, 2484, 0x1408a1c8},
				{8719, 2483, 0x14089c70},
				{8720, 2482, 0x14089722},
				{8721, 2481, 0x140891d4},
				{8722, 2480, 0x14088c8c},
				{8723, 2479, 0x14088744},
				{8724, 2478, 0x14088201},
				{8725, 2477, 0x14087cbe},
				{8726, 2476, 0x14087780},
				{8727, 2475, 0x14087248},
				{8728, 2474, 0x14086d10},
				{8729, 2473, 0x140867dd},
				{8730, 2472, 0x140862af},
				{8731, 2471, 0x14085d87},
				{8732, 2470, 0x1408585f},
				{8733, 2469, 0x14085337},
				{8734, 2468, 0x14084e14},
				{8735, 2467, 0x140848f1},
				{8736, 2466, 0x140843d4},
				{8737, 2465, 0x14083ebc},
				{8738, 2464, 0x140839a4},
				{8739, 2463, 0x14083491},
				{8740, 2462, 0x14082f7e},
				{8741, 2461, 0x14082a71},
				{8742, 2460, 0x14082569},
				{8743, 2459, 0x14082061},
				{8744, 2458, 0x14081b5f},
				{8745, 2457, 0x1408165c},
				{8746, 2456, 0x1408115f},
				{8747, 2455, 0x14080c6c},
				{8748, 2454, 0x14080775},
				{8749, 2453, 0x1408027d},
				{8750, 2452, 0x1407fd90},
				{8751, 2451, 0x1407f89d},
				{8752, 2450, 0x1407f3b5},
				{8753, 2449, 0x1407eece},
				{8754, 2448, 0x1407e9e6},
				{8755, 2447, 0x1407e503},
				{8756, 2446, 0x1407e026},
				{8757, 2445, 0x1407db49},
				{8758, 2444, 0x1407d672},
				{8759, 2443, 0x1407d19a},
				{8760, 2442, 0x1407ccc8},
				{8761, 2441, 0x1407c7fb},
				{8762, 2440, 0x1407c333},
				{8763, 2439, 0x1407be66},
				{8764, 2438, 0x1407b99f},
				{8765, 2437, 0x1407b4dd},
				{8766, 2436, 0x1407b01b},
				{8767, 2435, 0x1407ab5e},
				{8768, 2434, 0x1407a6a1},
				{8769, 2433, 0x1407a1e9},
				{8770, 2432, 0x14079d37},
				{8771, 2431, 0x14079885},
				{8772, 2430, 0x140793d3},
				{8773, 2429, 0x14078f26},
				{8774, 2428, 0x14078a7f},
				{8775, 2427, 0x140785d8},
				{8776, 2426, 0x14078136},
				{8777, 2425, 0x14077c94},
				{8778, 2424, 0x140777fd},
				{8779, 2423, 0x14077360},
				{8780, 2422, 0x14076ec9},
				{8781, 2421, 0x14076a31},
				{8782, 2420, 0x1407659f},
				{8783, 2419, 0x1407610e},
				{8784, 2418, 0x14075c81},
				{8785, 2417, 0x140757fa},
				{8786, 2416, 0x14075373},
				{8787, 2415, 0x14074eec},
				{8788, 2414, 0x14074a6f},
				{8789, 2413, 0x140745ed},
				{8790, 2412, 0x14074171},
				{8791, 2411, 0x14073cfa},
				{8792, 2410, 0x14073883},
				{8793, 2409, 0x14073417},
				{8794, 2408, 0x14072fa5},
				{8795, 2407, 0x14072b39},
				{8796, 2406, 0x140726cc},
				{8797, 2405, 0x14072265},
				{8798, 2404, 0x14071dfe},
				{8799, 2403, 0x1407199d},
				{8800, 2402, 0x1407153b},
				{8801, 2401, 0x140710df},
				{8802, 2400, 0x14070c83},
				{8803, 2399, 0x1407082c},
				{8804, 2398, 0x140703d5},
				{8805, 2397, 0x1406ff83},
				{8806, 2396, 0x1406fb32},
				{8807, 2395, 0x1406f6e6},
				{8808, 2394, 0x1406f29f},
				{8809, 2393, 0x1406ee58},
				{8810, 2392, 0x1406ea12},
				{8811, 2391, 0x1406e5d0},
				{8812, 2390, 0x1406e18f},
				{8813, 2389, 0x1406dd53},
				{8814, 2388, 0x1406d917},
				{8815, 2387, 0x1406d4e0},
				{8816, 2386, 0x1406d0a9},
				{8817, 2385, 0x1406cc78},
				{8818, 2384, 0x1406c847},
				{8819, 2383, 0x1406c41b},
				{8820, 2382, 0x1406bfef},
				{8821, 2381, 0x1406bbc8},
				{8822, 2380, 0x1406b7a2},
				{8823, 2379, 0x1406b37b},
				{8824, 2378, 0x1406af60},
				{8825, 2377, 0x1406ab44},
				{8826, 2376, 0x1406a728},
				{8827, 2375, 0x1406a311},
				{8828, 2374, 0x14069efb},
				{8829, 2373, 0x14069ae4},
				{8830, 2372, 0x140696d3},
				{8831, 2371, 0x140692c2},
				{8832, 2370, 0x14068eb7},
				{8833, 2369, 0x14068ab0},
				{8834, 2368, 0x140686a5},
				{8835, 2367, 0x140682a4},
				{8836, 2366, 0x14067e9d},
				{8837, 2365, 0x14067aa2},
				{8838, 2364, 0x140676a1},
				{8839, 2363, 0x140672aa},
				{8840, 2362, 0x14066eb4},
				{8841, 2361, 0x14066abe},
				{8842, 2360, 0x140666c7},
				{8843, 2359, 0x140662d7},
				{8844, 2358, 0x14065eeb},
				{8845, 2357, 0x14065afa},
				{8846, 2356, 0x14065714},
				{8847, 2355, 0x1406532e},
				{8848, 2354, 0x14064f48},
				{8849, 2353, 0x14064b62},
				{8850, 2352, 0x14064781},
				{8851, 2351, 0x140643a5},
				{8852, 2350, 0x14063fca},
				{8853, 2349, 0x14063bee},
				{8854, 2348, 0x1406381e},
				{8855, 2347, 0x14063448},
				{8856, 2346, 0x14063077},
				{8857, 2345, 0x14062ca6},
				{8858, 2344, 0x140628db},
				{8859, 2343, 0x14062510},
				{8860, 2342, 0x14062144},
				{8861, 2341, 0x14061d7e},
				{8862, 2340, 0x140619b8},
				{8863, 2339, 0x140615f8},
				{8864, 2338, 0x14061237},
				{8865, 2337, 0x14060e7c},
				{8866, 2336, 0x14060ac1},
				{8867, 2335, 0x14060706},
				{8868, 2334, 0x14060350},
				{8869, 2333, 0x1405ff9a},
				{8870, 2332, 0x1405fbef},
				{8871, 2331, 0x1405f83e},
				{8872, 2330, 0x1405f493},
				{8873, 2329, 0x1405f0e3},
				{8874, 2328, 0x1405ed3d},
				{8875, 2327, 0x1405e997},
				{8876, 2326, 0x1405e5f1},
				{8877, 2325, 0x1405e24b},
				{8878, 2324, 0x1405deab},
				{8879, 2323, 0x1405db10},
				{8880, 2322, 0x1405d775},
				{8881, 2321, 0x1405d3da},
				{8882, 2320, 0x1405d03f},
				{8883, 2319, 0x1405cca9},
				{8884, 2318, 0x1405c919},
				{8885, 2317, 0x1405c589},
				{8886, 2316, 0x1405c1fe},
				{8887, 2315, 0x1405be6d},
				{8888, 2314, 0x1405bae3},
				{8889, 2313, 0x1405b75d},
				{8890, 2312, 0x1405b3d2},
				{8891, 2311, 0x1405b052},
				{8892, 2310, 0x1405accc},
				{8893, 2309, 0x1405a94c},
				{8894, 2308, 0x1405a5d1},
				{8895, 2307, 0x1405a251},
				{8896, 2306, 0x14059ed6},
				{8897, 2305, 0x14059b61},
				{8898, 2304, 0x140597eb},
				{8899, 2303, 0x14059476},
				{8900, 2302, 0x14059106},
				{8901, 2301, 0x14058d9b},
				{8902, 2300, 0x14058a2b},
				{8903, 2299, 0x140586c0},
				{8904, 2298, 0x14058355},
				{8905, 2297, 0x14057feb},
				{8906, 2296, 0x14057c85},
				{8907, 2295, 0x14057925},
				{8908, 2294, 0x140575c0},
				{8909, 2293, 0x14057260},
				{8910, 2292, 0x14056f05},
				{8911, 2291, 0x14056ba5},
				{8912, 2290, 0x1405684b},
				{8913, 2289, 0x140564f5},
				{8914, 2288, 0x140561a0},
				{8915, 2287, 0x14055e4b},
				{8916, 2286, 0x14055b00},
				{8917, 2285, 0x140557b6},
				{8918, 2284, 0x1405546b},
				{8919, 2283, 0x14055126},
				{8920, 2282, 0x14054de1},
				{8921, 2281, 0x14054a9c},
				{8922, 2280, 0x14054759},
				{8923, 2279, 0x1405441c},
				{8924, 2278, 0x140540dc},
				{8925, 2277, 0x14053da2},
				{8926, 2276, 0x14053a68},
				{8927, 2275, 0x14053730},
				{8928, 2274, 0x140533fb},
				{8929, 2273, 0x140530c6},
				{8930, 2272, 0x14052d93},
				{8931, 2271, 0x14052a66},
				{8932, 2270, 0x14052739},
				{8933, 2269, 0x1405240c},
				{8934, 2268, 0x140520e2},
				{8935, 2267, 0x14051dba},
				{8936, 2266, 0x14051a93},
				{8937, 2265, 0x1405176e},
				{8938, 2264, 0x1405144b},
				{8939, 2263, 0x14051129},
				{8940, 2262, 0x14050e09},
				{8941, 2261, 0x14050aed},
				{8942, 2260, 0x140507d2},
				{8943, 2259, 0x140504b8},
				{8944, 2258, 0x140501a1},
				{8945, 2257, 0x1404fe8c},
				{8946, 2256, 0x1404fb79},
				{8947, 2255, 0x1404f867},
				{8948, 2254, 0x1404f558},
				{8949, 2253, 0x1404f248},
				{8950, 2252, 0x1404ef3e},
				{8951, 2251, 0x1404ec31},
				{8952, 2250, 0x1404e92a},
				{8953, 2249, 0x1404e622},
				{8954, 2248, 0x1404e31e},
				{8955, 2247, 0x1404e019},
				{8956, 2246, 0x1404dd17},
				{8957, 2245, 0x1404da17},
				{8958, 2244, 0x1404d71b},
				{8959, 2243, 0x1404d41e},
				{8960, 2242, 0x1404d124},
				{8961, 2241, 0x1404ce2a},
				{8962, 2240, 0x1404cb38},
				{8963, 2239, 0x1404c843},
				{8964, 2238, 0x1404c54f},
				{8965, 2237, 0x1404c25f},
				{8966, 2236, 0x1404bf70},
				{8967, 2235, 0x1404bc81},
				{8968, 2234, 0x1404b994},
				{8969, 2233, 0x1404b6aa},
				{8970, 2232, 0x1404b3c3},
				{8971, 2231, 0x1404b0dc},
				{8972, 2230, 0x1404adf7},
				{8973, 2229, 0x1404ab12},
				{8974, 2228, 0x1404a833},
				{8975, 2227, 0x1404a554},
				{8976, 2226, 0x1404a275},
				{8977, 2225, 0x14049f9b},
				{8978, 2224, 0x14049cc1},
				{8979, 2223, 0x140499ea},
				{8980, 2222, 0x14049713},
				{8981, 2221, 0x1404943c},
				{8982, 2220, 0x1404916a},
				{8983, 2219, 0x14048e98},
				{8984, 2218, 0x14048bc9},
				{8985, 2217, 0x140488fa},
				{8986, 2216, 0x1404862d},
				{8987, 2215, 0x14048361},
				{8988, 2214, 0x1404809a},
				{8989, 2213, 0x14047dd3},
				{8990, 2212, 0x14047b0c},
				{8991, 2211, 0x14047848},
				{8992, 2210, 0x14047588},
				{8993, 2209, 0x140472c9},
				{8994, 2208, 0x1404700a},
				{8995, 2207, 0x14046d4b},
				{8996, 2206, 0x14046a92},
				{8997, 2205, 0x140467d8},
				{8998, 2204, 0x1404651f},
				{8999, 2203, 0x14046268},
				{9000, 2202, 0x14045fb3},
				{9001, 2201, 0x14045d02},
				{9002, 2200, 0x14045a50},
				{9003, 2199, 0x1404579f},
				{9004, 2198, 0x140454f0},
				{9005, 2197, 0x14045243},
				{9006, 2196, 0x14044f9a},
				{9007, 2195, 0x14044cf0},
				{9008, 2194, 0x14044a4c},
				{9009, 2193, 0x140447a5},
				{9010, 2192, 0x14044501},
				{9011, 2191, 0x14044260},
				{9012, 2190, 0x14043fbe},
				{9013, 2189, 0x14043d1f},
				{9014, 2188, 0x14043a81},
				{9015, 2187, 0x140437e4},
				{9016, 2186, 0x1404354b},
				{9017, 2185, 0x140432b2},
				{9018, 2184, 0x1404301b},
				{9019, 2183, 0x14042d84},
				{9020, 2182, 0x14042af0},
				{9021, 2181, 0x1404285c},
				{9022, 2180, 0x140425cd},
				{9023, 2179, 0x1404233e},
				{9024, 2178, 0x140420b2},
				{9025, 2177, 0x14041e26},
				{9026, 2176, 0x14041b9a},
				{9027, 2175, 0x14041911},
				{9028, 2174, 0x1404168a},
				{9029, 2173, 0x14041404},
				{9030, 2172, 0x14041180},
				{9031, 2171, 0x14040efe},
				{9032, 2170, 0x14040c7d},
				{9033, 2169, 0x140409fc},
				{9034, 2168, 0x14040780},
				{9035, 2167, 0x14040501},
				{9036, 2166, 0x14040288},
				{9037, 2165, 0x1404000f},
				{9038, 2164, 0x1403fd98},
				{9039, 2163, 0x1403fb22},
				{9040, 2162, 0x1403f8ae},
				{9041, 2161, 0x1403f63a},
				{9042, 2160, 0x1403f3c9},
				{9043, 2159, 0x1403f158},
				{9044, 2158, 0x1403eee9},
				{9045, 2157, 0x1403ec7d},
				{9046, 2156, 0x1403ea11},
				{9047, 2155, 0x1403e7a6},
				{9048, 2154, 0x1403e53d},
				{9049, 2153, 0x1403e2d6},
				{9050, 2152, 0x1403e070},
				{9051, 2151, 0x1403de0c},
				{9052, 2150, 0x1403dbab},
				{9053, 2149, 0x1403d94a},
				{9054, 2148, 0x1403d6eb},
				{9055, 2147, 0x1403d48d},
				{9056, 2146, 0x1403d231},
				{9057, 2145, 0x1403cfd5},
				{9058, 2144, 0x1403cd7a},
				{9059, 2143, 0x1403cb23},
				{9060, 2142, 0x1403c8ca},
				{9061, 2141, 0x1403c677},
				{9062, 2140, 0x1403c423},
				{9063, 2139, 0x1403c1cf},
				{9064, 2138, 0x1403bf7e},
				{9065, 2137, 0x1403bd30},
				{9066, 2136, 0x1403bae2},
				{9067, 2135, 0x1403b893},
				{9068, 2134, 0x1403b648},
				{9069, 2133, 0x1403b401},
				{9070, 2132, 0x1403b1b8},
				{9071, 2131, 0x1403af72},
				{9072, 2130, 0x1403ad2c},
				{9073, 2129, 0x1403aae8},
				{9074, 2128, 0x1403a8a5},
				{9075, 2127, 0x1403a664},
				{9076, 2126, 0x1403a423},
				{9077, 2125, 0x1403a1e5},
				{9078, 2124, 0x14039fa7},
				{9079, 2123, 0x14039d6b},
				{9080, 2122, 0x14039b2f},
				{9081, 2121, 0x140398f7},
				{9082, 2120, 0x140396c0},
				{9083, 2119, 0x1403948a},
				{9084, 2118, 0x14039257},
				{9085, 2117, 0x14039023},
				{9086, 2116, 0x14038df3},
				{9087, 2115, 0x14038bc2},
				{9088, 2114, 0x14038991},
				{9089, 2113, 0x14038763},
				{9090, 2112, 0x14038537},
				{9091, 2111, 0x1403830c},
				{9092, 2110, 0x140380e1},
				{9093, 2109, 0x14037eb8},
				{9094, 2108, 0x14037c92},
				{9095, 2107, 0x14037a6c},
				{9096, 2106, 0x14037846},
				{9097, 2105, 0x14037622},
				{9098, 2104, 0x14037402},
				{9099, 2103, 0x140371e1},
				{9100, 2102, 0x14036fc3},
				{9101, 2101, 0x14036da5},
				{9102, 2100, 0x14036b8a},
				{9103, 2099, 0x1403696e},
				{9104, 2098, 0x14036753},
				{9105, 2097, 0x1403653a},
				{9106, 2096, 0x14036324},
				{9107, 2095, 0x1403610e},
				{9108, 2094, 0x14035ef8},
				{9109, 2093, 0x14035ce5},
				{9110, 2092, 0x14035ad2},
				{9111, 2091, 0x140358c1},
				{9112, 2090, 0x140356b3},
				{9113, 2089, 0x140354a3},
				{9114, 2088, 0x14035297},
				{9115, 2087, 0x1403508c},
				{9116, 2086, 0x14034e81},
				{9117, 2085, 0x14034c78},
				{9118, 2084, 0x14034a70},
				{9119, 2083, 0x1403486a},
				{9120, 2082, 0x14034664},
				{9121, 2081, 0x14034461},
				{9122, 2080, 0x1403425e},
				{9123, 2079, 0x1403405b},
				{9124, 2078, 0x14033e5a},
				{9125, 2077, 0x14033c5c},
				{9126, 2076, 0x14033a5c},
				{9127, 2075, 0x14033861},
				{9128, 2074, 0x14033666},
				{9129, 2073, 0x1403346a},
				{9130, 2072, 0x14033272},
				{9131, 2071, 0x1403307c},
				{9132, 2070, 0x14032e84},
				{9133, 2069, 0x14032c8e},
				{9134, 2068, 0x14032a9b},
				{9135, 2067, 0x140328a8},
				{9136, 2066, 0x140326b5},
				{9137, 2065, 0x140324c4},
				{9138, 2064, 0x140322d4},
				{9139, 2063, 0x140320e6},
				{9140, 2062, 0x14031ef9},
				{9141, 2061, 0x14031d0e},
				{9142, 2060, 0x14031b23},
				{9143, 2059, 0x14031937},
				{9144, 2058, 0x1403174f},
				{9145, 2057, 0x14031569},
				{9146, 2056, 0x14031384},
				{9147, 2055, 0x1403119e},
				{9148, 2054, 0x14030fbb},
				{9149, 2053, 0x14030dd8},
				{9150, 2052, 0x14030bf8},
				{9151, 2051, 0x14030a18},
				{9152, 2050, 0x14030837},
				{9153, 2049, 0x1403065a},
				{9154, 2048, 0x1403047c},
				{9155, 2047, 0x140302a1},
				{9156, 2046, 0x140300c6},
				{9157, 2045, 0x1402feeb},
				{9158, 2044, 0x1402fd13},
				{9159, 2043, 0x1402fb3b},
				{9160, 2042, 0x1402f965},
				{9161, 2041, 0x1402f792},
				{9162, 2040, 0x1402f5bf},
				{9163, 2039, 0x1402f3ea},
				{9164, 2038, 0x1402f21a},
				{9165, 2037, 0x1402f049},
				{9166, 2036, 0x1402ee79},
				{9167, 2035, 0x1402eca9},
				{9168, 2034, 0x1402eadc},
				{9169, 2033, 0x1402e911},
				{9170, 2032, 0x1402e746},
				{9171, 2031, 0x1402e57b},
				{9172, 2030, 0x1402e3b0},
				{9173, 2029, 0x1402e1e8},
				{9174, 2028, 0x1402e023},
				{9175, 2027, 0x1402de5d},
				{9176, 2026, 0x1402dc9a},
				{9177, 2025, 0x1402dad5},
				{9178, 2024, 0x1402d912},
				{9179, 2023, 0x1402d752},
				{9180, 2022, 0x1402d58f},
				{9181, 2021, 0x1402d3d2},
				{9182, 2020, 0x1402d212},
				{9183, 2019, 0x1402d054},
				{9184, 2018, 0x1402ce97},
				{9185, 2017, 0x1402ccdc},
				{9186, 2016, 0x1402cb21},
				{9187, 2015, 0x1402c967},
				{9188, 2014, 0x1402c7af},
				{9189, 2013, 0x1402c5f9},
				{9190, 2012, 0x1402c441},
				{9191, 2011, 0x1402c28c},
				{9192, 2010, 0x1402c0d9},
				{9193, 2009, 0x1402bf27},
				{9194, 2008, 0x1402bd74},
				{9195, 2007, 0x1402bbc1},
				{9196, 2006, 0x1402ba11},
				{9197, 2005, 0x1402b861},
				{9198, 2004, 0x1402b6b4},
				{9199, 2003, 0x1402b507},
				{9200, 2002, 0x1402b359},
				{9201, 2001, 0x1402b1af},
				{9202, 2000, 0x1402b004},
				{9203, 1999, 0x1402ae5b},
				{9204, 1998, 0x1402acb5},
				{9205, 1997, 0x1402ab10},
				{9206, 1996, 0x1402a96b},
				{9207, 1995, 0x1402a7ca},
				{9208, 1994, 0x1402a628},
				{9209, 1993, 0x1402a487},
				{9210, 1992, 0x1402a2e6},
				{9211, 1991, 0x1402a147},
				{9212, 1990, 0x14029fa9},
				{9213, 1989, 0x14029e0c},
				{9214, 1988, 0x14029c6e},
				{9215, 1987, 0x14029ad4},
				{9216, 1986, 0x14029939},
				{9217, 1985, 0x1402979f},
				{9218, 1984, 0x14029607},
				{9219, 1983, 0x1402946f},
				{9220, 1982, 0x140292d8},
				{9221, 1981, 0x14029142},
				{9222, 1980, 0x14028fb0},
				{9223, 1979, 0x14028e1c},
				{9224, 1978, 0x14028c89},
				{9225, 1977, 0x14028af7},
				{9226, 1976, 0x14028966},
				{9227, 1975, 0x140287d6},
				{9228, 1974, 0x14028646},
				{9229, 1973, 0x140284b8},
				{9230, 1972, 0x1402832a},
				{9231, 1971, 0x1402819f},
				{9232, 1970, 0x14028013},
				{9233, 1969, 0x14027e88},
				{9234, 1968, 0x14027cff},
				{9235, 1967, 0x14027b76},
				{9236, 1966, 0x140279ee},
				{9237, 1965, 0x14027868},
				{9238, 1964, 0x140276e4},
				{9239, 1963, 0x1402755f},
				{9240, 1962, 0x140273dc},
				{9241, 1961, 0x14027258},
				{9242, 1960, 0x140270d5},
				{9243, 1959, 0x14026f54},
				{9244, 1958, 0x14026dd5},
				{9245, 1957, 0x14026c55},
				{9246, 1956, 0x14026ad7},
				{9247, 1955, 0x14026958},
				{9248, 1954, 0x140267db},
				{9249, 1953, 0x14026660},
				{9250, 1952, 0x140264e5},
				{9251, 1951, 0x1402636b},
				{9252, 1950, 0x140261f2},
				{9253, 1949, 0x1402607c},
				{9254, 1948, 0x14025f04},
				{9255, 1947, 0x14025d8e},
				{9256, 1946, 0x14025c17},
				{9257, 1945, 0x14025aa4},
				{9258, 1944, 0x1402592f},
				{9259, 1943, 0x140257bc},
				{9260, 1942, 0x1402564a},
				{9261, 1941, 0x140254d9},
				{9262, 1940, 0x1402536a},
				{9263, 1939, 0x140251fa},
				{9264, 1938, 0x1402508a},
				{9265, 1937, 0x14024f1d},
				{9266, 1936, 0x14024db1},
				{9267, 1935, 0x14024c45},
				{9268, 1934, 0x14024adb},
				{9269, 1933, 0x14024971},
				{9270, 1932, 0x14024808},
				{9271, 1931, 0x140246a0},
				{9272, 1930, 0x14024539},
				{9273, 1929, 0x140243d1},
				{9274, 1928, 0x1402426b},
				{9275, 1927, 0x14024106},
				{9276, 1926, 0x14023fa2},
				{9277, 1925, 0x14023e3f},
				{9278, 1924, 0x14023cdd},
				{9279, 1923, 0x14023b7a},
				{9280, 1922, 0x14023a1a},
				{9281, 1921, 0x140238ba},
				{9282, 1920, 0x1402375c},
				{9283, 1919, 0x140235fc},
				{9284, 1918, 0x140234a1},
				{9285, 1917, 0x14023344},
				{9286, 1916, 0x140231e9},
				{9287, 1915, 0x1402308f},
				{9288, 1914, 0x14022f34},
				{9289, 1913, 0x14022ddc},
				{9290, 1912, 0x14022c83},
				{9291, 1911, 0x14022b2b},
				{9292, 1910, 0x140229d4},
				{9293, 1909, 0x1402287f},
				{9294, 1908, 0x14022729},
				{9295, 1907, 0x140225d6},
				{9296, 1906, 0x14022482},
				{9297, 1905, 0x14022330},
				{9298, 1904, 0x140221de},
				{9299, 1903, 0x1402208f},
				{9300, 1902, 0x14021f3e},
				{9301, 1901, 0x14021dee},
				{9302, 1900, 0x14021ca0},
				{9303, 1899, 0x14021b52},
				{9304, 1898, 0x14021a06},
				{9305, 1897, 0x140218ba},
				{9306, 1896, 0x1402176f},
				{9307, 1895, 0x14021623},
				{9308, 1894, 0x140214db},
				{9309, 1893, 0x14021391},
				{9310, 1892, 0x14021249},
				{9311, 1891, 0x14021102},
				{9312, 1890, 0x14020fbc},
				{9313, 1889, 0x14020e76},
				{9314, 1888, 0x14020d33},
				{9315, 1887, 0x14020bee},
				{9316, 1886, 0x14020aab},
				{9317, 1885, 0x14020969},
				{9318, 1884, 0x14020827},
				{9319, 1883, 0x140206e5},
				{9320, 1882, 0x140205a5},
				{9321, 1881, 0x14020466},
				{9322, 1880, 0x14020327},
				{9323, 1879, 0x140201e9},
				{9324, 1878, 0x140200ac},
				{9325, 1877, 0x1401ff70},
				{9326, 1876, 0x1401fe34},
				{9327, 1875, 0x1401fcf9},
				{9328, 1874, 0x1401fbbf},
				{9329, 1873, 0x1401fa85},
				{9330, 1872, 0x1401f94f},
				{9331, 1871, 0x1401f817},
				{9332, 1870, 0x1401f6df},
				{9333, 1869, 0x1401f5a9},
				{9334, 1868, 0x1401f474},
				{9335, 1867, 0x1401f33f},
				{9336, 1866, 0x1401f20c},
				{9337, 1865, 0x1401f0d9},
				{9338, 1864, 0x1401efa5},
				{9339, 1863, 0x1401ee73},
				{9340, 1862, 0x1401ed43},
				{9341, 1861, 0x1401ec12},
				{9342, 1860, 0x1401eae3},
				{9343, 1859, 0x1401e9b4},
				{9344, 1858, 0x1401e886},
				{9345, 1857, 0x1401e75a},
				{9346, 1856, 0x1401e62d},
				{9347, 1855, 0x1401e502},
				{9348, 1854, 0x1401e3d7},
				{9349, 1853, 0x1401e2ad},
				{9350, 1852, 0x1401e183},
				{9351, 1851, 0x1401e059},
				{9352, 1850, 0x1401df32},
				{9353, 1849, 0x1401de09},
				{9354, 1848, 0x1401dce4},
				{9355, 1847, 0x1401dbbd},
				{9356, 1846, 0x1401da98},
				{9357, 1845, 0x1401d974},
				{9358, 1844, 0x1401d84f},
				{9359, 1843, 0x1401d72c},
				{9360, 1842, 0x1401d60a},
				{9361, 1841, 0x1401d4e8},
				{9362, 1840, 0x1401d3c8},
				{9363, 1839, 0x1401d2a7},
				{9364, 1838, 0x1401d188},
				{9365, 1837, 0x1401d069},
				{9366, 1836, 0x1401cf4a},
				{9367, 1835, 0x1401ce2c},
				{9368, 1834, 0x1401cd10},
				{9369, 1833, 0x1401cbf4},
				{9370, 1832, 0x1401cad8},
				{9371, 1831, 0x1401c9bd},
				{9372, 1830, 0x1401c8a4},
				{9373, 1829, 0x1401c78a},
				{9374, 1828, 0x1401c672},
				{9375, 1827, 0x1401c559},
				{9376, 1826, 0x1401c443},
				{9377, 1825, 0x1401c32c},
				{9378, 1824, 0x1401c217},
				{9379, 1823, 0x1401c101},
				{9380, 1822, 0x1401bfed},
				{9381, 1821, 0x1401bed8},
				{9382, 1820, 0x1401bdc5},
				{9383, 1819, 0x1401bcb2},
				{9384, 1818, 0x1401bba0},
				{9385, 1817, 0x1401ba8f},
				{9386, 1816, 0x1401b97e},
				{9387, 1815, 0x1401b86e},
				{9388, 1814, 0x1401b75f},
				{9389, 1813, 0x1401b650},
				{9390, 1812, 0x1401b542},
				{9391, 1811, 0x1401b436},
				{9392, 1810, 0x1401b32a},
				{9393, 1809, 0x1401b21d},
				{9394, 1808, 0x1401b112},
				{9395, 1807, 0x1401b007},
				{9396, 1806, 0x1401aefc},
				{9397, 1805, 0x1401adf3},
				{9398, 1804, 0x1401aceb},
				{9399, 1803, 0x1401abe2},
				{9400, 1802, 0x1401aadb},
				{9401, 1801, 0x1401a9d4},
				{9402, 1800, 0x1401a8cd},
				{9403, 1799, 0x1401a7c8},
				{9404, 1798, 0x1401a6c3},
				{9405, 1797, 0x1401a5bf},
				{9406, 1796, 0x1401a4bc},
				{9407, 1795, 0x1401a3b9},
				{9408, 1794, 0x1401a2b6},
				{9409, 1793, 0x1401a1b5},
				{9410, 1792, 0x1401a0b3},
				{9411, 1791, 0x14019fb3},
				{9412, 1790, 0x14019eb3},
				{9413, 1789, 0x14019db4},
				{9414, 1788, 0x14019cb5},
				{9415, 1787, 0x14019bb6},
				{9416, 1786, 0x14019ab8},
				{9417, 1785, 0x140199bc},
				{9418, 1784, 0x140198c0},
				{9419, 1783, 0x140197c4},
				{9420, 1782, 0x140196c9},
				{9421, 1781, 0x140195ce},
				{9422, 1780, 0x140194d6},
				{9423, 1779, 0x140193dc},
				{9424, 1778, 0x140192e3},
				{9425, 1777, 0x140191ec},
				{9426, 1776, 0x140190f4},
				{9427, 1775, 0x14018ffd},
				{9428, 1774, 0x14018f06},
				{9429, 1773, 0x14018e10},
				{9430, 1772, 0x14018d1b},
				{9431, 1771, 0x14018c27},
				{9432, 1770, 0x14018b33},
				{9433, 1769, 0x14018a40},
				{9434, 1768, 0x1401894d},
				{9435, 1767, 0x1401885a},
				{9436, 1766, 0x14018769},
				{9437, 1765, 0x14018678},
				{9438, 1764, 0x14018588},
				{9439, 1763, 0x14018498},
				{9440, 1762, 0x140183a9},
				{9441, 1761, 0x140182bb},
				{9442, 1760, 0x140181cc},
				{9443, 1759, 0x140180de},
				{9444, 1758, 0x14017ff1},
				{9445, 1757, 0x14017f05},
				{9446, 1756, 0x14017e19},
				{9447, 1755, 0x14017d2c},
				{9448, 1754, 0x14017c42},
				{9449, 1753, 0x14017b58},
				{9450, 1752, 0x14017a6d},
				{9451, 1751, 0x14017985},
				{9452, 1750, 0x1401789c},
				{9453, 1749, 0x140177b5},
				{9454, 1748, 0x140176cd},
				{9455, 1747, 0x140175e6},
				{9456, 1746, 0x14017500},
				{9457, 1745, 0x1401741a},
				{9458, 1744, 0x14017335},
				{9459, 1743, 0x1401724f},
				{9460, 1742, 0x1401716b},
				{9461, 1741, 0x14017087},
				{9462, 1740, 0x14016fa4},
				{9463, 1739, 0x14016ec2},
				{9464, 1738, 0x14016ddf},
				{9465, 1737, 0x14016cfe},
				{9466, 1736, 0x14016c1c},
				{9467, 1735, 0x14016b3c},
				{9468, 1734, 0x14016a5d},
				{9469, 1733, 0x1401697d},
				{9470, 1732, 0x1401689f},
				{9471, 1731, 0x140167c0},
				{9472, 1730, 0x140166e1},
				{9473, 1729, 0x14016604},
				{9474, 1728, 0x14016526},
				{9475, 1727, 0x1401644a},
				{9476, 1726, 0x1401636e},
				{9477, 1725, 0x14016292},
				{9478, 1724, 0x140161b8},
				{9479, 1723, 0x140160dd},
				{9480, 1722, 0x14016004},
				{9481, 1721, 0x14015f2a},
				{9482, 1720, 0x14015e51},
				{9483, 1719, 0x14015d79},
				{9484, 1718, 0x14015ca1},
				{9485, 1717, 0x14015bca},
				{9486, 1716, 0x14015af2},
				{9487, 1715, 0x14015a1c},
				{9488, 1714, 0x14015946},
				{9489, 1713, 0x14015871},
				{9490, 1712, 0x1401579c},
				{9491, 1711, 0x140156c8},
				{9492, 1710, 0x140155f5},
				{9493, 1709, 0x14015523},
				{9494, 1708, 0x14015451},
				{9495, 1707, 0x14015380},
				{9496, 1706, 0x140152af},
				{9497, 1705, 0x140151df},
				{9498, 1704, 0x1401510f},
				{9499, 1703, 0x14015040},
				{9500, 1702, 0x14014f71},
				{9501, 1701, 0x14014ea3},
				{9502, 1700, 0x14014dd5},
				{9503, 1699, 0x14014d07},
				{9504, 1698, 0x14014c3a},
				{9505, 1697, 0x14014b6e},
				{9506, 1696, 0x14014aa1},
				{9507, 1695, 0x140149d6},
				{9508, 1694, 0x1401490b},
				{9509, 1693, 0x14014840},
				{9510, 1692, 0x14014776},
				{9511, 1691, 0x140146ac},
				{9512, 1690, 0x140145e3},
				{9513, 1689, 0x1401451b},
				{9514, 1688, 0x14014453},
				{9515, 1687, 0x1401438b},
				{9516, 1686, 0x140142c3},
				{9517, 1685, 0x140141fd},
				{9518, 1684, 0x14014136},
				{9519, 1683, 0x14014070},
				{9520, 1682, 0x14013fab},
				{9521, 1681, 0x14013ee6},
				{9522, 1680, 0x14013e21},
				{9523, 1679, 0x14013d5e},
				{9524, 1678, 0x14013c9a},
				{9525, 1677, 0x14013bd7},
				{9526, 1676, 0x14013b14},
				{9527, 1675, 0x14013a51},
				{9528, 1674, 0x14013990},
				{9529, 1673, 0x140138cf},
				{9530, 1672, 0x1401380f},
				{9531, 1671, 0x1401374e},
				{9532, 1670, 0x1401368e},
				{9533, 1669, 0x140135ce},
				{9534, 1668, 0x1401350f},
				{9535, 1667, 0x14013451},
				{9536, 1666, 0x14013393},
				{9537, 1665, 0x140132d5},
				{9538, 1664, 0x14013218},
				{9539, 1663, 0x1401315b},
				{9540, 1662, 0x1401309f},
				{9541, 1661, 0x14012fe3},
				{9542, 1660, 0x14012f27},
				{9543, 1659, 0x14012e6c},
				{9544, 1658, 0x14012db2},
				{9545, 1657, 0x14012cf8},
				{9546, 1656, 0x14012c3f},
				{9547, 1655, 0x14012b85},
				{9548, 1654, 0x14012acd},
				{9549, 1653, 0x14012a14},
				{9550, 1652, 0x1401295d},
				{9551, 1651, 0x140128a5},
				{9552, 1650, 0x140127ee},
				{9553, 1649, 0x14012737},
				{9554, 1648, 0x14012681},
				{9555, 1647, 0x140125cb},
				{9556, 1646, 0x14012516},
				{9557, 1645, 0x14012461},
				{9558, 1644, 0x140123ad},
				{9559, 1643, 0x140122f9},
				{9560, 1642, 0x14012246},
				{9561, 1641, 0x14012193},
				{9562, 1640, 0x140120e0},
				{9563, 1639, 0x1401202d},
				{9564, 1638, 0x14011f7c},
				{9565, 1637, 0x14011ecb},
				{9566, 1636, 0x14011e1a},
				{9567, 1635, 0x14011d69},
				{9568, 1634, 0x14011cb9},
				{9569, 1633, 0x14011c09},
				{9570, 1632, 0x14011b5a},
				{9571, 1631, 0x14011aab},
				{9572, 1630, 0x140119fc},
				{9573, 1629, 0x1401194e},
				{9574, 1628, 0x140118a0},
				{9575, 1627, 0x140117f4},
				{9576, 1626, 0x14011747},
				{9577, 1625, 0x1401169b},
				{9578, 1624, 0x140115ef},
				{9579, 1623, 0x14011544},
				{9580, 1622, 0x14011499},
				{9581, 1621, 0x140113ed},
				{9582, 1620, 0x14011343},
				{9583, 1619, 0x14011299},
				{9584, 1618, 0x140111f0},
				{9585, 1617, 0x14011147},
				{9586, 1616, 0x1401109e},
				{9587, 1615, 0x14010ff6},
				{9588, 1614, 0x14010f4e},
				{9589, 1613, 0x14010ea6},
				{9590, 1612, 0x14010dff},
				{9591, 1611, 0x14010d59},
				{9592, 1610, 0x14010cb3},
				{9593, 1609, 0x14010c0d},
				{9594, 1608, 0x14010b68},
				{9595, 1607, 0x14010ac3},
				{9596, 1606, 0x14010a1e},
				{9597, 1605, 0x1401097a},
				{9598, 1604, 0x140108d6},
				{9599, 1603, 0x14010832},
				{9600, 1602, 0x1401078f},
				{9601, 1601, 0x140106ed},
				{9602, 1600, 0x1401064b},
				{9603, 1599, 0x140105a9},
				{9604, 1598, 0x14010507},
				{9605, 1597, 0x14010466},
				{9606, 1596, 0x140103c6},
				{9607, 1595, 0x14010326},
				{9608, 1594, 0x14010286},
				{9609, 1593, 0x140101e6},
				{9610, 1592, 0x14010147},
				{9611, 1591, 0x140100a8},
				{9612, 1590, 0x1401000a},
				{9613, 1589, 0x1400ff6c},
				{9614, 1588, 0x1400fece},
				{9615, 1587, 0x1400fe31},
				{9616, 1586, 0x1400fd94},
				{9617, 1585, 0x1400fcf8},
				{9618, 1584, 0x1400fc5c},
				{9619, 1583, 0x1400fbc0},
				{9620, 1582, 0x1400fb25},
				{9621, 1581, 0x1400fa8b},
				{9622, 1580, 0x1400f9f0},
				{9623, 1579, 0x1400f955},
				{9624, 1578, 0x1400f8bc},
				{9625, 1577, 0x1400f822},
				{9626, 1576, 0x1400f789},
				{9627, 1575, 0x1400f6f1},
				{9628, 1574, 0x1400f658},
				{9629, 1573, 0x1400f5c0},
				{9630, 1572, 0x1400f529},
				{9631, 1571, 0x1400f492},
				{9632, 1570, 0x1400f3fb},
				{9633, 1569, 0x1400f364},
				{9634, 1568, 0x1400f2ce},
				{9635, 1567, 0x1400f238},
				{9636, 1566, 0x1400f1a3},
				{9637, 1565, 0x1400f10e},
				{9638, 1564, 0x1400f07a},
				{9639, 1563, 0x1400efe6},
				{9640, 1562, 0x1400ef52},
				{9641, 1561, 0x1400eebe},
				{9642, 1560, 0x1400ee2b},
				{9643, 1559, 0x1400ed98},
				{9644, 1558, 0x1400ed05},
				{9645, 1557, 0x1400ec74},
				{9646, 1556, 0x1400ebe2},
				{9647, 1555, 0x1400eb50},
				{9648, 1554, 0x1400eabf},
				{9649, 1553, 0x1400ea2e},
				{9650, 1552, 0x1400e99e},
				{9651, 1551, 0x1400e90e},
				{9652, 1550, 0x1400e87f},
				{9653, 1549, 0x1400e7f0},
				{9654, 1548, 0x1400e760},
				{9655, 1547, 0x1400e6d2},
				{9656, 1546, 0x1400e644},
				{9657, 1545, 0x1400e5b6},
				{9658, 1544, 0x1400e528},
				{9659, 1543, 0x1400e49b},
				{9660, 1542, 0x1400e40e},
				{9661, 1541, 0x1400e382},
				{9662, 1540, 0x1400e2f5},
				{9663, 1539, 0x1400e269},
				{9664, 1538, 0x1400e1de},
				{9665, 1537, 0x1400e153},
				{9666, 1536, 0x1400e0c8},
				{9667, 1535, 0x1400e03e},
				{9668, 1534, 0x1400dfb4},
				{9669, 1533, 0x1400df2a},
				{9670, 1532, 0x1400dea1},
				{9671, 1531, 0x1400de17},
				{9672, 1530, 0x1400dd8e},
				{9673, 1529, 0x1400dd06},
				{9674, 1528, 0x1400dc7e},
				{9675, 1527, 0x1400dbf6},
				{9676, 1526, 0x1400db6e},
				{9677, 1525, 0x1400dae7},
				{9678, 1524, 0x1400da61},
				{9679, 1523, 0x1400d9da},
				{9680, 1522, 0x1400d954},
				{9681, 1521, 0x1400d8ce},
				{9682, 1520, 0x1400d849},
				{9683, 1519, 0x1400d7c4},
				{9684, 1518, 0x1400d73f},
				{9685, 1517, 0x1400d6ba},
				{9686, 1516, 0x1400d636},
				{9687, 1515, 0x1400d5b2},
				{9688, 1514, 0x1400d52e},
				{9689, 1513, 0x1400d4ab},
				{9690, 1512, 0x1400d428},
				{9691, 1511, 0x1400d3a5},
				{9692, 1510, 0x1400d323},
				{9693, 1509, 0x1400d2a1},
				{9694, 1508, 0x1400d220},
				{9695, 1507, 0x1400d19e},
				{9696, 1506, 0x1400d11d},
				{9697, 1505, 0x1400d09c},
				{9698, 1504, 0x1400d01c},
				{9699, 1503, 0x1400cf9c},
				{9700, 1502, 0x1400cf1c},
				{9701, 1501, 0x1400ce9d},
				{9702, 1500, 0x1400ce1d},
				{9703, 1499, 0x1400cd9e},
				{9704, 1498, 0x1400cd20},
				{9705, 1497, 0x1400cca1},
				{9706, 1496, 0x1400cc23},
				{9707, 1495, 0x1400cba6},
				{9708, 1494, 0x1400cb28},
				{9709, 1493, 0x1400caab},
				{9710, 1492, 0x1400ca2e},
				{9711, 1491, 0x1400c9b2},
				{9712, 1490, 0x1400c935},
				{9713, 1489, 0x1400c8ba},
				{9714, 1488, 0x1400c83f},
				{9715, 1487, 0x1400c7c3},
				{9716, 1486, 0x1400c748},
				{9717, 1485, 0x1400c6ce},
				{9718, 1484, 0x1400c653},
				{9719, 1483, 0x1400c5d9},
				{9720, 1482, 0x1400c55f},
				{9721, 1481, 0x1400c4e5},
				{9722, 1480, 0x1400c46c},
				{9723, 1479, 0x1400c3f3},
				{9724, 1478, 0x1400c37a},
				{9725, 1477, 0x1400c302},
				{9726, 1476, 0x1400c28a},
				{9727, 1475, 0x1400c212},
				{9728, 1474, 0x1400c19b},
				{9729, 1473, 0x1400c124},
				{9730, 1472, 0x1400c0ad},
				{9731, 1471, 0x1400c036},
				{9732, 1470, 0x1400bfc0},
				{9733, 1469, 0x1400bf4a},
				{9734, 1468, 0x1400bed4},
				{9735, 1467, 0x1400be5e},
				{9736, 1466, 0x1400bde9},
				{9737, 1465, 0x1400bd74},
				{9738, 1464, 0x1400bcff},
				{9739, 1463, 0x1400bc8b},
				{9740, 1462, 0x1400bc16},
				{9741, 1461, 0x1400bba3},
				{9742, 1460, 0x1400bb2f},
				{9743, 1459, 0x1400babb},
				{9744, 1458, 0x1400ba49},
				{9745, 1457, 0x1400b9d6},
				{9746, 1456, 0x1400b963},
				{9747, 1455, 0x1400b8f1},
				{9748, 1454, 0x1400b87f},
				{9749, 1453, 0x1400b80d},
				{9750, 1452, 0x1400b79c},
				{9751, 1451, 0x1400b72a},
				{9752, 1450, 0x1400b6ba},
				{9753, 1449, 0x1400b649},
				{9754, 1448, 0x1400b5d8},
				{9755, 1447, 0x1400b568},
				{9756, 1446, 0x1400b4f8},
				{9757, 1445, 0x1400b489},
				{9758, 1444, 0x1400b419},
				{9759, 1443, 0x1400b3ab},
				{9760, 1442, 0x1400b33b},
				{9761, 1441, 0x1400b2cd},
				{9762, 1440, 0x1400b25f},
				{9763, 1439, 0x1400b1f1},
				{9764, 1438, 0x1400b183},
				{9765, 1437, 0x1400b115},
				{9766, 1436, 0x1400b0a7},
				{9767, 1435, 0x1400b03b},
				{9768, 1434, 0x1400afcd},
				{9769, 1433, 0x1400af61},
				{9770, 1432, 0x1400aef5},
				{9771, 1431, 0x1400ae89},
				{9772, 1430, 0x1400ae1d},
				{9773, 1429, 0x1400adb1},
				{9774, 1428, 0x1400ad45},
				{9775, 1427, 0x1400acdb},
				{9776, 1426, 0x1400ac70},
				{9777, 1425, 0x1400ac05},
				{9778, 1424, 0x1400ab9b},
				{9779, 1423, 0x1400ab31},
				{9780, 1422, 0x1400aac8},
				{9781, 1421, 0x1400aa5f},
				{9782, 1420, 0x1400a9f6},
				{9783, 1419, 0x1400a98e},
				{9784, 1418, 0x1400a925},
				{9785, 1417, 0x1400a8bd},
				{9786, 1416, 0x1400a855},
				{9787, 1415, 0x1400a7ee},
				{9788, 1414, 0x1400a787},
				{9789, 1413, 0x1400a71f},
				{9790, 1412, 0x1400a6b9},
				{9791, 1411, 0x1400a652},
				{9792, 1410, 0x1400a5ec},
				{9793, 1409, 0x1400a586},
				{9794, 1408, 0x1400a520},
				{9795, 1407, 0x1400a4ba},
				{9796, 1406, 0x1400a455},
				{9797, 1405, 0x1400a3f0},
				{9798, 1404, 0x1400a38b},
				{9799, 1403, 0x1400a326},
				{9800, 1402, 0x1400a2c1},
				{9801, 1401, 0x1400a25d},
				{9802, 1400, 0x1400a1f9},
				{9803, 1399, 0x1400a195},
				{9804, 1398, 0x1400a132},
				{9805, 1397, 0x1400a0cf},
				{9806, 1396, 0x1400a06c},
				{9807, 1395, 0x1400a009},
				{9808, 1394, 0x14009fa6},
				{9809, 1393, 0x14009f44},
				{9810, 1392, 0x14009ee2},
				{9811, 1391, 0x14009e80},
				{9812, 1390, 0x14009e1e},
				{9813, 1389, 0x14009dbc},
				{9814, 1388, 0x14009d5b},
				{9815, 1387, 0x14009cfa},
				{9816, 1386, 0x14009c99},
				{9817, 1385, 0x14009c39},
				{9818, 1384, 0x14009bd8},
				{9819, 1383, 0x14009b78},
				{9820, 1382, 0x14009b19},
				{9821, 1381, 0x14009ab9},
				{9822, 1380, 0x14009a5a},
				{9823, 1379, 0x140099fb},
				{9824, 1378, 0x1400999c},
				{9825, 1377, 0x1400993d},
				{9826, 1376, 0x140098de},
				{9827, 1375, 0x14009880},
				{9828, 1374, 0x14009822},
				{9829, 1373, 0x140097c4},
				{9830, 1372, 0x14009767},
				{9831, 1371, 0x14009709},
				{9832, 1370, 0x140096ac},
				{9833, 1369, 0x1400964f},
				{9834, 1368, 0x140095f2},
				{9835, 1367, 0x14009596},
				{9836, 1366, 0x1400953a},
				{9837, 1365, 0x140094de},
				{9838, 1364, 0x14009482},
				{9839, 1363, 0x14009426},
				{9840, 1362, 0x140093cb},
				{9841, 1361, 0x14009370},
				{9842, 1360, 0x14009315},
				{9843, 1359, 0x140092ba},
				{9844, 1358, 0x1400925f},
				{9845, 1357, 0x14009205},
				{9846, 1356, 0x140091ab},
				{9847, 1355, 0x14009151},
				{9848, 1354, 0x140090f8},
				{9849, 1353, 0x1400909e},
				{9850, 1352, 0x14009045},
				{9851, 1351, 0x14008fec},
				{9852, 1350, 0x14008f93},
				{9853, 1349, 0x14008f3a},
				{9854, 1348, 0x14008ee2},
				{9855, 1347, 0x14008e8a},
				{9856, 1346, 0x14008e32},
				{9857, 1345, 0x14008dda},
				{9858, 1344, 0x14008d83},
				{9859, 1343, 0x14008d2b},
				{9860, 1342, 0x14008cd4},
				{9861, 1341, 0x14008c7d},
				{9862, 1340, 0x14008c27},
				{9863, 1339, 0x14008bd0},
				{9864, 1338, 0x14008b7a},
				{9865, 1337, 0x14008b24},
				{9866, 1336, 0x14008ace},
				{9867, 1335, 0x14008a79},
				{9868, 1334, 0x14008a23},
				{9869, 1333, 0x140089ce},
				{9870, 1332, 0x14008979},
				{9871, 1331, 0x14008924},
				{9872, 1330, 0x140088cf},
				{9873, 1329, 0x1400887b},
				{9874, 1328, 0x14008827},
				{9875, 1327, 0x140087d2},
				{9876, 1326, 0x1400877f},
				{9877, 1325, 0x1400872b},
				{9878, 1324, 0x140086d8},
				{9879, 1323, 0x14008684},
				{9880, 1322, 0x14008631},
				{9881, 1321, 0x140085de},
				{9882, 1320, 0x1400858c},
				{9883, 1319, 0x1400853a},
				{9884, 1318, 0x140084e7},
				{9885, 1317, 0x14008496},
				{9886, 1316, 0x14008444},
				{9887, 1315, 0x140083f2},
				{9888, 1314, 0x140083a0},
				{9889, 1313, 0x1400834f},
				{9890, 1312, 0x140082fe},
				{9891, 1311, 0x140082ad},
				{9892, 1310, 0x1400825d},
				{9893, 1309, 0x1400820c},
				{9894, 1308, 0x140081bc},
				{9895, 1307, 0x1400816c},
				{9896, 1306, 0x1400811c},
				{9897, 1305, 0x140080cd},
				{9898, 1304, 0x1400807e},
				{9899, 1303, 0x1400802e},
				{9900, 1302, 0x137fdf60},
				{9901, 1301, 0x137f908e},
				{9902, 1300, 0x137f41bc},
				{9903, 1299, 0x137ef340},
				{9904, 1298, 0x137ea51a},
				{9905, 1297, 0x137e56f4},
				{9906, 1296, 0x137e08ce},
				{9907, 1295, 0x137dbafd},
				{9908, 1294, 0x137d6d83},
				{9909, 1293, 0x137d2008},
				{9910, 1292, 0x137cd2e4},
				{9911, 1291, 0x137c8615},
				{9912, 1290, 0x137c3947},
				{9913, 1289, 0x137becce},
				{9914, 1288, 0x137ba055},
				{9915, 1287, 0x137b5432},
				{9916, 1286, 0x137b080f},
				{9917, 1285, 0x137abc42},
				{9918, 1284, 0x137a7075},
				{9919, 1283, 0x137a24fd},
				{9920, 1282, 0x1379d986},
				{9921, 1281, 0x13798e65},
				{9922, 1280, 0x13794399},
				{9923, 1279, 0x1378f8ce},
				{9924, 1278, 0x1378ae02},
				{9925, 1277, 0x1378638d},
				{9926, 1276, 0x1378196d},
				{9927, 1275, 0x1377cf4d},
				{9928, 1274, 0x137785d9},
				{9929, 1273, 0x13773c0f},
				{9930, 1272, 0x1376f29b},
				{9931, 1271, 0x1376a927},
				{9932, 1270, 0x13766009},
				{9933, 1269, 0x13761740},
				{9934, 1268, 0x1375ce22},
				{9935, 1267, 0x137585b0},
				{9936, 1266, 0x13753d3d},
				{9937, 1265, 0x1374f4cb},
				{9938, 1264, 0x1374acae},
				{9939, 1263, 0x137464e8},
				{9940, 1262, 0x13741d21},
				{9941, 1261, 0x1373d5b0},
				{9942, 1260, 0x13738e3f},
				{9943, 1259, 0x13734724},
				{9944, 1258, 0x13730009},
				{9945, 1257, 0x1372b944},
				{9946, 1256, 0x1372727f},
				{9947, 1255, 0x13722c0f},
				{9948, 1254, 0x1371e5a0},
				{9949, 1253, 0x13719f87},
				{9950, 1252, 0x1371596d},
				{9951, 1251, 0x137113aa},
				{9952, 1250, 0x1370cde6},
				{9953, 1249, 0x13708878},
				{9954, 1248, 0x1370430b},
				{9955, 1247, 0x136ffdf3},
				{9956, 1246, 0x136fb8db},
				{9957, 1245, 0x136f7419},
				{9958, 1244, 0x136f2f57},
				{9959, 1243, 0x136eeb41},
				{9960, 1242, 0x136ea6d5},
				{9961, 1241, 0x136e62bf},
				{9962, 1240, 0x136e1ea8},
				{9963, 1239, 0x136ddae8},
				{9964, 1238, 0x136d9728},
				{9965, 1237, 0x136d53bd},
				{9966, 1236, 0x136d1052},
				{9967, 1235, 0x136ccd3e},
				{9968, 1234, 0x136c8a29},
				{9969, 1233, 0x136c4715},
				{9970, 1232, 0x136c04ac},
				{9971, 1231, 0x136bc1ed},
				{9972, 1230, 0x136b7f84},
				{9973, 1229, 0x136b3d71},
				{9974, 1228, 0x136afbb4},
				{9975, 1227, 0x136ab9a1},
				{9976, 1226, 0x136a77e3},
				{9977, 1225, 0x136a367c},
				{9978, 1224, 0x1369f515},
				{9979, 1223, 0x1369b3ad},
				{9980, 1222, 0x1369729c},
				{9981, 1221, 0x1369318a},
				{9982, 1220, 0x1368f0cf},
				{9983, 1219, 0x1368b069},
				{9984, 1218, 0x13686fad},
				{9985, 1217, 0x13682f9d},
				{9986, 1216, 0x1367ef38},
				{9987, 1215, 0x1367af28},
				{9988, 1214, 0x13676f6e},
				{9989, 1213, 0x1367300a},
				{9990, 1212, 0x1366f0a6},
				{9991, 1211, 0x1366b141},
				{9992, 1210, 0x136671dd},
				{9993, 1209, 0x136632cf},
				{9994, 1208, 0x1365f3c1},
				{9995, 1207, 0x1365b508},
				{9996, 1206, 0x13657650},
				{9997, 1205, 0x136537ed},
				{9998, 1204, 0x1364f98b},
				{9999, 1203, 0x1364bb7e},
				{10000, 1202, 0x13647d71},
			},
		},
	}

	b := BlockChain{
		chainParams: &chaincfg.MainNetParams,
	}
	b.chainParams.ReduceMinDifficulty = false
	for _, test := range testVectors {
		for i, block := range test.blocks {
			target, err := b.calcAsertRequiredDifficulty(&blockNode{height: int32(block.height), timestamp: block.timestamp}, int32(test.anchorHeight), test.anchorTime, test.anchorBits, time.Unix(block.timestamp, 0))
			if err != nil {
				t.Errorf("Run %d, block %d: %s", test.run, i, err)
			}
			if target != block.expectedTarget {
				t.Errorf("Run %d, block %d: Expected target %d got %d", test.run, i, block.expectedTarget, target)
			}
		}
	}
}
